/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/graph/Graph.java#13 $
 *
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.graph;

import java.io.Serializable;
import java.util.ArrayDeque;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.google.common.collect.Sets;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

public class Graph<T> implements Iterable<T>, Serializable {

    private Set<T> nodes;

    /**
     * The graphs edges represented as {@link Multimap} entries of the form
     * <pre>
     * { src => { dst1, dst2, dst3} }
     * </pre>
     */
    private Multimap<T, T> edges;

    /*
     * Only used for the external representation of an edge
     */
    public static class Edge<T> implements Serializable {
        private T src;
        private T dest;

        private Edge(T src, T dest) {
            this.src = src;
            this.dest = dest;
        }

        public T getSource() {
            return this.src;
        }

        public T getSrc() {
            return this.src;
        }

        public T getDest() {
            return this.dest;
        }

        @Override
        public String toString() {
            return String.format("%s ==> %s", this.src, this.dest);
        }
    }

    public Graph() {
        this.nodes = new HashSet<T>();
        this.edges = HashMultimap.create();
    }

    public Graph(Graph<T> other) {
        this.nodes = new HashSet<T>(other.nodes);
        this.edges = HashMultimap.create(other.edges);
    }

    private Graph(Set<T> nodes, Multimap<T, T> edges) {
        this.nodes = nodes;
        this.edges = edges;
    }

    /**
     * Static factory method that allows for cleaner construction using type inference.
     * For example:
     * <p>
     * {@code Graph<String> graph = Graph.create();}
     */
    public static <T> Graph<T> create() {
        return new Graph<T>();
    }

    public Set<Edge<T>> getEdges() {
        Set<Edge<T>> edgeSet = new HashSet<Edge<T>>();
        for (Map.Entry<T, T> edge : this.edges.entries()) {
            T src = edge.getKey();
            T dest = edge.getValue();

            edgeSet.add(new Edge<T>(src, dest));
        }
        return edgeSet;
    }

    public Set<T> getSuccessors(T node) {
        return ImmutableSet.copyOf(this.edges.get(node));
    }

    public Set<T> getPredecessors(T node) {
        Set<T> predecessors = new HashSet<T>();

        for (Map.Entry<T, T> edge : this.edges.entries()) {
            T src = edge.getKey();
            T dest = edge.getValue();

            if (node.equals(dest)) {
                predecessors.add(src);
            }
        }

        return Collections.unmodifiableSet(predecessors);
    }

    public Set<T> getDescendants(T node) {
        Set<T> descendants = Sets.newHashSet();
        for (T successor : this.getSuccessors(node)) {
            descendants.addAll(Sets.newHashSet(this.depthFirstPreorderIterator(successor)));
        }

        return descendants;
    }

    public Set<T> getAncestors(T node) {
        return this.reverse().getDescendants(node);
    }

    @SuppressWarnings("unchecked")
    public void addEdge(T src, T dest) {
        this.addEdges(src, dest);
    }

    public void addEdges(T src, T... dests) {
        this.addNode(src);
        for (T dest : dests) {
            this.addNode(dest);
            this.edges.put(src, dest);
        }
    }

    @SuppressWarnings("unchecked")
    public Graph<T> withEdge(T src, T dest) {
        return this.withEdges(src, dest);
    }

    public Graph<T> withEdges(T src, T... dests) {
        this.addEdges(src, dests);
        return this;
    }

    public void addNode(T node) {
        this.nodes.add(node);
    }

    public Graph<T> withNode(T node) {
        this.addNode(node);
        return this;
    }

    @Override
    public Iterator<T> iterator() {
        return this.getNodes().iterator();
    }

    public Set<T> getNodes() {
        return Collections.unmodifiableSet(this.nodes);
    }

    public Set<T> getSources() {
        /*
         * TODO
         * XXX
         * This is inefficient...
         */
        return this.reverse().getSinks();
    }

    public Set<T> getSinks() {
        return Sets.difference(this.nodes, this.edges.keySet());
    }

    /**
     * Remove the node from the graph.
     * All edges in which that node is a source or destination will also be removed.
     */
    public void removeNode(T node) {
        this.nodes.remove(node);
        this.edges.removeAll(node);

        for (Iterator<Map.Entry<T, T>> i = this.edges.entries().iterator(); i.hasNext();) {
            Map.Entry<T, T> edge = i.next();

            T dest = edge.getValue();

            if (node.equals(dest)) {
                i.remove();
            }
        }

    }

    public void removeEdge(T src, T dest) {
        for (Iterator<T> i = this.edges.get(src).iterator(); i.hasNext();) {
            if (i.next().equals(dest)) {
                i.remove();
            }
        }
    }

    public void replace(T orig, T replacement) {

        for (T successor : this.getSuccessors(orig)) {
            this.addEdge(replacement, successor);
        }

        for (T predecessor : this.getPredecessors(orig)) {
            this.addEdge(predecessor, replacement);
        }

        this.removeNode(orig);
    }

    /**
     * Return a <b>new</b> <tt>Graph</tt> with the same nodes as this graph but with the edges reversed.
     * @return the reversed graph
     */
    public Graph<T> reverse() {
        Multimap<T, T> reverseEdges = HashMultimap.create();
        Multimaps.invertFrom(this.edges, reverseEdges);

        return new Graph<T>(new HashSet<T>(this.nodes), reverseEdges);
    }

    public boolean pathExists(T source, T dest) {
        if (source.equals(dest)) {
            return true;
        }

        for (Iterator<T> i = this.depthFirstPreorderIterator(source); i.hasNext(); ) {
            if (dest.equals(i.next())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Iterator over the graph nodes in order of a depth-first <b>preorder</b> traversal.
     */
    public Iterator<T> depthFirstPreorderIterator(T startNode) {
        return new DepthFirstPreorderIterator<T>(startNode, this);
    }

    public void depthFirstVisit(T startNode, NodeVisitor<T> visitor) {
        this.depthFirstVisitHelper(startNode, visitor, new HashSet<T>());
    }

    private void depthFirstVisitHelper(T node, NodeVisitor<T> visitor, Set<T> nodesVisited) {

        if (nodesVisited.contains(node)) {
            return;
        }

        nodesVisited.add(node);

        if (visitor.preVisit(node)) {

            for (T child : this.getSuccessors(node)) {
                this.depthFirstVisitHelper(child, visitor, nodesVisited);
            }

            visitor.postVisit(node);

        }

    }

    /**
     * Return a list of the nodes in "topological sort" order.
     * An ordering in which each node is guaranteed to come before nodes to which it has edges.
     *
     * <p>
     * The results of this method are unspecified if the graph has cycles.
     */
    public List<T> topologicalSort() {
        final Set<T> visited = new HashSet<T>(this.nodes.size());
        final Deque<T> result = new ArrayDeque<T>(this.nodes.size());

        /*
         * The *reverse* of the post-order traversal list represents a topological sort of a graph.
         *
         * Since the graph could contain multiple sources, initiate a depth-first search from each source
         * short-circuiting each DFS if we've already hit a node that has been visited by a DFS originated
         * from a previous source.
         */
        for (T source : this.getSources()) {
            this.depthFirstVisit(source, new NodeVisitor<T>() {
                @Override
                public boolean preVisit(T node) {
                    boolean alreadyVisited = visited.contains(node);
                    visited.add(node);

                    return !alreadyVisited;
                }

                @Override
                public void postVisit(T node) {
                    /*
                     * Since we're building up a list in reverse order, *prepend* the node to the head of the Deque
                     */
                    result.addFirst(node);
                }
            });
        }

        return ImmutableList.copyOf(result);
    }

    public <R> R resultProvidingDepthFirstVisit(T startNode, ResultProvidingNodeVisitor<T, R> visitor) {
        return this.resultProvidingDepthFirstVisitHelper(startNode, visitor, new HashMap<T, R>());
    }

    private <R> R resultProvidingDepthFirstVisitHelper(T node, ResultProvidingNodeVisitor<T, R> visitor, Map<T, R> nodesVisited) {

        if (nodesVisited.containsKey(node)) {
            return nodesVisited.get(node);
        }

        visitor.preVisit(node);
        Set<R> successorResults = new HashSet<R>();
        for (T child : this.getSuccessors(node)) {
            successorResults.add(this.resultProvidingDepthFirstVisitHelper(child, visitor, nodesVisited));
        }

        R result = visitor.postVisit(node, successorResults);
        nodesVisited.put(node, result);

        return result;
    }

    /**
     * Return a <b>new</b> <tt>Graph</tt> with the result of applying the
     * given function to each node in this graph.
     */
    public <S> Graph<S> apply(Function<T, S> function) {
        Map<T, S> nodeMap = new HashMap<T, S>();
        Graph<S> transformedGraph = new Graph<S>();

        for (T node : this.nodes) {
            S transformedNode = function.apply(node);
            nodeMap.put(node, transformedNode);
            transformedGraph.addNode(transformedNode);
        }

        for (Map.Entry<T, T> edge : this.edges.entries()) {
            T src = edge.getKey();
            T dest = edge.getValue();

            transformedGraph.addEdge(nodeMap.get(src), nodeMap.get(dest));
        }

        return transformedGraph;
    }

    public void union(Graph<T> other) {
        this.nodes.addAll(other.nodes);
        this.edges.putAll(other.edges);
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder()
            .append(this.nodes)
            .append(this.edges)
            .toHashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof Graph)) {
            return false;
        }

        @SuppressWarnings("unchecked")
        Graph<T> that = (Graph<T>) obj;
        return new EqualsBuilder()
            .append(this.nodes, that.nodes)
            .append(this.edges, that.edges)
            .isEquals();
    }

    @Override
    public String toString() {
        return Joiner.on("\n").join(this.getEdges());
    }

    /**
     * Flatten a <tt>Graph</tt> of <tt>Graph</tt>s into a single, flat <tt>Graph</tt>.
     *
     * <p>
     * Edges are created from the {@link #getSinks() sinks} of each subGraph to
     * the {@link #getSources() sources} of its sucessor subGraphs.
     *
     * <p>
     * For example, the input <tt>Graph</tt>:
     * <pre>
     * (  A  )
     * ( / \ )
     * (B   C)
     *    |
     * (D   E)
     * ( \ / )
     * (  F  )
     * </pre>
     * is flattened into the output <tt>Graph</tt>:
     * <pre>
     *    A
     *   / \
     *  B   C
     *  | X |
     *  D   E
     *   \ /
     *    F
     * </pre>
     */
    public static <T> Graph<T> flatten(final Graph<Graph<T>> graph) {
        Graph<T> result = new Graph<T>();

        for (Graph<T> subGraph : graph.getNodes()) {
            for (T subGraphNode : subGraph.getNodes()) {
                result.addNode(subGraphNode);
            }

            for (Map.Entry<T, T> subGraphEdge : subGraph.edges.entries()) {
                result.addEdge(subGraphEdge.getKey(), subGraphEdge.getValue());
            }

            for (Graph<T> predecessorSubGraph : graph.getPredecessors(subGraph)) {
                for (T predecessorSubGraphSink : predecessorSubGraph.getSinks()) {
                    for (T subGraphSource : subGraph.getSources()) {
                        result.addEdge(predecessorSubGraphSink, subGraphSource);
                    }
                }
            }
        }

        return result;
    }

    /**
     * Encode this graph in Graphviz DOT format.
     * {@link BasicDisplayNodeFunction}
     */
    public String toGraphviz(String graphName) {
        return DisplayGraph.fromGraph(graphName, this).toGraphviz();
    }
}
