/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/logging/NDCConverter.java#2 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.common.logging;

import java.util.Deque;
import java.util.Iterator;

import ch.qos.logback.classic.pattern.ClassicConverter;
import ch.qos.logback.classic.spi.ILoggingEvent;

import com.netapp.common.logging.NDC.NDCElement;


/**
 * Custom converter for our {@link com.netapp.dfm.common.logging.NDC}.  This converter must be declared in
 * logback.xml prior to use as follows:
 *
 * {@code <conversionRule conversionWord="ndc" converterClass="com.netapp.common.logging.NDCConverter"/>}
 *
 * <p>
 * It can then be used in format specifiers like this: {@code %ndc{:}}
 *
 * <p>
 * The arguments to the conversion rule are:
 * <ol>
 * <li>The delimiter (e.g. ":", "|", etc.)
 * <li>The ANSI coloring setting - if the second argument is <tt>color</tt> then ANSI coloring will be enabled.
 * <blockquote>
 * Coloring is typically only recommended for use with {@link ch.qos.logback.core.ConsoleAppender}s.  Using the
 * coloring option with {@link ch.qos.logback.core.FileAppender}s will result in ANSI escape sequences in the log
 * files, which most text editors do not interpret correctly.
 * </blockquote>
 * </ol>
 *
 * <p>
 * For example, to include the NDC in a log format, delimited by "|" and using colors, use the following specifier:
 * {@code %ndc{|, color}}
 *
 * <p>
 * The <tt>color</tt> param controls NDC coloring on a per-conversion basis.
 * The property <tt>ndc.color</tt> in the logback context controls NDC coloring at a global level.
 * The property <tt>ndc.color</tt> is <tt>false</tt> by default.
 *
 * @author thomaswb
 * @version $Revision: #2 $
 *          Date: $Date: 2011/07/06 $
 */
public class NDCConverter extends ClassicConverter {
    private static final String COLOR_PARAM = "color";
    private static final String COLOR_PROPERTY = "ndc.color";

    private volatile boolean colorEnabled;
    private volatile String separator;

    @Override
    public void start() {
        this.separator = getFirstOption() == null ? "|" : getFirstOption();
        this.colorEnabled =
            Boolean.parseBoolean(getContext().getProperty(COLOR_PROPERTY))
            && getOptionList().size() > 1
            && COLOR_PARAM.equalsIgnoreCase(getOptionList().get(1));

        super.start();
    }

    @Override
    public String convert(ILoggingEvent event) {
        Deque<NDCElement> stack = NDC.get();
        if (stack.isEmpty()) {
            return "";
        }

        StringBuilder sb = new StringBuilder();

        for (Iterator<NDCElement> i = stack.descendingIterator(); i.hasNext();) {
            NDCElement next = i.next();
            sb.append(colorEnabled ? next.getColoredText() : next.getText());
            if (i.hasNext()) {
                sb.append(separator);
            }
        }
        String ndc = sb.toString();

        /*
         * We include the extra space and brackets to the NDC log string itself
         * so that log messages without an NDC do not have any extra whitespace.
         */

        return ndc.equals("") ? "" : String.format(" [%s]", ndc);
    }

}
