/*
 * $Id: //depot/prod/newport_news/main/modules/nn-platform/src/main/java/com/netapp/common/platform/linux/LinuxDeviceAnalyzer.java#3 $
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.execution.ShellExecutor;
import com.netapp.common.execution.ShellExecutorFactory;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.MsgKey;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.platform.PlatformException;

/**
 * The Linux DeviceAnalyzer is broken into two parts:
 * Subclasses of this class manage "raw" devices themselves. Older OSes should use the /dev/raw interface in
 * {@link RawDeviceAnalyzer}, while newer 2.6 kernels should use O_DIRECT managed by {@link DirectDeviceAnalyzer}.
 *
 * Block devices are managed by {@link BlockDeviceHandler}s that plugin to concrete instances of this class as
 * described above. Different distributions have different handlers for MPIO support.
 * @author sunday
 * @version $Revision: #3 $ $Date: 2011/06/24 $
 */
public abstract class LinuxDeviceAnalyzer implements DeviceAnalyzer {
    private static final NALogger logger = NALogger.getLogger(LinuxDeviceAnalyzer.class);
    protected transient OperatingSystem operatingSystem;
    protected transient ShellExecutor shellExecutor;
    protected transient ProcDevicesFile deviceMap;
    protected transient Map<String, BlockDeviceHandler> handlers = new HashMap<String, BlockDeviceHandler>();
    private static final Pattern PARTITION_DIGIT_PATTERN = Pattern.compile("\\D*(\\d+)");

    LinuxDeviceAnalyzer(OperatingSystem operatingSystem) {
        this.operatingSystem = operatingSystem;
    }

    void addBlockDeviceHandler(BlockDeviceHandler handler) {
        this.handlers.put(handler.driverName(), handler);
    }

    @Override
    public DeviceType determineDeviceType(File file) {
        for (BlockDeviceHandler handler : this.handlers.values()) {
            DeviceType result = handler.determineDeviceType(file);
            if (result != null && result != DeviceType.NON_DEVICE) {
                return result;
            }
        }
        return DeviceType.NON_DEVICE;
    }

    @Override
    public File deviceForPartition(File disk, int partitionNumber) {
        String partition = Integer.toString(partitionNumber);
        return this.deviceForPartition(disk, partition);
    }

    @Override
    public File deviceForPartition(File disk, String partition) {
        // The partition for any device created by *DeviceHandler classes should always only
        // be the numeric portion. Upgrades from SMO 2.2 though may be 'pX'.
        String canonicalPartition = canonicalPartitionNumber(partition);
        for (BlockDeviceHandler handler : this.handlers.values()) {
            File result = handler.deviceForPartition(disk, canonicalPartition);
            if (result != null) {
                return result;
            }
        }
        throw new PlatformException(logger.error(MsgKey.PLATFORM_DEVICE_ANALYZER_UNABLE_TO_DETERMINE_PARTITION_DEVICE, disk, partition));
    }

    @Override
    public String extractFullDiskPathElement(File file) {
        for (BlockDeviceHandler handler : this.handlers.values()) {
            String result = handler.extractFullDiskPathElement(file);
            if (result != null) {
                return result;
            }
        }
        throw new PlatformException(logger.error(MsgKey.PLATFORM_DEVICE_ANALYZER_UNABLE_TO_DETERMINE_BASE_DEVICE_FROM_PARTITION, file));
    }

    @Override
    public int extractPartitionNumber(File file) {
        return Integer.parseInt(extractPartitionPathElement(file));
    }

    @Override
    public String extractPartitionPathElement(File file) {
        for (BlockDeviceHandler handler : this.handlers.values()) {
            String result = handler.extractPartitionPathElement(file);
            if (result != null) {
                return result;
            }
        }
        throw new PlatformException(logger.error(MsgKey.PLATFORM_DEVICE_ANALYZER_UNABLE_TO_DETERMINE_PARTITION_NUMBER_FROM_PARTITION_DEVICE, file));
    }

    /**
     * Determine the canonical device node for a given major minor number.
     * @param major
     * @param minor
     * @return Device for the corresponding major/minor number
     */
    public File deviceFromMajorMinor(int major, int minor) {
        if (this.deviceMap == null) {
            this.deviceMap = new ProcDevicesFile(ProcDevicesFile.SYSTEM_DEVICES_FILE);
        }
        String driver = deviceMap.driverFromBlockMajor(major);

        if (driver != null) {
            BlockDeviceHandler handler = this.handlers.get(driver);
            if (handler != null) {
                File result = handler.deviceFromMajorMinor(major, minor);
                if (result != null) {
                    return result;
                }
            }
        }

        throw new PlatformException(logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_INVALID_DEVICE_MAJOR,
            "block", (driver != null) ? driver : "none", major, minor));
    }

    static String canonicalPartitionNumber(String partitionElement) {
        Matcher matcher = PARTITION_DIGIT_PATTERN.matcher(partitionElement);
        if (matcher.matches()) {
            return matcher.group(1);
        }
        return null;
    }

    protected synchronized ShellExecutor getShellExecutor() {
        if (this.shellExecutor == null) {
            if (this.operatingSystem != null) {
                this.shellExecutor = ShellExecutorFactory.createShellExecutor();
            } else {
                throw new RuntimeException(
                    logger.error(MsgKey.PLATFORM_NO_IMPLEMENTATION));
            }
        }
        return this.shellExecutor;
    }
}
