/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/util/Cache.java#4 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.lang.management.ManagementFactory;
import java.lang.ref.Reference;
import java.lang.ref.SoftReference;
import java.lang.ref.WeakReference;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicLong;

import javax.management.MBeanServer;
import javax.management.ObjectName;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Implement a soft reference cache. References that are expired are removed
 * via a scavenge thread shared among all instantiated Caches.
 * @author sunday
 * @version $Revision: #4 $ $Date: 2012/11/13 $
 */
public class Cache<K, V> implements CacheMBean {
    private static final String[] STRING_CAST = new String[0];
    private static final NALogger logger = NALogger.getLogger(Cache.class);
    private static final int SCAVENGE_SECONDS = (60 * 5);
    private static final ScavengerThread SCAVENGER;
    static {
        SCAVENGER = new ScavengerThread();
        SCAVENGER.setDaemon(true);
        SCAVENGER.start();
    }

    private final String name;
    private final Map<K, Reference<V>> cache;
    private volatile Date lastScavengeTime;
    private volatile long hitCount;
    private volatile long missCount;
    private volatile AtomicLong lifetimeCount = new AtomicLong(0);

    public Cache() {
        this(null);
    }
    /**
     * Construct a new cache with a specified name.
     * @param name The complete JMX name of this Cache: <samp>com.netapp.common.util:type=Cache,name=Foo</samp>
     */
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("NN")
    public Cache(String name) {
        this.name = name;
        this.cache = new HashMap<K, Reference<V>>();
        SCAVENGER.add(this);
        // Wake up the scavenger thread to get registered
        synchronized (SCAVENGER) {
            SCAVENGER.notify();
        }
    }

    /**
     * Insert a new value into the cache
     * @param key
     * @param value
     */
    public synchronized void put(K key, V value) {
        Reference<V> ref = this.cache.get(key);
        // Reference equality is the correct operation here, not equals()
        if (ref == null || ref.get() != value) {
            this.cache.put(key, new SoftReference<V>(value));
            this.lifetimeCount.incrementAndGet();
        }
    }

    /**
     * Retrieve a value from the cache
     * @param key
     * @return The value specified by key, or null if the value does not exist in this cache
     */
    public V get(Object key) {
        Reference<V> ref = this.cache.get(key);
        V value = null;
        if (ref != null) {
            value = ref.get();
        }
        hitCount += (value != null) ? 1 : 0;
        missCount += (value == null) ? 1 : 0;
        return value;
    }

    /**
     * Removes the entry for this key from this cache if it is present.
     * Returns the value to which the cache previously associated the key,
     * or null if the cache contained no entry for this key, or if the entry
     * for this key has been garbage collected.
     *
     * @param key key whose entry is to be removed from the cache.
     * @return previous value associated with specified key, or null if there was no mapping for key,
     * or if the entry for this key has been garbage collected.
     */
    public V remove(Object key) {
        Reference<V> ref = this.cache.remove(key);
        V value = null;
        if (ref != null) {
            value = ref.get();
        }
        return value;
    }

    /**
     * Return the size of this cache. This may include expired references that have not yet been scavenged.
     * @return size of the cache
     */
    public int size() {
        return this.cache.size();
    }

    /**
     * Scavenge the cache, removing any references that have been garbage collected.
     */
    @Override
    public synchronized void scavenge() {
        int removed = 0;
        Iterator<Reference<V>> iter = this.cache.values().iterator();
        while (iter.hasNext()) {
            Reference<V> ref = iter.next();
            if (ref.get() == null) {
                iter.remove();
                removed++;
            }
        }
        if (removed > 0) {
            logger.debug(MsgKey.UTIL_CACHE_SCAVENGE_COUNT, removed, this);
        }
        this.lastScavengeTime = new Date();
    }

    /**
     * Return the size of this cache. This may include expired references that havenot yet been scavenged.
     * @return size of the cache
     */
    @Override
    public int getCurrentCount() {
        return this.cache.size();
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#getLifetimeCount()
     */
    @Override
    public long getLifetimeCount() {
        return this.lifetimeCount.get();
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#getHitCount()
     */
    @Override
    public long getHitCount() {
        return this.hitCount;
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#getMissCount()
     */
    @Override
    public long getMissCount() {
        return this.missCount;
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#getLastScavengeTime()
     */
    @Override
    @edu.umd.cs.findbugs.annotations.SuppressWarnings("EI")
    public Date getLastScavengeTime() {
        return this.lastScavengeTime;
    }

    /**
     * Eject all items from the cache.
     */
    @Override
    public synchronized void clear() {
        this.cache.clear();
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#resetCounters()
     */
    @Override
    public void resetCounters() {
        this.lifetimeCount.set(0);
        this.hitCount = 0;
        this.missCount = 0;
    }

    /**
     *
     * @see com.netapp.common.util.CacheMBean#getKeys()
     */
    @Override
    public String[] getKeys() {
        Set<String> set = new HashSet<String>();

        for (Entry<K, Reference<V>> entry : this.cache.entrySet()) {
            if (entry.getValue().get() != null) {
                set.add(entry.getKey().toString());
            }
        }
        return set.toArray(STRING_CAST);
    }

    /**
     * Register the cache's MBean with JMX
     *
     */
    private void jmxRegister() {
        if (this.name == null) {
            return;
        }
        try {
            logger.debug(MsgKey.UTIL_CACHE_JMX_STARTING, this);
            MBeanServer mbeanServer;
            mbeanServer = ManagementFactory.getPlatformMBeanServer();
            ObjectName beanName = new ObjectName(this.name);

            mbeanServer.registerMBean(this, beanName);
        } catch (Exception ex) {
            logger.debug(MsgKey.UTIL_CACHE_JMX_START_EXCEPTION, ex, this.name, ex.getLocalizedMessage());
        }
    }

    @Override
    public String toString() {
        return this.name != null ? this.name : "Unknown";
    }

    protected static class ScavengerThread extends Thread {
        /**
         * Queue for new threads. We add them to this and then add
         * them to our working list inside our working thread so new
         * caches don't deadlock.
         */
        private final Queue<Cache<?, ?>> cacheQueue;
        private final List<Reference<Cache<?, ?>>> cacheList;
        private ScavengerThread() {
            super("CacheScavenger");
            cacheQueue = new ConcurrentLinkedQueue<Cache<?, ?>>();
            cacheList = new LinkedList<Reference<Cache<?, ?>>>();
        }

        private void add(Cache<?, ?> cache) {
            this.cacheQueue.add(cache);
        }

        @Override
        public void run() {
            while (true) {
                doRun();
                synchronized (this) {
                    try {
                        this.wait(SCAVENGE_SECONDS * 1000);
                    } catch (InterruptedException e) { }
                }
            }
        }

        private void doRun() {
            logger.debug(MsgKey.UTIL_CACHE_BEGIN_SCAVENGE);
            long time = System.currentTimeMillis();
            int scavengeCount = 0;
            Cache<?, ?> cache = cacheQueue.poll();
            while (cache != null) {
                if (cache.name != null) {
                    cache.jmxRegister();
                }
                this.cacheList.add(new WeakReference<Cache<?, ?>>(cache));
                cache = cacheQueue.poll();
            }

            Iterator<Reference<Cache<?, ?>>> iter = this.cacheList.iterator();
            while (iter.hasNext()) {
                Reference<Cache<?, ?>> ref = iter.next();
                cache = ref.get();
                if (cache == null) {
                    iter.remove();
                } else {
                    cache.scavenge();
                    scavengeCount++;
                }
            }
            time = System.currentTimeMillis() - time;
            logger.debug(MsgKey.UTIL_CACHE_END_SCAVENGE, scavengeCount, time);
        }
    }
}
