/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/xml/JaxpContext.java#3 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.xml;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.transform.TransformerFactory;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Throwables;
import net.jcip.annotations.Immutable;

/**
 * Holds JAXP-related constants and factory objects.
 *
 * Non threadsafe JAXP objects are held as {@link ThreadLocal}s.
 *
 * <p/>
 * See <a href="http://en.wikipedia.org/wiki/Java_API_for_XML_Processing">this link</a> for
 * a general explanation of the JAXP set of APIs.
 *
 * @author sirianni
 * @version $Revision: #3 $ $Date: 2012/11/05 $
 */
public class JaxpContext {

    private static Stax stax = new Stax();

    private JaxpContext() { }

    public static Stax stax() {
        return stax;
    }

    @Immutable
    public static class Stax {

        private final ThreadLocal<XMLInputFactory> xmlInputFactory = new ThreadLocal<XMLInputFactory>() {
            @Override
            protected XMLInputFactory initialValue() {
                XMLInputFactory f = XMLInputFactory.newInstance();
                f.setProperty(XMLInputFactory.SUPPORT_DTD, false);
                f.setProperty(XMLInputFactory.IS_COALESCING, true);
                return f;
            }
        };

        private final ThreadLocal<XMLOutputFactory> xmlOutputFactory = new ThreadLocal<XMLOutputFactory>() {
            @Override
            protected XMLOutputFactory initialValue() {
                return XMLOutputFactory.newInstance();
            }
        };

        private final ThreadLocal<XMLEventFactory> xmlEventFactory = new ThreadLocal<XMLEventFactory>() {
            @Override
            protected XMLEventFactory initialValue() {
                return XMLEventFactory.newInstance();
            }
        };

        private final StaxImplementation implementation = detectStaxImplementation();

        private Stax() { }

        /**
         * This {@link XMLInputFactory} is set with the following properties:
         * <ul>
         * <li>{@link XMLInputFactory#SUPPORT_DTD} = <tt>false</tt>
         * <li>{@link XMLInputFactory#IS_COALESCING} = <tt>true</tt>
         * </ul>
         */
        public XMLInputFactory xmlInputFactory() {
            return this.xmlInputFactory.get();
        }

        public XMLOutputFactory xmlOutputFactory() {
            return this.xmlOutputFactory.get();
        }

        public XMLEventFactory xmlEventFactory() {
            return this.xmlEventFactory.get();
        }

        public StaxImplementation getImplementation() {
            return this.implementation;
        }

        public enum StaxImplementation {

            /**
             * The <a href="http://woodstox.codehaus.org/">Woodstock StAX parser</a>.
             */
            WOODSTOX {
                @Override
                boolean isInstance(XMLInputFactory xmlInputFactory) {
                    try {
                        return Class.forName("com.ctc.wstx.stax.WstxInputFactory", true, Thread.currentThread().getContextClassLoader())
                            .isInstance(xmlInputFactory);
                    } catch (ClassNotFoundException e) {
                        return false;
                    }
                }

                @Override
                void setSystemProperties() {
                    System.setProperty("javax.xml.stream.XMLInputFactory", "com.ctc.wstx.stax.WstxInputFactory");
                    System.setProperty("javax.xml.stream.XMLOutputFactory", "com.ctc.wstx.stax.WstxOutputFactory");
                    System.setProperty("javax.xml.stream.XMLEventFactory", "com.ctc.wstx.stax.WstxEventFactory");
                }
            },

            /**
             * Sun Java Streaming XML Parser.
             * The StAX implementation built-in to Java 6.
             */
            SJSXP {
                @Override
                boolean isInstance(XMLInputFactory xmlInputFactory) {
                    try {
                        /*
                         * The javac compiler does not allow an import of this class
                         * So we have to Class.forName() it at runtime.
                         * See http://blogs.sun.com/ahe/entry/b86_compiler_fixes for more info.
                         */
                        return Class.forName("com.sun.xml.internal.stream.XMLInputFactoryImpl", true, Thread.currentThread().getContextClassLoader())
                            .isInstance(xmlInputFactory);
                    } catch (ClassNotFoundException e) {
                        return false;
                    }
                }

                @Override
                void setSystemProperties() {
                    System.setProperty("javax.xml.stream.XMLInputFactory", "com.sun.xml.internal.stream.XMLInputFactoryImpl");
                    System.setProperty("javax.xml.stream.XMLOutputFactory", "com.sun.xml.internal.stream.XMLOutputFactoryImpl");
                    System.setProperty("javax.xml.stream.XMLEventFactory", "com.sun.xml.internal.stream.events.XMLEventFactoryImpl");
                }
            };

            abstract boolean isInstance(XMLInputFactory xmlInputFactory);
            abstract void setSystemProperties();
        }

        private static StaxImplementation detectStaxImplementation() {
            XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();
            for (StaxImplementation staxImplementation : StaxImplementation.values()) {
                if (staxImplementation.isInstance(xmlInputFactory)) {
                    return staxImplementation;
                }
            }

            throw new IllegalStateException("Unable to determine StAX implementation: " + xmlInputFactory.getClass());
        }

        /**
         * Set the <b>global</b> StAX implementation for this JVM.
         *
         * This will change the StAX factories returned from these methods:
         * <ul>
         * <li>{@link XMLInputFactory#newFactory()}
         * <li>{@link XMLOutputFactory#newFactory()}
         * <li>{@link XMLEventFactory#newFactory()}
         * </ul>
         *
         * It is <b>highly inadvisable</b> to call this method anywhere <i>except</i> JUnit tests.
         */
        @VisibleForTesting
        public static synchronized void setImplementation(StaxImplementation staxImplementation) {
            staxImplementation.setSystemProperties();
            stax = new Stax();
        }

        /**
         * Reset the StAX implementation to the classpath-dependent implementation as returned by:
         * <ul>
         * <li>{@link XMLInputFactory#newFactory()}
         * <li>{@link XMLOutputFactory#newFactory()}
         * <li>{@link XMLEventFactory#newFactory()}
         * </ul>
         *
         * It is <b>highly inadvisable</b> to call this method anywhere <i>except</i> JUnit tests.
         */
        @VisibleForTesting
        public static synchronized void resetImplementation() {
            System.getProperties().remove("javax.xml.stream.XMLInputFactory");
            System.getProperties().remove("javax.xml.stream.XMLOutputFactory");
            System.getProperties().remove("javax.xml.stream.XMLEventFactory");
            stax = new Stax();
        }

    }

    //CHECKSTYLE:OFF DeclarationOrder
    private static Dom dom = new Dom();

    public static Dom dom() {
        return dom;
    }
    //CHECKSTYLE:ON DeclarationOrder

    public static class Dom {

        private final ThreadLocal<DocumentBuilder> documentBuilder = new ThreadLocal<DocumentBuilder>() {
            @Override
            protected DocumentBuilder initialValue() {
                try {
                    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                    factory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);

                    return factory.newDocumentBuilder();
                } catch (ParserConfigurationException e) {
                    throw Throwables.propagate(e);
                }
            }
        };

        private Dom() { }

        /**
         * Get a thread-local {@link DocumentBuilder}.
         * The returned <tt>DocumentBuilder</tt> has external DTD loading disabled.
         */
        public DocumentBuilder documentBuilder() {
            return this.documentBuilder.get();
        }

    }

    //CHECKSTYLE:OFF DeclarationOrder
    private static Transform transform = new Transform();

    public static Transform transform() {
        return transform;
    }
    //CHECKSTYLE:ON DeclarationOrder

    public static class Transform {

        private final ThreadLocal<TransformerFactory> transformerFactory = new ThreadLocal<TransformerFactory>() {
            @Override
            protected TransformerFactory initialValue() {
                return TransformerFactory.newInstance();
            }
        };

        private Transform() { }

        public TransformerFactory transformerFactory() {
            return this.transformerFactory.get();
        }

    }

}
