package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.ControlAccessParamsType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.OrgType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Organization extends VcloudEntity<OrgType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private Collection<ReferenceType> catalogRefs;
  private HashMap<String, ReferenceType> vdcRefsByName;
  private HashMap<String, ReferenceType> networkRefsByName;
  private ReferenceType tasksListRef;

  Organization(VcloudClient client, OrgType org_v1_5)
  {
    super(client, org_v1_5);
    sortRefs_v1_5();
  }

  public static Organization getOrganizationByReference(VcloudClient client, ReferenceType orgRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + orgRef.getHref());

    return new Organization(client, (OrgType)getResourceByReference(client, orgRef));
  }

  public static Organization getOrganizationById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    return new Organization(client, (OrgType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.org+xml"));
  }

  private static String getId(String href)
  {
    int i = href.lastIndexOf("/");
    return href.substring(i + 1);
  }

  public ControlAccessParamsType getCatalogControlAccessByReference(ReferenceType catalogRef)
    throws VCloudException
  {
    String url = getReference().getHref() + "/catalog/" + getId(catalogRef.getHref()) + "/controlAccess/";

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + url);

    return (ControlAccessParamsType)SdkUtil.get(getVcloudClient(), url, 200);
  }

  public ControlAccessParamsType updateCatalogControlAccessByReference(ReferenceType catalogRef, ControlAccessParamsType controlAccessParams)
    throws VCloudException
  {
    String url = getReference().getHref() + "/catalog/" + getId(catalogRef.getHref()) + "/action/controlAccess";

    ObjectFactory vcloudObjectFactory = new ObjectFactory();
    String requestString = JAXBUtil.marshal(vcloudObjectFactory.createControlAccessParams(controlAccessParams));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.POST_URL_MSG) + " - " + url);

    return (ControlAccessParamsType)SdkUtil.post(getVcloudClient(), url, requestString, "application/vnd.vmware.vcloud.controlAccess+xml", 200);
  }

  private void sortRefs_v1_5()
  {
    this.catalogRefs = new ArrayList();
    this.vdcRefsByName = new HashMap();
    this.networkRefsByName = new HashMap();
    this.tasksListRef = new ReferenceType();

    if (((OrgType)getResource()).getLink() != null) {
      List links = ((OrgType)getResource()).getLink();

      for (LinkType link : links) {
        ReferenceType sdkReference = link;
        if (link.getType().equals("application/vnd.vmware.vcloud.catalog+xml"))
          this.catalogRefs.add(sdkReference);
        else if (link.getType().equals("application/vnd.vmware.vcloud.vdc+xml"))
          this.vdcRefsByName.put(sdkReference.getName(), sdkReference);
        else if (link.getType().equals("application/vnd.vmware.vcloud.orgNetwork+xml"))
        {
          this.networkRefsByName.put(sdkReference.getName(), sdkReference);
        } else if (link.getType().equals("application/vnd.vmware.vcloud.tasksList+xml"))
        {
          this.tasksListRef = sdkReference;
        }
        else logger.log(Level.WARNING, SdkUtil.getI18nString(SdkMessage.UNKNOWN_REF_TYPE_MSG) + " - " + link.getType());
      }
    }
  }

  public ReferenceType getTasksListRef()
  {
    return this.tasksListRef;
  }

  public HashMap<String, ReferenceType> getVdcRefsByName()
  {
    return this.vdcRefsByName;
  }

  public HashMap<String, ReferenceType> getNetworkRefsByName()
  {
    return this.networkRefsByName;
  }

  public Collection<ReferenceType> getVdcRefs()
  {
    return this.vdcRefsByName.values();
  }

  public ReferenceType getVdcRefByName(String name)
  {
    return (ReferenceType)this.vdcRefsByName.get(name);
  }

  public Collection<ReferenceType> getCatalogRefs()
  {
    return this.catalogRefs;
  }

  public Collection<ReferenceType> getNetworkRefs()
  {
    return this.networkRefsByName.values();
  }

  public ReferenceType getNetworkRefByName(String name)
  {
    return (ReferenceType)this.networkRefsByName.get(name);
  }

  public Task createTask(TaskType taskType)
    throws VCloudException
  {
    ObjectFactory vcloudObjectFactory = new ObjectFactory();
    String requestString = JAXBUtil.marshal(vcloudObjectFactory.createTask(taskType));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.POST_URL_MSG) + " - " + getTasksListRef().getHref());

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), getTasksListRef().getHref(), requestString, "application/vnd.vmware.vcloud.task+xml", 200));
  }
}