package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.ControlAccessParamsType;
import com.vmware.vcloud.api.rest.schema.LeaseSettingsSectionType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.NetworkConfigSectionType;
import com.vmware.vcloud.api.rest.schema.OwnerType;
import com.vmware.vcloud.api.rest.schema.RecomposeVAppParamsType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.api.rest.schema.VAppChildrenType;
import com.vmware.vcloud.api.rest.schema.VAppNetworkConfigurationType;
import com.vmware.vcloud.api.rest.schema.VAppType;
import com.vmware.vcloud.api.rest.schema.VCloudExtensionType;
import com.vmware.vcloud.api.rest.schema.VmType;
import com.vmware.vcloud.api.rest.schema.extension.VimObjectRefType;
import com.vmware.vcloud.api.rest.schema.ovf.EnvelopeType;
import com.vmware.vcloud.api.rest.schema.ovf.NetworkSectionNetwork;
import com.vmware.vcloud.api.rest.schema.ovf.NetworkSectionType;
import com.vmware.vcloud.api.rest.schema.ovf.SectionType;
import com.vmware.vcloud.api.rest.schema.ovf.StartupSectionType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.constants.VappStatus;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class Vapp extends AbstractVapp<VAppType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private StartupSectionType startupSection;
  private NetworkSectionType networkSection;
  private NetworkConfigSectionType networkConfigSection;
  private LeaseSettingsSectionType leaseSettingsSection;
  private List<Vapp> childrenvApps;
  private List<VM> childrenVms;
  private ReferenceType vdcReference;
  private HashMap<String, ReferenceType> vappNetworkRefsByName;

  public Vapp(VcloudClient client, VAppType vAppType)
  {
    super(client, vAppType);
    sortOvfSectionsAndReferences();
  }

  public static Vapp getVappByReference(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + vappRef.getHref());

    VAppType vAppType = (VAppType)getResourceByReference(client, vappRef);
    Vapp vapp = new Vapp(client, vAppType);
    return vapp;
  }

  public static Vapp getVappById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    VAppType vAppType = (VAppType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.vApp+xml");

    Vapp vapp = new Vapp(client, vAppType);
    return vapp;
  }

  public Task updateVapp(VAppType vAppType)
    throws VCloudException
  {
    com.vmware.vcloud.api.rest.schema.ObjectFactory of = new com.vmware.vcloud.api.rest.schema.ObjectFactory();
    String vAppUpdateUrl = getReference().getHref();
    String vAppRequestXmlString = JAXBUtil.marshal(of.createVApp(vAppType));
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG) + " - " + vAppUpdateUrl);

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), vAppUpdateUrl, vAppRequestXmlString, "application/vnd.vmware.vcloud.vApp+xml", 202));
  }

  public VappStatus getVappStatus()
  {
    return VappStatus.fromValue(((VAppType)getResource()).getStatus().intValue());
  }

  public List<Vapp> getChildrenVapps()
    throws VCloudException
  {
    if (this.childrenvApps == null) {
      this.childrenvApps = new ArrayList();
      if (((VAppType)getResource()).getChildren() != null) {
        VAppChildrenType vAppChildrenType = ((VAppType)getResource()).getChildren();

        List vAppTypeList = vAppChildrenType.getVApp();
        for (VAppType vAppType : vAppTypeList) {
          this.childrenvApps.add(new Vapp(getVcloudClient(), vAppType));
        }
      }
    }

    return this.childrenvApps;
  }

  public List<VM> getChildrenVms()
    throws VCloudException
  {
    if (this.childrenVms == null) {
      this.childrenVms = new ArrayList();
      if (((VAppType)getResource()).getChildren() != null) {
        VAppChildrenType childrenType = ((VAppType)getResource()).getChildren();

        List vmTypeList = childrenType.getVm();
        for (VmType vmType : vmTypeList) {
          this.childrenVms.add(new VM(getVcloudClient(), vmType));
        }
      }
    }
    return this.childrenVms;
  }

  public LeaseSettingsSectionType getLeaseSettingsSection()
    throws VCloudException
  {
    if (this.leaseSettingsSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    return this.leaseSettingsSection;
  }

  public static LeaseSettingsSectionType getLeaseSettingsSection(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return (LeaseSettingsSectionType)SdkUtil.get(client, vappRef.getHref() + "/leaseSettingsSection/", 200);
  }

  public StartupSectionType getStartUpSection()
    throws VCloudException
  {
    if (this.startupSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    return this.startupSection;
  }

  public static StartupSectionType getStartUpSection(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return (StartupSectionType)SdkUtil.get(client, vappRef.getHref() + "/startupSection/", 200);
  }

  public NetworkSectionType getNetworkSection()
    throws VCloudException
  {
    if (this.networkSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    return this.networkSection;
  }

  public static NetworkSectionType getNetworkSection(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return (NetworkSectionType)SdkUtil.get(client, vappRef.getHref() + "/networkSection/", 200);
  }

  public HashMap<String, NetworkSectionNetwork> getNetworksByName()
    throws VCloudException
  {
    HashMap networks = new HashMap();
    if (this.networkSection != null) {
      for (NetworkSectionNetwork network : this.networkSection.getNetwork())
        networks.put(network.getName(), network);
    }
    return networks;
  }

  public NetworkSectionNetwork getNetworkByName(String networkName)
    throws VCloudException
  {
    return (NetworkSectionNetwork)getNetworksByName().get(networkName);
  }

  public Set<String> getNetworkNames()
    throws VCloudException
  {
    return getNetworksByName().keySet();
  }

  public Collection<NetworkSectionNetwork> getNetworks()
    throws VCloudException
  {
    return getNetworksByName().values();
  }

  public NetworkConfigSectionType getNetworkConfigSection()
    throws VCloudException
  {
    if (this.networkConfigSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    return this.networkConfigSection;
  }

  public static NetworkConfigSectionType getNetworkConfigSection(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return (NetworkConfigSectionType)SdkUtil.get(client, vappRef.getHref() + "/networkConfigSection/", 200);
  }

  public HashMap<String, ReferenceType> getVappNetworkRefsByName()
  {
    return this.vappNetworkRefsByName;
  }

  public HashMap<String, VAppNetworkConfigurationType> getVappNetworkConfigurationsByName()
    throws VCloudException
  {
    if (this.networkConfigSection == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
    }

    HashMap networks = new HashMap();
    List netConnections = this.networkConfigSection.getNetworkConfig();

    Iterator it = netConnections.iterator();
    while (it.hasNext()) {
      VAppNetworkConfigurationType conn = (VAppNetworkConfigurationType)it.next();
      networks.put(conn.getNetworkName(), conn);
    }

    return networks;
  }

  public VAppNetworkConfigurationType getVappNetworkConfigurationByName(String vappNetworkConfigurationName)
    throws VCloudException
  {
    return (VAppNetworkConfigurationType)getVappNetworkConfigurationsByName().get(vappNetworkConfigurationName);
  }

  public Set<String> getVappNetworkConfigurationNames()
    throws VCloudException
  {
    return getVappNetworkConfigurationsByName().keySet();
  }

  public Collection<VAppNetworkConfigurationType> getVappNetworkConfigurations()
    throws VCloudException
  {
    return getVappNetworkConfigurationsByName().values();
  }

  public Task updateSection(SectionType sectionType)
    throws VCloudException
  {
    com.vmware.vcloud.api.rest.schema.ovf.ObjectFactory ovfObjectFactory = new com.vmware.vcloud.api.rest.schema.ovf.ObjectFactory();
    com.vmware.vcloud.api.rest.schema.ObjectFactory vCloudObjectFactory_v1_5 = new com.vmware.vcloud.api.rest.schema.ObjectFactory();
    String sectionUrl = getReference().getHref();
    String sectionContentType = null;
    String sectionRequestString = null;
    if ((sectionType instanceof LeaseSettingsSectionType)) {
      JAXBElement section = vCloudObjectFactory_v1_5.createLeaseSettingsSection((LeaseSettingsSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/leaseSettingsSection/";

      sectionContentType = "application/vnd.vmware.vcloud.leaseSettingsSection+xml";
    } else if ((sectionType instanceof StartupSectionType)) {
      JAXBElement section = ovfObjectFactory.createStartupSection((StartupSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/startupSection/";
      sectionContentType = "application/vnd.vmware.vcloud.startupSection+xml";
    } else if ((sectionType instanceof NetworkConfigSectionType)) {
      JAXBElement section = vCloudObjectFactory_v1_5.createNetworkConfigSection((NetworkConfigSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/networkConfigSection/";

      sectionContentType = "application/vnd.vmware.vcloud.networkConfigSection+xml";
    } else {
      logger.log(Level.INFO, sectionType.getClass().getName() + " - " + SdkUtil.getI18nString(SdkMessage.NOT_VALID_SECTION_MSG));

      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NOT_VALID_SECTION_MSG));
    }

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), sectionUrl, sectionRequestString, sectionContentType, 202));
  }

  public ControlAccessParamsType getControlAccess()
    throws VCloudException
  {
    String url = getReference().getHref() + "/controlAccess/";

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + url);

    return (ControlAccessParamsType)SdkUtil.get(getVcloudClient(), url, 200);
  }

  public ControlAccessParamsType updateControlAccess(ControlAccessParamsType controlAccessParams)
    throws VCloudException
  {
    String url = getReference().getHref() + "/action/controlAccess";

    com.vmware.vcloud.api.rest.schema.ObjectFactory vcloudObjectFactory = new com.vmware.vcloud.api.rest.schema.ObjectFactory();
    String requestString = JAXBUtil.marshal(vcloudObjectFactory.createControlAccessParams(controlAccessParams));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.POST_URL_MSG) + " - " + url);

    return (ControlAccessParamsType)SdkUtil.post(getVcloudClient(), url, requestString, "application/vnd.vmware.vcloud.controlAccess+xml", 200);
  }

  public BigInteger getVappSize()
    throws VCloudException
  {
    BigInteger size = BigInteger.ZERO;
    for (VM childVm : getChildrenVms()) {
      size = size.add(childVm.getVmSize());
    }
    return size;
  }

  private void sortOvfSectionsAndReferences()
  {
    this.vappNetworkRefsByName = new HashMap();
    List sections = ((VAppType)getResource()).getSection();

    Iterator it = sections.iterator();
    while (it.hasNext()) {
      JAXBElement sec = (JAXBElement)it.next();
      if ((sec.getValue() instanceof NetworkConfigSectionType)) {
        this.networkConfigSection = ((NetworkConfigSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof NetworkSectionType))
        this.networkSection = ((NetworkSectionType)sec.getValue());
      else if ((sec.getValue() instanceof LeaseSettingsSectionType)) {
        this.leaseSettingsSection = ((LeaseSettingsSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof StartupSectionType))
        this.startupSection = ((StartupSectionType)sec.getValue());
      logger.log(Level.INFO, ((SectionType)sec.getValue()).getClass().getName());
    }

    for (LinkType vAppLink : ((VAppType)getResource()).getLink())
      if ((vAppLink.getRel() != null) && (vAppLink.getType() != null))
        if ((vAppLink.getRel().equals("up")) && (vAppLink.getType().equals("application/vnd.vmware.vcloud.vdc+xml")))
        {
          this.vdcReference = vAppLink;
        } else if ((vAppLink.getRel().equals("down")) && (vAppLink.getType().equals("application/vnd.vmware.vcloud.vAppNetwork+xml")))
        {
          this.vappNetworkRefsByName.put(vAppLink.getName(), vAppLink);
        }
  }

  public ReferenceType getVdcReference()
    throws VCloudException
  {
    if (this.vdcReference != null)
      return this.vdcReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public VimObjectRefType getNetworkVimRef(String vAppNetworkName)
    throws VCloudException
  {
    VAppNetworkConfigurationType vappNetworkConfig = getVappNetworkConfigurationByName(vAppNetworkName);

    if (vappNetworkConfig != null)
      for (VCloudExtensionType vappExtension : vappNetworkConfig.getVCloudExtension())
      {
        for (i$ = vappExtension.getAny().iterator(); i$.hasNext(); ) { Object jaxb = i$.next();
          if ((((JAXBElement)jaxb).getValue() instanceof VimObjectRefType))
            return (VimObjectRefType)((JAXBElement)jaxb).getValue();
        }
      }
    Iterator i$;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.VIM_OBJECT_REF_NOT_FOUND_MSG));
  }

  public Task recomposeVapp(RecomposeVAppParamsType recomposeVappParamsType)
    throws VCloudException
  {
    com.vmware.vcloud.api.rest.schema.ObjectFactory of = new com.vmware.vcloud.api.rest.schema.ObjectFactory();

    JAXBElement composeVappParams = of.createRecomposeVAppParams(recomposeVappParamsType);

    String requestBodyXml = JAXBUtil.marshal(composeVappParams);

    String postURL = getReference().getHref() + "/action/recomposeVApp";

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), postURL, requestBodyXml, "application/vnd.vmware.vcloud.recomposeVAppParams+xml", 202));
  }

  public ReferenceType getOwner()
    throws VCloudException
  {
    return ((VAppType)getResource()).getOwner().getUser();
  }

  public static ReferenceType getOwner(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    String url = vappRef.getHref() + "/owner";
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + url);

    return ((OwnerType)SdkUtil.get(client, url, 200)).getUser();
  }

  public static void changeOwner(VcloudClient client, ReferenceType vappRef, ReferenceType userReference)
    throws VCloudException
  {
    String url = vappRef.getHref() + "/owner";
    changeOwner(url, userReference, client);
  }

  public void changeOwner(ReferenceType userReference)
    throws VCloudException
  {
    String url = getReference().getHref() + "/owner";
    changeOwner(url, userReference, getVcloudClient());
  }

  private static void changeOwner(String url, ReferenceType userReference, VcloudClient client)
    throws VCloudException
  {
    String requestXML = null;
    com.vmware.vcloud.api.rest.schema.ObjectFactory objectFactory = new com.vmware.vcloud.api.rest.schema.ObjectFactory();
    OwnerType ownerType = new OwnerType();
    ownerType.setUser(userReference);
    JAXBElement changeOwnerParams = objectFactory.createOwner(ownerType);

    requestXML = JAXBUtil.marshal(changeOwnerParams);
    SdkUtil.put(client, url, requestXML, "application/vnd.vmware.vcloud.owner+xml", 204);
  }

  public Task resetvAppNetwork(String vAppNetworkName)
    throws VCloudException
  {
    VAppNetworkConfigurationType vappNetworkConfig = getVappNetworkConfigurationByName(vAppNetworkName);

    for (LinkType vappNetworkConfigLink : vappNetworkConfig.getLink()) {
      if (vappNetworkConfigLink.getRel().equals("repair")) {
        return AbstractVapp.executeAbstractVappAction(getVcloudClient(), vappNetworkConfigLink.getHref(), null, null, 202);
      }

    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.RESOURCE_NOT_FOUND_MSG));
  }

  public Task syncSyslogServer(String vAppNetworkName)
    throws VCloudException
  {
    VAppNetworkConfigurationType vappNetworkConfig = getVappNetworkConfigurationByName(vAppNetworkName);

    for (LinkType vappNetworkConfigLink : vappNetworkConfig.getLink()) {
      if (vappNetworkConfigLink.getRel().equals("syncSyslogSettings"))
      {
        return AbstractVapp.executeAbstractVappAction(getVcloudClient(), vappNetworkConfigLink.getHref(), null, null, 202);
      }

    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.RESOURCE_NOT_FOUND_MSG));
  }

  public void enableMaintenance()
    throws VCloudException
  {
    String url = getReference().getHref() + "/action/enterMaintenanceMode";

    SdkUtil.post(getVcloudClient(), url, null, null, 204);
  }

  public void disableMaintenance()
    throws VCloudException
  {
    String url = getReference().getHref() + "/action/exitMaintenanceMode";

    SdkUtil.post(getVcloudClient(), url, null, null, 204);
  }

  public static void enableMaintenance(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    String url = vappRef.getHref() + "/action/enterMaintenanceMode";

    SdkUtil.post(client, url, null, null, 204);
  }

  public static void disableMaintenance(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    String url = vappRef.getHref() + "/action/exitMaintenanceMode";

    SdkUtil.post(client, url, null, null, 204);
  }

  public static EnvelopeType getOvf(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    return getOvf(client, vappRef.getHref() + "/ovf");
  }

  public EnvelopeType getOvf()
    throws VCloudException
  {
    return getOvf(getVcloudClient(), getReference().getHref() + "/ovf");
  }

  private static EnvelopeType getOvf(VcloudClient client, String vAppOvfUrl)
    throws VCloudException
  {
    return (EnvelopeType)SdkUtil.get(client, vAppOvfUrl, 200);
  }
}