package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.CustomizationSectionType;
import com.vmware.vcloud.api.rest.schema.FilesListType;
import com.vmware.vcloud.api.rest.schema.GuestCustomizationSectionType;
import com.vmware.vcloud.api.rest.schema.LeaseSettingsSectionType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.NetworkConfigSectionType;
import com.vmware.vcloud.api.rest.schema.NetworkConnectionSectionType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.OwnerType;
import com.vmware.vcloud.api.rest.schema.ProductSectionListType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.RelocateParamsType;
import com.vmware.vcloud.api.rest.schema.TaskType;
import com.vmware.vcloud.api.rest.schema.VAppTemplateChildrenType;
import com.vmware.vcloud.api.rest.schema.VAppTemplateType;
import com.vmware.vcloud.api.rest.schema.VCloudExtensionType;
import com.vmware.vcloud.api.rest.schema.extension.VimObjectRefType;
import com.vmware.vcloud.api.rest.schema.extension.VmVimInfoType;
import com.vmware.vcloud.api.rest.schema.ovf.EnvelopeType;
import com.vmware.vcloud.api.rest.schema.ovf.NetworkSectionType;
import com.vmware.vcloud.api.rest.schema.ovf.ProductSectionType;
import com.vmware.vcloud.api.rest.schema.ovf.RASDType;
import com.vmware.vcloud.api.rest.schema.ovf.ResourceType;
import com.vmware.vcloud.api.rest.schema.ovf.SectionType;
import com.vmware.vcloud.api.rest.schema.ovf.VirtualHardwareSectionType;
import com.vmware.vcloud.api.rest.schema.ovf.VirtualSystemCollectionType;
import com.vmware.vcloud.api.rest.schema.ovf.VirtualSystemType;
import com.vmware.vcloud.sdk.constants.EntityType;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.constants.VappTemplateStatus;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class VappTemplate extends VcloudEntity<VAppTemplateType>
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private NetworkSectionType networkSection;
  private List<ProductSectionType> productSections;
  private List<VappTemplate> childrenvAppTemplates;
  private HashMap<String, com.vmware.vcloud.api.rest.schema.FileType> uploadFileMap;
  private HashMap<String, com.vmware.vcloud.api.rest.schema.FileType> uploadedFileMap;
  private NetworkConnectionSectionType networkConnectionSection;
  private NetworkConfigSectionType networkConfigSection;
  private LeaseSettingsSectionType leaseSettingsSection;
  private CustomizationSectionType customizationSection;
  private GuestCustomizationSectionType guestCustomizationSection;
  private ReferenceType vdcReference;
  private ReferenceType catalogItemReference;
  private ReferenceType parentVappTemplateReference;
  private ReferenceType storageProfileRef;
  private VmVimInfoType vmVimInfoType;

  public VappTemplate(VcloudClient client, VAppTemplateType vAppTemplateType_v1_5)
  {
    super(client, vAppTemplateType_v1_5);
    sortVAppTemplateFiles_v1_5();
    sortOvfSectionAndReferences_v1_5();
  }

  public Boolean isVm()
  {
    return Boolean.valueOf(getReference().getType().equals("application/vnd.vmware.vcloud.vm+xml"));
  }

  public List<VappTemplate> getChildren()
  {
    if (this.childrenvAppTemplates == null) {
      this.childrenvAppTemplates = new ArrayList();
      if (((VAppTemplateType)getResource()).getChildren() != null) {
        VAppTemplateChildrenType vAppTemplChildrenType = ((VAppTemplateType)getResource()).getChildren();

        List vAppTypeList = vAppTemplChildrenType.getVm();

        for (VAppTemplateType vAppTemplType : vAppTypeList) {
          this.childrenvAppTemplates.add(new VappTemplate(getVcloudClient(), vAppTemplType));
        }
      }
    }
    return this.childrenvAppTemplates;
  }

  public static VappTemplate getVappTemplateByReference(VcloudClient client, ReferenceType vappTemplRef)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_URL_MSG) + " - " + vappTemplRef.getHref());

    return new VappTemplate(client, (VAppTemplateType)getResourceByReference(client, vappTemplRef));
  }

  public static VappTemplate getVappTemplateById(VcloudClient client, String vCloudId)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_VCLOUD_ID_MSG) + " - " + vCloudId);

    if (vCloudId.contains(EntityType.VM.value()))
    {
      return new VappTemplate(client, (VAppTemplateType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.vm+xml"));
    }

    return new VappTemplate(client, (VAppTemplateType)getEntityById(client, vCloudId, "application/vnd.vmware.vcloud.vAppTemplate+xml"));
  }

  public ReferenceType getVdcReference()
    throws VCloudException
  {
    if (this.vdcReference != null)
      return this.vdcReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getVdcStorageProfileReference()
    throws VCloudException
  {
    if (this.storageProfileRef != null)
      return this.storageProfileRef;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getCatalogItemReference()
    throws VCloudException
  {
    if (this.catalogItemReference != null)
      return this.catalogItemReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public ReferenceType getParentVappTemplateReference()
    throws VCloudException
  {
    if (this.parentVappTemplateReference != null)
      return this.parentVappTemplateReference;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
  }

  public VimObjectRefType getVMDatastoreVimRef()
    throws VCloudException
  {
    if ((this.vmVimInfoType != null) && (this.vmVimInfoType.getDatastoreVimObjectRef() != null))
    {
      return this.vmVimInfoType.getDatastoreVimObjectRef();
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.VIM_OBJECT_REF_NOT_FOUND_MSG));
  }

  public VimObjectRefType getVMHostVimRef()
    throws VCloudException
  {
    if ((this.vmVimInfoType != null) && (this.vmVimInfoType.getHostVimObjectRef() != null))
    {
      return this.vmVimInfoType.getHostVimObjectRef();
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.VIM_OBJECT_REF_NOT_FOUND_MSG));
  }

  public Integer getVMDiskChainLength()
    throws VCloudException
  {
    if (this.vmVimInfoType != null) {
      return Integer.valueOf(this.vmVimInfoType.getVirtualDisksMaxChainLength());
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.DATA_NOT_FOUND));
  }

  public VimObjectRefType getVMVimRef()
    throws VCloudException
  {
    if ((this.vmVimInfoType != null) && (this.vmVimInfoType.getVmVimObjectRef() != null)) {
      return this.vmVimInfoType.getVmVimObjectRef();
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.VIM_OBJECT_REF_NOT_FOUND_MSG));
  }

  private void sortOvfSectionAndReferences_v1_5()
  {
    this.productSections = new ArrayList();
    List sections = ((VAppTemplateType)getResource()).getSection();

    Iterator it = sections.iterator();
    while (it.hasNext()) {
      JAXBElement sec = (JAXBElement)it.next();
      if ((sec.getValue() instanceof NetworkConfigSectionType)) {
        this.networkConfigSection = ((NetworkConfigSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof NetworkSectionType))
        this.networkSection = ((NetworkSectionType)sec.getValue());
      else if ((sec.getValue() instanceof NetworkConnectionSectionType)) {
        this.networkConnectionSection = ((NetworkConnectionSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof LeaseSettingsSectionType)) {
        this.leaseSettingsSection = ((LeaseSettingsSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof CustomizationSectionType)) {
        this.customizationSection = ((CustomizationSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof GuestCustomizationSectionType)) {
        this.guestCustomizationSection = ((GuestCustomizationSectionType)sec.getValue());
      }
      else if ((sec.getValue() instanceof ProductSectionType))
        this.productSections.add((ProductSectionType)sec.getValue());
      logger.log(Level.INFO, ((SectionType)sec.getValue()).getClass().getName());
    }

    for (LinkType vAppTemplateLink : ((VAppTemplateType)getResource()).getLink()) {
      if ((vAppTemplateLink.getRel().equals("up")) && (vAppTemplateLink.getType().equals("application/vnd.vmware.vcloud.vAppTemplate+xml")))
      {
        this.parentVappTemplateReference = vAppTemplateLink;
      } else if ((vAppTemplateLink.getRel().equals("catalogItem")) && (vAppTemplateLink.getType().equals("application/vnd.vmware.vcloud.catalogItem+xml")))
      {
        this.catalogItemReference = vAppTemplateLink;
      } else if ((vAppTemplateLink.getRel().equals("up")) && (vAppTemplateLink.getType().equals("application/vnd.vmware.vcloud.vdc+xml")))
      {
        this.vdcReference = vAppTemplateLink;
      } else if ((vAppTemplateLink.getRel().equals("storageProfile")) && (vAppTemplateLink.getType().equals("application/vnd.vmware.vcloud.vdcStorageProfile+xml")))
      {
        this.storageProfileRef = vAppTemplateLink;
      }

    }

    for (VCloudExtensionType vcloudExtension : ((VAppTemplateType)getResource()).getVCloudExtension())
    {
      for (i$ = vcloudExtension.getAny().iterator(); i$.hasNext(); ) { Object jaxbObject = i$.next();
        if ((((JAXBElement)jaxbObject).getValue() instanceof VmVimInfoType)) {
          this.vmVimInfoType = ((VmVimInfoType)((JAXBElement)jaxbObject).getValue());

          break;
        }
      }
    }
    Iterator i$;
  }

  public ReferenceType getOwner()
    throws VCloudException
  {
    return ((VAppTemplateType)getResource()).getOwner().getUser();
  }

  public VappTemplateStatus getVappTemplateStatus()
  {
    return VappTemplateStatus.fromValue(((VAppTemplateType)getResource()).getStatus().intValue());
  }

  public static EnvelopeType getOvf(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return getOvf(client, vappTemplateRef.getHref() + "/ovf");
  }

  public EnvelopeType getOvf()
    throws VCloudException
  {
    return getOvf(getVcloudClient(), getReference().getHref() + "/ovf");
  }

  public static EnvelopeType getOvf(VcloudClient client, String vAppTemplateOvfUrl)
    throws VCloudException
  {
    return (EnvelopeType)SdkUtil.get(client, vAppTemplateOvfUrl, 200);
  }

  private EnvelopeType downloadOvfEnvelope()
    throws VCloudException
  {
    for (LinkType link : ((VAppTemplateType)getResource()).getLink()) {
      if (link.getRel().equals("download:default")) {
        return (EnvelopeType)JAXBUtil.unmarshallResource(RestUtil.downloadFile(getVcloudClient(), link.getHref()));
      }
    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_DOWNLOAD_LINK_MSG));
  }

  public Task delete()
    throws VCloudException
  {
    String vAppTemplateUrl = getReference().getHref();
    return deletevAppTemplate(getVcloudClient(), vAppTemplateUrl);
  }

  public static Task delete(VcloudClient client, ReferenceType vAppTemplateRef)
    throws VCloudException
  {
    return deletevAppTemplate(client, vAppTemplateRef.getHref());
  }

  private static Task deletevAppTemplate(VcloudClient client, String vAppTemplateUrl)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.delete(client, vAppTemplateUrl, 202));
  }

  public Task deleteVappTemplate()
    throws VCloudException
  {
    String vAppTemplateUrl = getReference().getHref();
    if (isPartOfCatalogItem())
    {
      CatalogItem.delete(getVcloudClient(), getCatalogItemReference());
    }
    return new Task(getVcloudClient(), (TaskType)SdkUtil.delete(getVcloudClient(), vAppTemplateUrl, 202));
  }

  public static Task deleteVappTemplate(VcloudClient client, ReferenceType vAppTemplateRef)
    throws VCloudException
  {
    return deleteVappTemplateWithCatalogItem(client, vAppTemplateRef);
  }

  private static Task deleteVappTemplateWithCatalogItem(VcloudClient client, ReferenceType vappTemplRef)
    throws VCloudException
  {
    VappTemplate vappTemplate = getVappTemplateByReference(client, vappTemplRef);
    String vappTemplateUrl = vappTemplate.getReference().getHref();
    if (vappTemplate.isPartOfCatalogItem())
    {
      CatalogItem.delete(client, vappTemplate.getCatalogItemReference());
    }
    return new Task(client, (TaskType)SdkUtil.delete(client, vappTemplateUrl, 202));
  }

  public void disableDownload()
    throws VCloudException
  {
    String vAppTemplateDisableDownloadURL = getReference().getHref() + "/action/disableDownload";

    SdkUtil.post(getVcloudClient(), vAppTemplateDisableDownloadURL, null, null, 204);
  }

  public static void disableDownload(VcloudClient client, ReferenceType vAppTemplateRef)
    throws VCloudException
  {
    String vAppTemplateEnableDownloadURL = vAppTemplateRef.getHref() + "/action/disableDownload";

    SdkUtil.post(client, vAppTemplateEnableDownloadURL, null, null, 204);
  }

  public Task enableDownload()
    throws VCloudException
  {
    String vAppTemplateEnableDownloadURL = getReference().getHref() + "/action/enableDownload";

    return executeVappTemplateAction(getVcloudClient(), vAppTemplateEnableDownloadURL, null, null, 202);
  }

  /** @deprecated */
  public static Task relocate(VcloudClient client, ReferenceType vmRef, ReferenceType datastoreReference)
    throws VCloudException
  {
    RelocateParamsType paramsType = new RelocateParamsType();
    paramsType.setDatastore(datastoreReference);
    String request = JAXBUtil.marshal(new ObjectFactory().createRelocateParams(paramsType));

    return new Task(client, (TaskType)SdkUtil.post(client, vmRef.getHref() + "/action/relocate", request, "application/vnd.vmware.vcloud.relocateVmParams+xml", 202));
  }

  /** @deprecated */
  public Task relocate(ReferenceType datastoreReference)
    throws VCloudException
  {
    RelocateParamsType paramsType = new RelocateParamsType();
    paramsType.setDatastore(datastoreReference);
    String request = JAXBUtil.marshal(new ObjectFactory().createRelocateParams(paramsType));

    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), getReference().getHref() + "/action/relocate", request, "application/vnd.vmware.vcloud.relocateVmParams+xml", 202));
  }

  public Task consolidate()
    throws VCloudException
  {
    return new Task(getVcloudClient(), (TaskType)SdkUtil.post(getVcloudClient(), getReference().getHref() + "/action/consolidate", null, null, 202));
  }

  public static Task consolidate(VcloudClient client, ReferenceType vappTemplateVmRef)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, vappTemplateVmRef.getHref() + "/action/consolidate", null, null, 202));
  }

  public static Task enableDownload(VcloudClient client, ReferenceType vAppTemplateRef)
    throws VCloudException
  {
    String vAppTemplateEnableDownloadURL = vAppTemplateRef.getHref() + "/action/enableDownload";

    return executeVappTemplateAction(client, vAppTemplateEnableDownloadURL, null, null, 202);
  }

  private static Task executeVappTemplateAction(VcloudClient client, String vappTemplateActionUrl, String content, String contentType, int statusCode)
    throws VCloudException
  {
    return new Task(client, (TaskType)SdkUtil.post(client, vappTemplateActionUrl, content, contentType, statusCode));
  }

  public Set<String> getUploadFileNames()
    throws VCloudException
  {
    if (!((VAppTemplateType)getResource()).isOvfDescriptorUploaded().booleanValue()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.OVF_DESCRIPTOR_NOT_UPLOADED_NO_FILES_MSG));
    }

    return this.uploadFileMap.keySet();
  }

  public Set<String> getUploadedFileNames()
    throws VCloudException
  {
    return this.uploadedFileMap.keySet();
  }

  public void uploadOVFFile(InputStream is, long size)
    throws VCloudException
  {
    if (((VAppTemplateType)getResource()).isOvfDescriptorUploaded().booleanValue()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.OVF_DESCRIPTOR_ALREADY_UPLOADED_MSG));
    }

    if (this.uploadFileMap.size() != 1) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.INVALID_NO_FILES_MSG) + " - " + this.uploadFileMap.size());
    }

    Set keys = this.uploadFileMap.keySet();
    String fileName = (String)keys.iterator().next();
    com.vmware.vcloud.api.rest.schema.FileType file = (com.vmware.vcloud.api.rest.schema.FileType)this.uploadFileMap.get(fileName);
    RestUtil.uploadFile(getVcloudClient(), file, is, size, 0L, 0L);
  }

  public void uploadFile(String fileName, InputStream is, long size)
    throws VCloudException
  {
    if (!((VAppTemplateType)getResource()).isOvfDescriptorUploaded().booleanValue()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.OVF_DESCRIPTOR_NOT_UPLOADED_MSG));
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_FILES_TO_UPLOAD_MSG));
    }

    com.vmware.vcloud.api.rest.schema.FileType file = (com.vmware.vcloud.api.rest.schema.FileType)this.uploadFileMap.get(fileName);
    if (file != null)
      RestUtil.uploadFile(getVcloudClient(), file, is, size, 0L, 0L);
    else
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.FILE_NOT_FOUND_MSG));
  }

  public void uploadFile(String fileName, InputStream is, long startByte, long endByte)
    throws VCloudException
  {
    if (!((VAppTemplateType)getResource()).isOvfDescriptorUploaded().booleanValue()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.OVF_DESCRIPTOR_NOT_UPLOADED_MSG));
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_FILES_TO_UPLOAD_MSG));
    }

    com.vmware.vcloud.api.rest.schema.FileType file = (com.vmware.vcloud.api.rest.schema.FileType)this.uploadFileMap.get(fileName);
    if (file != null) {
      RestUtil.uploadFile(getVcloudClient(), file, is, file.getSize().longValue(), startByte, endByte);
    }
    else
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.FILE_NOT_FOUND_MSG));
  }

  private void writeToFile(String downloadLocation, String fileName, InputStream inputStream)
    throws VCloudException
  {
    File file = new File(downloadLocation + "/" + fileName);
    OutputStream out;
    try
    {
      out = new FileOutputStream(file);
    } catch (FileNotFoundException e) {
      throw new VCloudRuntimeException(e);
    }
    byte[] buf = new byte[1024];
    try
    {
      int len;
      while ((len = inputStream.read(buf)) > 0)
        out.write(buf, 0, len);
    }
    catch (IOException e) {
      throw new VCloudRuntimeException(e);
    } finally {
      try {
        if (out != null) {
          out.flush();
          out.close();
        }
        if (inputStream != null)
          inputStream.close();
      }
      catch (IOException e) {
        throw new VCloudRuntimeException(e);
      }
    }
  }

  public void downloadVappTemplate(String downloadLocation)
    throws VCloudException
  {
    Set downloadFileNames = getDownloadFileNames().keySet();
    InputStream ovfInputStream = downloadOVFFile();
    writeToFile(downloadLocation, "descriptor.ovf", ovfInputStream);
    for (String fileName : downloadFileNames) {
      InputStream vmdkInputStream = downloadFile(fileName);
      writeToFile(downloadLocation, fileName, vmdkInputStream);
    }
  }

  public InputStream downloadOVFFile()
    throws VCloudException
  {
    for (LinkType link : ((VAppTemplateType)getResource()).getLink()) {
      if (link.getRel().equals("download:default")) {
        return RestUtil.downloadFile(getVcloudClient(), link.getHref());
      }
    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_DOWNLOAD_LINK_MSG));
  }

  public String getOVFDownloadURL()
    throws VCloudException
  {
    for (LinkType link : ((VAppTemplateType)getResource()).getLink()) {
      if (link.getRel().equals("download:default"))
        return link.getHref();
    }
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_DOWNLOAD_LINK_MSG));
  }

  public HashMap<String, com.vmware.vcloud.api.rest.schema.ovf.FileType> getDownloadFileNames()
    throws VCloudException
  {
    HashMap downloadFileMap = new HashMap();
    com.vmware.vcloud.api.rest.schema.ovf.ReferencesType fileReferences = downloadOvfEnvelope().getReferences();
    if (fileReferences != null) {
      List files = fileReferences.getFile();

      for (com.vmware.vcloud.api.rest.schema.ovf.FileType file : files) {
        if (!downloadFileMap.containsKey(file.getHref()))
          downloadFileMap.put(file.getHref(), file);
      }
    }
    return downloadFileMap;
  }

  public InputStream downloadFile(String fileName)
    throws VCloudException
  {
    for (LinkType link : ((VAppTemplateType)getResource()).getLink()) {
      if (link.getRel().equals("download:default")) {
        return downloadFile(fileName, link.getHref());
      }
    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_DOWNLOAD_LINK_MSG));
  }

  private InputStream downloadFile(String fileName, String href)
    throws VCloudException
  {
    HashMap downloadFiles = getDownloadFileNames();
    if (!downloadFiles.containsKey(fileName)) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.FILE_NOT_FOUND_MSG));
    }
    return RestUtil.downloadFile(getVcloudClient(), href.replaceAll("descriptor.ovf", fileName));
  }

  public HashMap<String, com.vmware.vcloud.api.rest.schema.FileType> monitorUpload()
    throws VCloudException
  {
    HashMap filesStatusMap = new HashMap();
    VappTemplate vappTemplate = getVappTemplateByReference(getVcloudClient(), getReference());

    if (((VAppTemplateType)vappTemplate.getResource()).getFiles() != null) {
      FilesListType filesType = ((VAppTemplateType)vappTemplate.getResource()).getFiles();
      List files = filesType.getFile();
      for (com.vmware.vcloud.api.rest.schema.FileType file : files) {
        filesStatusMap.put(file.getName(), file);
      }
    }

    return filesStatusMap;
  }

  public Task updateVappTemplate(VAppTemplateType vappTemplateType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    String vappTemplateUpdateUrl = getReference().getHref();
    String vappTemplateRequestXmlString = JAXBUtil.marshal(of.createVAppTemplate(vappTemplateType));

    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.PUT_URL_MSG) + " - " + vappTemplateUpdateUrl);

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), vappTemplateUpdateUrl, vappTemplateRequestXmlString, "application/vnd.vmware.vcloud.vAppTemplate+xml", 202));
  }

  public List<ProductSectionType> getProductSections()
    throws VCloudException
  {
    return this.productSections;
  }

  public static List<ProductSectionType> getProductSections(VcloudClient client, ReferenceType vappRef)
    throws VCloudException
  {
    ProductSectionListType productSectionListType = (ProductSectionListType)SdkUtil.get(client, vappRef.getHref() + "/productSections/", 200);

    return productSectionListType.getProductSection();
  }

  public NetworkSectionType getNetworkSection()
  {
    return this.networkSection;
  }

  public static NetworkSectionType getNetworkSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (NetworkSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/networkSection/", 200);
  }

  public NetworkConfigSectionType getNetworkConfigSection()
  {
    return this.networkConfigSection;
  }

  public static NetworkConfigSectionType getNetworkConfigSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (NetworkConfigSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/networkConfigSection/", 200);
  }

  public LeaseSettingsSectionType getLeaseSettingsSection()
  {
    return this.leaseSettingsSection;
  }

  public static LeaseSettingsSectionType getLeaseSettingsSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (LeaseSettingsSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/leaseSettingsSection/", 200);
  }

  public NetworkConnectionSectionType getNetworkConnectionSection()
  {
    return this.networkConnectionSection;
  }

  public static NetworkConnectionSectionType getNetworkConnectionSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (NetworkConnectionSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/networkConnectionSection/", 200);
  }

  public GuestCustomizationSectionType getGuestCustomizationSection()
  {
    return this.guestCustomizationSection;
  }

  public static GuestCustomizationSectionType getGuestCustomizationSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (GuestCustomizationSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/guestCustomizationSection/", 200);
  }

  public CustomizationSectionType getCustomizationSection()
  {
    return this.customizationSection;
  }

  public static CustomizationSectionType getCustomizationSection(VcloudClient client, ReferenceType vappTemplateRef)
    throws VCloudException
  {
    return (CustomizationSectionType)SdkUtil.get(client, vappTemplateRef.getHref() + "/customizationSection/", 200);
  }

  /** @deprecated */
  public Task updateSection(SectionType sectionType)
    throws VCloudException
  {
    String sectionUrl = getReference().getHref();
    String sectionContentType = null;
    String sectionRequestString = null;
    ObjectFactory vCloudObjectFactory = new ObjectFactory();
    if ((sectionType instanceof LeaseSettingsSectionType)) {
      JAXBElement section = vCloudObjectFactory.createLeaseSettingsSection((LeaseSettingsSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/leaseSettingsSection/";

      sectionContentType = "application/vnd.vmware.vcloud.leaseSettingsSection+xml";
    } else if ((sectionType instanceof GuestCustomizationSectionType)) {
      JAXBElement section = vCloudObjectFactory.createGuestCustomizationSection((GuestCustomizationSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/guestCustomizationSection/";

      sectionContentType = "application/vnd.vmware.vcloud.guestCustomizationSection+xml";
    } else if ((sectionType instanceof CustomizationSectionType)) {
      JAXBElement section = vCloudObjectFactory.createCustomizationSection((CustomizationSectionType)sectionType);

      sectionRequestString = JAXBUtil.marshal(section);
      sectionUrl = sectionUrl + "/customizationSection/";
      sectionContentType = "application/vnd.vmware.vcloud.customizationSection+xml";
    } else {
      logger.log(Level.INFO, sectionType.getClass().getName() + " - " + SdkUtil.getI18nString(SdkMessage.NOT_VALID_SECTION_MSG));

      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NOT_VALID_SECTION_MSG));
    }

    return new Task(getVcloudClient(), (TaskType)SdkUtil.put(getVcloudClient(), sectionUrl, sectionRequestString, sectionContentType, 202));
  }

  private void sortVAppTemplateFiles_v1_5()
  {
    this.uploadFileMap = new HashMap();
    this.uploadedFileMap = new HashMap();
    if ((getResource() instanceof VAppTemplateType)) {
      FilesListType fileList = ((VAppTemplateType)getResource()).getFiles();
      if (fileList == null) {
        return;
      }
      List files = fileList.getFile();

      Iterator fileIt = files.iterator();

      while (fileIt.hasNext()) {
        com.vmware.vcloud.api.rest.schema.FileType file = (com.vmware.vcloud.api.rest.schema.FileType)fileIt.next();
        String fileName = file.getName();
        if (file.getSize().compareTo(file.getBytesTransferred()) == 0)
          this.uploadedFileMap.put(fileName, file);
        else
          this.uploadFileMap.put(fileName, file);
      }
    }
  }

  private static BigInteger getVmSize(VirtualSystemType virtualSystemType)
    throws VCloudException
  {
    BigInteger vAppTemplateVmSize = BigInteger.ZERO;
    for (JAXBElement section : virtualSystemType.getSection())
    {
      if ((section.getValue() instanceof VirtualHardwareSectionType)) {
        VirtualHardwareSectionType virtualHardwareSection = (VirtualHardwareSectionType)section.getValue();

        List items = virtualHardwareSection.getItem();
        for (RASDType item : items) {
          if (item.getResourceType().getValue().equals("17"))
          {
            VirtualDisk disk = new VirtualDisk(item);
            if (disk.getHardDiskSize() != null) {
              vAppTemplateVmSize = vAppTemplateVmSize.add(disk.getHardDiskSize());
            }
          }
        }
      }
    }
    return vAppTemplateVmSize;
  }

  public BigInteger getVappTemplateSize()
    throws VCloudException
  {
    BigInteger vAppTemplateSize = BigInteger.ZERO;
    EnvelopeType ovf = null;
    if (getReference().getType().equals("application/vnd.vmware.vcloud.vm+xml")) {
      VappTemplate vappTemplate = getVappTemplateByReference(getVcloudClient(), getParentVappTemplateReference());

      ovf = vappTemplate.getOvf();
      if (ovf.getContent() != null) {
        if ((ovf.getContent().getValue() instanceof VirtualSystemCollectionType)) {
          VirtualSystemCollectionType virtualSystemCollection = (VirtualSystemCollectionType)ovf.getContent().getValue();

          for (JAXBElement virtualSystems : virtualSystemCollection.getContent())
          {
            VirtualSystemType virtualSystem = (VirtualSystemType)virtualSystems.getValue();

            if (virtualSystem.getId().equals(getReference().getName()))
            {
              vAppTemplateSize = vAppTemplateSize.add(getVmSize(virtualSystem));
            }
          }
        }
        else if ((ovf.getContent().getValue() instanceof VirtualSystemType)) {
          vAppTemplateSize = vAppTemplateSize.add(getVmSize((VirtualSystemType)ovf.getContent().getValue()));
        }
      }
    }
    else
    {
      ovf = getOvf();
      if (ovf.getContent() != null) {
        if ((ovf.getContent().getValue() instanceof VirtualSystemCollectionType)) {
          VirtualSystemCollectionType virtualSystemCollection = (VirtualSystemCollectionType)ovf.getContent().getValue();

          for (JAXBElement virtualSystems : virtualSystemCollection.getContent())
          {
            vAppTemplateSize = vAppTemplateSize.add(getVmSize((VirtualSystemType)virtualSystems.getValue()));
          }

        }
        else if ((ovf.getContent().getValue() instanceof VirtualSystemType)) {
          vAppTemplateSize = vAppTemplateSize.add(getVmSize((VirtualSystemType)ovf.getContent().getValue()));
        }
      }

    }

    return vAppTemplateSize;
  }

  public Collection<ReferenceType> getShadowVmReferences()
    throws VCloudException
  {
    if (!isVm().booleanValue()) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.OPERATION_NOT_SUPPORTED));
    }
    String shadowVmsUrl = getReference().getHref() + "/shadowVms";

    com.vmware.vcloud.api.rest.schema.ReferencesType shadowVmReferencesType = (com.vmware.vcloud.api.rest.schema.ReferencesType)SdkUtil.get(getVcloudClient(), shadowVmsUrl, 200);

    Collection shadowVmReferences = new ArrayList();
    for (JAXBElement shadowVmReferenceType : shadowVmReferencesType.getReference())
    {
      shadowVmReferences.add(shadowVmReferenceType.getValue());
    }
    return shadowVmReferences;
  }

  public Task updateGoldMaster(Boolean goldMasterFlag)
    throws VCloudException
  {
    ((VAppTemplateType)getResource()).setGoldMaster(goldMasterFlag);
    return updateVappTemplate((VAppTemplateType)getResource());
  }

  public boolean isPartOfCatalogItem()
    throws VCloudException
  {
    return this.catalogItemReference != null;
  }

  public InputStream downloadLosslessOVF()
    throws VCloudException
  {
    for (LinkType link : ((VAppTemplateType)getResource()).getLink()) {
      if (link.getRel().equals("download:identity")) {
        return RestUtil.downloadFile(getVcloudClient(), link.getHref());
      }
    }

    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.NO_DOWNLOAD_LINK_MSG));
  }
}