package com.vmware.vcloud.sdk_v1_0;

import com.vmware.vcloud.api.rest.schema_v1_0.FileType;
import com.vmware.vcloud.api.rest.schema_v1_0.FilesListType;
import com.vmware.vcloud.api.rest.schema_v1_0.LinkType;
import com.vmware.vcloud.api.rest.schema_v1_0.MediaType;
import com.vmware.vcloud.api.rest.schema_v1_0.ObjectFactory;
import com.vmware.vcloud.api.rest.schema_v1_0.ReferenceType;
import com.vmware.vcloud.api.rest.schema_v1_0.RelationType;
import com.vmware.vcloud.api.rest.schema_v1_0.TaskType;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Media extends VcloudEntity<MediaType>
{
  private static Logger logger = Logger.getLogger(Media.class.getSimpleName());
  private HashMap<String, FileType> uploadFileMap;
  private HashMap<String, FileType> uploadedFileMap;
  private ReferenceType catalogItemReference;
  private ReferenceType vdcReference;

  Media(VcloudClient client, MediaType mediaResource)
  {
    super(client, mediaResource);
    sortMediaFilesAndReferences();
  }

  public static Media getMediaByReference(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    return getMedia(client, mediaRef.getHref());
  }

  private static Media getMedia(VcloudClient client, String url)
    throws VCloudException
  {
    Response response = RestUtil.get(client, url);
    MediaType mediaType = null;

    if (!response.isExpected(200))
    {
      response.handleUnExpectedResponse();
    }
    mediaType = (MediaType)response.getResource();
    Media media = new Media(client, mediaType);
    return media;
  }

  public Task updateMedia(MediaType mediaType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    String mediaUpdateUrl = ((MediaType)getResource()).getHref();
    String mediaRequestXmlString = JAXBUtil.marshal(of.createMedia(mediaType));

    logger.log(Level.INFO, "MEDIA URL : " + mediaUpdateUrl);

    Response response = RestUtil.put(getVcloudClient(), mediaUpdateUrl, mediaRequestXmlString, "application/vnd.vmware.vcloud.media+xml", null);

    TaskType taskType = null;

    if (response.isExpected(202)) {
      taskType = (TaskType)response.getResource();
    }
    else {
      response.handleUnExpectedResponse();
    }

    Task task = new Task(getVcloudClient(), taskType);
    return task;
  }

  public Task delete()
    throws VCloudException
  {
    String mediaUrl = ((MediaType)getResource()).getHref();
    Task task = deleteMedia(getVcloudClient(), mediaUrl);
    return task;
  }

  public static Task delete(VcloudClient client, ReferenceType mediaRef)
    throws VCloudException
  {
    return deleteMedia(client, mediaRef.getHref());
  }

  private static Task deleteMedia(VcloudClient client, String mediaUrl)
    throws VCloudException
  {
    Task task = null;
    Response response = RestUtil.delete(client, mediaUrl);
    TaskType taskType = null;

    if (response.isExpected(202)) {
      taskType = (TaskType)response.getResource();
    }
    else {
      response.handleUnExpectedResponse();
    }
    task = new Task(client, taskType);

    return task;
  }

  public Set<String> getUploadFileNames()
  {
    return this.uploadFileMap.keySet();
  }

  public Set<String> getUploadedFileNames()
  {
    return this.uploadedFileMap.keySet();
  }

  public void uploadFile(String fileName, InputStream is, long size)
    throws VCloudException
  {
    if (((MediaType)getResource()).getStatus().intValue() == 1) {
      throw new VCloudException("Media Resolved State: Files are already uploaded");
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException("No file(s) to upload ");
    }
    FileType file = (FileType)this.uploadFileMap.get(fileName);
    if (file != null)
      RestUtil.uploadFile(getVcloudClient(), file, is, size, 0L, 0L);
    else
      throw new VCloudException("File Name Not found");
  }

  public void uploadFile(String fileName, InputStream is, long startByte, long endByte)
    throws VCloudException
  {
    if (((MediaType)getResource()).getStatus().intValue() == 1) {
      throw new VCloudException("Media Resolved State: Files are already uploaded");
    }

    if (this.uploadFileMap.isEmpty()) {
      throw new VCloudException("No file(s) to upload ");
    }
    FileType file = (FileType)this.uploadFileMap.get(fileName);
    if (file != null) {
      RestUtil.uploadFile(getVcloudClient(), file, is, file.getSize().longValue(), startByte, endByte);
    }
    else
      throw new VCloudException("File Name Not found");
  }

  public HashMap<String, FileType> monitorUpload()
    throws VCloudException
  {
    Media media = getMedia(getVcloudClient(), ((MediaType)getResource()).getHref());

    FilesListType filesType = ((MediaType)media.getResource()).getFiles();
    List files = filesType.getFile();
    HashMap filesStatusMap = new HashMap();
    for (FileType file : files) {
      filesStatusMap.put(file.getName(), file);
    }
    return filesStatusMap;
  }

  private void sortMediaFilesAndReferences()
  {
    this.uploadFileMap = new HashMap();
    this.uploadedFileMap = new HashMap();
    FilesListType fileList = ((MediaType)getResource()).getFiles();
    if (fileList != null) {
      List files = fileList.getFile();

      Iterator fileIt = files.iterator();

      while (fileIt.hasNext()) {
        FileType file = (FileType)fileIt.next();
        String fileName = file.getName();
        if (file.getSize().compareTo(file.getBytesTransferred()) == 0)
          this.uploadedFileMap.put(fileName, file);
        else {
          this.uploadFileMap.put(fileName, file);
        }
      }
    }
    for (LinkType mediaLink : ((MediaType)getResource()).getLink()) {
      if ((mediaLink.getRel().equals(RelationType.CATALOG_ITEM)) && (mediaLink.getType().equals("application/vnd.vmware.vcloud.catalogItem+xml")))
      {
        this.catalogItemReference = mediaLink;
      }if ((mediaLink.getRel().equals(RelationType.UP)) && (mediaLink.getType().equals("application/vnd.vmware.vcloud.vdc+xml")))
      {
        this.vdcReference = mediaLink;
      }
    }
  }

  public ReferenceType getCatalogItemReference()
    throws VCloudException
  {
    if (this.catalogItemReference != null)
      return this.catalogItemReference;
    throw new VCloudException("Reference Not Found");
  }

  public ReferenceType getVdcReference()
    throws VCloudException
  {
    if (this.vdcReference != null)
      return this.vdcReference;
    throw new VCloudException("Reference Not Found");
  }
}