package com.vmware.vcloud.sdk_v1_0;

import com.vmware.vcloud.api.rest.schema_v1_0.ControlAccessParamsType;
import com.vmware.vcloud.api.rest.schema_v1_0.LinkType;
import com.vmware.vcloud.api.rest.schema_v1_0.ObjectFactory;
import com.vmware.vcloud.api.rest.schema_v1_0.OrgType;
import com.vmware.vcloud.api.rest.schema_v1_0.ReferenceType;
import com.vmware.vcloud.api.rest.schema_v1_0.RelationType;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Organization extends VcloudEntity<OrgType>
{
  private static Logger logger = Logger.getLogger(Organization.class.getSimpleName());
  private HashMap<String, ReferenceType> catalogRefsByName;
  private HashMap<String, ReferenceType> catalogRefsById;
  private HashMap<String, ReferenceType> vdcRefsByName;
  private HashMap<String, ReferenceType> networkRefsByName;
  private ReferenceType tasksListRef;
  private HashMap<String, ReferenceType> catalogsControlAccessRefsByName;
  private HashMap<String, ReferenceType> catalogsControlAccessActionRefsByName;

  Organization(VcloudClient client, OrgType org)
  {
    super(client, org);
    sortRefs();
  }

  public static Organization getOrganizationByReference(VcloudClient client, ReferenceType orgRef)
    throws VCloudException
  {
    return getOrganization(client, orgRef.getHref());
  }

  private static Organization getOrganization(VcloudClient client, String url)
    throws VCloudException
  {
    logger.log(Level.INFO, "GET: Organization  " + url);
    Response response = RestUtil.get(client, url);
    OrgType orgType = null;

    if (response.isExpected(200))
      orgType = (OrgType)response.getResource();
    else {
      response.handleUnExpectedResponse();
    }
    Organization org = new Organization(client, orgType);
    return org;
  }

  private static String getId(String href)
  {
    int i = href.lastIndexOf("/");
    return href.substring(i + 1);
  }

  private static String getId(String href, String type)
  {
    int startSlash = href.indexOf("/", href.lastIndexOf(type)) + 1;
    int endSlash = href.indexOf("/", startSlash);
    return href.substring(startSlash, endSlash);
  }

  public ControlAccessParamsType getCatalogControlAccessByName(String catalogName)
    throws VCloudException
  {
    String url = ((ReferenceType)this.catalogsControlAccessRefsByName.get(catalogName)).getHref();
    logger.log(Level.INFO, "GET: Catalog's Control Access " + url);
    Response response = RestUtil.get(getVcloudClient(), url);
    ControlAccessParamsType controlAccessParamsType = null;

    if (!response.isExpected(200))
    {
      response.handleUnExpectedResponse();
    }

    controlAccessParamsType = (ControlAccessParamsType)response.getResource();

    return controlAccessParamsType;
  }

  public ControlAccessParamsType updateCatalogControlAccessByName(String catalogName, ControlAccessParamsType controlAccessParams)
    throws VCloudException
  {
    String url = ((ReferenceType)this.catalogsControlAccessActionRefsByName.get(catalogName)).getHref();

    logger.log(Level.INFO, "POST: Catalog's Control Access " + url);
    ObjectFactory vcloudObjectFactory = new ObjectFactory();
    String requestString = JAXBUtil.marshal(vcloudObjectFactory.createControlAccessParams(controlAccessParams));

    Response response = RestUtil.post(getVcloudClient(), url, requestString, "application/vnd.vmware.vcloud.controlAccess+xml", null);

    ControlAccessParamsType controlAccessParamsType = null;

    if (response.isExpected(200)) {
      controlAccessParamsType = (ControlAccessParamsType)response.getResource();
    }
    else
    {
      response.handleUnExpectedResponse();
    }

    return controlAccessParamsType;
  }

  private void sortRefs()
  {
    this.catalogRefsByName = new HashMap();
    this.catalogRefsById = new HashMap();
    this.vdcRefsByName = new HashMap();
    this.networkRefsByName = new HashMap();
    this.catalogsControlAccessRefsByName = new HashMap();
    this.catalogsControlAccessActionRefsByName = new HashMap();
    this.tasksListRef = new ReferenceType();

    if (((OrgType)getResource()).getLink() != null) {
      List links = ((OrgType)getResource()).getLink();

      for (LinkType link : links) {
        String id = getId(link.getHref());
        if (link.getType().equals("application/vnd.vmware.vcloud.catalog+xml")) {
          this.catalogRefsById.put(id, link);
          this.catalogRefsByName.put(link.getName(), link);
        } else if (link.getType().equals("application/vnd.vmware.vcloud.vdc+xml")) {
          this.vdcRefsByName.put(link.getName(), link);
        } else if (link.getType().equals("application/vnd.vmware.vcloud.network+xml"))
        {
          this.networkRefsByName.put(link.getName(), link);
        } else if (link.getType().equals("application/vnd.vmware.vcloud.tasksList+xml"))
        {
          this.tasksListRef = link;
        } else {
          logger.log(Level.WARNING, "Unknown Link Type: " + link.getType());
        }

      }

      for (LinkType link : links) {
        String id = getId(link.getHref());
        if ((link.getType().equals("application/vnd.vmware.vcloud.controlAccess+xml")) && (link.getRel().equals(RelationType.DOWN)))
        {
          id = getId(link.getHref(), "catalog");
          this.catalogsControlAccessRefsByName.put(((ReferenceType)this.catalogRefsById.get(id)).getName(), link);
        }

        if ((link.getType().equals("application/vnd.vmware.vcloud.controlAccess+xml")) && (link.getRel().equals(RelationType.CONTROL_ACCESS)))
        {
          id = getId(link.getHref(), "catalog");
          this.catalogsControlAccessActionRefsByName.put(((ReferenceType)this.catalogRefsById.get(id)).getName(), link);
        }
      }
    }
  }

  public ReferenceType getTasksListRef()
  {
    return this.tasksListRef;
  }

  public HashMap<String, ReferenceType> getCatalogRefsByName()
  {
    return this.catalogRefsByName;
  }

  public HashMap<String, ReferenceType> getVdcRefsByName()
  {
    return this.vdcRefsByName;
  }

  public HashMap<String, ReferenceType> getNetworkRefsByName()
  {
    return this.networkRefsByName;
  }

  public Collection<ReferenceType> getVdcRefs()
  {
    return this.vdcRefsByName.values();
  }

  public ReferenceType getVdcRefByName(String name)
  {
    return (ReferenceType)this.vdcRefsByName.get(name);
  }

  public Collection<ReferenceType> getCatalogRefs()
  {
    return this.catalogRefsByName.values();
  }

  public ReferenceType getCatalogRefByName(String name)
  {
    return (ReferenceType)this.catalogRefsByName.get(name);
  }

  public Collection<ReferenceType> getNetworkRefs()
  {
    return this.networkRefsByName.values();
  }

  public ReferenceType getNetworkRefByName(String name)
  {
    return (ReferenceType)this.networkRefsByName.get(name);
  }
}