package com.vmware.vcloud.sdk_v1_0.admin;

import com.vmware.vcloud.api.rest.schema_v1_0.CatalogItemType;
import com.vmware.vcloud.api.rest.schema_v1_0.CatalogType;
import com.vmware.vcloud.api.rest.schema_v1_0.LinkType;
import com.vmware.vcloud.api.rest.schema_v1_0.ObjectFactory;
import com.vmware.vcloud.api.rest.schema_v1_0.PublishCatalogParamsType;
import com.vmware.vcloud.api.rest.schema_v1_0.ReferenceType;
import com.vmware.vcloud.api.rest.schema_v1_0.RelationType;
import com.vmware.vcloud.sdk_v1_0.Catalog;
import com.vmware.vcloud.sdk_v1_0.CatalogItem;
import com.vmware.vcloud.sdk_v1_0.JAXBUtil;
import com.vmware.vcloud.sdk_v1_0.Response;
import com.vmware.vcloud.sdk_v1_0.RestUtil;
import com.vmware.vcloud.sdk_v1_0.VCloudException;
import com.vmware.vcloud.sdk_v1_0.VcloudClient;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class AdminCatalog extends Catalog
{
  private static Logger logger = Logger.getLogger(AdminCatalog.class.getSimpleName());
  private ReferenceType adminOrgReference;
  private ReferenceType catalogReference;

  AdminCatalog(VcloudClient client, CatalogType catalogType)
  {
    super(client, catalogType);
    sortRefs();
  }

  public ReferenceType getAdminOrganizationReference()
    throws VCloudException
  {
    if (this.adminOrgReference != null)
      return this.adminOrgReference;
    throw new VCloudException("Reference Not Found");
  }

  public ReferenceType getCatalogReference()
    throws VCloudException
  {
    if (this.catalogReference != null)
      return this.catalogReference;
    throw new VCloudException("Reference Not Found");
  }

  private void sortRefs()
  {
    for (LinkType adminCatalogLink : ((CatalogType)getResource()).getLink()) {
      if ((adminCatalogLink.getRel().equals(RelationType.UP)) && (adminCatalogLink.getType().equals("application/vnd.vmware.admin.organization+xml")))
      {
        this.adminOrgReference = adminCatalogLink;
      }if ((adminCatalogLink.getRel().equals(RelationType.ALTERNATE)) && (adminCatalogLink.getType().equals("application/vnd.vmware.vcloud.catalog+xml")))
      {
        this.catalogReference = adminCatalogLink;
      }
    }
  }

  public static AdminCatalog getCatalogByReference(VcloudClient client, ReferenceType catalogRef)
    throws VCloudException
  {
    String url = catalogRef.getHref();
    return getCatalog(client, url);
  }

  private static AdminCatalog getCatalog(VcloudClient client, String url)
    throws VCloudException
  {
    Response response = RestUtil.get(client, url);
    CatalogType catalogType = null;

    if (response.isExpected(200))
      catalogType = (CatalogType)response.getResource();
    else
      response.handleUnExpectedResponse();
    AdminCatalog catalog = new AdminCatalog(client, catalogType);
    return catalog;
  }

  public CatalogItem addCatalogItem(CatalogItemType catalogItemType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    JAXBElement catalogItemParams = of.createCatalogItem(catalogItemType);

    String requestBodyXml = JAXBUtil.marshal(catalogItemParams);
    String postURL = ((CatalogType)getResource()).getHref().replaceAll("/admin", "") + "/catalogItems";

    Response response = RestUtil.post(getVcloudClient(), postURL, requestBodyXml, "application/vnd.vmware.vcloud.catalogItem+xml", null);

    if (!response.isExpected(201))
    {
      response.handleUnExpectedResponse();
    }
    CatalogItemType newCatalogItemType = (CatalogItemType)response.getResource();

    CatalogItem catalogItem = new CatalogItem(getVcloudClient(), newCatalogItemType);

    return catalogItem;
  }

  public AdminCatalog updateAdminCatalog(CatalogType catalogType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    String adminCatalogUpdateUrl = ((CatalogType)getResource()).getHref();
    String adminCatalogUpdateRequestXmlString = JAXBUtil.marshal(of.createCatalog(catalogType));

    logger.log(Level.INFO, "Update ADMIN Catalog URL : " + adminCatalogUpdateUrl);

    Response response = RestUtil.put(getVcloudClient(), adminCatalogUpdateUrl, adminCatalogUpdateRequestXmlString, "application/vnd.vmware.admin.catalog+xml", null);

    CatalogType updatedAdminCatalogType = null;
    if (response.isExpected(200))
      updatedAdminCatalogType = (CatalogType)response.getResource();
    else
      response.handleUnExpectedResponse();
    return new AdminCatalog(getVcloudClient(), updatedAdminCatalogType);
  }

  public void delete()
    throws VCloudException
  {
    String adminCatalogUrl = ((CatalogType)getResource()).getHref();
    deleteAdminCatalog(getVcloudClient(), adminCatalogUrl);
  }

  public static void delete(VcloudClient client, ReferenceType adminCatalogRef)
    throws VCloudException
  {
    deleteAdminCatalog(client, adminCatalogRef.getHref());
  }

  private static void deleteAdminCatalog(VcloudClient client, String adminCatalogUrl)
    throws VCloudException
  {
    Response response = RestUtil.delete(client, adminCatalogUrl);
    if (!response.isExpected(204))
      response.handleUnExpectedResponse();
  }

  public void publishCatalog(PublishCatalogParamsType publishCatalogParamsType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    JAXBElement catalogItemParams = of.createPublishCatalogParams(publishCatalogParamsType);

    String requestBodyXml = JAXBUtil.marshal(catalogItemParams);
    String postURL = ((CatalogType)getResource()).getHref() + "/action" + "/publish";

    Response response = RestUtil.post(getVcloudClient(), postURL, requestBodyXml, "application/vnd.vmware.admin.publishCatalogParams+xml", null);

    if (!response.isExpected(204))
    {
      response.handleUnExpectedResponse();
    }
  }
}