package org.eclipse.jdt.internal.compiler;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.env.IBinaryType;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.env.ISourceType;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ITypeRequestor;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilationUnit;
import org.eclipse.jdt.internal.compiler.problem.DefaultProblem;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;
import org.eclipse.jdt.internal.compiler.util.Messages;

public class Compiler
  implements ITypeRequestor, ProblemSeverities
{
  public Parser parser;
  public ICompilerRequestor requestor;
  public CompilerOptions options;
  public ProblemReporter problemReporter;
  protected PrintWriter out;
  public CompilationUnitDeclaration[] unitsToProcess;
  public int totalUnits;
  public LookupEnvironment lookupEnvironment;
  public static boolean DEBUG = false;
  public int parseThreshold = -1;
  public AbstractAnnotationProcessorManager annotationProcessorManager;
  public ReferenceBinding[] referenceBindings;
  public static IDebugRequestor DebugRequestor = null;

  /** @deprecated */
  public Compiler(INameEnvironment environment, IErrorHandlingPolicy policy, Map settings, ICompilerRequestor requestor, IProblemFactory problemFactory)
  {
    this(environment, policy, new CompilerOptions(settings), requestor, problemFactory, null);
  }

  /** @deprecated */
  public Compiler(INameEnvironment environment, IErrorHandlingPolicy policy, Map settings, ICompilerRequestor requestor, IProblemFactory problemFactory, boolean parseLiteralExpressionsAsConstants)
  {
    this(environment, policy, new CompilerOptions(settings, parseLiteralExpressionsAsConstants), requestor, problemFactory, null);
  }

  public Compiler(INameEnvironment environment, IErrorHandlingPolicy policy, CompilerOptions options, ICompilerRequestor requestor, IProblemFactory problemFactory)
  {
    this(environment, policy, options, requestor, problemFactory, null);
  }

  public Compiler(INameEnvironment environment, IErrorHandlingPolicy policy, CompilerOptions options, ICompilerRequestor requestor, IProblemFactory problemFactory, PrintWriter out)
  {
    this.options = options;

    if (DebugRequestor == null)
      this.requestor = requestor;
    else {
      this.requestor = new ICompilerRequestor() { private final ICompilerRequestor val$requestor;

        public void acceptResult(CompilationResult result) { if (Compiler.DebugRequestor.isActive()) {
            Compiler.DebugRequestor.acceptDebugResult(result);
          }
          this.val$requestor.acceptResult(result);
        }
      };
    }
    this.problemReporter = new ProblemReporter(policy, this.options, problemFactory);
    this.lookupEnvironment = new LookupEnvironment(this, this.options, this.problemReporter, environment);
    this.out = (out == null ? new PrintWriter(System.out, true) : out);
    initializeParser();
  }

  public void accept(IBinaryType binaryType, PackageBinding packageBinding, AccessRestriction accessRestriction)
  {
    if (this.options.verbose) {
      this.out.println(
        Messages.bind(Messages.compilation_loadBinary, new String(binaryType.getName())));
    }

    this.lookupEnvironment.createBinaryTypeFrom(binaryType, packageBinding, accessRestriction);
  }

  public void accept(ICompilationUnit sourceUnit, AccessRestriction accessRestriction)
  {
    CompilationResult unitResult = 
      new CompilationResult(sourceUnit, this.totalUnits, this.totalUnits, this.options.maxProblemsPerUnit);
    try {
      if (this.options.verbose) {
        String count = String.valueOf(this.totalUnits + 1);
        this.out.println(
          Messages.bind(Messages.compilation_request, 
          new String[] { 
          count, 
          count, 
          new String(sourceUnit.getFileName()) }));
      }
      CompilationUnitDeclaration parsedUnit;
      CompilationUnitDeclaration parsedUnit;
      if (this.totalUnits < this.parseThreshold)
        parsedUnit = this.parser.parse(sourceUnit, unitResult);
      else {
        parsedUnit = this.parser.dietParse(sourceUnit, unitResult);
      }
      parsedUnit.bits |= 1;

      this.lookupEnvironment.buildTypeBindings(parsedUnit, accessRestriction);
      addCompilationUnit(sourceUnit, parsedUnit);

      this.lookupEnvironment.completeTypeBindings(parsedUnit);
    }
    catch (AbortCompilationUnit e)
    {
      if (unitResult.compilationUnit == sourceUnit)
        this.requestor.acceptResult(unitResult.tagAsAccepted());
      else
        throw e;
    }
  }

  public void accept(ISourceType[] sourceTypes, PackageBinding packageBinding, AccessRestriction accessRestriction)
  {
    this.problemReporter.abortDueToInternalError(
      Messages.bind(Messages.abort_againstSourceModel, new String[] { String.valueOf(sourceTypes[0].getName()), String.valueOf(sourceTypes[0].getFileName()) }));
  }

  protected void addCompilationUnit(ICompilationUnit sourceUnit, CompilationUnitDeclaration parsedUnit)
  {
    int size = this.unitsToProcess.length;
    if (this.totalUnits == size)
    {
      System.arraycopy(
        this.unitsToProcess, 
        0, 
        this.unitsToProcess = new CompilationUnitDeclaration[size * 2], 
        0, 
        this.totalUnits);
    }this.unitsToProcess[(this.totalUnits++)] = parsedUnit;
  }

  protected void beginToCompile(ICompilationUnit[] sourceUnits)
  {
    int maxUnits = sourceUnits.length;
    this.totalUnits = 0;
    this.unitsToProcess = new CompilationUnitDeclaration[maxUnits];

    internalBeginToCompile(sourceUnits, maxUnits);
  }

  public void compile(ICompilationUnit[] sourceUnits)
  {
    CompilationUnitDeclaration unit = null;
    int i = 0;
    try
    {
      beginToCompile(sourceUnits);

      if (this.annotationProcessorManager != null) {
        processAnnotations();
        if (!this.options.generateClassFiles)
        {
          return;
        }
      }

      for (; i < this.totalUnits; i++) {
        unit = this.unitsToProcess[i];
        try {
          if (this.options.verbose) {
            this.out.println(
              Messages.bind(Messages.compilation_process, 
              new String[] { 
              String.valueOf(i + 1), 
              String.valueOf(this.totalUnits), 
              new String(this.unitsToProcess[i].getFileName()) }));
          }
          process(unit, i);
        }
        finally {
          unit.cleanUp();
        }
        this.unitsToProcess[i] = null;
        this.requestor.acceptResult(unit.compilationResult.tagAsAccepted());
        if (this.options.verbose)
          this.out.println(
            Messages.bind(Messages.compilation_done, 
            new String[] { 
            String.valueOf(i + 1), 
            String.valueOf(this.totalUnits), 
            new String(unit.getFileName()) }));
      }
    }
    catch (AbortCompilation e) {
      handleInternalException(e, unit);
    } catch (Error e) {
      handleInternalException(e, unit, null);
      throw e;
    } catch (RuntimeException e) {
      handleInternalException(e, unit, null);
      throw e;
    } finally {
      reset();
    }
    if (this.options.verbose)
      if (this.totalUnits > 1)
        this.out.println(
          Messages.bind(Messages.compilation_units, String.valueOf(this.totalUnits)));
      else
        this.out.println(
          Messages.bind(Messages.compilation_unit, String.valueOf(this.totalUnits)));
  }

  public void setBinaryTypes(ReferenceBinding[] binaryTypes)
  {
    this.referenceBindings = binaryTypes;
  }

  protected void handleInternalException(Throwable internalException, CompilationUnitDeclaration unit, CompilationResult result)
  {
    if ((result == null) && (unit != null)) {
      result = unit.compilationResult;
    }

    if ((result == null) && (this.lookupEnvironment.unitBeingCompleted != null)) {
      result = this.lookupEnvironment.unitBeingCompleted.compilationResult;
    }

    if ((result == null) && (this.lookupEnvironment.unitBeingCompleted != null)) {
      result = this.lookupEnvironment.unitBeingCompleted.compilationResult;
    }
    if ((result == null) && (this.unitsToProcess != null) && (this.totalUnits > 0)) {
      result = this.unitsToProcess[(this.totalUnits - 1)].compilationResult;
    }

    boolean needToPrint = true;
    if (result != null)
    {
      StringWriter stringWriter = new StringWriter();
      PrintWriter writer = new PrintWriter(stringWriter);
      internalException.printStackTrace(writer);
      StringBuffer buffer = stringWriter.getBuffer();

      String[] pbArguments = { 
        Messages.compilation_internalError + 
        "\n" + 
        buffer.toString() };

      result
        .record(
        this.problemReporter
        .createProblem(
        result.getFileName(), 
        0, 
        pbArguments, 
        pbArguments, 
        1, 
        0, 
        0, 
        0, 
        0), 
        unit);

      if (!result.hasBeenAccepted) {
        this.requestor.acceptResult(result.tagAsAccepted());
        needToPrint = false;
      }
    }
    if (needToPrint)
    {
      internalException.printStackTrace();
    }
  }

  protected void handleInternalException(AbortCompilation abortException, CompilationUnitDeclaration unit)
  {
    if (abortException.isSilent) {
      if (abortException.silentException == null) {
        return;
      }
      throw abortException.silentException;
    }

    CompilationResult result = abortException.compilationResult;
    if ((result == null) && (unit != null)) {
      result = unit.compilationResult;
    }

    if ((result == null) && (this.lookupEnvironment.unitBeingCompleted != null)) {
      result = this.lookupEnvironment.unitBeingCompleted.compilationResult;
    }
    if ((result == null) && (this.unitsToProcess != null) && (this.totalUnits > 0)) {
      result = this.unitsToProcess[(this.totalUnits - 1)].compilationResult;
    }
    if ((result != null) && (!result.hasBeenAccepted))
    {
      if (abortException.problem != null)
      {
        CategorizedProblem distantProblem = abortException.problem;
        CategorizedProblem[] knownProblems = result.problems;
        for (int i = 0; i < result.problemCount; i++) {
          if (knownProblems[i] == distantProblem) {
            break;
          }
        }
        if ((distantProblem instanceof DefaultProblem)) {
          ((DefaultProblem)distantProblem).setOriginatingFileName(result.getFileName());
        }
        result.record(distantProblem, unit);
      }
      else if (abortException.exception != null) {
        handleInternalException(abortException.exception, null, result);
        return;
      }

      if (!result.hasBeenAccepted)
        this.requestor.acceptResult(result.tagAsAccepted());
    }
    else {
      abortException.printStackTrace();
    }
  }

  public void initializeParser()
  {
    this.parser = new Parser(this.problemReporter, this.options.parseLiteralExpressionsAsConstants);
  }

  protected void internalBeginToCompile(ICompilationUnit[] sourceUnits, int maxUnits)
  {
    for (int i = 0; i < maxUnits; i++) {
      CompilationResult unitResult = 
        new CompilationResult(sourceUnits[i], i, maxUnits, this.options.maxProblemsPerUnit);
      CompilationUnitDeclaration parsedUnit;
      try { if (this.options.verbose)
          this.out.println(
            Messages.bind(Messages.compilation_request, 
            new String[] { 
            String.valueOf(i + 1), 
            String.valueOf(maxUnits), 
            new String(sourceUnits[i].getFileName()) }));
        CompilationUnitDeclaration parsedUnit;
        CompilationUnitDeclaration parsedUnit;
        if (this.totalUnits < this.parseThreshold)
          parsedUnit = this.parser.parse(sourceUnits[i], unitResult);
        else {
          parsedUnit = this.parser.dietParse(sourceUnits[i], unitResult);
        }

        this.lookupEnvironment.buildTypeBindings(parsedUnit, null);
        addCompilationUnit(sourceUnits[i], parsedUnit);
        ImportReference currentPackage = parsedUnit.currentPackage;
        if (currentPackage != null) {
          unitResult.recordPackageName(currentPackage.tokens);
        }
      }
      finally
      {
        sourceUnits[i] = null;
      }
    }

    this.lookupEnvironment.completeTypeBindings();
  }

  public void process(CompilationUnitDeclaration unit, int i)
  {
    this.lookupEnvironment.unitBeingCompleted = unit;

    this.parser.getMethodBodies(unit);

    if (unit.scope != null) {
      unit.scope.faultInTypes();
    }

    if (unit.scope != null) {
      unit.scope.verifyMethods(this.lookupEnvironment.methodVerifier());
    }

    unit.resolve();

    unit.analyseCode();

    unit.generateCode();

    if ((this.options.produceReferenceInfo) && (unit.scope != null)) {
      unit.scope.storeDependencyInfo();
    }

    unit.compilationResult.totalUnitsKnown = this.totalUnits;

    this.lookupEnvironment.unitBeingCompleted = null;
  }

  protected void processAnnotations() {
    int newUnitSize = 0;
    int newClassFilesSize = 0;
    int bottom = 0;
    int top = this.unitsToProcess.length;
    ReferenceBinding[] binaryTypeBindingsTemp = this.referenceBindings;
    if ((top == 0) && (binaryTypeBindingsTemp == null)) return;
    this.referenceBindings = null;
    do
    {
      int length = top - bottom;
      CompilationUnitDeclaration[] currentUnits = new CompilationUnitDeclaration[length];
      int index = 0;
      for (int i = bottom; i < top; i++) {
        CompilationUnitDeclaration currentUnit = this.unitsToProcess[i];
        if ((currentUnit.bits & 0x1) == 0) {
          currentUnits[(index++)] = currentUnit;
        }
      }
      if (index != length) {
        System.arraycopy(currentUnits, 0, currentUnits = new CompilationUnitDeclaration[index], 0, index);
      }
      this.annotationProcessorManager.processAnnotations(currentUnits, binaryTypeBindingsTemp, false);
      ICompilationUnit[] newUnits = this.annotationProcessorManager.getNewUnits();
      newUnitSize = newUnits.length;
      ReferenceBinding[] newClassFiles = this.annotationProcessorManager.getNewClassFiles();
      binaryTypeBindingsTemp = newClassFiles;
      newClassFilesSize = newClassFiles.length;
      if (newUnitSize != 0)
      {
        internalBeginToCompile(newUnits, newUnitSize);
        bottom = top;
        top = this.totalUnits;
      } else {
        bottom = top;
      }
      this.annotationProcessorManager.reset();
    }while ((newUnitSize != 0) || (newClassFilesSize != 0));

    this.annotationProcessorManager.processAnnotations(null, null, true);
  }

  public void reset()
  {
    this.lookupEnvironment.reset();
    this.parser.scanner.source = null;
    this.unitsToProcess = null;
    if (DebugRequestor != null) DebugRequestor.reset();
    this.problemReporter.reset();
  }

  public CompilationUnitDeclaration resolve(CompilationUnitDeclaration unit, ICompilationUnit sourceUnit, boolean verifyMethods, boolean analyzeCode, boolean generateCode)
  {
    try
    {
      if (unit == null)
      {
        this.parseThreshold = 0;
        beginToCompile(new ICompilationUnit[] { sourceUnit });

        unit = this.unitsToProcess[0];
      }
      else {
        this.lookupEnvironment.buildTypeBindings(unit, null);

        this.lookupEnvironment.completeTypeBindings();
      }
      this.lookupEnvironment.unitBeingCompleted = unit;
      this.parser.getMethodBodies(unit);
      if (unit.scope != null)
      {
        unit.scope.faultInTypes();
        if ((unit.scope != null) && (verifyMethods))
        {
          unit.scope.verifyMethods(this.lookupEnvironment.methodVerifier());
        }

        unit.resolve();

        if (analyzeCode) unit.analyseCode();

        if (generateCode) unit.generateCode();
      }
      if (this.unitsToProcess != null) this.unitsToProcess[0] = null;
      this.requestor.acceptResult(unit.compilationResult.tagAsAccepted());
      return unit;
    } catch (AbortCompilation e) {
      handleInternalException(e, unit);
      return unit == null ? this.unitsToProcess[0] : unit;
    } catch (Error e) {
      handleInternalException(e, unit, null);
      throw e;
    } catch (RuntimeException e) {
      handleInternalException(e, unit, null);
      throw e;
    }
  }

  public CompilationUnitDeclaration resolve(ICompilationUnit sourceUnit, boolean verifyMethods, boolean analyzeCode, boolean generateCode)
  {
    return resolve(
      null, 
      sourceUnit, 
      verifyMethods, 
      analyzeCode, 
      generateCode);
  }
}