package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilationUnit;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.AbortType;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;

public abstract class AbstractMethodDeclaration extends ASTNode
  implements ProblemSeverities, ReferenceContext
{
  public MethodScope scope;
  public char[] selector;
  public int declarationSourceStart;
  public int declarationSourceEnd;
  public int modifiers;
  public int modifiersSourceStart;
  public Annotation[] annotations;
  public Argument[] arguments;
  public TypeReference[] thrownExceptions;
  public Statement[] statements;
  public int explicitDeclarations;
  public MethodBinding binding;
  public boolean ignoreFurtherInvestigation = false;
  public Javadoc javadoc;
  public int bodyStart;
  public int bodyEnd = -1;
  public CompilationResult compilationResult;

  AbstractMethodDeclaration(CompilationResult compilationResult)
  {
    this.compilationResult = compilationResult;
  }

  public void abort(int abortLevel, CategorizedProblem problem)
  {
    switch (abortLevel) {
    case 2:
      throw new AbortCompilation(this.compilationResult, problem);
    case 4:
      throw new AbortCompilationUnit(this.compilationResult, problem);
    case 8:
      throw new AbortType(this.compilationResult, problem);
    case 3:
    case 5:
    case 6:
    case 7: } throw new AbortMethod(this.compilationResult, problem);
  }

  public abstract void analyseCode(ClassScope paramClassScope, InitializationFlowContext paramInitializationFlowContext, FlowInfo paramFlowInfo);

  public void bindArguments()
  {
    if (this.arguments != null)
    {
      if (this.binding == null) {
        int i = 0; for (int length = this.arguments.length; i < length; i++) {
          this.arguments[i].bind(this.scope, null, true);
        }
        return;
      }
      boolean used = (this.binding.isAbstract()) || (this.binding.isNative());
      AnnotationBinding[][] paramAnnotations = (AnnotationBinding[][])null;
      int i = 0; for (int length = this.arguments.length; i < length; i++) {
        Argument argument = this.arguments[i];
        argument.bind(this.scope, this.binding.parameters[i], used);
        if (argument.annotations != null) {
          this.binding.tagBits |= 1024L;
          if (paramAnnotations == null)
            paramAnnotations = new AnnotationBinding[length][];
          paramAnnotations[i] = argument.binding.getAnnotations();
        }
      }
      if (paramAnnotations != null)
        this.binding.setParameterAnnotations(paramAnnotations);
    }
  }

  public void bindThrownExceptions()
  {
    if ((this.thrownExceptions != null) && 
      (this.binding != null) && 
      (this.binding.thrownExceptions != null)) {
      int thrownExceptionLength = this.thrownExceptions.length;
      int length = this.binding.thrownExceptions.length;
      if (length == thrownExceptionLength) {
        for (int i = 0; i < length; i++)
          this.thrownExceptions[i].resolvedType = this.binding.thrownExceptions[i];
      }
      else {
        int bindingIndex = 0;
        for (int i = 0; (i < thrownExceptionLength) && (bindingIndex < length); i++) {
          TypeReference thrownException = this.thrownExceptions[i];
          ReferenceBinding thrownExceptionBinding = this.binding.thrownExceptions[bindingIndex];
          char[][] bindingCompoundName = thrownExceptionBinding.compoundName;
          if (bindingCompoundName != null)
            if ((thrownException instanceof SingleTypeReference))
            {
              int lengthName = bindingCompoundName.length;
              char[] thrownExceptionTypeName = thrownException.getTypeName()[0];
              if (CharOperation.equals(thrownExceptionTypeName, bindingCompoundName[(lengthName - 1)])) {
                thrownException.resolvedType = thrownExceptionBinding;
                bindingIndex++;
              }

            }
            else if (CharOperation.equals(thrownException.getTypeName(), bindingCompoundName)) {
              thrownException.resolvedType = thrownExceptionBinding;
              bindingIndex++;
            }
        }
      }
    }
  }

  public CompilationResult compilationResult()
  {
    return this.compilationResult;
  }

  public void generateCode(ClassScope classScope, ClassFile classFile)
  {
    int problemResetPC = 0;
    classFile.codeStream.wideMode = false;
    if (this.ignoreFurtherInvestigation)
    {
      if (this.binding == null) {
        return;
      }
      CategorizedProblem[] problems = 
        this.scope.referenceCompilationUnit().compilationResult.getProblems();
      int problemsLength;
      CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
      System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
      classFile.addProblemMethod(this, this.binding, problemsCopy);
      return;
    }
    try
    {
      problemResetPC = classFile.contentsOffset;
      generateCode(classFile);
    }
    catch (AbortMethod e) {
      if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE)
      {
        try {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount -= 1;
          classFile.codeStream.wideMode = true;
          generateCode(classFile);
        }
        catch (AbortMethod localAbortMethod1) {
          CategorizedProblem[] problems = 
            this.scope.referenceCompilationUnit().compilationResult.getAllProblems();
          int problemsLength;
          CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
          System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
          classFile.addProblemMethod(this, this.binding, problemsCopy, problemResetPC);
        }
      }
      else
      {
        CategorizedProblem[] problems = 
          this.scope.referenceCompilationUnit().compilationResult.getAllProblems();
        int problemsLength;
        CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
        System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
        classFile.addProblemMethod(this, this.binding, problemsCopy, problemResetPC);
      }
    }
  }

  public void generateCode(ClassFile classFile)
  {
    classFile.generateMethodInfoHeader(this.binding);
    int methodAttributeOffset = classFile.contentsOffset;
    int attributeNumber = classFile.generateMethodInfoAttribute(this.binding);
    if ((!this.binding.isNative()) && (!this.binding.isAbstract())) {
      int codeAttributeOffset = classFile.contentsOffset;
      classFile.generateCodeAttributeHeader();
      CodeStream codeStream = classFile.codeStream;
      codeStream.reset(this, classFile);

      this.scope.computeLocalVariablePositions(this.binding.isStatic() ? 0 : 1, codeStream);

      if (this.arguments != null) {
        int i = 0; for (int max = this.arguments.length; i < max; i++)
        {
          LocalVariableBinding argBinding;
          codeStream.addVisibleLocalVariable(argBinding = this.arguments[i].binding);
          argBinding.recordInitializationStartPC(0);
        }
      }
      if (this.statements != null) {
        int i = 0; for (int max = this.statements.length; i < max; i++)
          this.statements[i].generateCode(this.scope, codeStream);
      }
      if ((this.bits & 0x40) != 0) {
        codeStream.return_();
      }

      codeStream.exitUserScope(this.scope);
      codeStream.recordPositionsFrom(0, this.declarationSourceEnd);
      classFile.completeCodeAttribute(codeAttributeOffset);
      attributeNumber++;
    } else {
      checkArgumentsSize();
    }
    classFile.completeMethodInfo(methodAttributeOffset, attributeNumber);

    if (this.ignoreFurtherInvestigation)
      throw new AbortMethod(this.scope.referenceCompilationUnit().compilationResult, null);
  }

  private void checkArgumentsSize()
  {
    TypeBinding[] parameters = this.binding.parameters;
    int size = 1;
    int i = 0; for (int max = parameters.length; i < max; i++) {
      TypeBinding parameter = parameters[i];
      if ((parameter == TypeBinding.LONG) || (parameter == TypeBinding.DOUBLE))
        size += 2;
      else {
        size++;
      }
      if (size > 255)
        this.scope.problemReporter().noMoreAvailableSpaceForArgument(this.scope.locals[i], this.scope.locals[i].declaration);
    }
  }

  public boolean hasErrors()
  {
    return this.ignoreFurtherInvestigation;
  }

  public boolean isAbstract()
  {
    if (this.binding != null)
      return this.binding.isAbstract();
    return (this.modifiers & 0x400) != 0;
  }

  public boolean isAnnotationMethod()
  {
    return false;
  }

  public boolean isClinit()
  {
    return false;
  }

  public boolean isConstructor()
  {
    return false;
  }

  public boolean isDefaultConstructor()
  {
    return false;
  }

  public boolean isInitializationMethod()
  {
    return false;
  }

  public boolean isMethod()
  {
    return false;
  }

  public boolean isNative()
  {
    if (this.binding != null)
      return this.binding.isNative();
    return (this.modifiers & 0x100) != 0;
  }

  public boolean isStatic()
  {
    if (this.binding != null)
      return this.binding.isStatic();
    return (this.modifiers & 0x8) != 0;
  }

  public abstract void parseStatements(Parser paramParser, CompilationUnitDeclaration paramCompilationUnitDeclaration);

  public StringBuffer print(int tab, StringBuffer output)
  {
    if (this.javadoc != null) {
      this.javadoc.print(tab, output);
    }
    printIndent(tab, output);
    printModifiers(this.modifiers, output);
    if (this.annotations != null) printAnnotations(this.annotations, output);

    TypeParameter[] typeParams = typeParameters();
    if (typeParams != null) {
      output.append('<');
      int max = typeParams.length - 1;
      for (int j = 0; j < max; j++) {
        typeParams[j].print(0, output);
        output.append(", ");
      }
      typeParams[max].print(0, output);
      output.append('>');
    }

    printReturnType(0, output).append(this.selector).append('(');
    if (this.arguments != null) {
      for (int i = 0; i < this.arguments.length; i++) {
        if (i > 0) output.append(", ");
        this.arguments[i].print(0, output);
      }
    }
    output.append(')');
    if (this.thrownExceptions != null) {
      output.append(" throws ");
      for (int i = 0; i < this.thrownExceptions.length; i++) {
        if (i > 0) output.append(", ");
        this.thrownExceptions[i].print(0, output);
      }
    }
    printBody(tab + 1, output);
    return output;
  }

  public StringBuffer printBody(int indent, StringBuffer output)
  {
    if ((isAbstract()) || ((this.modifiers & 0x1000000) != 0)) {
      return output.append(';');
    }
    output.append(" {");
    if (this.statements != null) {
      for (int i = 0; i < this.statements.length; i++) {
        output.append('\n');
        this.statements[i].printStatement(indent, output);
      }
    }
    output.append('\n');
    printIndent(indent == 0 ? 0 : indent - 1, output).append('}');
    return output;
  }

  public StringBuffer printReturnType(int indent, StringBuffer output)
  {
    return output;
  }

  public void resolve(ClassScope upperScope)
  {
    if (this.binding == null) {
      this.ignoreFurtherInvestigation = true;
    }
    try
    {
      bindArguments();
      bindThrownExceptions();
      resolveJavadoc();
      resolveAnnotations(this.scope, this.annotations, this.binding);
      resolveStatements();

      if ((this.binding != null) && 
        ((this.binding.getAnnotationTagBits() & 0x0) == 0L) && 
        ((this.binding.modifiers & 0x100000) != 0) && 
        (this.scope.compilerOptions().sourceLevel >= 3211264L))
        this.scope.problemReporter().missingDeprecatedAnnotationForMethod(this);
    }
    catch (AbortMethod localAbortMethod) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public void resolveJavadoc()
  {
    if (this.binding == null) return;
    if (this.javadoc != null) {
      this.javadoc.resolve(this.scope);
      return;
    }
    if ((this.binding.declaringClass != null) && (!this.binding.declaringClass.isLocalType()))
      this.scope.problemReporter().javadocMissing(this.sourceStart, this.sourceEnd, this.binding.modifiers);
  }

  public void resolveStatements()
  {
    if (this.statements != null) {
      int i = 0; for (int length = this.statements.length; i < length; i++)
        this.statements[i].resolve(this.scope);
    }
    else if ((this.bits & 0x8) != 0) {
      this.scope.problemReporter().undocumentedEmptyBlock(this.bodyStart - 1, this.bodyEnd + 1);
    }
  }

  public void tagAsHavingErrors()
  {
    this.ignoreFurtherInvestigation = true;
  }

  public void traverse(ASTVisitor visitor, ClassScope classScope)
  {
  }

  public TypeParameter[] typeParameters()
  {
    return null;
  }
}