package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ConditionalExpression extends OperatorExpression
{
  public Expression condition;
  public Expression valueIfTrue;
  public Expression valueIfFalse;
  public Constant optimizedBooleanConstant;
  public Constant optimizedIfTrueConstant;
  public Constant optimizedIfFalseConstant;
  int trueInitStateIndex = -1;
  int falseInitStateIndex = -1;
  int mergedInitStateIndex = -1;

  public ConditionalExpression(Expression condition, Expression valueIfTrue, Expression valueIfFalse)
  {
    this.condition = condition;
    this.valueIfTrue = valueIfTrue;
    this.valueIfFalse = valueIfFalse;
    this.sourceStart = condition.sourceStart;
    this.sourceEnd = valueIfFalse.sourceEnd;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    Constant cst = this.condition.optimizedBooleanConstant();
    boolean isConditionOptimizedTrue = (cst != Constant.NotAConstant) && (cst.booleanValue());
    boolean isConditionOptimizedFalse = (cst != Constant.NotAConstant) && (!cst.booleanValue());

    int mode = flowInfo.reachMode();
    flowInfo = this.condition.analyseCode(currentScope, flowContext, flowInfo, cst == Constant.NotAConstant);

    FlowInfo trueFlowInfo = flowInfo.initsWhenTrue().copy();
    if (isConditionOptimizedFalse) {
      trueFlowInfo.setReachMode(1);
    }
    this.trueInitStateIndex = currentScope.methodScope().recordInitializationStates(trueFlowInfo);
    trueFlowInfo = this.valueIfTrue.analyseCode(currentScope, flowContext, trueFlowInfo);

    FlowInfo falseFlowInfo = flowInfo.initsWhenFalse().copy();
    if (isConditionOptimizedTrue) {
      falseFlowInfo.setReachMode(1);
    }
    this.falseInitStateIndex = currentScope.methodScope().recordInitializationStates(falseFlowInfo);
    falseFlowInfo = this.valueIfFalse.analyseCode(currentScope, flowContext, falseFlowInfo);
    FlowInfo mergedInfo;
    FlowInfo mergedInfo;
    if (isConditionOptimizedTrue) {
      mergedInfo = trueFlowInfo.addPotentialInitializationsFrom(falseFlowInfo);
    }
    else
    {
      FlowInfo mergedInfo;
      if (isConditionOptimizedFalse) {
        mergedInfo = falseFlowInfo.addPotentialInitializationsFrom(trueFlowInfo);
      }
      else {
        cst = this.optimizedIfTrueConstant;
        boolean isValueIfTrueOptimizedTrue = (cst != null) && (cst != Constant.NotAConstant) && (cst.booleanValue());
        boolean isValueIfTrueOptimizedFalse = (cst != null) && (cst != Constant.NotAConstant) && (!cst.booleanValue());

        cst = this.optimizedIfFalseConstant;
        boolean isValueIfFalseOptimizedTrue = (cst != null) && (cst != Constant.NotAConstant) && (cst.booleanValue());
        boolean isValueIfFalseOptimizedFalse = (cst != null) && (cst != Constant.NotAConstant) && (!cst.booleanValue());

        UnconditionalFlowInfo trueInfoWhenTrue = trueFlowInfo.initsWhenTrue().unconditionalCopy();
        UnconditionalFlowInfo falseInfoWhenTrue = falseFlowInfo.initsWhenTrue().unconditionalCopy();
        UnconditionalFlowInfo trueInfoWhenFalse = trueFlowInfo.initsWhenFalse().unconditionalInits();
        UnconditionalFlowInfo falseInfoWhenFalse = falseFlowInfo.initsWhenFalse().unconditionalInits();
        if (isValueIfTrueOptimizedFalse) trueInfoWhenTrue.setReachMode(1);
        if (isValueIfFalseOptimizedFalse) falseInfoWhenTrue.setReachMode(1);
        if (isValueIfTrueOptimizedTrue) trueInfoWhenFalse.setReachMode(1);
        if (isValueIfFalseOptimizedTrue) falseInfoWhenFalse.setReachMode(1);

        mergedInfo = 
          FlowInfo.conditional(
          trueInfoWhenTrue.mergedWith(falseInfoWhenTrue), 
          trueInfoWhenFalse.mergedWith(falseInfoWhenFalse));
      }
    }
    this.mergedInitStateIndex = 
      currentScope.methodScope().recordInitializationStates(mergedInfo);
    mergedInfo.setReachMode(mode);
    return mergedInfo;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;

    if (this.constant != Constant.NotAConstant) {
      if (valueRequired)
        codeStream.generateConstant(this.constant, this.implicitConversion);
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }
    Constant cst = this.condition.optimizedBooleanConstant();
    boolean needTruePart = (cst == Constant.NotAConstant) || (cst.booleanValue());
    boolean needFalsePart = (cst == Constant.NotAConstant) || (!cst.booleanValue());
    BranchLabel endifLabel = new BranchLabel(codeStream);

    BranchLabel falseLabel = new BranchLabel(codeStream);
    falseLabel.tagBits |= 2;
    this.condition.generateOptimizedBoolean(
      currentScope, 
      codeStream, 
      null, 
      falseLabel, 
      cst == Constant.NotAConstant);

    if (this.trueInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(
        currentScope, 
        this.trueInitStateIndex);
      codeStream.addDefinitelyAssignedVariables(currentScope, this.trueInitStateIndex);
    }

    if (needTruePart) {
      this.valueIfTrue.generateCode(currentScope, codeStream, valueRequired);
      if (needFalsePart)
      {
        int position = codeStream.position;
        codeStream.goto_(endifLabel);
        codeStream.updateLastRecordedEndPC(currentScope, position);

        if (valueRequired) {
          codeStream.decrStackSize((this.resolvedType == TypeBinding.LONG) || (this.resolvedType == TypeBinding.DOUBLE) ? 2 : 1);
        }
      }
    }
    if (needFalsePart) {
      if (this.falseInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables(
          currentScope, 
          this.falseInitStateIndex);
        codeStream.addDefinitelyAssignedVariables(currentScope, this.falseInitStateIndex);
      }
      if (falseLabel.forwardReferenceCount() > 0) {
        falseLabel.place();
      }
      this.valueIfFalse.generateCode(currentScope, codeStream, valueRequired);
      if (valueRequired) {
        codeStream.recordExpressionType(this.resolvedType);
      }
      if (needTruePart)
      {
        endifLabel.place();
      }
    }

    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(
        currentScope, 
        this.mergedInitStateIndex);
    }

    if (valueRequired)
      codeStream.generateImplicitConversion(this.implicitConversion);
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public void generateOptimizedBoolean(BlockScope currentScope, CodeStream codeStream, BranchLabel trueLabel, BranchLabel falseLabel, boolean valueRequired)
  {
    if (((this.constant != Constant.NotAConstant) && (this.constant.typeID() == 5)) || 
      ((this.valueIfTrue.implicitConversion & 0xFF) >> 4 != 5)) {
      super.generateOptimizedBoolean(currentScope, codeStream, trueLabel, falseLabel, valueRequired);
      return;
    }
    Constant cst = this.condition.constant;
    Constant condCst = this.condition.optimizedBooleanConstant();
    boolean needTruePart = 
      ((cst == Constant.NotAConstant) || (cst.booleanValue())) && (
      (condCst == Constant.NotAConstant) || (
      condCst.booleanValue()));
    boolean needFalsePart = 
      ((cst == Constant.NotAConstant) || (!cst.booleanValue())) && (
      (condCst == Constant.NotAConstant) || (!
      condCst.booleanValue()));

    BranchLabel endifLabel = new BranchLabel(codeStream);

    boolean needConditionValue = (cst == Constant.NotAConstant) && (condCst == Constant.NotAConstant);
    BranchLabel internalFalseLabel;
    this.condition.generateOptimizedBoolean(
      currentScope, 
      codeStream, 
      null, 
      internalFalseLabel = new BranchLabel(codeStream), 
      needConditionValue);

    if (this.trueInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(
        currentScope, 
        this.trueInitStateIndex);
      codeStream.addDefinitelyAssignedVariables(currentScope, this.trueInitStateIndex);
    }

    if (needTruePart) {
      this.valueIfTrue.generateOptimizedBoolean(currentScope, codeStream, trueLabel, falseLabel, valueRequired);

      if (needFalsePart)
      {
        if (falseLabel == null) {
          if (trueLabel != null)
          {
            cst = this.optimizedIfTrueConstant;
            boolean isValueIfTrueOptimizedTrue = (cst != null) && (cst != Constant.NotAConstant) && (cst.booleanValue());
            if (isValueIfTrueOptimizedTrue) break label368;
          }

        }
        else if (trueLabel == null) {
          cst = this.optimizedIfTrueConstant;
          boolean isValueIfTrueOptimizedFalse = (cst != null) && (cst != Constant.NotAConstant) && (!cst.booleanValue());
          if (isValueIfTrueOptimizedFalse)
          {
            break label368;
          }
        }
        int position = codeStream.position;
        codeStream.goto_(endifLabel);
        codeStream.updateLastRecordedEndPC(currentScope, position);
      }

    }

    label368: if (needFalsePart) {
      internalFalseLabel.place();
      if (this.falseInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.falseInitStateIndex);
        codeStream.addDefinitelyAssignedVariables(currentScope, this.falseInitStateIndex);
      }
      this.valueIfFalse.generateOptimizedBoolean(currentScope, codeStream, trueLabel, falseLabel, valueRequired);

      endifLabel.place();
    }

    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
    }

    codeStream.updateLastRecordedEndPC(currentScope, codeStream.position);
  }

  public int nullStatus(FlowInfo flowInfo) {
    Constant cst = this.condition.optimizedBooleanConstant();
    if (cst != Constant.NotAConstant) {
      if (cst.booleanValue()) {
        return this.valueIfTrue.nullStatus(flowInfo);
      }
      return this.valueIfFalse.nullStatus(flowInfo);
    }
    int ifTrueNullStatus = this.valueIfTrue.nullStatus(flowInfo);
    int ifFalseNullStatus = this.valueIfFalse.nullStatus(flowInfo);
    if (ifTrueNullStatus == ifFalseNullStatus) {
      return ifTrueNullStatus;
    }
    return 0;
  }

  public Constant optimizedBooleanConstant()
  {
    return this.optimizedBooleanConstant == null ? this.constant : this.optimizedBooleanConstant;
  }

  public StringBuffer printExpressionNoParenthesis(int indent, StringBuffer output)
  {
    this.condition.printExpression(indent, output).append(" ? ");
    this.valueIfTrue.printExpression(0, output).append(" : ");
    return this.valueIfFalse.printExpression(0, output);
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    LookupEnvironment env = scope.environment();
    boolean use15specifics = scope.compilerOptions().sourceLevel >= 3211264L;
    TypeBinding conditionType = this.condition.resolveTypeExpecting(scope, TypeBinding.BOOLEAN);
    this.condition.computeConversion(scope, TypeBinding.BOOLEAN, conditionType);

    if ((this.valueIfTrue instanceof CastExpression)) this.valueIfTrue.bits |= 32;
    TypeBinding originalValueIfTrueType = this.valueIfTrue.resolveType(scope);

    if ((this.valueIfFalse instanceof CastExpression)) this.valueIfFalse.bits |= 32;
    TypeBinding originalValueIfFalseType = this.valueIfFalse.resolveType(scope);

    if ((conditionType == null) || (originalValueIfTrueType == null) || (originalValueIfFalseType == null)) {
      return null;
    }
    TypeBinding valueIfTrueType = originalValueIfTrueType;
    TypeBinding valueIfFalseType = originalValueIfFalseType;
    if ((use15specifics) && (valueIfTrueType != valueIfFalseType))
      if (valueIfTrueType.isBaseType()) {
        if (valueIfFalseType.isBaseType())
        {
          if (valueIfTrueType == TypeBinding.NULL)
            valueIfFalseType = env.computeBoxingType(valueIfFalseType);
          else if (valueIfFalseType == TypeBinding.NULL)
            valueIfTrueType = env.computeBoxingType(valueIfTrueType);
        }
        else
        {
          TypeBinding unboxedIfFalseType = valueIfFalseType.isBaseType() ? valueIfFalseType : env.computeBoxingType(valueIfFalseType);
          if ((valueIfTrueType.isNumericType()) && (unboxedIfFalseType.isNumericType()))
            valueIfFalseType = unboxedIfFalseType;
          else if (valueIfTrueType != TypeBinding.NULL)
            valueIfFalseType = env.computeBoxingType(valueIfFalseType);
        }
      }
      else if (valueIfFalseType.isBaseType())
      {
        TypeBinding unboxedIfTrueType = valueIfTrueType.isBaseType() ? valueIfTrueType : env.computeBoxingType(valueIfTrueType);
        if ((unboxedIfTrueType.isNumericType()) && (valueIfFalseType.isNumericType()))
          valueIfTrueType = unboxedIfTrueType;
        else if (valueIfFalseType != TypeBinding.NULL)
          valueIfTrueType = env.computeBoxingType(valueIfTrueType);
      }
      else
      {
        TypeBinding unboxedIfTrueType = env.computeBoxingType(valueIfTrueType);
        TypeBinding unboxedIfFalseType = env.computeBoxingType(valueIfFalseType);
        if ((unboxedIfTrueType.isNumericType()) && (unboxedIfFalseType.isNumericType())) {
          valueIfTrueType = unboxedIfTrueType;
          valueIfFalseType = unboxedIfFalseType;
        }
      }
    Constant condConstant;
    Constant trueConstant;
    Constant falseConstant;
    if (((condConstant = this.condition.constant) != Constant.NotAConstant) && 
      ((trueConstant = this.valueIfTrue.constant) != Constant.NotAConstant) && 
      ((falseConstant = this.valueIfFalse.constant) != Constant.NotAConstant))
    {
      this.constant = (condConstant.booleanValue() ? trueConstant : falseConstant);
    }
    if (valueIfTrueType == valueIfFalseType) {
      this.valueIfTrue.computeConversion(scope, valueIfTrueType, originalValueIfTrueType);
      this.valueIfFalse.computeConversion(scope, valueIfFalseType, originalValueIfFalseType);
      if (valueIfTrueType == TypeBinding.BOOLEAN) {
        this.optimizedIfTrueConstant = this.valueIfTrue.optimizedBooleanConstant();
        this.optimizedIfFalseConstant = this.valueIfFalse.optimizedBooleanConstant();
        if ((this.optimizedIfTrueConstant != Constant.NotAConstant) && 
          (this.optimizedIfFalseConstant != Constant.NotAConstant) && 
          (this.optimizedIfTrueConstant.booleanValue() == this.optimizedIfFalseConstant.booleanValue()))
        {
          this.optimizedBooleanConstant = this.optimizedIfTrueConstant;
        } else if ((condConstant = this.condition.optimizedBooleanConstant()) != Constant.NotAConstant) {
          this.optimizedBooleanConstant = (condConstant.booleanValue() ? 
            this.optimizedIfTrueConstant : 
            this.optimizedIfFalseConstant);
        }
      }
      return this.resolvedType = valueIfTrueType;
    }

    if ((valueIfTrueType.isNumericType()) && (valueIfFalseType.isNumericType()))
    {
      if (((valueIfTrueType == TypeBinding.BYTE) && (valueIfFalseType == TypeBinding.SHORT)) || (
        (valueIfTrueType == TypeBinding.SHORT) && (valueIfFalseType == TypeBinding.BYTE))) {
        this.valueIfTrue.computeConversion(scope, TypeBinding.SHORT, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, TypeBinding.SHORT, originalValueIfFalseType);
        return this.resolvedType = TypeBinding.SHORT;
      }

      if (((valueIfTrueType == TypeBinding.BYTE) || (valueIfTrueType == TypeBinding.SHORT) || (valueIfTrueType == TypeBinding.CHAR)) && 
        (valueIfFalseType == TypeBinding.INT) && 
        (this.valueIfFalse.isConstantValueOfTypeAssignableToType(valueIfFalseType, valueIfTrueType))) {
        this.valueIfTrue.computeConversion(scope, valueIfTrueType, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, valueIfTrueType, originalValueIfFalseType);
        return this.resolvedType = valueIfTrueType;
      }
      if (((valueIfFalseType == TypeBinding.BYTE) || 
        (valueIfFalseType == TypeBinding.SHORT) || 
        (valueIfFalseType == TypeBinding.CHAR)) && 
        (valueIfTrueType == TypeBinding.INT) && 
        (this.valueIfTrue.isConstantValueOfTypeAssignableToType(valueIfTrueType, valueIfFalseType))) {
        this.valueIfTrue.computeConversion(scope, valueIfFalseType, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, valueIfFalseType, originalValueIfFalseType);
        return this.resolvedType = valueIfFalseType;
      }

      if ((BaseTypeBinding.isNarrowing(valueIfTrueType.id, 10)) && 
        (BaseTypeBinding.isNarrowing(valueIfFalseType.id, 10))) {
        this.valueIfTrue.computeConversion(scope, TypeBinding.INT, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, TypeBinding.INT, originalValueIfFalseType);
        return this.resolvedType = TypeBinding.INT;
      }

      if ((BaseTypeBinding.isNarrowing(valueIfTrueType.id, 7)) && 
        (BaseTypeBinding.isNarrowing(valueIfFalseType.id, 7))) {
        this.valueIfTrue.computeConversion(scope, TypeBinding.LONG, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, TypeBinding.LONG, originalValueIfFalseType);
        return this.resolvedType = TypeBinding.LONG;
      }

      if ((BaseTypeBinding.isNarrowing(valueIfTrueType.id, 9)) && 
        (BaseTypeBinding.isNarrowing(valueIfFalseType.id, 9))) {
        this.valueIfTrue.computeConversion(scope, TypeBinding.FLOAT, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, TypeBinding.FLOAT, originalValueIfFalseType);
        return this.resolvedType = TypeBinding.FLOAT;
      }

      this.valueIfTrue.computeConversion(scope, TypeBinding.DOUBLE, originalValueIfTrueType);
      this.valueIfFalse.computeConversion(scope, TypeBinding.DOUBLE, originalValueIfFalseType);
      return this.resolvedType = TypeBinding.DOUBLE;
    }

    if ((valueIfTrueType.isBaseType()) && (valueIfTrueType != TypeBinding.NULL)) {
      if (use15specifics) {
        valueIfTrueType = env.computeBoxingType(valueIfTrueType);
      } else {
        scope.problemReporter().conditionalArgumentsIncompatibleTypes(this, valueIfTrueType, valueIfFalseType);
        return null;
      }
    }
    if ((valueIfFalseType.isBaseType()) && (valueIfFalseType != TypeBinding.NULL)) {
      if (use15specifics) {
        valueIfFalseType = env.computeBoxingType(valueIfFalseType);
      } else {
        scope.problemReporter().conditionalArgumentsIncompatibleTypes(this, valueIfTrueType, valueIfFalseType);
        return null;
      }
    }
    if (use15specifics)
    {
      TypeBinding commonType = null;
      if (valueIfTrueType == TypeBinding.NULL)
        commonType = valueIfFalseType;
      else if (valueIfFalseType == TypeBinding.NULL)
        commonType = valueIfTrueType;
      else {
        commonType = scope.lowerUpperBound(new TypeBinding[] { valueIfTrueType, valueIfFalseType });
      }
      if (commonType != null) {
        this.valueIfTrue.computeConversion(scope, commonType, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, commonType, originalValueIfFalseType);
        return this.resolvedType = commonType.capture(scope, this.sourceEnd);
      }
    }
    else {
      if (valueIfFalseType.isCompatibleWith(valueIfTrueType)) {
        this.valueIfTrue.computeConversion(scope, valueIfTrueType, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, valueIfTrueType, originalValueIfFalseType);
        return this.resolvedType = valueIfTrueType;
      }if (valueIfTrueType.isCompatibleWith(valueIfFalseType)) {
        this.valueIfTrue.computeConversion(scope, valueIfFalseType, originalValueIfTrueType);
        this.valueIfFalse.computeConversion(scope, valueIfFalseType, originalValueIfFalseType);
        return this.resolvedType = valueIfFalseType;
      }
    }
    scope.problemReporter().conditionalArgumentsIncompatibleTypes(
      this, 
      valueIfTrueType, 
      valueIfFalseType);
    return null;
  }

  public void traverse(ASTVisitor visitor, BlockScope scope) {
    if (visitor.visit(this, scope)) {
      this.condition.traverse(visitor, scope);
      this.valueIfTrue.traverse(visitor, scope);
      this.valueIfFalse.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}