package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class FieldReference extends Reference
  implements InvocationSite
{
  public static final int READ = 0;
  public static final int WRITE = 1;
  public Expression receiver;
  public char[] token;
  public FieldBinding binding;
  protected FieldBinding codegenBinding;
  public MethodBinding[] syntheticAccessors;
  public long nameSourcePosition;
  public TypeBinding receiverType;
  public TypeBinding genericCast;

  public FieldReference(char[] source, long pos)
  {
    this.token = source;
    this.nameSourcePosition = pos;

    this.sourceStart = ((int)(pos >>> 32));
    this.sourceEnd = ((int)(pos & 0xFFFFFFFF));
    this.bits |= 1;
  }

  public FlowInfo analyseAssignment(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, Assignment assignment, boolean isCompound)
  {
    if (isCompound) {
      if ((this.binding.isBlankFinal()) && 
        (this.receiver.isThis()) && 
        (currentScope.allowBlankFinalFieldAssignment(this.binding)) && 
        (!flowInfo.isDefinitelyAssigned(this.binding))) {
        currentScope.problemReporter().uninitializedBlankFinalField(this.binding, this);
      }

      manageSyntheticAccessIfNecessary(currentScope, flowInfo, true);
    }
    flowInfo = 
      this.receiver
      .analyseCode(currentScope, flowContext, flowInfo, !this.binding.isStatic())
      .unconditionalInits();
    if (assignment.expression != null) {
      flowInfo = 
        assignment.expression
        .analyseCode(currentScope, flowContext, flowInfo)
        .unconditionalInits();
    }
    manageSyntheticAccessIfNecessary(currentScope, flowInfo, false);

    if (this.binding.isFinal())
    {
      if ((this.binding.isBlankFinal()) && 
        (!isCompound) && 
        (this.receiver.isThis()) && 
        (!(this.receiver instanceof QualifiedThisReference)) && 
        ((this.receiver.bits & 0x1FE00000) == 0) && 
        (currentScope.allowBlankFinalFieldAssignment(this.binding))) {
        if (flowInfo.isPotentiallyAssigned(this.binding))
          currentScope.problemReporter().duplicateInitializationOfBlankFinalField(
            this.binding, 
            this);
        else {
          flowContext.recordSettingFinal(this.binding, this, flowInfo);
        }
        flowInfo.markAsDefinitelyAssigned(this.binding);
      }
      else {
        currentScope.problemReporter().cannotAssignToFinalField(this.binding, this);
      }
    }
    return flowInfo;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    return analyseCode(currentScope, flowContext, flowInfo, true);
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo, boolean valueRequired) {
    boolean nonStatic = !this.binding.isStatic();
    this.receiver.analyseCode(currentScope, flowContext, flowInfo, nonStatic);
    if (nonStatic) {
      this.receiver.checkNPE(currentScope, flowContext, flowInfo);
    }

    if ((valueRequired) || (currentScope.compilerOptions().complianceLevel >= 3145728L)) {
      manageSyntheticAccessIfNecessary(currentScope, flowInfo, true);
    }
    return flowInfo;
  }

  public void computeConversion(Scope scope, TypeBinding runtimeTimeType, TypeBinding compileTimeType)
  {
    if ((runtimeTimeType == null) || (compileTimeType == null)) {
      return;
    }
    if ((this.binding != null) && (this.binding.isValidBinding())) {
      FieldBinding originalBinding = this.binding.original();
      TypeBinding originalType = originalBinding.type;

      if ((originalBinding != this.binding) && 
        (originalType != this.binding.type) && 
        (runtimeTimeType.id != 1) && 
        ((originalType.tagBits & 0x20000000) != 0L)) {
        TypeBinding targetType = (!compileTimeType.isBaseType()) && (runtimeTimeType.isBaseType()) ? 
          compileTimeType : 
          runtimeTimeType;
        this.genericCast = originalBinding.type.genericCast(targetType);
      }
    }
    super.computeConversion(scope, runtimeTimeType, compileTimeType);
  }

  public FieldBinding fieldBinding() {
    return this.binding;
  }

  public void generateAssignment(BlockScope currentScope, CodeStream codeStream, Assignment assignment, boolean valueRequired) {
    int pc = codeStream.position;
    this.receiver.generateCode(
      currentScope, 
      codeStream, 
      !this.codegenBinding.isStatic());
    codeStream.recordPositionsFrom(pc, this.sourceStart);
    assignment.expression.generateCode(currentScope, codeStream, true);
    fieldStore(
      codeStream, 
      this.codegenBinding, 
      this.syntheticAccessors == null ? null : this.syntheticAccessors[1], 
      valueRequired);
    if (valueRequired)
      codeStream.generateImplicitConversion(assignment.implicitConversion);
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;
    if (this.constant != Constant.NotAConstant) {
      if (valueRequired) {
        codeStream.generateConstant(this.constant, this.implicitConversion);
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }
    boolean isStatic = this.codegenBinding.isStatic();
    boolean isThisReceiver = this.receiver instanceof ThisReference;
    Constant fieldConstant = this.codegenBinding.constant();
    if (fieldConstant != Constant.NotAConstant) {
      if (!isThisReceiver) {
        this.receiver.generateCode(currentScope, codeStream, !isStatic);
        if (!isStatic) {
          codeStream.invokeObjectGetClass();
          codeStream.pop();
        }
      }
      if (valueRequired) {
        codeStream.generateConstant(fieldConstant, this.implicitConversion);
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }
    boolean isUnboxing;
    if ((valueRequired) || 
      ((!isThisReceiver) && (currentScope.compilerOptions().complianceLevel >= 3145728L)) || 
      ((this.implicitConversion & 0x400) != 0) || 
      (this.genericCast != null)) {
      this.receiver.generateCode(currentScope, codeStream, !isStatic);
      pc = codeStream.position;
      if (this.codegenBinding.declaringClass == null) {
        codeStream.arraylength();
        if (valueRequired) {
          codeStream.generateImplicitConversion(this.implicitConversion); break label566;
        }

        codeStream.pop(); break label566;
      }

      if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null)) {
        if (isStatic)
          codeStream.getstatic(this.codegenBinding);
        else
          codeStream.getfield(this.codegenBinding);
      }
      else {
        codeStream.invokestatic(this.syntheticAccessors[0]);
      }

      if (this.genericCast != null) codeStream.checkcast(this.genericCast);
      if (valueRequired) {
        codeStream.generateImplicitConversion(this.implicitConversion); break label566;
      }
      isUnboxing = (this.implicitConversion & 0x400) != 0;

      if (isUnboxing) codeStream.generateImplicitConversion(this.implicitConversion); 
    }
    switch (isUnboxing ? postConversionType(currentScope).id : this.codegenBinding.type.id) {
    case 7:
    case 8:
      codeStream.pop2();
      break;
    default:
      codeStream.pop(); break;

      if (isThisReceiver) {
        if (isStatic)
        {
          if (this.binding.original().declaringClass != this.receiverType.erasure()) {
            MethodBinding accessor = this.syntheticAccessors == null ? null : this.syntheticAccessors[0];
            if (accessor == null)
              codeStream.getstatic(this.codegenBinding);
            else {
              codeStream.invokestatic(accessor);
            }
            switch (this.codegenBinding.type.id) {
            case 7:
            case 8:
              codeStream.pop2();
              break;
            default:
              codeStream.pop(); break;
            }
          }
        }
      } else {
        this.receiver.generateCode(currentScope, codeStream, !isStatic);
        if (!isStatic) {
          codeStream.invokeObjectGetClass();
          codeStream.pop();
        }
      }
      break;
    }
    label566: codeStream.recordPositionsFrom(pc, this.sourceEnd);
  }

  public void generateCompoundAssignment(BlockScope currentScope, CodeStream codeStream, Expression expression, int operator, int assignmentImplicitConversion, boolean valueRequired)
  {
    boolean isStatic;
    this.receiver.generateCode(
      currentScope, 
      codeStream, 
      !(isStatic = this.codegenBinding.isStatic()));
    if (isStatic) {
      if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
        codeStream.getstatic(this.codegenBinding);
      else
        codeStream.invokestatic(this.syntheticAccessors[0]);
    }
    else {
      codeStream.dup();
      if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
        codeStream.getfield(this.codegenBinding);
      else
        codeStream.invokestatic(this.syntheticAccessors[0]);
    }
    int operationTypeID;
    switch (operationTypeID = (this.implicitConversion & 0xFF) >> 4) {
    case 0:
    case 1:
    case 11:
      codeStream.generateStringConcatenationAppend(currentScope, null, expression);
      break;
    default:
      if (this.genericCast != null) {
        codeStream.checkcast(this.genericCast);
      }
      codeStream.generateImplicitConversion(this.implicitConversion);

      if (expression == IntLiteral.One)
        codeStream.generateConstant(expression.constant, this.implicitConversion);
      else {
        expression.generateCode(currentScope, codeStream, true);
      }

      codeStream.sendOperator(operator, operationTypeID);

      codeStream.generateImplicitConversion(assignmentImplicitConversion);
    }
    fieldStore(
      codeStream, 
      this.codegenBinding, 
      this.syntheticAccessors == null ? null : this.syntheticAccessors[1], 
      valueRequired);
  }

  public void generatePostIncrement(BlockScope currentScope, CodeStream codeStream, CompoundAssignment postIncrement, boolean valueRequired)
  {
    boolean isStatic;
    this.receiver.generateCode(
      currentScope, 
      codeStream, 
      !(isStatic = this.codegenBinding.isStatic()));
    if (isStatic) {
      if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
        codeStream.getstatic(this.codegenBinding);
      else
        codeStream.invokestatic(this.syntheticAccessors[0]);
    }
    else {
      codeStream.dup();
      if ((this.syntheticAccessors == null) || (this.syntheticAccessors[0] == null))
        codeStream.getfield(this.codegenBinding);
      else {
        codeStream.invokestatic(this.syntheticAccessors[0]);
      }
    }
    if (valueRequired) {
      if (isStatic) {
        if ((this.codegenBinding.type == TypeBinding.LONG) || 
          (this.codegenBinding.type == TypeBinding.DOUBLE))
          codeStream.dup2();
        else {
          codeStream.dup();
        }
      }
      else if ((this.codegenBinding.type == TypeBinding.LONG) || 
        (this.codegenBinding.type == TypeBinding.DOUBLE))
        codeStream.dup2_x1();
      else {
        codeStream.dup_x1();
      }
    }

    if (this.genericCast != null)
      codeStream.checkcast(this.genericCast);
    codeStream.generateImplicitConversion(this.implicitConversion);
    codeStream.generateConstant(
      postIncrement.expression.constant, 
      this.implicitConversion);
    codeStream.sendOperator(postIncrement.operator, this.implicitConversion & 0xF);
    codeStream.generateImplicitConversion(
      postIncrement.preAssignImplicitConversion);
    fieldStore(codeStream, this.codegenBinding, this.syntheticAccessors == null ? null : this.syntheticAccessors[1], false);
  }

  public TypeBinding[] genericTypeArguments()
  {
    return null;
  }
  public boolean isSuperAccess() {
    return this.receiver.isSuper();
  }

  public boolean isTypeAccess() {
    return (this.receiver != null) && (this.receiver.isTypeReference());
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo, boolean isReadAccess)
  {
    if ((flowInfo.tagBits & 0x1) != 0) return;

    this.codegenBinding = this.binding.original();

    if (this.binding.isPrivate()) {
      if ((currentScope.enclosingSourceType() != this.codegenBinding.declaringClass) && 
        (this.binding.constant() == Constant.NotAConstant)) {
        if (this.syntheticAccessors == null)
          this.syntheticAccessors = new MethodBinding[2];
        this.syntheticAccessors[(isReadAccess ? 0 : 1)] = 
          ((SourceTypeBinding)this.codegenBinding.declaringClass).addSyntheticMethod(this.codegenBinding, isReadAccess);
        currentScope.problemReporter().needToEmulateFieldAccess(this.codegenBinding, this, isReadAccess);
      }
    }
    else {
      if ((this.receiver instanceof QualifiedSuperReference))
      {
        SourceTypeBinding destinationType = 
          (SourceTypeBinding)((QualifiedSuperReference)this.receiver).currentCompatibleType;

        if (this.syntheticAccessors == null)
          this.syntheticAccessors = new MethodBinding[2];
        this.syntheticAccessors[(isReadAccess ? 0 : 1)] = destinationType.addSyntheticMethod(this.codegenBinding, isReadAccess);
        currentScope.problemReporter().needToEmulateFieldAccess(this.codegenBinding, this, isReadAccess);
        return;
      }
      if (this.binding.isProtected())
      {
        SourceTypeBinding enclosingSourceType;
        if (((this.bits & 0x1FE0) != 0) && 
          (this.binding.declaringClass.getPackage() != 
          (enclosingSourceType = currentScope.enclosingSourceType()).getPackage()))
        {
          SourceTypeBinding currentCompatibleType = 
            (SourceTypeBinding)enclosingSourceType.enclosingTypeAt(
            (this.bits & 0x1FE0) >> 5);
          if (this.syntheticAccessors == null)
            this.syntheticAccessors = new MethodBinding[2];
          this.syntheticAccessors[(isReadAccess ? 0 : 1)] = currentCompatibleType.addSyntheticMethod(this.codegenBinding, isReadAccess);
          currentScope.problemReporter().needToEmulateFieldAccess(this.codegenBinding, this, isReadAccess);
          return;
        }

      }

    }

    if ((this.binding.declaringClass != this.receiverType) && 
      (!this.receiverType.isArrayType()) && 
      (this.binding.declaringClass != null) && 
      (this.binding.constant() == Constant.NotAConstant)) {
      CompilerOptions options = currentScope.compilerOptions();
      if (((options.targetJDK >= 3014656L) && 
        ((options.complianceLevel >= 3145728L) || (!this.receiver.isImplicitThis()) || (!this.codegenBinding.isStatic())) && 
        (this.binding.declaringClass.id != 1)) || 
        (!this.binding.declaringClass.canBeSeenBy(currentScope)))
      {
        this.codegenBinding = 
          currentScope.enclosingSourceType().getUpdatedFieldBinding(
          this.codegenBinding, 
          (ReferenceBinding)this.receiverType.erasure());
      }
    }
  }

  public int nullStatus(FlowInfo flowInfo) {
    return 0;
  }

  public Constant optimizedBooleanConstant() {
    switch (this.resolvedType.id) {
    case 5:
    case 33:
      return this.constant != Constant.NotAConstant ? this.constant : this.binding.constant();
    }
    return Constant.NotAConstant;
  }

  public TypeBinding postConversionType(Scope scope)
  {
    TypeBinding convertedType = this.resolvedType;
    if (this.genericCast != null)
      convertedType = this.genericCast;
    int runtimeType = (this.implicitConversion & 0xFF) >> 4;
    switch (runtimeType) {
    case 5:
      convertedType = TypeBinding.BOOLEAN;
      break;
    case 3:
      convertedType = TypeBinding.BYTE;
      break;
    case 4:
      convertedType = TypeBinding.SHORT;
      break;
    case 2:
      convertedType = TypeBinding.CHAR;
      break;
    case 10:
      convertedType = TypeBinding.INT;
      break;
    case 9:
      convertedType = TypeBinding.FLOAT;
      break;
    case 7:
      convertedType = TypeBinding.LONG;
      break;
    case 8:
      convertedType = TypeBinding.DOUBLE;
    case 6:
    }

    if ((this.implicitConversion & 0x200) != 0) {
      convertedType = scope.environment().computeBoxingType(convertedType);
    }
    return convertedType;
  }

  public StringBuffer printExpression(int indent, StringBuffer output) {
    return this.receiver.printExpression(0, output).append('.').append(this.token);
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    boolean receiverCast = false;
    if ((this.receiver instanceof CastExpression)) {
      this.receiver.bits |= 32;
      receiverCast = true;
    }
    this.receiverType = this.receiver.resolveType(scope);
    if (this.receiverType == null) {
      this.constant = Constant.NotAConstant;
      return null;
    }
    if (receiverCast)
    {
      if (((CastExpression)this.receiver).expression.resolvedType == this.receiverType) {
        scope.problemReporter().unnecessaryCast((CastExpression)this.receiver);
      }
    }

    FieldBinding fieldBinding = this.codegenBinding = this.binding = scope.getField(this.receiverType, this.token, this);
    if (!fieldBinding.isValidBinding()) {
      this.constant = Constant.NotAConstant;
      scope.problemReporter().invalidField(this, this.receiverType);
      return null;
    }
    TypeBinding receiverErasure = this.receiverType.erasure();
    if (((receiverErasure instanceof ReferenceBinding)) && 
      (receiverErasure.findSuperTypeWithSameErasure(fieldBinding.declaringClass) == null)) {
      this.receiverType = fieldBinding.declaringClass;
    }

    this.receiver.computeConversion(scope, this.receiverType, this.receiverType);
    if (isFieldUseDeprecated(fieldBinding, scope, (this.bits & 0x2000) != 0)) {
      scope.problemReporter().deprecatedField(fieldBinding, this);
    }
    boolean isImplicitThisRcv = this.receiver.isImplicitThis();
    this.constant = (isImplicitThisRcv ? fieldBinding.constant() : Constant.NotAConstant);
    if (fieldBinding.isStatic())
    {
      if ((!isImplicitThisRcv) && (
        (!(this.receiver instanceof NameReference)) || 
        ((((NameReference)this.receiver).bits & 0x4) == 0))) {
        scope.problemReporter().nonStaticAccessToStaticField(this, fieldBinding);
      }
      if ((!isImplicitThisRcv) && 
        (fieldBinding.declaringClass != this.receiverType) && 
        (fieldBinding.declaringClass.canBeSeenBy(scope))) {
        scope.problemReporter().indirectAccessToStaticField(this, fieldBinding);
      }
    }

    return this.resolvedType = 
      (this.bits & 0x2000) == 0 ? 
      fieldBinding.type.capture(scope, this.sourceEnd) : 
      fieldBinding.type;
  }

  public void setActualReceiverType(ReferenceBinding receiverType)
  {
  }

  public void setDepth(int depth) {
    this.bits &= -8161;
    if (depth > 0)
      this.bits |= (depth & 0xFF) << 5;
  }

  public void setFieldIndex(int index)
  {
  }

  public void traverse(ASTVisitor visitor, BlockScope scope)
  {
    if (visitor.visit(this, scope)) {
      this.receiver.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}