package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class QualifiedAllocationExpression extends AllocationExpression
{
  public Expression enclosingInstance;
  public TypeDeclaration anonymousType;
  public ReferenceBinding superTypeBinding;

  public QualifiedAllocationExpression()
  {
  }

  public QualifiedAllocationExpression(TypeDeclaration anonymousType)
  {
    this.anonymousType = anonymousType;
    anonymousType.allocation = this;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    if (this.enclosingInstance != null) {
      flowInfo = this.enclosingInstance.analyseCode(currentScope, flowContext, flowInfo);
    }

    ReferenceBinding allocatedType = this.superTypeBinding == null ? this.binding.declaringClass : this.superTypeBinding;
    checkCapturedLocalInitializationIfNecessary(
      (ReferenceBinding)allocatedType.erasure(), 
      currentScope, 
      flowInfo);

    if (this.arguments != null) {
      int i = 0; for (int count = this.arguments.length; i < count; i++) {
        flowInfo = this.arguments[i].analyseCode(currentScope, flowContext, flowInfo);
      }

    }

    if (this.anonymousType != null)
      flowInfo = this.anonymousType.analyseCode(currentScope, flowContext, flowInfo);
    ReferenceBinding[] thrownExceptions;
    if ((thrownExceptions = this.binding.thrownExceptions).length != 0)
    {
      flowContext.checkExceptionHandlers(
        thrownExceptions, 
        this, 
        flowInfo.unconditionalCopy(), 
        currentScope);
    }
    manageEnclosingInstanceAccessIfNecessary(currentScope, flowInfo);
    manageSyntheticAccessIfNecessary(currentScope, flowInfo);
    return flowInfo;
  }

  public Expression enclosingInstance()
  {
    return this.enclosingInstance;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;
    ReferenceBinding allocatedType = this.codegenBinding.declaringClass;
    codeStream.new_(allocatedType);
    boolean isUnboxing = (this.implicitConversion & 0x400) != 0;
    if ((valueRequired) || (isUnboxing)) {
      codeStream.dup();
    }

    if (this.type != null) {
      codeStream.recordPositionsFrom(pc, this.type.sourceStart);
    }
    else {
      codeStream.ldc(String.valueOf(this.enumConstant.name));
      codeStream.generateInlinedValue(this.enumConstant.binding.id);
    }

    if (allocatedType.isNestedType()) {
      codeStream.generateSyntheticEnclosingInstanceValues(
        currentScope, 
        allocatedType, 
        enclosingInstance(), 
        this);
    }

    generateArguments(this.binding, this.arguments, currentScope, codeStream);

    if (allocatedType.isNestedType()) {
      codeStream.generateSyntheticOuterArgumentValues(
        currentScope, 
        allocatedType, 
        this);
    }

    if (this.syntheticAccessor == null) {
      codeStream.invokespecial(this.codegenBinding);
    }
    else {
      int i = 0;
      int max = this.syntheticAccessor.parameters.length - this.codegenBinding.parameters.length;
      for (; i < max; 
        i++) {
        codeStream.aconst_null();
      }
      codeStream.invokespecial(this.syntheticAccessor);
    }
    if (valueRequired) {
      codeStream.generateImplicitConversion(this.implicitConversion);
    } else if (isUnboxing)
    {
      codeStream.generateImplicitConversion(this.implicitConversion);
      switch (postConversionType(currentScope).id) {
      case 7:
      case 8:
        codeStream.pop2();
        break;
      default:
        codeStream.pop();
      }
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);

    if (this.anonymousType != null)
      this.anonymousType.generateCode(currentScope, codeStream);
  }

  public boolean isSuperAccess()
  {
    return this.anonymousType != null;
  }

  public void manageEnclosingInstanceAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) == 0) {
      ReferenceBinding allocatedTypeErasure = (ReferenceBinding)this.binding.declaringClass.erasure();

      if ((allocatedTypeErasure.isNestedType()) && 
        (currentScope.enclosingSourceType().isLocalType()))
      {
        if (allocatedTypeErasure.isLocalType()) {
          ((LocalTypeBinding)allocatedTypeErasure).addInnerEmulationDependent(currentScope, this.enclosingInstance != null);
        }
        else
          currentScope.propagateInnerEmulation(allocatedTypeErasure, this.enclosingInstance != null);
      }
    }
  }

  public StringBuffer printExpression(int indent, StringBuffer output)
  {
    if (this.enclosingInstance != null)
      this.enclosingInstance.printExpression(0, output).append('.');
    super.printExpression(0, output);
    if (this.anonymousType != null) {
      this.anonymousType.print(indent, output);
    }
    return output;
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    if ((this.anonymousType == null) && (this.enclosingInstance == null)) {
      return super.resolveType(scope);
    }

    this.constant = Constant.NotAConstant;
    TypeBinding enclosingInstanceType = null;
    TypeBinding receiverType = null;
    boolean hasError = false;
    boolean enclosingInstanceContainsCast = false;
    boolean argsContainCast = false;

    if (this.enclosingInstance != null) {
      if ((this.enclosingInstance instanceof CastExpression)) {
        this.enclosingInstance.bits |= 32;
        enclosingInstanceContainsCast = true;
      }
      if ((enclosingInstanceType = this.enclosingInstance.resolveType(scope)) == null) {
        hasError = true;
      } else if ((enclosingInstanceType.isBaseType()) || (enclosingInstanceType.isArrayType())) {
        scope.problemReporter().illegalPrimitiveOrArrayTypeForEnclosingInstance(
          enclosingInstanceType, 
          this.enclosingInstance);
        hasError = true;
      } else if ((this.type instanceof QualifiedTypeReference)) {
        scope.problemReporter().illegalUsageOfQualifiedTypeReference((QualifiedTypeReference)this.type);
        hasError = true;
      } else {
        receiverType = ((SingleTypeReference)this.type).resolveTypeEnclosing(scope, (ReferenceBinding)enclosingInstanceType);
        if ((receiverType != null) && (enclosingInstanceContainsCast)) {
          CastExpression.checkNeedForEnclosingInstanceCast(scope, this.enclosingInstance, enclosingInstanceType, receiverType);
        }
      }
    }
    else if (this.type == null)
    {
      receiverType = scope.enclosingSourceType();
    } else {
      receiverType = this.type.resolveType(scope, true);

      if ((receiverType != null) && 
        ((this.type instanceof ParameterizedQualifiedTypeReference))) {
        ReferenceBinding currentType = (ReferenceBinding)receiverType;

        while (((currentType.modifiers & 0x8) == 0) && 
          (!currentType.isRawType())) {
          if ((currentType = currentType.enclosingType()) == null) {
            ParameterizedQualifiedTypeReference qRef = (ParameterizedQualifiedTypeReference)this.type;
            for (int i = qRef.typeArguments.length - 2; i >= 0; i--) {
              if (qRef.typeArguments[i] != null) {
                scope.problemReporter().illegalQualifiedParameterizedTypeAllocation(this.type, receiverType);
                break;
              }
            }
          }
        }
      }
    }
    if (receiverType == null)
      hasError = true;
    else if (((ReferenceBinding)receiverType).isFinal()) {
      if (this.anonymousType != null) {
        if (!receiverType.isEnum()) {
          scope.problemReporter().anonymousClassCannotExtendFinalClass(this.type, receiverType);
          hasError = true;
        }
      } else if (!receiverType.canBeInstantiated()) {
        scope.problemReporter().cannotInstantiate(this.type, receiverType);
        return this.resolvedType = receiverType;
      }
    }

    if (this.typeArguments != null) {
      int length = this.typeArguments.length;
      this.genericTypeArguments = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        TypeReference typeReference = this.typeArguments[i];
        TypeBinding argType = typeReference.resolveType(scope, true);
        if (argType == null) {
          if ((typeReference instanceof Wildcard)) {
            scope.problemReporter().illegalUsageOfWildcard(typeReference);
          }
          return null;
        }
        this.genericTypeArguments[i] = argType;
      }

    }

    TypeBinding[] argumentTypes = Binding.NO_PARAMETERS;
    if (this.arguments != null) {
      int length = this.arguments.length;
      argumentTypes = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        Expression argument = this.arguments[i];
        if ((argument instanceof CastExpression)) {
          argument.bits |= 32;
          argsContainCast = true;
        }
        if ((argumentTypes[i] =  = argument.resolveType(scope)) == null) {
          hasError = true;
        }
      }
    }

    if (hasError) {
      if ((receiverType instanceof ReferenceBinding))
      {
        int length = this.arguments == null ? 0 : this.arguments.length;
        TypeBinding[] pseudoArgs = new TypeBinding[length];
        int i = length;
        do { pseudoArgs[i] = (argumentTypes[i] == null ? TypeBinding.NULL : argumentTypes[i]);

          i--; } while (i >= 0);

        this.binding = scope.findMethod((ReferenceBinding)receiverType, TypeConstants.INIT, pseudoArgs, this);
        if ((this.binding != null) && (!this.binding.isValidBinding())) {
          MethodBinding closestMatch = ((ProblemMethodBinding)this.binding).closestMatch;

          if (closestMatch != null) {
            if (closestMatch.original().typeVariables != Binding.NO_TYPE_VARIABLES)
            {
              closestMatch = scope.environment().createParameterizedGenericMethod(closestMatch.original(), null);
            }
            this.binding = closestMatch;
            MethodBinding closestMatchOriginal = closestMatch.original();
            if (((closestMatchOriginal.isPrivate()) || (closestMatchOriginal.declaringClass.isLocalType())) && (!scope.isDefinedInMethod(closestMatchOriginal)))
            {
              closestMatchOriginal.modifiers |= 134217728;
            }
          }
        }
      }

      return this.resolvedType = receiverType;
    }
    if (this.anonymousType == null)
    {
      if (!receiverType.canBeInstantiated()) {
        scope.problemReporter().cannotInstantiate(this.type, receiverType);
        return this.resolvedType = receiverType;
      }
      ReferenceBinding allocationType = (ReferenceBinding)receiverType;
      if ((this.binding = scope.getConstructor(allocationType, argumentTypes, this)).isValidBinding()) {
        if (isMethodUseDeprecated(this.binding, scope, true)) {
          scope.problemReporter().deprecatedMethod(this.binding, this);
        }
        checkInvocationArguments(scope, null, allocationType, this.binding, this.arguments, argumentTypes, argsContainCast, this);
      } else {
        if (this.binding.declaringClass == null) {
          this.binding.declaringClass = allocationType;
        }
        scope.problemReporter().invalidConstructor(this, this.binding);
        return this.resolvedType = receiverType;
      }

      ReferenceBinding expectedType = this.binding.declaringClass.enclosingType();
      if (expectedType != enclosingInstanceType)
        scope.compilationUnitScope().recordTypeConversion(expectedType, enclosingInstanceType);
      if ((enclosingInstanceType.isCompatibleWith(expectedType)) || (scope.isBoxingCompatibleWith(enclosingInstanceType, expectedType))) {
        this.enclosingInstance.computeConversion(scope, expectedType, enclosingInstanceType);
        return this.resolvedType = receiverType;
      }
      scope.problemReporter().typeMismatchError(enclosingInstanceType, expectedType, this.enclosingInstance);
      return this.resolvedType = receiverType;
    }

    if (receiverType.isTypeVariable()) {
      receiverType = new ProblemReferenceBinding(receiverType.sourceName(), (ReferenceBinding)receiverType, 9);
      scope.problemReporter().invalidType(this, receiverType);
      return null;
    }if ((this.type != null) && (receiverType.isEnum())) {
      scope.problemReporter().cannotInstantiate(this.type, receiverType);
      return this.resolvedType = receiverType;
    }

    this.superTypeBinding = (receiverType.isInterface() ? scope.getJavaLangObject() : (ReferenceBinding)receiverType);

    scope.addAnonymousType(this.anonymousType, (ReferenceBinding)receiverType);
    this.anonymousType.resolve(scope);
    if (this.superTypeBinding.erasure().id == 41) {
      scope.problemReporter().cannotExtendEnum(this.anonymousType.binding, this.type, this.superTypeBinding);
    }

    if ((receiverType.tagBits & 0x40000000) != 0L) {
      scope.problemReporter().superTypeCannotUseWildcard(this.anonymousType.binding, this.type, receiverType);
    }

    MethodBinding inheritedBinding = scope.getConstructor(this.superTypeBinding, argumentTypes, this);
    if (!inheritedBinding.isValidBinding()) {
      if (inheritedBinding.declaringClass == null) {
        inheritedBinding.declaringClass = this.superTypeBinding;
      }
      scope.problemReporter().invalidConstructor(this, inheritedBinding);
      return this.resolvedType = this.anonymousType.binding;
    }
    if (this.enclosingInstance != null) {
      ReferenceBinding targetEnclosing = inheritedBinding.declaringClass.enclosingType();
      if (targetEnclosing == null) {
        scope.problemReporter().unnecessaryEnclosingInstanceSpecification(this.enclosingInstance, (ReferenceBinding)receiverType);
        return this.resolvedType = this.anonymousType.binding;
      }if ((!enclosingInstanceType.isCompatibleWith(targetEnclosing)) && (!scope.isBoxingCompatibleWith(enclosingInstanceType, targetEnclosing))) {
        scope.problemReporter().typeMismatchError(enclosingInstanceType, targetEnclosing, this.enclosingInstance);
        return this.resolvedType = this.anonymousType.binding;
      }
      this.enclosingInstance.computeConversion(scope, targetEnclosing, enclosingInstanceType);
    }
    if (this.arguments != null) {
      checkInvocationArguments(scope, null, this.superTypeBinding, inheritedBinding, this.arguments, argumentTypes, argsContainCast, this);
    }

    this.binding = this.anonymousType.createDefaultConstructorWithBinding(inheritedBinding);
    return this.resolvedType = this.anonymousType.binding;
  }

  public void traverse(ASTVisitor visitor, BlockScope scope)
  {
    if (visitor.visit(this, scope)) {
      if (this.enclosingInstance != null)
        this.enclosingInstance.traverse(visitor, scope);
      if (this.typeArguments != null) {
        int i = 0; for (int typeArgumentsLength = this.typeArguments.length; i < typeArgumentsLength; i++) {
          this.typeArguments[i].traverse(visitor, scope);
        }
      }
      if (this.type != null)
        this.type.traverse(visitor, scope);
      if (this.arguments != null) {
        int argumentsLength = this.arguments.length;
        for (int i = 0; i < argumentsLength; i++)
          this.arguments[i].traverse(visitor, scope);
      }
      if (this.anonymousType != null)
        this.anonymousType.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}