package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CaseLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.SwitchFlowContext;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class SwitchStatement extends Statement
{
  public Expression expression;
  public Statement[] statements;
  public BlockScope scope;
  public int explicitDeclarations;
  public BranchLabel breakLabel;
  public CaseStatement[] cases;
  public CaseStatement defaultCase;
  public int blockStart;
  public int caseCount;
  int[] constants;
  public static final int CASE = 0;
  public static final int FALLTHROUGH = 1;
  public static final int ESCAPING = 2;
  public SyntheticMethodBinding synthetic;
  int preSwitchInitStateIndex = -1;
  int mergedInitStateIndex = -1;

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    try
    {
      flowInfo = this.expression.analyseCode(currentScope, flowContext, flowInfo);
      SwitchFlowContext switchContext = 
        new SwitchFlowContext(flowContext, this, this.breakLabel = new BranchLabel());

      FlowInfo caseInits = FlowInfo.DEAD_END;

      this.preSwitchInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(flowInfo);
      int caseIndex = 0;
      if (this.statements != null) {
        boolean didAlreadyComplain = false;
        int fallThroughState = 0;
        int i = 0; for (int max = this.statements.length; i < max; i++) {
          Statement statement = this.statements[i];
          if ((caseIndex < this.caseCount) && (statement == this.cases[caseIndex])) {
            this.scope.enclosingCase = this.cases[caseIndex];
            caseIndex++;
            if ((fallThroughState == 1) && 
              ((statement.bits & 0x20000000) == 0)) {
              this.scope.problemReporter().possibleFallThroughCase(this.scope.enclosingCase);
            }
            caseInits = caseInits.mergedWith(flowInfo.unconditionalInits());
            didAlreadyComplain = false;
            fallThroughState = 0;
          } else if (statement == this.defaultCase) {
            this.scope.enclosingCase = this.defaultCase;
            if ((fallThroughState == 1) && 
              ((statement.bits & 0x20000000) == 0)) {
              this.scope.problemReporter().possibleFallThroughCase(this.scope.enclosingCase);
            }
            caseInits = caseInits.mergedWith(flowInfo.unconditionalInits());
            didAlreadyComplain = false;
            fallThroughState = 0;
          } else {
            fallThroughState = 1;
          }
          if (!statement.complainIfUnreachable(caseInits, this.scope, didAlreadyComplain)) {
            caseInits = statement.analyseCode(this.scope, switchContext, caseInits);
            if (caseInits == FlowInfo.DEAD_END)
              fallThroughState = 2;
          }
          else {
            didAlreadyComplain = true;
          }
        }
      }

      TypeBinding resolvedTypeBinding = this.expression.resolvedType;
      if ((this.caseCount > 0) && (resolvedTypeBinding.isEnum())) {
        SourceTypeBinding sourceTypeBinding = this.scope.classScope().referenceContext.binding;
        this.synthetic = sourceTypeBinding.addSyntheticMethodForSwitchEnum(resolvedTypeBinding);
      }
      FlowInfo localFlowInfo1;
      if (this.defaultCase == null)
      {
        flowInfo.addPotentialInitializationsFrom(
          caseInits.mergedWith(switchContext.initsOnBreak));
        this.mergedInitStateIndex = 
          currentScope.methodScope().recordInitializationStates(flowInfo);
        return flowInfo;
      }
      TypeBinding resolvedTypeBinding;
      int caseIndex;
      FlowInfo caseInits;
      SwitchFlowContext switchContext;
      FlowInfo mergedInfo = caseInits.mergedWith(switchContext.initsOnBreak);
      this.mergedInitStateIndex = 
        currentScope.methodScope().recordInitializationStates(mergedInfo);
      return mergedInfo;
    } finally {
      if (this.scope != null) this.scope.enclosingCase = null;
    }
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    try
    {
      if ((this.bits & 0x80000000) == 0) {
        return;
      }
      int pc = codeStream.position;

      this.breakLabel.initialize(codeStream);
      CaseLabel[] caseLabels = new CaseLabel[this.caseCount];
      boolean needSwitch = this.caseCount != 0;
      for (int i = 0; i < this.caseCount; i++) {
        this.cases[i].targetLabel = (caseLabels[i] =  = new CaseLabel(codeStream));
        caseLabels[i].tagBits |= 2;
      }
      CaseLabel defaultLabel = new CaseLabel(codeStream);
      if (needSwitch) defaultLabel.tagBits |= 2;
      if (this.defaultCase != null) {
        this.defaultCase.targetLabel = defaultLabel;
      }

      TypeBinding resolvedType = this.expression.resolvedType;
      if (resolvedType.isEnum()) {
        if (needSwitch)
        {
          codeStream.invokestatic(this.synthetic);
          this.expression.generateCode(currentScope, codeStream, true);

          codeStream.invokeEnumOrdinal(resolvedType.constantPoolName());
          codeStream.iaload();
        }
        else {
          this.expression.generateCode(currentScope, codeStream, false);
        }
      }
      else {
        this.expression.generateCode(currentScope, codeStream, needSwitch);
      }

      if (needSwitch) {
        int[] sortedIndexes = new int[this.caseCount];

        for (int i = 0; i < this.caseCount; i++)
          sortedIndexes[i] = i;
        int[] localKeysCopy;
        System.arraycopy(this.constants, 0, localKeysCopy = new int[this.caseCount], 0, this.caseCount);
        CodeStream.sort(localKeysCopy, 0, this.caseCount - 1, sortedIndexes);

        int max = localKeysCopy[(this.caseCount - 1)];
        int min = localKeysCopy[0];
        if (()(this.caseCount * 2.5D) > max - min)
        {
          if ((max > 2147418112) && (currentScope.compilerOptions().complianceLevel < 3145728L)) {
            codeStream.lookupswitch(defaultLabel, this.constants, sortedIndexes, caseLabels);
          }
          else
            codeStream.tableswitch(
              defaultLabel, 
              min, 
              max, 
              this.constants, 
              sortedIndexes, 
              caseLabels);
        }
        else {
          codeStream.lookupswitch(defaultLabel, this.constants, sortedIndexes, caseLabels);
        }
        codeStream.updateLastRecordedEndPC(this.scope, codeStream.position);
      }

      int caseIndex = 0;
      if (this.statements != null) {
        int i = 0; for (int maxCases = this.statements.length; i < maxCases; i++) {
          Statement statement = this.statements[i];
          if ((caseIndex < this.caseCount) && (statement == this.cases[caseIndex])) {
            this.scope.enclosingCase = this.cases[caseIndex];
            if (this.preSwitchInitStateIndex != -1) {
              codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.preSwitchInitStateIndex);
            }
            caseIndex++;
          }
          else if (statement == this.defaultCase) {
            this.scope.enclosingCase = this.defaultCase;
            if (this.preSwitchInitStateIndex != -1) {
              codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.preSwitchInitStateIndex);
            }
          }

          statement.generateCode(this.scope, codeStream);
        }
      }

      if (this.mergedInitStateIndex != -1) {
        codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
        codeStream.addDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
      }
      if (this.scope != currentScope) {
        codeStream.exitUserScope(this.scope);
      }

      this.breakLabel.place();
      if (this.defaultCase == null)
      {
        codeStream.recordPositionsFrom(codeStream.position, this.sourceEnd, true);
        defaultLabel.place();
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
    } finally {
      if (this.scope != null) this.scope.enclosingCase = null;  } if (this.scope != null) this.scope.enclosingCase = null;
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    printIndent(indent, output).append("switch (");
    this.expression.printExpression(0, output).append(") {");
    if (this.statements != null) {
      for (int i = 0; i < this.statements.length; i++) {
        output.append('\n');
        if ((this.statements[i] instanceof CaseStatement))
          this.statements[i].printStatement(indent, output);
        else {
          this.statements[i].printStatement(indent + 2, output);
        }
      }
    }
    output.append("\n");
    return printIndent(indent, output).append('}');
  }

  public void resolve(BlockScope upperScope)
  {
    label645: 
    try { boolean isEnumSwitch = false;
      TypeBinding expressionType = this.expression.resolveType(upperScope);
      if (expressionType != null) {
        this.expression.computeConversion(upperScope, expressionType, expressionType);

        if (expressionType.isBaseType()) {
          if (this.expression.isConstantValueOfTypeAssignableToType(expressionType, TypeBinding.INT))
            break label114;
          if (expressionType.isCompatibleWith(TypeBinding.INT))
            break label114; 
        } else { if (expressionType.isEnum()) {
            isEnumSwitch = true;
            break label114;
          }if (upperScope.isBoxingCompatibleWith(expressionType, TypeBinding.INT)) {
            this.expression.computeConversion(upperScope, TypeBinding.INT, expressionType);
            break label114;
          } }
        upperScope.problemReporter().incorrectSwitchType(this.expression, expressionType);
        expressionType = null;
      }

      label114: if (this.statements != null) {
        this.scope = new BlockScope(upperScope);
        int length;
        this.cases = new CaseStatement[length = this.statements.length];
        this.constants = new int[length];
        CaseStatement[] duplicateCaseStatements = (CaseStatement[])null;
        int duplicateCaseStatementsCounter = 0;
        int counter = 0;
        for (int i = 0; i < length; i++)
        {
          Statement statement = this.statements[i];
          Constant constant;
          if ((constant = statement.resolveCase(this.scope, expressionType, this)) != Constant.NotAConstant) {
            int key = constant.intValue();

            for (int j = 0; j < counter; j++) {
              if (this.constants[j] == key) {
                CaseStatement currentCaseStatement = (CaseStatement)statement;
                if (duplicateCaseStatements == null) {
                  this.scope.problemReporter().duplicateCase(this.cases[j]);
                  this.scope.problemReporter().duplicateCase(currentCaseStatement);
                  duplicateCaseStatements = new CaseStatement[length];
                  duplicateCaseStatements[(duplicateCaseStatementsCounter++)] = this.cases[j];
                  duplicateCaseStatements[(duplicateCaseStatementsCounter++)] = currentCaseStatement;
                } else {
                  boolean found = false;
                  for (int k = 2; k < duplicateCaseStatementsCounter; k++) {
                    if (duplicateCaseStatements[k] == statement) {
                      found = true;
                      break;
                    }
                  }
                  if (!found) {
                    this.scope.problemReporter().duplicateCase(currentCaseStatement);
                    duplicateCaseStatements[(duplicateCaseStatementsCounter++)] = currentCaseStatement;
                  }
                }
              }
            }
            this.constants[(counter++)] = key;
          }
        }
        if (length != counter) {
          System.arraycopy(this.constants, 0, this.constants = new int[counter], 0, counter);
        }
      }
      else if ((this.bits & 0x8) != 0) {
        upperScope.problemReporter().undocumentedEmptyBlock(this.blockStart, this.sourceEnd);
      }

      if ((isEnumSwitch) && (this.defaultCase == null)) {
        if (upperScope.compilerOptions().getSeverity(2199023255552L) == -1) break label645;
        int constantCount = this.constants == null ? 0 : this.constants.length;
        if ((constantCount != this.caseCount) || 
          (this.caseCount == ((ReferenceBinding)expressionType).enumConstantCount())) break label645;
        FieldBinding[] enumFields = ((ReferenceBinding)expressionType.erasure()).fields();
        int i = 0; for (int max = enumFields.length; i < max; i++) {
          FieldBinding enumConstant = enumFields[i];
          if ((enumConstant.modifiers & 0x4000) != 0)
          {
            int j = 0;
            while (enumConstant.id + 1 != this.constants[j])
            {
              j++; if (j >= this.caseCount)
              {
                upperScope.problemReporter().missingEnumConstantCase(this, enumConstant);
              }
            }
          }
        }
      } } finally {
      if (this.scope != null) this.scope.enclosingCase = null;
    }
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope)
  {
    if (visitor.visit(this, blockScope)) {
      this.expression.traverse(visitor, this.scope);
      if (this.statements != null) {
        int statementsLength = this.statements.length;
        for (int i = 0; i < statementsLength; i++)
          this.statements[i].traverse(visitor, this.scope);
      }
    }
    visitor.endVisit(this, blockScope);
  }

  public void branchChainTo(BranchLabel label)
  {
    if (this.breakLabel.forwardReferenceCount() > 0)
      label.becomeDelegateFor(this.breakLabel);
  }
}