package org.eclipse.jdt.internal.compiler.classfmt;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.codegen.AttributeNamesConstants;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import org.eclipse.jdt.internal.compiler.env.IBinaryField;
import org.eclipse.jdt.internal.compiler.env.IBinaryMethod;
import org.eclipse.jdt.internal.compiler.env.IBinaryNestedType;
import org.eclipse.jdt.internal.compiler.env.IBinaryType;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.util.Util;

public class ClassFileReader extends ClassFileStruct
  implements IBinaryType
{
  private int accessFlags;
  private char[] classFileName;
  private char[] className;
  private int classNameIndex;
  private int constantPoolCount;
  private AnnotationInfo[] annotations;
  private FieldInfo[] fields;
  private int fieldsCount;
  private InnerClassInfo innerInfo;
  private int innerInfoIndex;
  private InnerClassInfo[] innerInfos;
  private char[][] interfaceNames;
  private int interfacesCount;
  private MethodInfo[] methods;
  private int methodsCount;
  private char[] signature;
  private char[] sourceName;
  private char[] sourceFileName;
  private char[] superclassName;
  private long tagBits;
  private long version;
  private char[] enclosingTypeName;

  private static String printTypeModifiers(int modifiers)
  {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    PrintWriter print = new PrintWriter(out);

    if ((modifiers & 0x1) != 0) print.print("public ");
    if ((modifiers & 0x2) != 0) print.print("private ");
    if ((modifiers & 0x10) != 0) print.print("final ");
    if ((modifiers & 0x20) != 0) print.print("super ");
    if ((modifiers & 0x200) != 0) print.print("interface ");
    if ((modifiers & 0x400) != 0) print.print("abstract ");
    print.flush();
    return out.toString();
  }
  public static ClassFileReader read(InputStream stream, String fileName) throws ClassFormatException, IOException {
    return read(stream, fileName, false);
  }
  public static ClassFileReader read(File file) throws ClassFormatException, IOException {
    return read(file, false);
  }
  public static ClassFileReader read(InputStream stream, String fileName, boolean fullyInitialize) throws ClassFormatException, IOException {
    byte[] classFileBytes = Util.getInputStreamAsByteArray(stream, -1);
    ClassFileReader classFileReader = new ClassFileReader(classFileBytes, fileName.toCharArray());
    if (fullyInitialize) {
      classFileReader.initialize();
    }
    return classFileReader;
  }
  public static ClassFileReader read(File file, boolean fullyInitialize) throws ClassFormatException, IOException {
    byte[] classFileBytes = Util.getFileByteContent(file);
    ClassFileReader classFileReader = new ClassFileReader(classFileBytes, file.getAbsolutePath().toCharArray());
    if (fullyInitialize) {
      classFileReader.initialize();
    }
    return classFileReader;
  }

  public static ClassFileReader read(ZipFile zip, String filename)
    throws ClassFormatException, IOException
  {
    return read(zip, filename, false);
  }

  public static ClassFileReader read(ZipFile zip, String filename, boolean fullyInitialize)
    throws ClassFormatException, IOException
  {
    ZipEntry ze = zip.getEntry(filename);
    if (ze == null)
      return null;
    byte[] classFileBytes = Util.getZipEntryByteContent(ze, zip);
    ClassFileReader classFileReader = new ClassFileReader(classFileBytes, filename.toCharArray());
    if (fullyInitialize) {
      classFileReader.initialize();
    }
    return classFileReader;
  }
  public static ClassFileReader read(String fileName) throws ClassFormatException, IOException {
    return read(fileName, false);
  }
  public static ClassFileReader read(String fileName, boolean fullyInitialize) throws ClassFormatException, IOException {
    return read(new File(fileName), fullyInitialize);
  }

  public ClassFileReader(byte[] classFileBytes, char[] fileName)
    throws ClassFormatException
  {
    this(classFileBytes, fileName, false);
  }

  public ClassFileReader(byte[] classFileBytes, char[] fileName, boolean fullyInitialize)
    throws ClassFormatException
  {
    super(classFileBytes, null, 0);
    this.classFileName = fileName;
    int readOffset = 10;
    try {
      this.version = ((u2At(6) << 16) + u2At(4));
      this.constantPoolCount = u2At(8);

      this.constantPoolOffsets = new int[this.constantPoolCount];
      for (int i = 1; i < this.constantPoolCount; i++) {
        int tag = u1At(readOffset);
        switch (tag) {
        case 1:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += u2At(readOffset + 1);
          readOffset += 3;
          break;
        case 3:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
          break;
        case 4:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
          break;
        case 5:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 9;
          i++;
          break;
        case 6:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 9;
          i++;
          break;
        case 7:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 3;
          break;
        case 8:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 3;
          break;
        case 9:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
          break;
        case 10:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
          break;
        case 11:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
          break;
        case 12:
          this.constantPoolOffsets[i] = readOffset;
          readOffset += 5;
        case 2:
        }
      }
      this.accessFlags = u2At(readOffset);
      readOffset += 2;

      this.classNameIndex = u2At(readOffset);
      this.className = getConstantClassNameAt(this.classNameIndex);
      readOffset += 2;

      int superclassNameIndex = u2At(readOffset);
      readOffset += 2;

      if (superclassNameIndex != 0) {
        this.superclassName = getConstantClassNameAt(superclassNameIndex);
      }

      this.interfacesCount = u2At(readOffset);
      readOffset += 2;
      if (this.interfacesCount != 0) {
        this.interfaceNames = new char[this.interfacesCount][];
        for (int i = 0; i < this.interfacesCount; i++) {
          this.interfaceNames[i] = getConstantClassNameAt(u2At(readOffset));
          readOffset += 2;
        }
      }

      this.fieldsCount = u2At(readOffset);
      readOffset += 2;
      if (this.fieldsCount != 0)
      {
        this.fields = new FieldInfo[this.fieldsCount];
        for (int i = 0; i < this.fieldsCount; i++) {
          FieldInfo field = FieldInfo.createField(this.reference, this.constantPoolOffsets, readOffset);
          this.fields[i] = field;
          readOffset += field.sizeInBytes();
        }
      }

      this.methodsCount = u2At(readOffset);
      readOffset += 2;
      if (this.methodsCount != 0) {
        this.methods = new MethodInfo[this.methodsCount];
        boolean isAnnotationType = (this.accessFlags & 0x2000) != 0;
        for (int i = 0; i < this.methodsCount; i++) {
          this.methods[i] = (isAnnotationType ? 
            AnnotationMethodInfo.createAnnotationMethod(this.reference, this.constantPoolOffsets, readOffset) : 
            MethodInfo.createMethod(this.reference, this.constantPoolOffsets, readOffset));
          readOffset += this.methods[i].sizeInBytes();
        }

      }

      int attributesCount = u2At(readOffset);
      readOffset += 2;

      for (int i = 0; i < attributesCount; i++) {
        int utf8Offset = this.constantPoolOffsets[u2At(readOffset)];
        char[] attributeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
        if (attributeName.length == 0) {
          readOffset = (int)(readOffset + (6L + u4At(readOffset + 2)));
        }
        else {
          switch (attributeName[0]) {
          case 'E':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.EnclosingMethodName)) {
              utf8Offset = 
                this.constantPoolOffsets[u2At(this.constantPoolOffsets[u2At(readOffset + 6)] - this.structOffset + 1)] - this.structOffset;
              this.enclosingTypeName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
            }
            break;
          case 'D':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.DeprecatedName)) {
              this.accessFlags |= 1048576;
            }
            break;
          case 'I':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.InnerClassName)) {
              int innerOffset = readOffset + 6;
              int number_of_classes = u2At(innerOffset);
              if (number_of_classes != 0) {
                innerOffset += 2;
                this.innerInfos = new InnerClassInfo[number_of_classes];
                for (int j = 0; j < number_of_classes; j++) {
                  this.innerInfos[j] = 
                    new InnerClassInfo(this.reference, this.constantPoolOffsets, innerOffset);
                  if (this.classNameIndex == this.innerInfos[j].innerClassNameIndex) {
                    this.innerInfo = this.innerInfos[j];
                    this.innerInfoIndex = j;
                  }
                  innerOffset += 8;
                }
                if (this.innerInfo != null) {
                  char[] enclosingType = this.innerInfo.getEnclosingTypeName();
                  if (enclosingType != null)
                    this.enclosingTypeName = enclosingType;
                }
              }
            }
            else if (CharOperation.equals(attributeName, AttributeNamesConstants.InconsistentHierarchy)) {
              this.tagBits |= 128L;
            }
            break;
          case 'S':
            if (attributeName.length > 2)
              switch (attributeName[1]) {
              case 'o':
                if (CharOperation.equals(attributeName, AttributeNamesConstants.SourceName)) {
                  utf8Offset = this.constantPoolOffsets[u2At(readOffset + 6)];
                  this.sourceFileName = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
                }
                break;
              case 'y':
                if (CharOperation.equals(attributeName, AttributeNamesConstants.SyntheticName)) {
                  this.accessFlags |= 4096;
                }
                break;
              case 'i':
                if (CharOperation.equals(attributeName, AttributeNamesConstants.SignatureName)) {
                  utf8Offset = this.constantPoolOffsets[u2At(readOffset + 6)];
                  this.signature = utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
                }
                break;
              }
            break;
          case 'R':
            if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeVisibleAnnotationsName))
              decodeAnnotations(readOffset, true);
            else if (CharOperation.equals(attributeName, AttributeNamesConstants.RuntimeInvisibleAnnotationsName)) {
              decodeAnnotations(readOffset, false);
            }
            break;
          }
          readOffset = (int)(readOffset + (
            6L + u4At(readOffset + 2)));
        }
      }
      if (fullyInitialize)
        initialize();
    }
    catch (ClassFormatException e) {
      throw e;
    } catch (Exception localException) {
      throw new ClassFormatException(
        21, 
        readOffset);
    }
  }

  public int accessFlags()
  {
    return this.accessFlags;
  }
  private void decodeAnnotations(int offset, boolean runtimeVisible) {
    int numberOfAnnotations = u2At(offset + 6);
    if (numberOfAnnotations > 0) {
      int readOffset = offset + 8;
      AnnotationInfo[] newInfos = (AnnotationInfo[])null;
      int newInfoCount = 0;
      for (int i = 0; i < numberOfAnnotations; i++)
      {
        AnnotationInfo newInfo = new AnnotationInfo(this.reference, this.constantPoolOffsets, readOffset, runtimeVisible, false);
        readOffset += newInfo.readOffset;
        long standardTagBits = newInfo.standardAnnotationTagBits;
        if (standardTagBits != 0L) {
          this.tagBits |= standardTagBits;
        } else {
          if (newInfos == null)
            newInfos = new AnnotationInfo[numberOfAnnotations - i];
          newInfos[(newInfoCount++)] = newInfo;
        }
      }
      if (newInfos == null) {
        return;
      }
      if (this.annotations == null) {
        if (newInfoCount != newInfos.length)
          System.arraycopy(newInfos, 0, newInfos = new AnnotationInfo[newInfoCount], 0, newInfoCount);
        this.annotations = newInfos;
      } else {
        int length = this.annotations.length;
        AnnotationInfo[] temp = new AnnotationInfo[length + newInfoCount];
        System.arraycopy(this.annotations, 0, temp, 0, length);
        System.arraycopy(newInfos, 0, temp, length, newInfoCount);
        this.annotations = temp;
      }
    }
  }

  private char[] getConstantClassNameAt(int constantPoolIndex)
  {
    int utf8Offset = this.constantPoolOffsets[u2At(this.constantPoolOffsets[constantPoolIndex] + 1)];
    return utf8At(utf8Offset + 3, u2At(utf8Offset + 1));
  }

  public int[] getConstantPoolOffsets()
  {
    return this.constantPoolOffsets;
  }

  public char[] getEnclosingTypeName()
  {
    return this.enclosingTypeName;
  }

  public IBinaryField[] getFields()
  {
    return this.fields;
  }

  public char[] getFileName()
  {
    return this.classFileName;
  }
  public char[] getGenericSignature() {
    return this.signature;
  }

  public char[] getInnerSourceName()
  {
    if (this.innerInfo != null)
      return this.innerInfo.getSourceName();
    return null;
  }

  public char[][] getInterfaceNames()
  {
    return this.interfaceNames;
  }

  public IBinaryNestedType[] getMemberTypes()
  {
    if (this.innerInfos == null) return null;

    int length = this.innerInfos.length;
    int startingIndex = this.innerInfo != null ? this.innerInfoIndex + 1 : 0;
    if (length != startingIndex) {
      IBinaryNestedType[] memberTypes = 
        new IBinaryNestedType[length - this.innerInfoIndex];
      int memberTypeIndex = 0;
      for (int i = startingIndex; i < length; i++) {
        InnerClassInfo currentInnerInfo = this.innerInfos[i];
        int outerClassNameIdx = currentInnerInfo.outerClassNameIndex;
        int innerNameIndex = currentInnerInfo.innerNameIndex;

        if ((outerClassNameIdx != 0) && 
          (innerNameIndex != 0) && 
          (outerClassNameIdx == this.classNameIndex) && 
          (currentInnerInfo.getSourceName().length != 0)) {
          memberTypes[(memberTypeIndex++)] = currentInnerInfo;
        }
      }
      if (memberTypeIndex == 0) return null;
      if (memberTypeIndex != memberTypes.length)
      {
        System.arraycopy(
          memberTypes, 
          0, 
          memberTypes = new IBinaryNestedType[memberTypeIndex], 
          0, 
          memberTypeIndex);
      }
      return memberTypes;
    }
    return null;
  }

  public IBinaryMethod[] getMethods()
  {
    return this.methods;
  }

  public IBinaryAnnotation[] getAnnotations()
  {
    return this.annotations;
  }

  public int getModifiers()
  {
    if (this.innerInfo != null) {
      return this.innerInfo.getModifiers() | this.accessFlags & 0x100000;
    }
    return this.accessFlags;
  }

  public char[] getName()
  {
    return this.className;
  }
  public char[] getSourceName() {
    if (this.sourceName != null) {
      return this.sourceName;
    }
    char[] name = getInnerSourceName();
    if (name == null) {
      name = getName();
      int start;
      int start;
      if (isAnonymous())
        start = CharOperation.indexOf('$', name, CharOperation.lastIndexOf('/', name) + 1) + 1;
      else {
        start = CharOperation.lastIndexOf('/', name) + 1;
      }
      if (start > 0) {
        char[] newName = new char[name.length - start];
        System.arraycopy(name, start, newName, 0, newName.length);
        name = newName;
      }
    }
    return this.sourceName = name;
  }

  public char[] getSuperclassName()
  {
    return this.superclassName;
  }
  public long getTagBits() {
    return this.tagBits;
  }

  public long getVersion()
  {
    return this.version;
  }
  private boolean hasNonSyntheticFieldChanges(FieldInfo[] currentFieldInfos, FieldInfo[] otherFieldInfos) {
    int length1 = currentFieldInfos == null ? 0 : currentFieldInfos.length;
    int length2 = otherFieldInfos == null ? 0 : otherFieldInfos.length;
    int index1 = 0;
    int index2 = 0;
    do
    {
      index1++;

      while (currentFieldInfos[index1].isSynthetic())
        if (index1 >= length1)
          break;
      for (; otherFieldInfos[index2].isSynthetic(); 
        index2 >= length2) index2++;

      if (hasStructuralFieldChanges(currentFieldInfos[(index1++)], otherFieldInfos[(index2++)]))
        return true;
      if (index1 >= length1) break;  } while (index2 < length2);

    while (index1 < length1) {
      if (!currentFieldInfos[(index1++)].isSynthetic()) return true;
    }
    while (index2 < length2) {
      if (!otherFieldInfos[(index2++)].isSynthetic()) return true;
    }
    return false;
  }
  private boolean hasNonSyntheticMethodChanges(MethodInfo[] currentMethodInfos, MethodInfo[] otherMethodInfos) {
    int length1 = currentMethodInfos == null ? 0 : currentMethodInfos.length;
    int length2 = otherMethodInfos == null ? 0 : otherMethodInfos.length;
    int index1 = 0;
    int index2 = 0;
    do
    {
      index1++;
      MethodInfo m;
      while (((m = currentMethodInfos[index1]).isSynthetic()) || (m.isClinit()))
      {
        MethodInfo m;
        if (index1 >= length1) break;
      }
      for (; ((m = otherMethodInfos[index2]).isSynthetic()) || (m.isClinit()); 
        index2 >= length2) index2++;

      if (hasStructuralMethodChanges(currentMethodInfos[(index1++)], otherMethodInfos[(index2++)]))
        return true;
      if (index1 >= length1) break;  } while (index2 < length2);

    label167: for (; index1 < length1; 
      return true)
    {
      MethodInfo m;
      if (((m = currentMethodInfos[(index1++)]).isSynthetic()) || (m.isClinit())) break label167;
    }
    label202: for (; index2 < length2; 
      return true)
    {
      MethodInfo m;
      if (((m = otherMethodInfos[(index2++)]).isSynthetic()) || (m.isClinit())) break label202;
    }
    return false;
  }

  public boolean hasStructuralChanges(byte[] newBytes)
  {
    return hasStructuralChanges(newBytes, true, true);
  }

  public boolean hasStructuralChanges(byte[] newBytes, boolean orderRequired, boolean excludesSynthetic)
  {
    try
    {
      ClassFileReader newClassFile = 
        new ClassFileReader(newBytes, this.classFileName);

      if (getModifiers() != newClassFile.getModifiers()) {
        return true;
      }

      long OnlyStructuralTagBits = 140703128617088L;

      if ((getTagBits() & OnlyStructuralTagBits) != (newClassFile.getTagBits() & OnlyStructuralTagBits)) {
        return true;
      }

      if (!CharOperation.equals(getGenericSignature(), newClassFile.getGenericSignature())) {
        return true;
      }
      if (!CharOperation.equals(getSuperclassName(), newClassFile.getSuperclassName())) {
        return true;
      }
      char[][] newInterfacesNames = newClassFile.getInterfaceNames();
      if (this.interfaceNames != newInterfacesNames) {
        int newInterfacesLength = newInterfacesNames == null ? 0 : newInterfacesNames.length;
        if (newInterfacesLength != this.interfacesCount)
          return true;
        int i = 0; for (int max = this.interfacesCount; i < max; i++) {
          if (!CharOperation.equals(this.interfaceNames[i], newInterfacesNames[i])) {
            return true;
          }
        }
      }
      IBinaryNestedType[] currentMemberTypes = getMemberTypes();
      IBinaryNestedType[] otherMemberTypes = newClassFile.getMemberTypes();
      if (currentMemberTypes != otherMemberTypes) {
        int currentMemberTypeLength = currentMemberTypes == null ? 0 : currentMemberTypes.length;
        int otherMemberTypeLength = otherMemberTypes == null ? 0 : otherMemberTypes.length;
        if (currentMemberTypeLength != otherMemberTypeLength)
          return true;
        for (int i = 0; i < currentMemberTypeLength; i++) {
          if ((!CharOperation.equals(currentMemberTypes[i].getName(), otherMemberTypes[i].getName())) || 
            (currentMemberTypes[i].getModifiers() != otherMemberTypes[i].getModifiers())) {
            return true;
          }
        }
      }
      FieldInfo[] otherFieldInfos = (FieldInfo[])newClassFile.getFields();
      int otherFieldInfosLength = otherFieldInfos == null ? 0 : otherFieldInfos.length;
      boolean compareFields = true;
      if (this.fieldsCount == otherFieldInfosLength) {
        for (int i = 0; 
          i < this.fieldsCount; i++)
          if (hasStructuralFieldChanges(this.fields[i], otherFieldInfos[i])) break;
        if (((compareFields = i != this.fieldsCount ? 1 : 0) != 0) && (!orderRequired) && (!excludesSynthetic))
          return true;
      }
      if (compareFields) {
        if ((this.fieldsCount != otherFieldInfosLength) && (!excludesSynthetic))
          return true;
        if (orderRequired) {
          if (this.fieldsCount != 0)
            Arrays.sort(this.fields);
          if (otherFieldInfosLength != 0)
            Arrays.sort(otherFieldInfos);
        }
        if (excludesSynthetic) {
          if (hasNonSyntheticFieldChanges(this.fields, otherFieldInfos))
            return true;
        }
        else for (int i = 0; i < this.fieldsCount; i++) {
            if (hasStructuralFieldChanges(this.fields[i], otherFieldInfos[i])) {
              return true;
            }
          }
      }

      MethodInfo[] otherMethodInfos = (MethodInfo[])newClassFile.getMethods();
      int otherMethodInfosLength = otherMethodInfos == null ? 0 : otherMethodInfos.length;
      boolean compareMethods = true;
      if (this.methodsCount == otherMethodInfosLength) {
        for (int i = 0; 
          i < this.methodsCount; i++)
          if (hasStructuralMethodChanges(this.methods[i], otherMethodInfos[i])) break;
        if (((compareMethods = i != this.methodsCount ? 1 : 0) != 0) && (!orderRequired) && (!excludesSynthetic))
          return true;
      }
      if (compareMethods) {
        if ((this.methodsCount != otherMethodInfosLength) && (!excludesSynthetic))
          return true;
        if (orderRequired) {
          if (this.methodsCount != 0)
            Arrays.sort(this.methods);
          if (otherMethodInfosLength != 0)
            Arrays.sort(otherMethodInfos);
        }
        if (excludesSynthetic) {
          if (hasNonSyntheticMethodChanges(this.methods, otherMethodInfos))
            return true;
        }
        else for (int i = 0; i < this.methodsCount; i++) {
            if (hasStructuralMethodChanges(this.methods[i], otherMethodInfos[i])) {
              return true;
            }
          }
      }
      return false; } catch (ClassFormatException localClassFormatException) {
    }
    return true;
  }

  private boolean hasStructuralFieldChanges(FieldInfo currentFieldInfo, FieldInfo otherFieldInfo)
  {
    if (!CharOperation.equals(currentFieldInfo.getGenericSignature(), otherFieldInfo.getGenericSignature()))
      return true;
    if (currentFieldInfo.getModifiers() != otherFieldInfo.getModifiers())
      return true;
    if ((currentFieldInfo.getTagBits() & 0x0) != (otherFieldInfo.getTagBits() & 0x0))
      return true;
    if (!CharOperation.equals(currentFieldInfo.getName(), otherFieldInfo.getName()))
      return true;
    if (!CharOperation.equals(currentFieldInfo.getTypeName(), otherFieldInfo.getTypeName()))
      return true;
    if (currentFieldInfo.hasConstant() != otherFieldInfo.hasConstant())
      return true;
    if (currentFieldInfo.hasConstant()) {
      Constant currentConstant = currentFieldInfo.getConstant();
      Constant otherConstant = otherFieldInfo.getConstant();
      if (currentConstant.typeID() != otherConstant.typeID())
        return true;
      if (!currentConstant.getClass().equals(otherConstant.getClass()))
        return true;
      switch (currentConstant.typeID()) {
      case 10:
        return currentConstant.intValue() != otherConstant.intValue();
      case 3:
        return currentConstant.byteValue() != otherConstant.byteValue();
      case 4:
        return currentConstant.shortValue() != otherConstant.shortValue();
      case 2:
        return currentConstant.charValue() != otherConstant.charValue();
      case 7:
        return currentConstant.longValue() != otherConstant.longValue();
      case 9:
        return currentConstant.floatValue() != otherConstant.floatValue();
      case 8:
        return currentConstant.doubleValue() != otherConstant.doubleValue();
      case 5:
        return currentConstant.booleanValue() ^ otherConstant.booleanValue();
      case 11:
        return !currentConstant.stringValue().equals(otherConstant.stringValue());
      case 6:
      }
    }
    return false;
  }

  private boolean hasStructuralMethodChanges(MethodInfo currentMethodInfo, MethodInfo otherMethodInfo) {
    if (!CharOperation.equals(currentMethodInfo.getGenericSignature(), otherMethodInfo.getGenericSignature()))
      return true;
    if (currentMethodInfo.getModifiers() != otherMethodInfo.getModifiers())
      return true;
    if ((currentMethodInfo.getTagBits() & 0x0) != (otherMethodInfo.getTagBits() & 0x0))
      return true;
    if (!CharOperation.equals(currentMethodInfo.getSelector(), otherMethodInfo.getSelector()))
      return true;
    if (!CharOperation.equals(currentMethodInfo.getMethodDescriptor(), otherMethodInfo.getMethodDescriptor()))
      return true;
    if (!CharOperation.equals(currentMethodInfo.getGenericSignature(), otherMethodInfo.getGenericSignature())) {
      return true;
    }
    char[][] currentThrownExceptions = currentMethodInfo.getExceptionTypeNames();
    char[][] otherThrownExceptions = otherMethodInfo.getExceptionTypeNames();
    if (currentThrownExceptions != otherThrownExceptions) {
      int currentThrownExceptionsLength = currentThrownExceptions == null ? 0 : currentThrownExceptions.length;
      int otherThrownExceptionsLength = otherThrownExceptions == null ? 0 : otherThrownExceptions.length;
      if (currentThrownExceptionsLength != otherThrownExceptionsLength)
        return true;
      for (int k = 0; k < currentThrownExceptionsLength; k++)
        if (!CharOperation.equals(currentThrownExceptions[k], otherThrownExceptions[k]))
          return true;
    }
    return false;
  }

  private void initialize()
    throws ClassFormatException
  {
    try
    {
      int i = 0; for (int max = this.fieldsCount; i < max; i++) {
        this.fields[i].initialize();
      }
      int i = 0; for (int max = this.methodsCount; i < max; i++) {
        this.methods[i].initialize();
      }
      if (this.innerInfos != null) {
        int i = 0; for (int max = this.innerInfos.length; i < max; i++) {
          this.innerInfos[i].initialize();
        }
      }
      if (this.annotations != null) {
        int i = 0; for (int max = this.annotations.length; i < max; i++) {
          this.annotations[i].initialize();
        }
      }
      reset();
    } catch (RuntimeException e) {
      ClassFormatException exception = new ClassFormatException(e, this.classFileName);
      throw exception;
    }
  }

  public boolean isAnonymous()
  {
    if (this.innerInfo == null) return false;
    char[] innerSourceName = this.innerInfo.getSourceName();
    return (innerSourceName == null) || (innerSourceName.length == 0);
  }

  public boolean isBinaryType()
  {
    return true;
  }

  public boolean isLocal()
  {
    if (this.innerInfo == null) return false;
    if (this.innerInfo.getEnclosingTypeName() != null) return false;
    char[] innerSourceName = this.innerInfo.getSourceName();
    return (innerSourceName != null) && (innerSourceName.length > 0);
  }

  public boolean isMember()
  {
    if (this.innerInfo == null) return false;
    if (this.innerInfo.getEnclosingTypeName() == null) return false;
    char[] innerSourceName = this.innerInfo.getSourceName();
    return (innerSourceName != null) && (innerSourceName.length > 0);
  }

  public boolean isNestedType()
  {
    return this.innerInfo != null;
  }

  public char[] sourceFileName()
  {
    return this.sourceFileName;
  }
  public String toString() {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    PrintWriter print = new PrintWriter(out);
    print.println(getClass().getName() + "{");
    print.println(" this.className: " + new String(getName()));
    print.println(" this.superclassName: " + (getSuperclassName() == null ? "null" : new String(getSuperclassName())));
    print.println(" access_flags: " + printTypeModifiers(accessFlags()) + "(" + accessFlags() + ")");
    print.flush();
    return out.toString();
  }
}