package org.eclipse.jdt.internal.compiler.flow;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class FinallyFlowContext extends FlowContext
{
  Reference[] finalAssignments;
  VariableBinding[] finalVariables;
  int assignCount;
  LocalVariableBinding[] nullLocals;
  Expression[] nullReferences;
  int[] nullCheckTypes;
  int nullCount;

  public FinallyFlowContext(FlowContext parent, ASTNode associatedNode)
  {
    super(parent, associatedNode);
  }

  public void complainOnDeferredChecks(FlowInfo flowInfo, BlockScope scope)
  {
    for (int i = 0; i < this.assignCount; i++) {
      VariableBinding variable = this.finalVariables[i];
      if (variable != null)
      {
        boolean complained = false;
        if ((variable instanceof FieldBinding))
        {
          if (flowInfo.isPotentiallyAssigned((FieldBinding)variable)) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfBlankFinalField((FieldBinding)variable, this.finalAssignments[i]);
          }

        }
        else if (flowInfo.isPotentiallyAssigned((LocalVariableBinding)variable)) {
          complained = true;
          scope.problemReporter().duplicateInitializationOfFinalLocal(
            (LocalVariableBinding)variable, 
            this.finalAssignments[i]);
        }

        if (complained) {
          FlowContext currentContext = this.parent;
          while (currentContext != null)
          {
            currentContext.removeFinalAssignmentIfAny(this.finalAssignments[i]);

            currentContext = currentContext.parent;
          }
        }
      }
    }

    if (this.deferNullDiagnostic) {
      for (int i = 0; i < this.nullCount; i++) {
        this.parent.recordUsingNullReference(scope, this.nullLocals[i], 
          this.nullReferences[i], this.nullCheckTypes[i], flowInfo);
      }
    }
    else
      for (int i = 0; i < this.nullCount; i++) {
        Expression expression = this.nullReferences[i];

        LocalVariableBinding local = this.nullLocals[i];
        switch (this.nullCheckTypes[i]) {
        case 256:
        case 512:
          if (flowInfo.isDefinitelyNonNull(local)) {
            if (this.nullCheckTypes[i] == 512)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, expression);
            else
              scope.problemReporter().localVariableNonNullComparedToNull(local, expression);
          }
          break;
        case 257:
        case 513:
        case 769:
        case 1025:
          if (flowInfo.isDefinitelyNull(local)) {
            switch (this.nullCheckTypes[i] & 0xFFFFFF00) {
            case 256:
              scope.problemReporter().localVariableRedundantCheckOnNull(local, expression);
              break;
            case 512:
              scope.problemReporter().localVariableNullComparedToNonNull(local, expression);
              break;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, expression);
              break;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, expression);
            }
          }

          break;
        case 3:
          if (flowInfo.isDefinitelyNull(local)) {
            scope.problemReporter().localVariableNullReference(local, expression);
          }
          else if (flowInfo.isPotentiallyNull(local))
            scope.problemReporter().localVariablePotentialNullReference(local, expression);
          break;
        }
      }
  }

  public String individualToString()
  {
    StringBuffer buffer = new StringBuffer("Finally flow context");
    buffer.append("[finalAssignments count - ").append(this.assignCount).append(']');
    buffer.append("[nullReferences count - ").append(this.nullCount).append(']');
    return buffer.toString();
  }

  public boolean isSubRoutine() {
    return true;
  }

  protected boolean recordFinalAssignment(VariableBinding binding, Reference finalAssignment)
  {
    if (this.assignCount == 0) {
      this.finalAssignments = new Reference[5];
      this.finalVariables = new VariableBinding[5];
    } else {
      if (this.assignCount == this.finalAssignments.length)
        System.arraycopy(
          this.finalAssignments, 
          0, 
          this.finalAssignments = new Reference[this.assignCount * 2], 
          0, 
          this.assignCount);
      System.arraycopy(
        this.finalVariables, 
        0, 
        this.finalVariables = new VariableBinding[this.assignCount * 2], 
        0, 
        this.assignCount);
    }
    this.finalAssignments[this.assignCount] = finalAssignment;
    this.finalVariables[(this.assignCount++)] = binding;
    return true;
  }

  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, Expression reference, int checkType, FlowInfo flowInfo)
  {
    if (((flowInfo.tagBits & 0x1) == 0) && (!flowInfo.isDefinitelyUnknown(local))) {
      if (this.deferNullDiagnostic) {
        switch (checkType) {
        case 256:
        case 257:
        case 512:
        case 513:
        case 769:
        case 1025:
          if (flowInfo.cannotBeNull(local)) {
            if (checkType == 512)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, reference);
            else {
              scope.problemReporter().localVariableNonNullComparedToNull(local, reference);
            }
            return;
          }
          if (!flowInfo.canOnlyBeNull(local)) break;
          switch (checkType & 0xFFFFFF00) {
          case 256:
            scope.problemReporter().localVariableRedundantCheckOnNull(local, reference);
            return;
          case 512:
            scope.problemReporter().localVariableNullComparedToNonNull(local, reference);
            return;
          case 768:
            scope.problemReporter().localVariableRedundantNullAssignment(local, reference);
            return;
          case 1024:
            scope.problemReporter().localVariableNullInstanceof(local, reference);
            return;
          }

          break;
        case 3:
          if (flowInfo.cannotBeNull(local)) {
            return;
          }
          if (!flowInfo.canOnlyBeNull(local)) break;
          scope.problemReporter().localVariableNullReference(local, reference);
          return;
        default:
          break;
        }

      }
      else
      {
        switch (checkType) {
        case 256:
        case 512:
          if (flowInfo.isDefinitelyNonNull(local)) {
            if (checkType == 512)
              scope.problemReporter().localVariableRedundantCheckOnNonNull(local, reference);
            else {
              scope.problemReporter().localVariableNonNullComparedToNull(local, reference);
            }
            return;
          }
        case 257:
        case 513:
        case 769:
        case 1025:
          if (flowInfo.isDefinitelyNull(local)) {
            switch (checkType & 0xFFFFFF00) {
            case 256:
              scope.problemReporter().localVariableRedundantCheckOnNull(local, reference);
              return;
            case 512:
              scope.problemReporter().localVariableNullComparedToNonNull(local, reference);
              return;
            case 768:
              scope.problemReporter().localVariableRedundantNullAssignment(local, reference);
              return;
            case 1024:
              scope.problemReporter().localVariableNullInstanceof(local, reference);
              return;
            }
          }
          break;
        case 3:
          if (flowInfo.isDefinitelyNull(local)) {
            scope.problemReporter().localVariableNullReference(local, reference);
            return;
          }
          if (flowInfo.isPotentiallyNull(local)) {
            scope.problemReporter().localVariablePotentialNullReference(local, reference);
            return;
          }
          if (flowInfo.isDefinitelyNonNull(local))
          {
            return;
          }
          break;
        }
      }

      recordNullReference(local, reference, checkType);
    }
  }

  void removeFinalAssignmentIfAny(Reference reference)
  {
    for (int i = 0; i < this.assignCount; i++)
      if (this.finalAssignments[i] == reference) {
        this.finalAssignments[i] = null;
        this.finalVariables[i] = null;
        return;
      }
  }

  protected void recordNullReference(LocalVariableBinding local, Expression expression, int status)
  {
    if (this.nullCount == 0) {
      this.nullLocals = new LocalVariableBinding[5];
      this.nullReferences = new Expression[5];
      this.nullCheckTypes = new int[5];
    }
    else if (this.nullCount == this.nullLocals.length) {
      int newLength = this.nullCount * 2;
      System.arraycopy(this.nullLocals, 0, 
        this.nullLocals = new LocalVariableBinding[newLength], 0, 
        this.nullCount);
      System.arraycopy(this.nullReferences, 0, 
        this.nullReferences = new Expression[newLength], 0, 
        this.nullCount);
      System.arraycopy(this.nullCheckTypes, 0, 
        this.nullCheckTypes = new int[newLength], 0, 
        this.nullCount);
    }
    this.nullLocals[this.nullCount] = local;
    this.nullReferences[this.nullCount] = expression;
    this.nullCheckTypes[(this.nullCount++)] = status;
  }
}