package org.eclipse.jdt.internal.compiler.lookup;

import java.util.HashMap;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.QualifiedAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;

public class ClassScope extends Scope
{
  public TypeDeclaration referenceContext;
  public TypeReference superTypeReference;

  public ClassScope(Scope parent, TypeDeclaration context)
  {
    super(3, parent);
    this.referenceContext = context;
  }

  void buildAnonymousTypeBinding(SourceTypeBinding enclosingType, ReferenceBinding supertype) {
    LocalTypeBinding anonymousType = buildLocalType(enclosingType, enclosingType.fPackage);
    SourceTypeBinding sourceType = this.referenceContext.binding;
    if (supertype.isInterface()) {
      sourceType.superclass = getJavaLangObject();
      sourceType.superInterfaces = new ReferenceBinding[] { supertype };
    } else {
      sourceType.superclass = supertype;
      sourceType.superInterfaces = Binding.NO_SUPERINTERFACES;
    }
    connectMemberTypes();
    buildFieldsAndMethods();
    anonymousType.faultInTypesForFieldsAndMethods();
    sourceType.verifyMethods(environment().methodVerifier());
  }

  private void buildFields() {
    if (this.referenceContext.fields == null) {
      this.referenceContext.binding.setFields(Binding.NO_FIELDS);
      return;
    }

    FieldDeclaration[] fields = this.referenceContext.fields;
    int size = fields.length;
    int count = 0;
    for (int i = 0; i < size; i++) {
      switch (fields[i].getKind()) {
      case 1:
      case 3:
        count++;
      case 2:
      }
    }

    FieldBinding[] fieldBindings = new FieldBinding[count];
    HashtableOfObject knownFieldNames = new HashtableOfObject(count);
    boolean duplicate = false;
    count = 0;
    for (int i = 0; i < size; i++) {
      FieldDeclaration field = fields[i];
      if (field.getKind() == 2) {
        if (this.referenceContext.binding.isInterface())
          problemReporter().interfaceCannotHaveInitializers(this.referenceContext.binding, field);
      } else {
        FieldBinding fieldBinding = new FieldBinding(field, null, field.modifiers | 0x2000000, this.referenceContext.binding);
        fieldBinding.id = count;

        checkAndSetModifiersForField(fieldBinding, field);

        if (knownFieldNames.containsKey(field.name)) {
          duplicate = true;
          FieldBinding previousBinding = (FieldBinding)knownFieldNames.get(field.name);
          if (previousBinding != null) {
            for (int f = 0; f < i; f++) {
              FieldDeclaration previousField = fields[f];
              if (previousField.binding == previousBinding) {
                problemReporter().duplicateFieldInType(this.referenceContext.binding, previousField);
                previousField.binding = null;
                break;
              }
            }
          }
          knownFieldNames.put(field.name, null);
          problemReporter().duplicateFieldInType(this.referenceContext.binding, field);
          field.binding = null;
        } else {
          knownFieldNames.put(field.name, fieldBinding);

          fieldBindings[(count++)] = fieldBinding;
        }
      }
    }

    if (duplicate) {
      FieldBinding[] newFieldBindings = new FieldBinding[fieldBindings.length];

      size = count;
      count = 0;
      for (int i = 0; i < size; i++) {
        FieldBinding fieldBinding = fieldBindings[i];
        if (knownFieldNames.get(fieldBinding.name) != null) {
          fieldBinding.id = count;
          newFieldBindings[(count++)] = fieldBinding;
        }
      }
      fieldBindings = newFieldBindings;
    }
    if (count != fieldBindings.length)
      System.arraycopy(fieldBindings, 0, fieldBindings = new FieldBinding[count], 0, count);
    this.referenceContext.binding.setFields(fieldBindings);
  }

  void buildFieldsAndMethods() {
    buildFields();
    buildMethods();

    SourceTypeBinding sourceType = this.referenceContext.binding;
    if ((sourceType.isMemberType()) && (!sourceType.isLocalType())) {
      ((MemberTypeBinding)sourceType).checkSyntheticArgsAndFields();
    }
    ReferenceBinding[] memberTypes = sourceType.memberTypes;
    int i = 0; for (int length = memberTypes.length; i < length; i++)
      ((SourceTypeBinding)memberTypes[i]).scope.buildFieldsAndMethods();
  }

  private LocalTypeBinding buildLocalType(SourceTypeBinding enclosingType, PackageBinding packageBinding)
  {
    this.referenceContext.scope = this;
    this.referenceContext.staticInitializerScope = new MethodScope(this, this.referenceContext, true);
    this.referenceContext.initializerScope = new MethodScope(this, this.referenceContext, false);

    LocalTypeBinding localType = new LocalTypeBinding(this, enclosingType, innermostSwitchCase());
    this.referenceContext.binding = localType;
    checkAndSetModifiers();
    buildTypeVariables();

    ReferenceBinding[] memberTypeBindings = Binding.NO_MEMBER_TYPES;
    if (this.referenceContext.memberTypes != null) {
      int size = this.referenceContext.memberTypes.length;
      memberTypeBindings = new ReferenceBinding[size];
      int count = 0;
      for (int i = 0; i < size; i++) {
        TypeDeclaration memberContext = this.referenceContext.memberTypes[i];
        switch (TypeDeclaration.kind(memberContext.modifiers)) {
        case 2:
        case 4:
          problemReporter().illegalLocalTypeDeclaration(memberContext);
          break;
        case 3:
        default:
          ReferenceBinding type = localType;
          do
          {
            if (CharOperation.equals(type.sourceName, memberContext.name)) {
              problemReporter().typeCollidesWithEnclosingType(memberContext);
              break;
            }
            type = type.enclosingType();
          }while (type != null);

          for (int j = 0; j < i; j++) {
            if (CharOperation.equals(this.referenceContext.memberTypes[j].name, memberContext.name)) {
              problemReporter().duplicateNestedType(memberContext);
              break;
            }
          }
          ClassScope memberScope = new ClassScope(this, this.referenceContext.memberTypes[i]);
          LocalTypeBinding memberBinding = memberScope.buildLocalType(localType, packageBinding);
          memberBinding.setAsMemberType();
          memberTypeBindings[(count++)] = memberBinding;
        }
      }
      if (count != size)
        System.arraycopy(memberTypeBindings, 0, memberTypeBindings = new ReferenceBinding[count], 0, count);
    }
    localType.memberTypes = memberTypeBindings;
    return localType;
  }

  void buildLocalTypeBinding(SourceTypeBinding enclosingType)
  {
    LocalTypeBinding localType = buildLocalType(enclosingType, enclosingType.fPackage);
    connectTypeHierarchy();
    buildFieldsAndMethods();
    localType.faultInTypesForFieldsAndMethods();

    this.referenceContext.binding.verifyMethods(environment().methodVerifier());
  }

  private void buildMemberTypes(AccessRestriction accessRestriction) {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    ReferenceBinding[] memberTypeBindings = Binding.NO_MEMBER_TYPES;
    if (this.referenceContext.memberTypes != null) {
      int length = this.referenceContext.memberTypes.length;
      memberTypeBindings = new ReferenceBinding[length];
      int count = 0;
      for (int i = 0; i < length; i++) {
        TypeDeclaration memberContext = this.referenceContext.memberTypes[i];
        switch (TypeDeclaration.kind(memberContext.modifiers)) {
        case 2:
        case 4:
          if ((sourceType.isNestedType()) && 
            (sourceType.isClass()) && 
            (!sourceType.isStatic()))
            problemReporter().illegalLocalTypeDeclaration(memberContext);
          break;
        case 3:
        }

        ReferenceBinding type = sourceType;
        do
        {
          if (CharOperation.equals(type.sourceName, memberContext.name)) {
            problemReporter().typeCollidesWithEnclosingType(memberContext);
            break;
          }
          type = type.enclosingType();
        }while (type != null);

        for (int j = 0; j < i; j++) {
          if (CharOperation.equals(this.referenceContext.memberTypes[j].name, memberContext.name)) {
            problemReporter().duplicateNestedType(memberContext);
            break;
          }
        }

        ClassScope memberScope = new ClassScope(this, memberContext);
        memberTypeBindings[(count++)] = memberScope.buildType(sourceType, sourceType.fPackage, accessRestriction);
      }
      if (count != length)
        System.arraycopy(memberTypeBindings, 0, memberTypeBindings = new ReferenceBinding[count], 0, count);
    }
    sourceType.memberTypes = memberTypeBindings;
  }

  private void buildMethods() {
    boolean isEnum = TypeDeclaration.kind(this.referenceContext.modifiers) == 3;
    if ((this.referenceContext.methods == null) && (!isEnum)) {
      this.referenceContext.binding.setMethods(Binding.NO_METHODS);
      return;
    }

    AbstractMethodDeclaration[] methods = this.referenceContext.methods;
    int size = methods == null ? 0 : methods.length;

    int clinitIndex = -1;
    for (int i = 0; i < size; i++) {
      if (methods[i].isClinit()) {
        clinitIndex = i;
        break;
      }
    }

    int count = isEnum ? 2 : 0;
    MethodBinding[] methodBindings = new MethodBinding[(clinitIndex == -1 ? size : size - 1) + count];

    SourceTypeBinding sourceType = this.referenceContext.binding;
    if (isEnum) {
      methodBindings[0] = sourceType.addSyntheticEnumMethod(TypeConstants.VALUES);
      methodBindings[1] = sourceType.addSyntheticEnumMethod(TypeConstants.VALUEOF);
    }

    for (int i = 0; i < size; i++) {
      if (i != clinitIndex) {
        MethodScope scope = new MethodScope(this, methods[i], false);
        MethodBinding methodBinding = scope.createMethod(methods[i]);
        if (methodBinding != null)
          methodBindings[(count++)] = methodBinding;
      }
    }
    if (count != methodBindings.length)
      System.arraycopy(methodBindings, 0, methodBindings = new MethodBinding[count], 0, count);
    sourceType.tagBits &= -16385L;
    sourceType.setMethods(methodBindings);
  }

  SourceTypeBinding buildType(SourceTypeBinding enclosingType, PackageBinding packageBinding, AccessRestriction accessRestriction)
  {
    this.referenceContext.scope = this;
    this.referenceContext.staticInitializerScope = new MethodScope(this, this.referenceContext, true);
    this.referenceContext.initializerScope = new MethodScope(this, this.referenceContext, false);

    if (enclosingType == null) {
      char[][] className = CharOperation.arrayConcat(packageBinding.compoundName, this.referenceContext.name);
      this.referenceContext.binding = new SourceTypeBinding(className, packageBinding, this);
    } else {
      char[][] className = CharOperation.deepCopy(enclosingType.compoundName);
      className[(className.length - 1)] = 
        CharOperation.concat(className[(className.length - 1)], this.referenceContext.name, '$');
      ReferenceBinding existingType = packageBinding.getType0(className[(className.length - 1)]);
      if ((existingType != null) && 
        (!(existingType instanceof UnresolvedReferenceBinding)))
      {
        this.parent.problemReporter().duplicateNestedType(this.referenceContext);
      }

      this.referenceContext.binding = new MemberTypeBinding(className, this, enclosingType);
    }

    SourceTypeBinding sourceType = this.referenceContext.binding;
    environment().setAccessRestriction(sourceType, accessRestriction);
    sourceType.fPackage.addType(sourceType);
    checkAndSetModifiers();
    buildTypeVariables();
    buildMemberTypes(accessRestriction);
    return sourceType;
  }

  private void buildTypeVariables()
  {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    TypeParameter[] typeParameters = this.referenceContext.typeParameters;

    if ((typeParameters == null) || (compilerOptions().sourceLevel < 3211264L)) {
      sourceType.typeVariables = Binding.NO_TYPE_VARIABLES;
      return;
    }
    sourceType.typeVariables = Binding.NO_TYPE_VARIABLES;

    if (sourceType.id == 1) {
      problemReporter().objectCannotBeGeneric(this.referenceContext);
      return;
    }
    sourceType.typeVariables = createTypeVariables(typeParameters, sourceType);
    sourceType.modifiers |= 1073741824;
  }

  private void checkAndSetModifiers() {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    int modifiers = sourceType.modifiers;
    if ((modifiers & 0x400000) != 0)
      problemReporter().duplicateModifierForType(sourceType);
    ReferenceBinding enclosingType = sourceType.enclosingType();
    boolean isMemberType = sourceType.isMemberType();
    if (isMemberType) {
      modifiers |= enclosingType.modifiers & 0x40000800;

      if (enclosingType.isInterface())
        modifiers |= 1;
      if (sourceType.isEnum()) {
        if (!enclosingType.isStatic())
          problemReporter().nonStaticContextForEnumMemberType(sourceType);
        else
          modifiers |= 8;
      }
      if ((enclosingType.isViewedAsDeprecated()) && (!sourceType.isDeprecated()))
        modifiers |= 2097152;
    } else if (sourceType.isLocalType()) {
      if (sourceType.isEnum()) {
        problemReporter().illegalLocalTypeDeclaration(this.referenceContext);
        sourceType.modifiers = 0;
        return;
      }
      if (sourceType.isAnonymousType()) {
        modifiers |= 16;

        if (this.referenceContext.allocation.type == null)
          modifiers |= 16384;
      }
      Scope scope = this;
      do {
        switch (scope.kind) {
        case 2:
          MethodScope methodScope = (MethodScope)scope;
          if (methodScope.isInsideInitializer()) {
            SourceTypeBinding type = ((TypeDeclaration)methodScope.referenceContext).binding;

            if (methodScope.initializedField != null)
            {
              if ((methodScope.initializedField.isViewedAsDeprecated()) && (!sourceType.isDeprecated()))
                modifiers |= 2097152;
            } else {
              if (type.isStrictfp())
                modifiers |= 2048;
              if ((type.isViewedAsDeprecated()) && (!sourceType.isDeprecated()))
                modifiers |= 2097152;
            }
          } else {
            MethodBinding method = ((AbstractMethodDeclaration)methodScope.referenceContext).binding;
            if (method != null) {
              if (method.isStrictfp())
                modifiers |= 2048;
              if ((method.isViewedAsDeprecated()) && (!sourceType.isDeprecated()))
                modifiers |= 2097152;
            }
          }
          break;
        case 3:
          if (enclosingType.isStrictfp())
            modifiers |= 2048;
          if ((enclosingType.isViewedAsDeprecated()) && (!sourceType.isDeprecated()))
            modifiers |= 2097152;
          break;
        }
        scope = scope.parent;
      }
      while (scope != null);
    }

    int realModifiers = modifiers & 0xFFFF;

    if ((realModifiers & 0x200) != 0)
    {
      if (isMemberType)
      {
        if ((realModifiers & 0xFFFFD1F0) != 0) {
          if ((realModifiers & 0x2000) != 0)
            problemReporter().illegalModifierForAnnotationMemberType(sourceType);
          else {
            problemReporter().illegalModifierForMemberInterface(sourceType);
          }

        }

      }
      else if ((realModifiers & 0xFFFFD1FE) != 0) {
        if ((realModifiers & 0x2000) != 0)
          problemReporter().illegalModifierForAnnotationType(sourceType);
        else {
          problemReporter().illegalModifierForInterface(sourceType);
        }
      }
      modifiers |= 1024;
    } else if ((realModifiers & 0x4000) != 0)
    {
      if (isMemberType)
      {
        if ((realModifiers & 0xFFFFB7F0) != 0)
          problemReporter().illegalModifierForMemberEnum(sourceType);
      } else if (sourceType.isLocalType())
      {
        if ((realModifiers & 0xFFFFB7EF) != 0) {
          problemReporter().illegalModifierForLocalEnum(sourceType);
        }
      }
      else if ((realModifiers & 0xFFFFB7FE) != 0) {
        problemReporter().illegalModifierForEnum(sourceType);
      }
      if (!sourceType.isAnonymousType())
      {
        if ((this.referenceContext.bits & 0x800) != 0) {
          modifiers |= 1024;
        }
        else
        {
          TypeDeclaration typeDeclaration = this.referenceContext;
          FieldDeclaration[] fields = typeDeclaration.fields;
          int fieldsLength = fields == null ? 0 : fields.length;
          if (fieldsLength != 0) {
            AbstractMethodDeclaration[] methods = typeDeclaration.methods;
            int methodsLength = methods == null ? 0 : methods.length;

            boolean definesAbstractMethod = typeDeclaration.superInterfaces != null;
            for (int i = 0; (i < methodsLength) && (!definesAbstractMethod); i++)
              definesAbstractMethod = methods[i].isAbstract();
            if (definesAbstractMethod) {
              boolean needAbstractBit = false;
              for (int i = 0; i < fieldsLength; i++) {
                FieldDeclaration fieldDecl = fields[i];
                if (fieldDecl.getKind() == 3) {
                  if (!(fieldDecl.initialization instanceof QualifiedAllocationExpression)) break;
                  needAbstractBit = true;
                }

              }

              if (needAbstractBit) {
                modifiers |= 1024;
              }
            }
          }
        }
        TypeDeclaration typeDeclaration = this.referenceContext;
        FieldDeclaration[] fields = typeDeclaration.fields;
        if (fields != null) {
          int i = 0; for (int fieldsLength = fields.length; i < fieldsLength; i++) {
            FieldDeclaration fieldDecl = fields[i];
            if ((fieldDecl.getKind() == 3) && 
              ((fieldDecl.initialization instanceof QualifiedAllocationExpression)))
            {
              break;
            }
          }
        }
        modifiers |= 16;
      }
    }
    else
    {
      if (isMemberType)
      {
        if ((realModifiers & 0xFFFFF3E0) != 0)
          problemReporter().illegalModifierForMemberClass(sourceType);
      } else if (sourceType.isLocalType())
      {
        if ((realModifiers & 0xFFFFF3EF) != 0) {
          problemReporter().illegalModifierForLocalClass(sourceType);
        }
      }
      else if ((realModifiers & 0xFFFFF3EE) != 0) {
        problemReporter().illegalModifierForClass(sourceType);
      }

      if ((realModifiers & 0x410) == 1040) {
        problemReporter().illegalModifierCombinationFinalAbstractForClass(sourceType);
      }
    }
    if (isMemberType)
    {
      if (enclosingType.isInterface()) {
        if ((realModifiers & 0x6) != 0) {
          problemReporter().illegalVisibilityModifierForInterfaceMemberType(sourceType);

          if ((realModifiers & 0x4) != 0)
            modifiers &= -5;
          if ((realModifiers & 0x2) != 0)
            modifiers &= -3;
        }
      } else {
        int accessorBits = realModifiers & 0x7;
        if ((accessorBits & accessorBits - 1) > 1) {
          problemReporter().illegalVisibilityModifierCombinationForMemberType(sourceType);

          if ((accessorBits & 0x1) != 0) {
            if ((accessorBits & 0x4) != 0)
              modifiers &= -5;
            if ((accessorBits & 0x2) != 0)
              modifiers &= -3;
          } else if (((accessorBits & 0x4) != 0) && ((accessorBits & 0x2) != 0)) {
            modifiers &= -3;
          }
        }

      }

      if ((realModifiers & 0x8) == 0) {
        if (enclosingType.isInterface())
          modifiers |= 8;
      } else if (!enclosingType.isStatic())
      {
        problemReporter().illegalStaticModifierForMemberType(sourceType);
      }
    }

    sourceType.modifiers = modifiers;
  }

  private void checkAndSetModifiersForField(FieldBinding fieldBinding, FieldDeclaration fieldDecl)
  {
    int modifiers = fieldBinding.modifiers;
    ReferenceBinding declaringClass = fieldBinding.declaringClass;
    if ((modifiers & 0x400000) != 0) {
      problemReporter().duplicateModifierForField(declaringClass, fieldDecl);
    }
    if (declaringClass.isInterface())
    {
      modifiers |= 25;

      if ((modifiers & 0xFFFF) != 25) {
        if ((declaringClass.modifiers & 0x2000) != 0)
          problemReporter().illegalModifierForAnnotationField(fieldDecl);
        else
          problemReporter().illegalModifierForInterfaceField(fieldDecl);
      }
      fieldBinding.modifiers = modifiers;
      return;
    }if (fieldDecl.getKind() == 3)
    {
      if ((modifiers & 0xFFFF) != 0) {
        problemReporter().illegalModifierForEnumConstant(declaringClass, fieldDecl);
      }

      fieldBinding.modifiers |= 16409;
      return;
    }

    int realModifiers = modifiers & 0xFFFF;

    if ((realModifiers & 0xFFFFFF20) != 0) {
      problemReporter().illegalModifierForField(declaringClass, fieldDecl);
      modifiers &= -65313;
    }

    int accessorBits = realModifiers & 0x7;
    if ((accessorBits & accessorBits - 1) > 1) {
      problemReporter().illegalVisibilityModifierCombinationForField(declaringClass, fieldDecl);

      if ((accessorBits & 0x1) != 0) {
        if ((accessorBits & 0x4) != 0)
          modifiers &= -5;
        if ((accessorBits & 0x2) != 0)
          modifiers &= -3;
      } else if (((accessorBits & 0x4) != 0) && ((accessorBits & 0x2) != 0)) {
        modifiers &= -3;
      }
    }

    if ((realModifiers & 0x50) == 80) {
      problemReporter().illegalModifierCombinationFinalVolatileForField(declaringClass, fieldDecl);
    }
    if ((fieldDecl.initialization == null) && ((modifiers & 0x10) != 0))
      modifiers |= 67108864;
    fieldBinding.modifiers = modifiers;
  }

  public void checkParameterizedSuperTypeCollisions()
  {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    ReferenceBinding[] interfaces = sourceType.superInterfaces;
    Map invocations = new HashMap(2);
    ReferenceBinding itsSuperclass = sourceType.isInterface() ? null : sourceType.superclass;
    int i = 0; for (int length = interfaces.length; i < length; i++) {
      ReferenceBinding one = interfaces[i];
      if ((one != null) && (
        (itsSuperclass == null) || (!hasErasedCandidatesCollisions(itsSuperclass, one, invocations, sourceType, this.referenceContext))))
      {
        for (int j = 0; j < i; j++) {
          ReferenceBinding two = interfaces[j];
          if ((two != null) && 
            (hasErasedCandidatesCollisions(one, two, invocations, sourceType, this.referenceContext)))
            break;
        }
      }
    }
    TypeParameter[] typeParameters = this.referenceContext.typeParameters;
    int i = 0; for (int paramLength = typeParameters == null ? 0 : typeParameters.length; i < paramLength; i++) {
      TypeParameter typeParameter = typeParameters[i];
      TypeVariableBinding typeVariable = typeParameter.binding;
      if ((typeVariable != null) && (typeVariable.isValidBinding()))
      {
        TypeReference[] boundRefs = typeParameter.bounds;
        if (boundRefs != null) {
          boolean checkSuperclass = typeVariable.firstBound == typeVariable.superclass;
          int j = 0; for (int boundLength = boundRefs.length; j < boundLength; j++) {
            TypeReference typeRef = boundRefs[j];
            TypeBinding superType = typeRef.resolvedType;
            if ((superType != null) && (superType.isValidBinding()))
            {
              if ((checkSuperclass) && 
                (hasErasedCandidatesCollisions(superType, typeVariable.superclass, invocations, typeVariable, typeRef))) {
                break;
              }
              int index = typeVariable.superInterfaces.length;
              do { if (hasErasedCandidatesCollisions(superType, typeVariable.superInterfaces[index], invocations, typeVariable, typeRef))
                  break;
                index--; } while (index >= 0);
            }
          }
        }
      }
    }

    ReferenceBinding[] memberTypes = this.referenceContext.binding.memberTypes;
    if ((memberTypes != null) && (memberTypes != Binding.NO_MEMBER_TYPES)) {
      int i = 0; for (int size = memberTypes.length; i < size; i++)
        ((SourceTypeBinding)memberTypes[i]).scope.checkParameterizedSuperTypeCollisions();
    }
  }

  private void checkForInheritedMemberTypes(SourceTypeBinding sourceType)
  {
    ReferenceBinding currentType = sourceType;
    ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    do {
      if (currentType.hasMemberTypes()) {
        return;
      }
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();

      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES))
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
    }
    while (((currentType = currentType.superclass()) != null) && ((currentType.tagBits & 0x10000) == 0L));

    if (interfacesToVisit != null)
    {
      boolean needToTag = false;
      for (int i = 0; i < nextPosition; i++) {
        ReferenceBinding anInterface = interfacesToVisit[i];
        if ((anInterface.tagBits & 0x10000) == 0L) {
          if (anInterface.hasMemberTypes()) {
            return;
          }
          needToTag = true;
          ReferenceBinding[] itsInterfaces = anInterface.superInterfaces();
          if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
            int itsLength = itsInterfaces.length;
            if (nextPosition + itsLength >= interfacesToVisit.length)
              System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
            for (int a = 0; a < itsLength; a++) {
              ReferenceBinding next = itsInterfaces[a];
              int b = 0;
              while (next != interfacesToVisit[b])
              {
                b++; if (b >= nextPosition)
                {
                  interfacesToVisit[(nextPosition++)] = next;
                }
              }
            }
          }
        }
      }
      if (needToTag) {
        for (int i = 0; i < nextPosition; i++) {
          interfacesToVisit[i].tagBits |= 65536L;
        }
      }
    }

    currentType = sourceType;
    do
      currentType.tagBits |= 65536L;
    while (((currentType = currentType.superclass()) != null) && ((currentType.tagBits & 0x10000) == 0L));
  }

  public void checkParameterizedTypeBounds()
  {
    TypeReference superclass = this.referenceContext.superclass;
    if (superclass != null) {
      superclass.checkBounds(this);
    }
    TypeReference[] superinterfaces = this.referenceContext.superInterfaces;
    if (superinterfaces != null) {
      int i = 0; for (int length = superinterfaces.length; i < length; i++)
        superinterfaces[i].checkBounds(this);
    }
    TypeParameter[] typeParameters = this.referenceContext.typeParameters;
    if (typeParameters != null) {
      int i = 0; for (int paramLength = typeParameters.length; i < paramLength; i++)
        typeParameters[i].checkBounds(this);
    }
    ReferenceBinding[] memberTypes = this.referenceContext.binding.memberTypes;
    if ((memberTypes != null) && (memberTypes != Binding.NO_MEMBER_TYPES)) {
      int i = 0; for (int size = memberTypes.length; i < size; i++)
        ((SourceTypeBinding)memberTypes[i]).scope.checkParameterizedTypeBounds(); 
    }
  }

  private void connectMemberTypes() { SourceTypeBinding sourceType = this.referenceContext.binding;
    ReferenceBinding[] memberTypes = sourceType.memberTypes;
    if ((memberTypes != null) && (memberTypes != Binding.NO_MEMBER_TYPES)) {
      int i = 0; for (int size = memberTypes.length; i < size; i++)
        ((SourceTypeBinding)memberTypes[i]).scope.connectTypeHierarchy();
    }
  }

  private boolean connectSuperclass()
  {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    if (sourceType.id == 1) {
      sourceType.superclass = null;
      sourceType.superInterfaces = Binding.NO_SUPERINTERFACES;
      if (!sourceType.isClass())
        problemReporter().objectMustBeClass(sourceType);
      if ((this.referenceContext.superclass != null) || ((this.referenceContext.superInterfaces != null) && (this.referenceContext.superInterfaces.length > 0)))
        problemReporter().objectCannotHaveSuperTypes(sourceType);
      return true;
    }
    if (this.referenceContext.superclass == null) {
      if ((sourceType.isEnum()) && (compilerOptions().sourceLevel >= 3211264L))
        return connectEnumSuperclass();
      sourceType.superclass = getJavaLangObject();
      return !detectHierarchyCycle(sourceType, sourceType.superclass, null);
    }
    TypeReference superclassRef = this.referenceContext.superclass;
    ReferenceBinding superclass = findSupertype(superclassRef);
    if (superclass != null) {
      if (!superclass.isClass()) {
        problemReporter().superclassMustBeAClass(sourceType, superclassRef, superclass);
      } else if (superclass.isFinal()) {
        problemReporter().classExtendFinalClass(sourceType, superclassRef, superclass);
      } else if ((superclass.tagBits & 0x40000000) != 0L) {
        problemReporter().superTypeCannotUseWildcard(sourceType, superclassRef, superclass);
      } else if (superclass.erasure().id == 41) {
        problemReporter().cannotExtendEnum(sourceType, superclassRef, superclass);
      }
      else {
        sourceType.superclass = superclass;
        return true;
      }
    }
    sourceType.tagBits |= 131072L;
    sourceType.superclass = getJavaLangObject();
    if ((sourceType.superclass.tagBits & 0x100) == 0L)
      detectHierarchyCycle(sourceType, sourceType.superclass, null);
    return false;
  }

  private boolean connectEnumSuperclass()
  {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    ReferenceBinding rootEnumType = getJavaLangEnum();
    boolean foundCycle = detectHierarchyCycle(sourceType, rootEnumType, null);

    TypeVariableBinding[] refTypeVariables = rootEnumType.typeVariables();
    if (refTypeVariables == Binding.NO_TYPE_VARIABLES) {
      problemReporter().nonGenericTypeCannotBeParameterized(null, rootEnumType, new TypeBinding[] { sourceType });
      return false;
    }if (1 != refTypeVariables.length) {
      problemReporter().incorrectArityForParameterizedType(null, rootEnumType, new TypeBinding[] { sourceType });
      return false;
    }

    ParameterizedTypeBinding superType = environment().createParameterizedType(rootEnumType, new TypeBinding[] { environment().convertToRawType(sourceType) }, null);
    sourceType.superclass = superType;

    if (refTypeVariables[0].boundCheck(superType, sourceType) != 0) {
      problemReporter().typeMismatchError(rootEnumType, refTypeVariables[0], sourceType, null);
    }
    return !foundCycle;
  }

  private boolean connectSuperInterfaces()
  {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    sourceType.superInterfaces = Binding.NO_SUPERINTERFACES;
    if (this.referenceContext.superInterfaces == null) {
      if ((sourceType.isAnnotationType()) && (compilerOptions().sourceLevel >= 3211264L)) {
        ReferenceBinding annotationType = getJavaLangAnnotationAnnotation();
        boolean foundCycle = detectHierarchyCycle(sourceType, annotationType, null);
        sourceType.superInterfaces = new ReferenceBinding[] { annotationType };
        return !foundCycle;
      }
      return true;
    }
    if (sourceType.id == 1) {
      return true;
    }
    boolean noProblems = true;
    int length = this.referenceContext.superInterfaces.length;
    ReferenceBinding[] interfaceBindings = new ReferenceBinding[length];
    int count = 0;
    for (int i = 0; i < length; i++) {
      TypeReference superInterfaceRef = this.referenceContext.superInterfaces[i];
      ReferenceBinding superInterface = findSupertype(superInterfaceRef);
      if (superInterface == null) {
        sourceType.tagBits |= 131072L;
        noProblems = false;
      }
      else {
        superInterfaceRef.resolvedType = superInterface;

        for (int j = 0; j < i; j++) {
          if (interfaceBindings[j] == superInterface) {
            problemReporter().duplicateSuperinterface(sourceType, superInterfaceRef, superInterface);
            break;
          }
        }
        if (!superInterface.isInterface()) {
          problemReporter().superinterfaceMustBeAnInterface(sourceType, superInterfaceRef, superInterface);
          sourceType.tagBits |= 131072L;
          noProblems = false;
        } else {
          if (superInterface.isAnnotationType()) {
            problemReporter().annotationTypeUsedAsSuperinterface(sourceType, superInterfaceRef, superInterface);
          }
          if ((superInterface.tagBits & 0x40000000) != 0L) {
            problemReporter().superTypeCannotUseWildcard(sourceType, superInterfaceRef, superInterface);
            sourceType.tagBits |= 131072L;
            noProblems = false;
          }
          else
          {
            interfaceBindings[(count++)] = superInterface;
          }
        }
      }
    }
    if (count > 0) {
      if (count != length)
        System.arraycopy(interfaceBindings, 0, interfaceBindings = new ReferenceBinding[count], 0, count);
      sourceType.superInterfaces = interfaceBindings;
    }
    return noProblems;
  }

  void connectTypeHierarchy() {
    SourceTypeBinding sourceType = this.referenceContext.binding;
    if ((sourceType.tagBits & 0x100) == 0L) {
      sourceType.tagBits |= 256L;
      boolean noProblems = connectSuperclass();
      noProblems &= connectSuperInterfaces();
      sourceType.tagBits |= 512L;
      noProblems &= connectTypeVariables(this.referenceContext.typeParameters, false);
      sourceType.tagBits |= 262144L;
      if ((noProblems) && (sourceType.isHierarchyInconsistent()))
        problemReporter().hierarchyHasProblems(sourceType);
    }
    connectMemberTypes();
    LookupEnvironment env = environment();
    try {
      env.missingClassFileLocation = this.referenceContext;
      checkForInheritedMemberTypes(sourceType);
    } catch (AbortCompilation e) {
      e.updateContext(this.referenceContext, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  private void connectTypeHierarchyWithoutMembers()
  {
    if ((this.parent instanceof CompilationUnitScope)) {
      if (((CompilationUnitScope)this.parent).imports == null)
        ((CompilationUnitScope)this.parent).checkAndSetImports();
    } else if ((this.parent instanceof ClassScope))
    {
      ((ClassScope)this.parent).connectTypeHierarchyWithoutMembers();
    }

    SourceTypeBinding sourceType = this.referenceContext.binding;
    if ((sourceType.tagBits & 0x100) != 0L) {
      return;
    }
    sourceType.tagBits |= 256L;
    boolean noProblems = connectSuperclass();
    noProblems &= connectSuperInterfaces();
    sourceType.tagBits |= 512L;
    noProblems &= connectTypeVariables(this.referenceContext.typeParameters, false);
    sourceType.tagBits |= 262144L;
    if ((noProblems) && (sourceType.isHierarchyInconsistent()))
      problemReporter().hierarchyHasProblems(sourceType);
  }

  public boolean detectHierarchyCycle(TypeBinding superType, TypeReference reference) {
    if (!(superType instanceof ReferenceBinding)) return false;

    if (reference == this.superTypeReference) {
      if (superType.isTypeVariable()) {
        return false;
      }

      if (superType.isParameterizedType())
        superType = ((ParameterizedTypeBinding)superType).genericType();
      compilationUnitScope().recordSuperTypeReference(superType);
      return detectHierarchyCycle(this.referenceContext.binding, (ReferenceBinding)superType, reference);
    }

    if (((superType.tagBits & 0x100) == 0L) && ((superType instanceof SourceTypeBinding)))
    {
      ((SourceTypeBinding)superType).scope.connectTypeHierarchyWithoutMembers();
    }return false;
  }

  private boolean detectHierarchyCycle(SourceTypeBinding sourceType, ReferenceBinding superType, TypeReference reference)
  {
    if (superType.isRawType()) {
      superType = ((RawTypeBinding)superType).genericType();
    }

    if (sourceType == superType) {
      problemReporter().hierarchyCircularity(sourceType, superType, reference);
      sourceType.tagBits |= 131072L;
      return true;
    }

    if (superType.isMemberType()) {
      ReferenceBinding current = superType.enclosingType();
      do
        if ((current.isHierarchyBeingConnected()) && (current == sourceType)) {
          problemReporter().hierarchyCircularity(sourceType, current, reference);
          sourceType.tagBits |= 131072L;
          current.tagBits |= 131072L;
          return true;
        }
      while ((current = current.enclosingType()) != null);
    }

    if (superType.isBinaryBinding())
    {
      boolean hasCycle = false;
      ReferenceBinding parentType = superType.superclass();
      if (parentType != null) {
        if (sourceType == parentType) {
          problemReporter().hierarchyCircularity(sourceType, superType, reference);
          sourceType.tagBits |= 131072L;
          superType.tagBits |= 131072L;
          return true;
        }
        if (parentType.isParameterizedType())
          parentType = ((ParameterizedTypeBinding)parentType).genericType();
        hasCycle |= detectHierarchyCycle(sourceType, parentType, reference);
        if ((parentType.tagBits & 0x20000) != 0L) {
          sourceType.tagBits |= 131072L;
          parentType.tagBits |= 131072L;
        }
      }

      ReferenceBinding[] itsInterfaces = superType.superInterfaces();
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
        int i = 0; for (int length = itsInterfaces.length; i < length; i++) {
          ReferenceBinding anInterface = itsInterfaces[i];
          if (sourceType == anInterface) {
            problemReporter().hierarchyCircularity(sourceType, superType, reference);
            sourceType.tagBits |= 131072L;
            superType.tagBits |= 131072L;
            return true;
          }
          if (anInterface.isParameterizedType())
            anInterface = ((ParameterizedTypeBinding)anInterface).genericType();
          hasCycle |= detectHierarchyCycle(sourceType, anInterface, reference);
          if ((anInterface.tagBits & 0x20000) != 0L) {
            sourceType.tagBits |= 131072L;
            superType.tagBits |= 131072L;
          }
        }
      }
      return hasCycle;
    }

    if (superType.isHierarchyBeingConnected()) {
      TypeReference ref = ((SourceTypeBinding)superType).scope.superTypeReference;

      if ((ref != null) && ((ref.resolvedType == null) || (((ReferenceBinding)ref.resolvedType).isHierarchyBeingConnected()))) {
        problemReporter().hierarchyCircularity(sourceType, superType, reference);
        sourceType.tagBits |= 131072L;
        superType.tagBits |= 131072L;
        return true;
      }
    }
    if ((superType.tagBits & 0x100) == 0L)
    {
      ((SourceTypeBinding)superType).scope.connectTypeHierarchyWithoutMembers();
    }if ((superType.tagBits & 0x20000) != 0L)
      sourceType.tagBits |= 131072L;
    return false;
  }

  private ReferenceBinding findSupertype(TypeReference typeReference) {
    CompilationUnitScope unitScope = compilationUnitScope();
    LookupEnvironment env = unitScope.environment;
    try {
      env.missingClassFileLocation = typeReference;
      typeReference.aboutToResolve(this);
      unitScope.recordQualifiedReference(typeReference.getTypeName());
      this.superTypeReference = typeReference;
      ReferenceBinding superType = (ReferenceBinding)typeReference.resolveSuperType(this);
      return superType;
    } catch (AbortCompilation e) {
      SourceTypeBinding sourceType = this.referenceContext.binding;
      if (sourceType.superInterfaces == null) sourceType.superInterfaces = Binding.NO_SUPERINTERFACES;
      e.updateContext(typeReference, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
      this.superTypeReference = null;
    }
  }

  public ProblemReporter problemReporter()
  {
    MethodScope outerMethodScope;
    if ((outerMethodScope = outerMostMethodScope()) == null) {
      ProblemReporter problemReporter = referenceCompilationUnit().problemReporter;
      problemReporter.referenceContext = this.referenceContext;
      return problemReporter;
    }
    return outerMethodScope.problemReporter();
  }

  public TypeDeclaration referenceType()
  {
    return this.referenceContext;
  }

  public String toString() {
    if (this.referenceContext != null)
      return "--- Class Scope ---\n\n" + 
        this.referenceContext.binding.toString();
    return "--- Class Scope ---\n\n Binding not initialized";
  }
}