package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;

public final class LocalTypeBinding extends NestedTypeBinding
{
  static final char[] LocalTypePrefix = { '$', 'L', 'o', 'c', 'a', 'l', '$' };
  private InnerEmulationDependency[] dependents;
  public ArrayBinding[] localArrayBindings;
  public CaseStatement enclosingCase;
  public int sourceStart;
  public MethodBinding enclosingMethod;

  public LocalTypeBinding(ClassScope scope, SourceTypeBinding enclosingType, CaseStatement switchCase)
  {
    super(new char[][] { CharOperation.concat(LocalTypePrefix, scope.referenceContext.name) }, 
      scope, 
      enclosingType);

    if (this.sourceName == CharOperation.NO_CHAR)
      this.tagBits |= 52L;
    else
      this.tagBits |= 20L;
    this.enclosingCase = switchCase;
    this.sourceStart = scope.referenceContext.sourceStart;
    MethodScope methodScope = scope.enclosingMethodScope();
    AbstractMethodDeclaration declaration = methodScope.referenceMethod();
    if (declaration != null)
      this.enclosingMethod = declaration.binding;
  }

  public void addInnerEmulationDependent(BlockScope dependentScope, boolean wasEnclosingInstanceSupplied)
  {
    int index;
    if (this.dependents == null) {
      int index = 0;
      this.dependents = new InnerEmulationDependency[1];
    } else {
      index = this.dependents.length;
      for (int i = 0; i < index; i++)
        if (this.dependents[i].scope == dependentScope)
          return;
      System.arraycopy(this.dependents, 0, this.dependents = new InnerEmulationDependency[index + 1], 0, index);
    }
    this.dependents[index] = new InnerEmulationDependency(dependentScope, wasEnclosingInstanceSupplied);
  }

  public char[] computeUniqueKey(boolean isLeaf) {
    char[] outerKey = outermostEnclosingType().computeUniqueKey(isLeaf);
    int semicolon = CharOperation.lastIndexOf(';', outerKey);

    StringBuffer sig = new StringBuffer();
    sig.append(outerKey, 0, semicolon);

    sig.append('$');
    sig.append(String.valueOf(this.sourceStart));

    if (!isAnonymousType()) {
      sig.append('$');
      sig.append(this.sourceName);
    }

    sig.append(outerKey, semicolon, outerKey.length - semicolon);

    int sigLength = sig.length();
    char[] uniqueKey = new char[sigLength];
    sig.getChars(0, sigLength, uniqueKey, 0);
    return uniqueKey;
  }

  public char[] constantPoolName() {
    return this.constantPoolName;
  }

  ArrayBinding createArrayType(int dimensionCount, LookupEnvironment lookupEnvironment) {
    if (this.localArrayBindings == null) {
      this.localArrayBindings = new ArrayBinding[] { new ArrayBinding(this, dimensionCount, lookupEnvironment) };
      return this.localArrayBindings[0];
    }

    int length = this.localArrayBindings.length;
    for (int i = 0; i < length; i++) {
      if (this.localArrayBindings[i].dimensions == dimensionCount) {
        return this.localArrayBindings[i];
      }
    }
    System.arraycopy(this.localArrayBindings, 0, this.localArrayBindings = new ArrayBinding[length + 1], 0, length);
    return this.localArrayBindings[length] =  = new ArrayBinding(this, dimensionCount, lookupEnvironment);
  }

  public char[] genericTypeSignature()
  {
    if ((this.genericReferenceTypeSignature == null) && (constantPoolName() == null)) {
      if (isAnonymousType())
        setConstantPoolName(superclass().sourceName());
      else
        setConstantPoolName(sourceName());
    }
    return super.genericTypeSignature();
  }

  public char[] readableName()
  {
    char[] readableName;
    char[] readableName;
    if (isAnonymousType())
    {
      char[] readableName;
      if (this.superInterfaces == Binding.NO_SUPERINTERFACES)
        readableName = CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superclass.readableName(), TypeConstants.ANONYM_SUFFIX);
      else
        readableName = CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superInterfaces[0].readableName(), TypeConstants.ANONYM_SUFFIX);
    }
    else
    {
      char[] readableName;
      if (isMemberType())
        readableName = CharOperation.concat(enclosingType().readableName(), this.sourceName, '.');
      else
        readableName = this.sourceName;
    }
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(readableName).append('<');
      int i = 0; for (int length = typeVars.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(typeVars[i].readableName());
      }
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      readableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, readableName, 0);
    }
    return readableName;
  }

  public char[] shortReadableName()
  {
    char[] shortReadableName;
    char[] shortReadableName;
    if (isAnonymousType())
    {
      char[] shortReadableName;
      if (this.superInterfaces == Binding.NO_SUPERINTERFACES)
        shortReadableName = CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superclass.shortReadableName(), TypeConstants.ANONYM_SUFFIX);
      else
        shortReadableName = CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superInterfaces[0].shortReadableName(), TypeConstants.ANONYM_SUFFIX);
    }
    else
    {
      char[] shortReadableName;
      if (isMemberType())
        shortReadableName = CharOperation.concat(enclosingType().shortReadableName(), this.sourceName, '.');
      else
        shortReadableName = this.sourceName;
    }
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(shortReadableName).append('<');
      int i = 0; for (int length = typeVars.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(typeVars[i].shortReadableName());
      }
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      shortReadableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, shortReadableName, 0);
    }
    return shortReadableName;
  }

  public void setAsMemberType()
  {
    this.tagBits |= 12L;
  }

  public void setConstantPoolName(char[] computedConstantPoolName) {
    this.constantPoolName = computedConstantPoolName;
  }

  public char[] signature()
  {
    if ((this.signature == null) && (constantPoolName() == null)) {
      if (isAnonymousType())
        setConstantPoolName(superclass().sourceName());
      else
        setConstantPoolName(sourceName());
    }
    return super.signature();
  }
  public char[] sourceName() {
    if (isAnonymousType()) {
      if (this.superInterfaces == Binding.NO_SUPERINTERFACES) {
        return CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superclass.sourceName(), TypeConstants.ANONYM_SUFFIX);
      }
      return CharOperation.concat(TypeConstants.ANONYM_PREFIX, this.superInterfaces[0].sourceName(), TypeConstants.ANONYM_SUFFIX);
    }

    return this.sourceName;
  }
  public String toString() {
    if (isAnonymousType())
      return "Anonymous type : " + super.toString();
    if (isMemberType())
      return "Local member type : " + new String(sourceName()) + " " + super.toString();
    return "Local type : " + new String(sourceName()) + " " + super.toString();
  }

  public void updateInnerEmulationDependents()
  {
    if (this.dependents != null)
      for (int i = 0; i < this.dependents.length; i++) {
        InnerEmulationDependency dependency = this.dependents[i];

        dependency.scope.propagateInnerEmulation(this, dependency.wasEnclosingInstanceSupplied);
      }
  }
}