package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;

public class MethodVerifier
{
  SourceTypeBinding type;
  HashtableOfObject inheritedMethods;
  HashtableOfObject currentMethods;
  LookupEnvironment environment;
  private boolean allowCompatibleReturnTypes;

  MethodVerifier(LookupEnvironment environment)
  {
    this.type = null;
    this.inheritedMethods = null;
    this.currentMethods = null;
    this.environment = environment;
    this.allowCompatibleReturnTypes = 
      ((environment.globalOptions.complianceLevel >= 3211264L) && 
      (environment.globalOptions.sourceLevel < 3211264L));
  }
  boolean areMethodsCompatible(MethodBinding one, MethodBinding two) {
    return (doesMethodOverride(one, two)) && (areReturnTypesCompatible(one, two));
  }
  boolean areParametersEqual(MethodBinding one, MethodBinding two) {
    TypeBinding[] oneArgs = one.parameters;
    TypeBinding[] twoArgs = two.parameters;
    if (oneArgs == twoArgs) return true;

    int length = oneArgs.length;
    if (length != twoArgs.length) return false;

    for (int i = 0; i < length; i++)
      if (!areTypesEqual(oneArgs[i], twoArgs[i])) return false;
    return true;
  }
  boolean areReturnTypesCompatible(MethodBinding one, MethodBinding two) {
    if (one.returnType == two.returnType) return true;

    if (areTypesEqual(one.returnType, two.returnType)) return true;

    if ((this.allowCompatibleReturnTypes) && 
      ((one.declaringClass instanceof BinaryTypeBinding)) && 
      ((two.declaringClass instanceof BinaryTypeBinding))) {
      return areReturnTypesCompatible0(one, two);
    }
    return false;
  }

  boolean areReturnTypesCompatible0(MethodBinding one, MethodBinding two) {
    if (one.returnType.isBaseType()) return false;

    if (!one.declaringClass.isInterface()) {
      if (one.declaringClass.id == 1)
        return two.returnType.isCompatibleWith(one.returnType);
      return one.returnType.isCompatibleWith(two.returnType);
    }

    if (two.declaringClass.id == 1) {
      return one.returnType.isCompatibleWith(two.returnType);
    }

    if (one.declaringClass.implementsInterface(two.declaringClass, true))
      return one.returnType.isCompatibleWith(two.returnType);
    if (two.declaringClass.implementsInterface(one.declaringClass, true)) {
      return two.returnType.isCompatibleWith(one.returnType);
    }

    return (one.returnType.isCompatibleWith(two.returnType)) || 
      (two.returnType.isCompatibleWith(one.returnType));
  }
  boolean areTypesEqual(TypeBinding one, TypeBinding two) {
    if (one == two) return true;

    if ((one instanceof UnresolvedReferenceBinding))
      return ((UnresolvedReferenceBinding)one).resolvedType == two;
    if ((two instanceof UnresolvedReferenceBinding))
      return ((UnresolvedReferenceBinding)two).resolvedType == one;
    return false;
  }
  boolean canSkipInheritedMethods() {
    if ((this.type.superclass() != null) && (this.type.superclass().isAbstract()))
      return false;
    return this.type.superInterfaces() == Binding.NO_SUPERINTERFACES;
  }

  boolean canSkipInheritedMethods(MethodBinding one, MethodBinding two) {
    return (two == null) || 
      (one.declaringClass == two.declaringClass);
  }
  void checkAbstractMethod(MethodBinding abstractMethod) {
    if (mustImplementAbstractMethod(abstractMethod.declaringClass)) {
      TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
      if (typeDeclaration != null) {
        MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(abstractMethod);
        missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
      } else {
        problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
      }
    }
  }

  void checkAgainstInheritedMethods(MethodBinding currentMethod, MethodBinding[] methods, int length, MethodBinding[] allInheritedMethods) { if (this.type.isAnnotationType()) {
      problemReporter().annotationCannotOverrideMethod(currentMethod, methods[(length - 1)]);
      return; } CompilerOptions options = this.type.scope.compilerOptions();

    int[] overriddenInheritedMethods = length > 1 ? findOverriddenInheritedMethods(methods, length) : null;
    int i = length;
    label348: 
    do { MethodBinding inheritedMethod = methods[i];
      if ((overriddenInheritedMethods == null) || (overriddenInheritedMethods[i] == 0)) {
        if (currentMethod.isStatic() != inheritedMethod.isStatic()) {
          problemReporter(currentMethod).staticAndInstanceConflict(currentMethod, inheritedMethod);
          break label348;
        }

        if (inheritedMethod.isAbstract()) {
          if (inheritedMethod.declaringClass.isInterface())
            currentMethod.modifiers |= 536870912;
          else {
            currentMethod.modifiers |= 805306368;
          }

        }
        else
        {
          currentMethod.modifiers |= 268435456;
        }

        if ((!areReturnTypesCompatible(currentMethod, inheritedMethod)) && 
          (reportIncompatibleReturnTypeError(currentMethod, inheritedMethod))) {
          break label348;
        }
        if (currentMethod.thrownExceptions != Binding.NO_EXCEPTIONS)
          checkExceptions(currentMethod, inheritedMethod);
        if (inheritedMethod.isFinal())
          problemReporter(currentMethod).finalMethodCannotBeOverridden(currentMethod, inheritedMethod);
        if (!isAsVisible(currentMethod, inheritedMethod))
          problemReporter(currentMethod).visibilityConflict(currentMethod, inheritedMethod);
        if ((options.reportDeprecationWhenOverridingDeprecatedMethod) && (inheritedMethod.isViewedAsDeprecated()) && (
          (!currentMethod.isViewedAsDeprecated()) || (options.reportDeprecationInsideDeprecatedCode)))
        {
          ReferenceBinding declaringClass = inheritedMethod.declaringClass;
          if (declaringClass.isInterface()) {
            int j = length;
            do { if ((i != j) && (methods[j].declaringClass.implementsInterface(declaringClass, false)))
                break;
              j--; } while (j >= 0);
          }

          problemReporter(currentMethod).overridesDeprecatedMethod(currentMethod, inheritedMethod);
        }
      }

      checkForBridgeMethod(currentMethod, inheritedMethod, allInheritedMethods);

      i--; } while (i >= 0);
  }

  void checkConcreteInheritedMethod(MethodBinding concreteMethod, MethodBinding[] abstractMethods)
  {
    if (concreteMethod.isStatic())
    {
      problemReporter().staticInheritedMethodConflicts(this.type, concreteMethod, abstractMethods);
    }if (!concreteMethod.isPublic()) {
      int index = 0; int length = abstractMethods.length;
      if (concreteMethod.isProtected())
        for (; index < length; index++)
          if (abstractMethods[index].isPublic()) break;
      else if (concreteMethod.isDefault())
        for (; index < length; index++)
          if (!abstractMethods[index].isDefault())
            break;
      if (index < length)
        problemReporter().inheritedMethodReducesVisibility(this.type, concreteMethod, abstractMethods);
    }
    if (concreteMethod.thrownExceptions != Binding.NO_EXCEPTIONS) {
      int i = abstractMethods.length;
      do { checkExceptions(concreteMethod, abstractMethods[i]);

        i--; } while (i >= 0);
    }
  }

  void checkExceptions(MethodBinding newMethod, MethodBinding inheritedMethod)
  {
    ReferenceBinding[] newExceptions = resolvedExceptionTypesFor(newMethod);
    ReferenceBinding[] inheritedExceptions = resolvedExceptionTypesFor(inheritedMethod);
    int i = newExceptions.length;
    do { ReferenceBinding newException = newExceptions[i];
      int j = inheritedExceptions.length;
      do j--; while ((j > -1) && (!isSameClassOrSubclassOf(newException, inheritedExceptions[j])));
      if ((j == -1) && 
        (!newException.isUncheckedException(false)))
        problemReporter(newMethod).incompatibleExceptionInThrowsClause(this.type, newMethod, inheritedMethod, newException);
      i--; } while (i >= 0);
  }

  void checkForBridgeMethod(MethodBinding currentMethod, MethodBinding inheritedMethod, MethodBinding[] allInheritedMethods)
  {
  }

  void checkInheritedMethods(MethodBinding[] methods, int length)
  {
    if (length > 1) {
      int[] overriddenInheritedMethods = findOverriddenInheritedMethods(methods, length);
      if (overriddenInheritedMethods != null)
      {
        int index = 0;
        MethodBinding[] closestMethods = new MethodBinding[length];
        for (int i = 0; i < length; i++)
          if (overriddenInheritedMethods[i] == 0)
            closestMethods[(index++)] = methods[i];
        if ((index <= 1) || (checkInheritedReturnTypes(closestMethods, index)));
      }
      else if (!checkInheritedReturnTypes(methods, length)) {
        return;
      }
    }

    MethodBinding concreteMethod = null;
    if (!this.type.isInterface()) {
      int i = length;
      do { if (!methods[i].isAbstract()) {
          concreteMethod = methods[i];
          break;
        }
        i--; } while (i >= 0);
    }

    if (concreteMethod == null) {
      if (!this.type.isAbstract()) {
        int i = length;
        do { if (mustImplementAbstractMethod(methods[i].declaringClass)) {
            TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
            if (typeDeclaration != null) {
              MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(methods[0]);
              missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, methods[0]);
            } else {
              problemReporter().abstractMethodMustBeImplemented(this.type, methods[0]);
            }
            return;
          }
          i--; } while (i >= 0);
      }

      return;
    }

    if (length > 1) {
      MethodBinding[] abstractMethods = new MethodBinding[length - 1];
      int index = 0;
      int i = length;
      do { if (methods[i] != concreteMethod)
          abstractMethods[(index++)] = methods[i];
        i--; } while (i >= 0);

      checkConcreteInheritedMethod(concreteMethod, abstractMethods);
    }
  }

  boolean checkInheritedReturnTypes(MethodBinding[] methods, int length) { MethodBinding first = methods[0];
    int index = length;
    do index--; while ((index > 0) && (areReturnTypesCompatible(first, methods[index])));
    if (index == 0) {
      return true;
    }

    if (this.type.isInterface()) {
      int i = length;
      do { if (methods[i].declaringClass.id == 1)
          return false;
        i--; } while (i >= 0);
    }

    problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length);
    return false;
  }

  void checkMethods()
  {
    boolean mustImplementAbstractMethods = mustImplementAbstractMethods();
    boolean skipInheritedMethods = (mustImplementAbstractMethods) && (canSkipInheritedMethods());
    char[][] methodSelectors = this.inheritedMethods.keyTable;
    int s = methodSelectors.length;
    do { if (methodSelectors[s] != null)
      {
        MethodBinding[] current = (MethodBinding[])this.currentMethods.get(methodSelectors[s]);
        if ((current != null) || (!skipInheritedMethods))
        {
          MethodBinding[] inherited = (MethodBinding[])this.inheritedMethods.valueTable[s];
          if ((inherited.length == 1) && (current == null)) {
            if ((mustImplementAbstractMethods) && (inherited[0].isAbstract()))
              checkAbstractMethod(inherited[0]);
          }
          else
          {
            int index = -1;
            MethodBinding[] matchingInherited = new MethodBinding[inherited.length];
            if (current != null) {
              int i = 0; for (int length1 = current.length; i < length1; i++) {
                MethodBinding currentMethod = current[i];
                int j = 0; for (int length2 = inherited.length; j < length2; j++) {
                  MethodBinding inheritedMethod = computeSubstituteMethod(inherited[j], currentMethod);
                  if ((inheritedMethod != null) && 
                    (doesMethodOverride(currentMethod, inheritedMethod))) {
                    matchingInherited[(++index)] = inheritedMethod;
                    inherited[j] = null;
                  }
                }

                if (index >= 0) {
                  checkAgainstInheritedMethods(currentMethod, matchingInherited, index + 1, inherited);
                  while (index >= 0) matchingInherited[(index--)] = null;
                }
              }
            }

            int i = 0; for (int length = inherited.length; i < length; i++) {
              MethodBinding inheritedMethod = inherited[i];
              if (inheritedMethod != null)
              {
                matchingInherited[(++index)] = inheritedMethod;
                for (int j = i + 1; j < length; j++) {
                  MethodBinding otherInheritedMethod = inherited[j];
                  if (!canSkipInheritedMethods(inheritedMethod, otherInheritedMethod))
                  {
                    otherInheritedMethod = computeSubstituteMethod(otherInheritedMethod, inheritedMethod);
                    if ((otherInheritedMethod != null) && 
                      (doesMethodOverride(inheritedMethod, otherInheritedMethod))) {
                      matchingInherited[(++index)] = otherInheritedMethod;
                      inherited[j] = null;
                    }
                  }
                }
                if (index != -1) {
                  if (index > 0)
                    checkInheritedMethods(matchingInherited, index + 1);
                  else if ((mustImplementAbstractMethods) && (matchingInherited[0].isAbstract()))
                    checkAbstractMethod(matchingInherited[0]);
                  while (index >= 0) matchingInherited[(index--)] = null;
                }
              }
            }
          }
        }
      }
      s--; } while (s >= 0);
  }

  void checkPackagePrivateAbstractMethod(MethodBinding abstractMethod)
  {
    PackageBinding necessaryPackage = abstractMethod.declaringClass.fPackage;
    if (necessaryPackage == this.type.fPackage) return;

    ReferenceBinding superType = this.type.superclass();
    char[] selector = abstractMethod.selector;
    do {
      if (!superType.isValidBinding()) return;
      if (!superType.isAbstract()) return;

      if (necessaryPackage == superType.fPackage) {
        MethodBinding[] methods = superType.getMethods(selector);
        int m = methods.length;
        do { MethodBinding method = methods[m];
          if ((!method.isPrivate()) && (!method.isConstructor()) && (!method.isDefaultAbstract()))
          {
            if (areMethodsCompatible(method, abstractMethod))
              return;
          }
          m--; } while (m >= 0);
      }

    }

    while ((superType = superType.superclass()) != abstractMethod.declaringClass);

    problemReporter().abstractMethodCannotBeOverridden(this.type, abstractMethod);
  }
  void computeInheritedMethods() {
    ReferenceBinding superclass = this.type.isInterface() ? 
      this.type.scope.getJavaLangObject() : 
      this.type.superclass();
    computeInheritedMethods(superclass, this.type.superInterfaces());
  }

  void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] superInterfaces)
  {
    this.inheritedMethods = new HashtableOfObject(51);
    ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    ReferenceBinding[] itsInterfaces = superInterfaces;
    if (itsInterfaces != Binding.NO_SUPERINTERFACES) {
      nextPosition = itsInterfaces.length;
      interfacesToVisit = itsInterfaces;
    }

    ReferenceBinding superType = superclass;
    HashtableOfObject nonVisibleDefaultMethods = new HashtableOfObject(3);
    boolean allSuperclassesAreAbstract = true;

    while ((superType != null) && (superType.isValidBinding())) {
      if (allSuperclassesAreAbstract) {
        if (superType.isAbstract())
        {
          if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES)
            if (interfacesToVisit == null) {
              interfacesToVisit = itsInterfaces;
              nextPosition = interfacesToVisit.length;
            } else {
              int itsLength = itsInterfaces.length;
              if (nextPosition + itsLength >= interfacesToVisit.length)
                System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
              for (int a = 0; a < itsLength; a++) {
                ReferenceBinding next = itsInterfaces[a];
                int b = 0;
                while (next != interfacesToVisit[b])
                {
                  b++; if (b >= nextPosition)
                  {
                    interfacesToVisit[(nextPosition++)] = next;
                  }
                }
              }
            }
        } else allSuperclassesAreAbstract = false;

      }

      MethodBinding[] methods = superType.unResolvedMethods();
      int m = methods.length;
      do { MethodBinding inheritedMethod = methods[m];
        if ((!inheritedMethod.isPrivate()) && (!inheritedMethod.isConstructor()) && (!inheritedMethod.isDefaultAbstract()))
        {
          MethodBinding[] existingMethods = (MethodBinding[])this.inheritedMethods.get(inheritedMethod.selector);
          if (existingMethods != null) {
            int i = 0; for (int length = existingMethods.length; i < length; i++) {
              if ((existingMethods[i].declaringClass != inheritedMethod.declaringClass) && (areMethodsCompatible(existingMethods[i], inheritedMethod))) {
                if ((!inheritedMethod.isDefault()) || (!inheritedMethod.isAbstract())) break;
                checkPackagePrivateAbstractMethod(inheritedMethod);
                break;
              }
            }
          }
          MethodBinding[] nonVisible = (MethodBinding[])nonVisibleDefaultMethods.get(inheritedMethod.selector);
          if (nonVisible != null) {
            int i = 0; for (int l = nonVisible.length; i < l; i++)
              if (areMethodsCompatible(nonVisible[i], inheritedMethod))
                break;
          }
          if ((!inheritedMethod.isDefault()) || (inheritedMethod.declaringClass.fPackage == this.type.fPackage)) {
            if (existingMethods == null) {
              existingMethods = new MethodBinding[] { inheritedMethod };
            } else {
              int length = existingMethods.length;
              System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
              existingMethods[length] = inheritedMethod;
            }
            this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
          } else {
            if (nonVisible == null) {
              nonVisible = new MethodBinding[] { inheritedMethod };
            } else {
              int length = nonVisible.length;
              System.arraycopy(nonVisible, 0, nonVisible = new MethodBinding[length + 1], 0, length);
              nonVisible[length] = inheritedMethod;
            }
            nonVisibleDefaultMethods.put(inheritedMethod.selector, nonVisible);

            if ((inheritedMethod.isAbstract()) && (!this.type.isAbstract())) {
              problemReporter().abstractMethodCannotBeOverridden(this.type, inheritedMethod);
            }
            MethodBinding[] current = (MethodBinding[])this.currentMethods.get(inheritedMethod.selector);
            if (current != null) {
              int i = 0; for (int length = current.length; i < length; i++)
                if (areMethodsCompatible(current[i], inheritedMethod)) {
                  problemReporter().overridesPackageDefaultMethod(current[i], inheritedMethod);
                  break;
                }
            }
          }
        }
        m--; } while (m >= 0);

      superType = superType.superclass();
    }
    if (nextPosition == 0) return;

    SimpleSet skip = findSuperinterfaceCollisions(superclass, superInterfaces);
    for (int i = 0; i < nextPosition; i++) {
      superType = interfacesToVisit[i];
      if ((superType.isValidBinding()) && (
        (skip == null) || (!skip.includes(superType)))) {
        if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
        MethodBinding[] methods = superType.unResolvedMethods();
        int m = methods.length;
        do { MethodBinding inheritedMethod = methods[m];
          MethodBinding[] existingMethods = (MethodBinding[])this.inheritedMethods.get(inheritedMethod.selector);
          if (existingMethods == null) {
            existingMethods = new MethodBinding[] { inheritedMethod };
          } else {
            int length = existingMethods.length;

            for (int e = 0; e < length; e++)
              if (isInterfaceMethodImplemented(inheritedMethod, existingMethods[e], superType))
                break;
            System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
            existingMethods[length] = inheritedMethod;
          }
          this.inheritedMethods.put(inheritedMethod.selector, existingMethods);

          m--; } while (m >= 0);
      }
    }
  }

  void computeMethods()
  {
    MethodBinding[] methods = this.type.methods();
    int size = methods.length;
    this.currentMethods = new HashtableOfObject(size == 0 ? 1 : size);
    int m = size;
    do { MethodBinding method = methods[m];
      if ((!method.isConstructor()) && (!method.isDefaultAbstract())) {
        MethodBinding[] existingMethods = (MethodBinding[])this.currentMethods.get(method.selector);
        if (existingMethods == null)
          existingMethods = new MethodBinding[1];
        else
          System.arraycopy(existingMethods, 0, 
            existingMethods = new MethodBinding[existingMethods.length + 1], 0, existingMethods.length - 1);
        existingMethods[(existingMethods.length - 1)] = method;
        this.currentMethods.put(method.selector, existingMethods);
      }
      m--; } while (m >= 0);
  }

  MethodBinding computeSubstituteMethod(MethodBinding inheritedMethod, MethodBinding currentMethod)
  {
    if (inheritedMethod == null) return null;
    if (currentMethod.parameters.length != inheritedMethod.parameters.length) return null;
    return inheritedMethod;
  }
  public boolean doesMethodOverride(MethodBinding method, MethodBinding inheritedMethod) {
    return areParametersEqual(method, inheritedMethod);
  }
  SimpleSet findSuperinterfaceCollisions(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
    return null;
  }

  int[] findOverriddenInheritedMethods(MethodBinding[] methods, int length)
  {
    int[] toSkip = (int[])null;
    int i = 0;
    ReferenceBinding declaringClass = methods[i].declaringClass;
    if (!declaringClass.isInterface())
    {
      ReferenceBinding declaringClass2 = methods[(++i)].declaringClass;
      while (declaringClass == declaringClass2) {
        i++; if (i == length) return null;
        declaringClass2 = methods[i].declaringClass;
      }
      if (!declaringClass2.isInterface())
      {
        toSkip = new int[length];
        do {
          toSkip[i] = -1;
          i++; if (i == length) return toSkip;
          declaringClass2 = methods[i].declaringClass;
        }
        while (!
          declaringClass2.isInterface());
      }

    }

    for (; i < length; i++)
      if ((toSkip == null) || (toSkip[i] != -1)) {
        declaringClass = methods[i].declaringClass;
        for (int j = i + 1; j < length; j++)
          if ((toSkip == null) || (toSkip[j] != -1)) {
            ReferenceBinding declaringClass2 = methods[j].declaringClass;
            if (declaringClass != declaringClass2)
              if (declaringClass.implementsInterface(declaringClass2, true)) {
                if (toSkip == null)
                  toSkip = new int[length];
                toSkip[j] = -1;
              } else if (declaringClass2.implementsInterface(declaringClass, true)) {
                if (toSkip == null)
                  toSkip = new int[length];
                toSkip[i] = -1;
                break;
              }
          }
      }
    return toSkip;
  }
  boolean isAsVisible(MethodBinding newMethod, MethodBinding inheritedMethod) {
    if (inheritedMethod.modifiers == newMethod.modifiers) return true;

    if (newMethod.isPublic()) return true;
    if (inheritedMethod.isPublic()) return false;

    if (newMethod.isProtected()) return true;
    if (inheritedMethod.isProtected()) return false;

    return !newMethod.isPrivate();
  }

  boolean isInterfaceMethodImplemented(MethodBinding inheritedMethod, MethodBinding existingMethod, ReferenceBinding superType) {
    return (areParametersEqual(existingMethod, inheritedMethod)) && (existingMethod.declaringClass.implementsInterface(superType, true));
  }

  boolean isSameClassOrSubclassOf(ReferenceBinding testClass, ReferenceBinding superclass) {
    do if (testClass == superclass) return true;
    while ((testClass = testClass.superclass()) != null);
    return false;
  }

  boolean mustImplementAbstractMethod(ReferenceBinding declaringClass)
  {
    ReferenceBinding superclass = this.type.superclass();
    if (declaringClass.isClass()) {
      do {
        superclass = superclass.superclass();

        if (!superclass.isAbstract()) break;  } while (superclass != declaringClass);
    }
    else {
      if (this.type.implementsInterface(declaringClass, false)) {
        if (this.type.isAbstract()) return false;
        if (!superclass.implementsInterface(declaringClass, true))
          return true;
      }
      while ((superclass.isAbstract()) && (!superclass.implementsInterface(declaringClass, false)))
        superclass = superclass.superclass();
    }
    return superclass.isAbstract();
  }
  boolean mustImplementAbstractMethods() {
    return (!this.type.isInterface()) && (!this.type.isAbstract());
  }
  ProblemReporter problemReporter() {
    return this.type.scope.problemReporter();
  }
  ProblemReporter problemReporter(MethodBinding currentMethod) {
    ProblemReporter reporter = problemReporter();
    if ((currentMethod.declaringClass == this.type) && (currentMethod.sourceMethod() != null))
      reporter.referenceContext = currentMethod.sourceMethod();
    return reporter;
  }

  boolean reportIncompatibleReturnTypeError(MethodBinding currentMethod, MethodBinding inheritedMethod)
  {
    problemReporter(currentMethod).incompatibleReturnType(currentMethod, inheritedMethod);
    return true;
  }
  ReferenceBinding[] resolvedExceptionTypesFor(MethodBinding method) {
    ReferenceBinding[] exceptions = method.thrownExceptions;
    if ((method.modifiers & 0x2000000) == 0) {
      return exceptions;
    }
    if (!(method.declaringClass instanceof BinaryTypeBinding)) {
      return Binding.NO_EXCEPTIONS;
    }
    int i = exceptions.length;
    do { exceptions[i] = BinaryTypeBinding.resolveType(exceptions[i], this.environment, true);

      i--; } while (i >= 0);

    return exceptions;
  }
  void verify(SourceTypeBinding someType) {
    this.type = someType;
    computeMethods();
    computeInheritedMethods();
    checkMethods();
  }
  public String toString() {
    StringBuffer buffer = new StringBuffer(10);
    buffer.append("MethodVerifier for type: ");
    buffer.append(this.type.readableName());
    buffer.append('\n');
    buffer.append("\t-inherited methods: ");
    buffer.append(this.inheritedMethods);
    return buffer.toString();
  }
}