package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class NestedTypeBinding extends SourceTypeBinding
{
  public SourceTypeBinding enclosingType;
  public SyntheticArgumentBinding[] enclosingInstances;
  public SyntheticArgumentBinding[] outerLocalVariables;
  public int enclosingInstancesSlotSize;
  public int outerLocalVariablesSlotSize;

  public NestedTypeBinding(char[][] typeName, ClassScope scope, SourceTypeBinding enclosingType)
  {
    super(typeName, enclosingType.fPackage, scope);
    this.tagBits |= 4L;
    this.enclosingType = enclosingType;
  }

  public SyntheticArgumentBinding addSyntheticArgument(LocalVariableBinding actualOuterLocalVariable)
  {
    SyntheticArgumentBinding synthLocal = null;

    if (this.outerLocalVariables == null) {
      synthLocal = new SyntheticArgumentBinding(actualOuterLocalVariable);
      this.outerLocalVariables = new SyntheticArgumentBinding[] { synthLocal };
    } else {
      int size = this.outerLocalVariables.length;
      int newArgIndex = size;
      int i = size;
      do { if (this.outerLocalVariables[i].actualOuterLocalVariable == actualOuterLocalVariable)
          return this.outerLocalVariables[i];
        if (this.outerLocalVariables[i].id > actualOuterLocalVariable.id)
          newArgIndex = i;
        i--; } while (i >= 0);

      SyntheticArgumentBinding[] synthLocals = new SyntheticArgumentBinding[size + 1];
      System.arraycopy(this.outerLocalVariables, 0, synthLocals, 0, newArgIndex);
      void tmp132_129 = new SyntheticArgumentBinding(actualOuterLocalVariable); synthLocal = tmp132_129; synthLocals[newArgIndex] = tmp132_129;
      System.arraycopy(this.outerLocalVariables, newArgIndex, synthLocals, newArgIndex + 1, size - newArgIndex);
      this.outerLocalVariables = synthLocals;
    }

    if (this.scope.referenceCompilationUnit().isPropagatingInnerClassEmulation)
      updateInnerEmulationDependents();
    return synthLocal;
  }

  public SyntheticArgumentBinding addSyntheticArgument(ReferenceBinding targetEnclosingType)
  {
    SyntheticArgumentBinding synthLocal = null;
    if (this.enclosingInstances == null) {
      synthLocal = new SyntheticArgumentBinding(targetEnclosingType);
      this.enclosingInstances = new SyntheticArgumentBinding[] { synthLocal };
    } else {
      int size = this.enclosingInstances.length;
      int newArgIndex = size;
      int i = size;
      do { if (this.enclosingInstances[i].type == targetEnclosingType)
          return this.enclosingInstances[i];
        if (enclosingType() == targetEnclosingType)
          newArgIndex = 0;
        i--; } while (i >= 0);

      SyntheticArgumentBinding[] newInstances = new SyntheticArgumentBinding[size + 1];
      System.arraycopy(this.enclosingInstances, 0, newInstances, newArgIndex == 0 ? 1 : 0, size);
      void tmp130_127 = new SyntheticArgumentBinding(targetEnclosingType); synthLocal = tmp130_127; newInstances[newArgIndex] = tmp130_127;
      this.enclosingInstances = newInstances;
    }

    if (this.scope.referenceCompilationUnit().isPropagatingInnerClassEmulation)
      updateInnerEmulationDependents();
    return synthLocal;
  }

  public SyntheticArgumentBinding addSyntheticArgumentAndField(LocalVariableBinding actualOuterLocalVariable)
  {
    SyntheticArgumentBinding synthLocal = addSyntheticArgument(actualOuterLocalVariable);
    if (synthLocal == null) return null;

    if (synthLocal.matchingField == null)
      synthLocal.matchingField = addSyntheticFieldForInnerclass(actualOuterLocalVariable);
    return synthLocal;
  }

  public SyntheticArgumentBinding addSyntheticArgumentAndField(ReferenceBinding targetEnclosingType)
  {
    SyntheticArgumentBinding synthLocal = addSyntheticArgument(targetEnclosingType);
    if (synthLocal == null) return null;

    if (synthLocal.matchingField == null)
      synthLocal.matchingField = addSyntheticFieldForInnerclass(targetEnclosingType);
    return synthLocal;
  }

  public final void computeSyntheticArgumentSlotSizes()
  {
    int slotSize = 0;

    int enclosingInstancesCount = this.enclosingInstances == null ? 0 : this.enclosingInstances.length;
    for (int i = 0; i < enclosingInstancesCount; i++) {
      SyntheticArgumentBinding argument = this.enclosingInstances[i];

      argument.resolvedPosition = (slotSize + 1);
      if (slotSize + 1 > 255) {
        this.scope.problemReporter().noMoreAvailableSpaceForArgument(argument, this.scope.referenceType());
      }
      if ((argument.type == TypeBinding.LONG) || (argument.type == TypeBinding.DOUBLE))
        slotSize += 2;
      else {
        slotSize++;
      }
    }
    this.enclosingInstancesSlotSize = slotSize;

    slotSize = 0;
    int outerLocalsCount = this.outerLocalVariables == null ? 0 : this.outerLocalVariables.length;
    for (int i = 0; i < outerLocalsCount; i++) {
      SyntheticArgumentBinding argument = this.outerLocalVariables[i];

      if ((argument.type == TypeBinding.LONG) || (argument.type == TypeBinding.DOUBLE))
        slotSize += 2;
      else {
        slotSize++;
      }
    }
    this.outerLocalVariablesSlotSize = slotSize;
  }

  public ReferenceBinding enclosingType()
  {
    return this.enclosingType;
  }

  public SyntheticArgumentBinding getSyntheticArgument(LocalVariableBinding actualOuterLocalVariable)
  {
    if (this.outerLocalVariables == null) return null;

    int i = this.outerLocalVariables.length;
    do { if (this.outerLocalVariables[i].actualOuterLocalVariable == actualOuterLocalVariable)
        return this.outerLocalVariables[i];
      i--; } while (i >= 0);

    return null;
  }

  public SyntheticArgumentBinding[] syntheticEnclosingInstances() {
    return this.enclosingInstances;
  }

  public ReferenceBinding[] syntheticEnclosingInstanceTypes() {
    if (this.enclosingInstances == null) {
      return null;
    }
    int length = this.enclosingInstances.length;
    ReferenceBinding[] types = new ReferenceBinding[length];
    for (int i = 0; i < length; i++)
      types[i] = ((ReferenceBinding)this.enclosingInstances[i].type);
    return types;
  }

  public SyntheticArgumentBinding[] syntheticOuterLocalVariables()
  {
    return this.outerLocalVariables;
  }

  public void updateInnerEmulationDependents()
  {
  }

  public SyntheticArgumentBinding getSyntheticArgument(ReferenceBinding targetEnclosingType, boolean onlyExactMatch)
  {
    if (this.enclosingInstances == null) return null;

    int i = this.enclosingInstances.length;
    do { if ((this.enclosingInstances[i].type == targetEnclosingType) && 
        (this.enclosingInstances[i].actualOuterLocalVariable == null))
        return this.enclosingInstances[i];
      i--; } while (i >= 0);

    if (!onlyExactMatch) {
      int i = this.enclosingInstances.length;
      do { if ((this.enclosingInstances[i].actualOuterLocalVariable == null) && 
          (this.enclosingInstances[i].type.findSuperTypeWithSameErasure(targetEnclosingType) != null))
          return this.enclosingInstances[i];
        i--; } while (i >= 0);
    }

    return null;
  }
}