package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ParameterizedTypeBinding extends ReferenceBinding
  implements Substitution
{
  private ReferenceBinding type;
  public TypeBinding[] arguments;
  public LookupEnvironment environment;
  public char[] genericTypeSignature;
  public ReferenceBinding superclass;
  public ReferenceBinding[] superInterfaces;
  public FieldBinding[] fields;
  public ReferenceBinding[] memberTypes;
  public MethodBinding[] methods;
  private ReferenceBinding enclosingType;

  public ParameterizedTypeBinding(ReferenceBinding type, TypeBinding[] arguments, ReferenceBinding enclosingType, LookupEnvironment environment)
  {
    this.environment = environment;
    this.enclosingType = enclosingType;

    initialize(type, arguments);
    if ((type instanceof UnresolvedReferenceBinding))
      ((UnresolvedReferenceBinding)type).addWrapper(this, environment);
    if (arguments != null) {
      int i = 0; for (int l = arguments.length; i < l; i++)
        if ((arguments[i] instanceof UnresolvedReferenceBinding))
          ((UnresolvedReferenceBinding)arguments[i]).addWrapper(this, environment);
    }
    this.tagBits |= 16777216L;
  }

  protected ReferenceBinding actualType()
  {
    return this.type;
  }

  public void boundCheck(Scope scope, TypeReference[] argumentReferences)
  {
    if ((this.tagBits & 0x400000) == 0L) {
      boolean hasErrors = false;
      TypeVariableBinding[] typeVariables = this.type.typeVariables();
      if ((this.arguments != null) && (typeVariables != null)) {
        int i = 0; for (int length = typeVariables.length; i < length; i++) {
          if (typeVariables[i].boundCheck(this, this.arguments[i]) != 0) {
            hasErrors = true;
            scope.problemReporter().typeMismatchError(this.arguments[i], typeVariables[i], this.type, argumentReferences[i]);
          }
        }
      }
      if (!hasErrors) this.tagBits |= 4194304L;
    }
  }

  public boolean canBeInstantiated()
  {
    return ((this.tagBits & 0x40000000) == 0L) && (super.canBeInstantiated());
  }

  public TypeBinding capture(Scope scope, int position)
  {
    if ((this.tagBits & 0x40000000) == 0L) {
      return this;
    }
    TypeBinding[] originalArguments = this.arguments;
    int length = originalArguments.length;
    TypeBinding[] capturedArguments = new TypeBinding[length];

    ReferenceBinding contextType = scope.enclosingSourceType();
    if (contextType != null) contextType = contextType.outermostEnclosingType();

    for (int i = 0; i < length; i++) {
      TypeBinding argument = originalArguments[i];
      if ((argument.kind() == 516) && (((WildcardBinding)argument).otherBounds == null))
        capturedArguments[i] = new CaptureBinding((WildcardBinding)argument, contextType, position, scope.compilationUnitScope().nextCaptureID());
      else {
        capturedArguments[i] = argument;
      }
    }
    ParameterizedTypeBinding capturedParameterizedType = this.environment.createParameterizedType(this.type, capturedArguments, enclosingType());
    for (int i = 0; i < length; i++) {
      TypeBinding argument = capturedArguments[i];
      if (argument.isCapture()) {
        ((CaptureBinding)argument).initializeBounds(scope, capturedParameterizedType);
      }
    }
    return capturedParameterizedType;
  }

  public void collectSubstitutes(Scope scope, TypeBinding actualType, InferenceContext inferenceContext, int constraint)
  {
    if ((this.tagBits & 0x20000000) == 0L) return;
    if (actualType == TypeBinding.NULL) return;

    if (!(actualType instanceof ReferenceBinding))
      return;
    TypeBinding formalEquivalent;
    TypeBinding actualEquivalent;
    switch (constraint) {
    case 0:
    case 1:
      TypeBinding formalEquivalent = this;
      TypeBinding actualEquivalent = actualType.findSuperTypeWithSameErasure(this.type);
      if (actualEquivalent == null) return;
      break;
    case 2:
    default:
      formalEquivalent = findSuperTypeWithSameErasure(actualType);
      if (formalEquivalent == null) return;
      actualEquivalent = actualType;
    }

    ReferenceBinding formalEnclosingType = formalEquivalent.enclosingType();
    if (formalEnclosingType != null) {
      formalEnclosingType.collectSubstitutes(scope, actualEquivalent.enclosingType(), inferenceContext, constraint);
    }

    if (this.arguments == null)
      return;
    TypeBinding[] formalArguments;
    TypeBinding[] formalArguments;
    switch (formalEquivalent.kind()) {
    case 2052:
      formalArguments = formalEquivalent.typeVariables();
      break;
    case 260:
      formalArguments = ((ParameterizedTypeBinding)formalEquivalent).arguments;
      break;
    case 1028:
      if (!inferenceContext.checkRawSubstitution())
        inferenceContext.status = 1;
      return;
    default:
      return;
    }
    TypeBinding[] formalArguments;
    TypeBinding[] actualArguments;
    TypeBinding[] actualArguments;
    switch (actualEquivalent.kind()) {
    case 2052:
      actualArguments = actualEquivalent.typeVariables();
      break;
    case 260:
      actualArguments = ((ParameterizedTypeBinding)actualEquivalent).arguments;
      break;
    case 1028:
      if (!inferenceContext.checkRawSubstitution())
        inferenceContext.status = 1;
      return;
    default:
      return;
    }
    TypeBinding[] actualArguments;
    inferenceContext.depth += 1;
    int i = 0; for (int length = formalArguments.length; i < length; i++) {
      TypeBinding formalArgument = formalArguments[i];
      TypeBinding actualArgument = actualArguments[i];
      if (formalArgument.isWildcard()) {
        formalArgument.collectSubstitutes(scope, actualArgument, inferenceContext, constraint);
      } else {
        if (actualArgument.isWildcard()) {
          WildcardBinding actualWildcardArgument = (WildcardBinding)actualArgument;
          if (actualWildcardArgument.otherBounds == null) {
            if (constraint != 2) continue;
            switch (actualWildcardArgument.boundKind) {
            case 1:
              formalArgument.collectSubstitutes(scope, actualWildcardArgument.bound, inferenceContext, 2);
              break;
            case 2:
              formalArgument.collectSubstitutes(scope, actualWildcardArgument.bound, inferenceContext, 1);
              break;
            default:
              break;
            }

          }

        }

        formalArgument.collectSubstitutes(scope, actualArgument, inferenceContext, 0);
      }
    }
    inferenceContext.depth -= 1;
  }

  public void computeId()
  {
    this.id = 2147483647;
  }

  public char[] computeUniqueKey(boolean isLeaf) {
    StringBuffer sig = new StringBuffer(10);
    ReferenceBinding enclosing;
    if ((isMemberType()) && (((enclosing = enclosingType()).isParameterizedType()) || (enclosing.isRawType()))) {
      char[] typeSig = enclosing.computeUniqueKey(false);
      for (int i = 0; i < typeSig.length - 1; i++) sig.append(typeSig[i]);
      sig.append('.').append(sourceName());
    } else if (this.type.isLocalType()) {
      LocalTypeBinding localTypeBinding = (LocalTypeBinding)this.type;
      ReferenceBinding enclosing = localTypeBinding.enclosingType();
      ReferenceBinding temp;
      while ((temp = enclosing.enclosingType()) != null)
      {
        ReferenceBinding temp;
        enclosing = temp;
      }char[] typeSig = enclosing.computeUniqueKey(false);
      for (int i = 0; i < typeSig.length - 1; i++) sig.append(typeSig[i]);
      sig.append('$');
      sig.append(localTypeBinding.sourceStart);
    } else {
      char[] typeSig = this.type.computeUniqueKey(false);
      for (int i = 0; i < typeSig.length - 1; i++) sig.append(typeSig[i]);
    }
    ReferenceBinding captureSourceType = null;
    if (this.arguments != null) {
      sig.append('<');
      int i = 0; for (int length = this.arguments.length; i < length; i++) {
        TypeBinding typeBinding = this.arguments[i];
        sig.append(typeBinding.computeUniqueKey(false));
        if ((typeBinding instanceof CaptureBinding))
          captureSourceType = ((CaptureBinding)typeBinding).sourceType;
      }
      sig.append('>');
    }
    sig.append(';');
    if ((captureSourceType != null) && (captureSourceType != this.type))
    {
      sig.insert(0, "&");
      sig.insert(0, captureSourceType.computeUniqueKey(false));
    }

    int sigLength = sig.length();
    char[] uniqueKey = new char[sigLength];
    sig.getChars(0, sigLength, uniqueKey, 0);
    return uniqueKey;
  }

  public char[] constantPoolName()
  {
    return this.type.constantPoolName();
  }

  public ParameterizedMethodBinding createParameterizedMethod(MethodBinding originalMethod) {
    return new ParameterizedMethodBinding(this, originalMethod);
  }

  public String debugName()
  {
    StringBuffer nameBuffer = new StringBuffer(10);
    nameBuffer.append(this.type.sourceName());
    if (this.arguments != null) {
      nameBuffer.append('<');
      int i = 0; for (int length = this.arguments.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(this.arguments[i].debugName());
      }
      nameBuffer.append('>');
    }
    return nameBuffer.toString();
  }

  public ReferenceBinding enclosingType()
  {
    return this.enclosingType;
  }

  public LookupEnvironment environment()
  {
    return this.environment;
  }

  public TypeBinding erasure()
  {
    return this.type.erasure();
  }

  public int fieldCount()
  {
    return this.type.fieldCount();
  }

  public FieldBinding[] fields()
  {
    if ((this.tagBits & 0x2000) != 0L)
      return this.fields;
    try
    {
      FieldBinding[] originalFields = this.type.fields();
      int length = originalFields.length;
      FieldBinding[] parameterizedFields = new FieldBinding[length];
      for (int i = 0; i < length; i++)
      {
        parameterizedFields[i] = new ParameterizedFieldBinding(this, originalFields[i]);
      }this.fields = parameterizedFields;
    }
    finally {
      if (this.fields == null)
        this.fields = Binding.NO_FIELDS;
      this.tagBits |= 8192L;
    }
    return this.fields;
  }

  public ReferenceBinding genericType()
  {
    if ((this.type instanceof UnresolvedReferenceBinding))
      ((UnresolvedReferenceBinding)this.type).resolve(this.environment, false);
    return this.type;
  }

  public char[] genericTypeSignature()
  {
    if (this.genericTypeSignature == null) {
      StringBuffer sig = new StringBuffer(10);
      if ((isMemberType()) && (enclosingType().isParameterizedType())) {
        char[] typeSig = enclosingType().genericTypeSignature();
        for (int i = 0; i < typeSig.length - 1; i++) sig.append(typeSig[i]);
        sig.append('.').append(sourceName());
      } else {
        char[] typeSig = this.type.signature();
        for (int i = 0; i < typeSig.length - 1; i++) sig.append(typeSig[i]);
      }
      if (this.arguments != null) {
        sig.append('<');
        int i = 0; for (int length = this.arguments.length; i < length; i++) {
          sig.append(this.arguments[i].genericTypeSignature());
        }
        sig.append('>');
      }
      sig.append(';');
      int sigLength = sig.length();
      this.genericTypeSignature = new char[sigLength];
      sig.getChars(0, sigLength, this.genericTypeSignature, 0);
    }
    return this.genericTypeSignature;
  }

  public long getAnnotationTagBits()
  {
    return this.type.getAnnotationTagBits();
  }

  public MethodBinding getExactConstructor(TypeBinding[] argumentTypes)
  {
    int argCount = argumentTypes.length;
    MethodBinding match = null;

    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(TypeConstants.INIT, this.methods)) >= 0L) {
        int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          if (method.parameters.length == argCount) {
            TypeBinding[] toMatch = method.parameters;
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                if (match != null) return null;
                match = method;
              }
            }
          }
        }
      } } else { MethodBinding[] matchingMethods = getMethods(TypeConstants.INIT);
      int m = matchingMethods.length;
      do { MethodBinding method = matchingMethods[m];
        TypeBinding[] toMatch = method.parameters;
        if (toMatch.length == argCount) {
          int p = 0;
          while (toMatch[p] == argumentTypes[p])
          {
            p++; if (p >= argCount)
            {
              if (match != null) return null;
              match = method;
            }
          }
        }
        m--; } while (m >= 0);
    }

    return match;
  }

  public MethodBinding getExactMethod(char[] selector, TypeBinding[] argumentTypes, CompilationUnitScope refScope)
  {
    int argCount = argumentTypes.length;
    boolean foundNothing = true;
    MethodBinding match = null;

    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
        int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          foundNothing = false;
          if (method.parameters.length == argCount) {
            TypeBinding[] toMatch = method.parameters;
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                if (match != null) return null;
                match = method;
              }
            }
          }
        }
      } } else { MethodBinding[] matchingMethods = getMethods(selector);
      foundNothing = matchingMethods == Binding.NO_METHODS;
      int m = matchingMethods.length;
      do { MethodBinding method = matchingMethods[m];
        TypeBinding[] toMatch = method.parameters;
        if (toMatch.length == argCount) {
          int p = 0;
          while (toMatch[p] == argumentTypes[p])
          {
            p++; if (p >= argCount)
            {
              if (match != null) return null;
              match = method;
            }
          }
        }
        m--; } while (m >= 0);
    }

    if (match != null)
    {
      if (match.hasSubstitutedParameters()) return null;
      return match;
    }

    if ((foundNothing) && ((this.arguments == null) || (this.arguments.length <= 1))) {
      if (isInterface()) {
        if (superInterfaces().length == 1) {
          if (refScope != null)
            refScope.recordTypeReference(this.superInterfaces[0]);
          return this.superInterfaces[0].getExactMethod(selector, argumentTypes, refScope);
        }
      } else if (superclass() != null) {
        if (refScope != null)
          refScope.recordTypeReference(this.superclass);
        return this.superclass.getExactMethod(selector, argumentTypes, refScope);
      }
    }
    return null;
  }

  public FieldBinding getField(char[] fieldName, boolean needResolve)
  {
    fields();
    return ReferenceBinding.binarySearch(fieldName, this.fields);
  }

  public ReferenceBinding getMemberType(char[] typeName)
  {
    memberTypes();
    int typeLength = typeName.length;
    int i = this.memberTypes.length;
    do { ReferenceBinding memberType = this.memberTypes[i];
      if ((memberType.sourceName.length == typeLength) && (CharOperation.equals(memberType.sourceName, typeName)))
        return memberType;
      i--; } while (i >= 0);

    return null; } 
  // ERROR //
  public MethodBinding[] getMethods(char[] selector) { // Byte code:
    //   0: aload_0
    //   1: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   4: ifnull +57 -> 61
    //   7: aload_1
    //   8: aload_0
    //   9: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   12: invokestatic 676	org/eclipse/jdt/internal/compiler/lookup/ReferenceBinding:binarySearch	([C[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;)J
    //   15: dup2
    //   16: lstore_2
    //   17: lconst_0
    //   18: lcmp
    //   19: iflt +42 -> 61
    //   22: lload_2
    //   23: l2i
    //   24: istore 4
    //   26: lload_2
    //   27: bipush 32
    //   29: lshr
    //   30: l2i
    //   31: iload 4
    //   33: isub
    //   34: iconst_1
    //   35: iadd
    //   36: istore 5
    //   38: aload_0
    //   39: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   42: iload 4
    //   44: iload 5
    //   46: anewarray 335	org/eclipse/jdt/internal/compiler/lookup/MethodBinding
    //   49: dup
    //   50: astore 6
    //   52: iconst_0
    //   53: iload 5
    //   55: invokestatic 607	java/lang/System:arraycopy	(Ljava/lang/Object;ILjava/lang/Object;II)V
    //   58: aload 6
    //   60: areturn
    //   61: aload_0
    //   62: getfield 568	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:tagBits	J
    //   65: ldc2_w 286
    //   68: land
    //   69: lconst_0
    //   70: lcmp
    //   71: ifeq +7 -> 78
    //   74: getstatic 556	org/eclipse/jdt/internal/compiler/lookup/Binding:NO_METHODS	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   77: areturn
    //   78: aconst_null
    //   79: checkcast 323	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   82: astore_2
    //   83: aload_0
    //   84: getfield 580	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:type	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   87: aload_1
    //   88: invokevirtual 675	org/eclipse/jdt/internal/compiler/lookup/ReferenceBinding:getMethods	([C)[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   91: astore_3
    //   92: aload_3
    //   93: arraylength
    //   94: istore 4
    //   96: iload 4
    //   98: ifne +24 -> 122
    //   101: getstatic 556	org/eclipse/jdt/internal/compiler/lookup/Binding:NO_METHODS	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   104: astore 8
    //   106: aload_2
    //   107: ifnonnull +12 -> 119
    //   110: aload_0
    //   111: getstatic 556	org/eclipse/jdt/internal/compiler/lookup/Binding:NO_METHODS	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   114: dup
    //   115: astore_2
    //   116: putfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   119: aload 8
    //   121: areturn
    //   122: iload 4
    //   124: anewarray 335	org/eclipse/jdt/internal/compiler/lookup/MethodBinding
    //   127: astore_2
    //   128: iconst_0
    //   129: istore 5
    //   131: goto +18 -> 149
    //   134: aload_2
    //   135: iload 5
    //   137: aload_0
    //   138: aload_3
    //   139: iload 5
    //   141: aaload
    //   142: invokevirtual 650	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:createParameterizedMethod	(Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;)Lorg/eclipse/jdt/internal/compiler/lookup/ParameterizedMethodBinding;
    //   145: aastore
    //   146: iinc 5 1
    //   149: iload 5
    //   151: iload 4
    //   153: if_icmplt -19 -> 134
    //   156: aload_0
    //   157: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   160: ifnonnull +29 -> 189
    //   163: iload 4
    //   165: anewarray 335	org/eclipse/jdt/internal/compiler/lookup/MethodBinding
    //   168: astore 5
    //   170: aload_2
    //   171: iconst_0
    //   172: aload 5
    //   174: iconst_0
    //   175: iload 4
    //   177: invokestatic 607	java/lang/System:arraycopy	(Ljava/lang/Object;ILjava/lang/Object;II)V
    //   180: aload_0
    //   181: aload 5
    //   183: putfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   186: goto +67 -> 253
    //   189: iload 4
    //   191: aload_0
    //   192: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   195: arraylength
    //   196: iadd
    //   197: istore 5
    //   199: iload 5
    //   201: anewarray 335	org/eclipse/jdt/internal/compiler/lookup/MethodBinding
    //   204: astore 6
    //   206: aload_2
    //   207: iconst_0
    //   208: aload 6
    //   210: iconst_0
    //   211: iload 4
    //   213: invokestatic 607	java/lang/System:arraycopy	(Ljava/lang/Object;ILjava/lang/Object;II)V
    //   216: aload_0
    //   217: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   220: iconst_0
    //   221: aload 6
    //   223: iload 4
    //   225: aload_0
    //   226: getfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   229: arraylength
    //   230: invokestatic 607	java/lang/System:arraycopy	(Ljava/lang/Object;ILjava/lang/Object;II)V
    //   233: iload 5
    //   235: iconst_1
    //   236: if_icmple +11 -> 247
    //   239: aload 6
    //   241: iconst_0
    //   242: iload 5
    //   244: invokestatic 677	org/eclipse/jdt/internal/compiler/lookup/ReferenceBinding:sortMethods	([Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;II)V
    //   247: aload_0
    //   248: aload 6
    //   250: putfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   253: aload_2
    //   254: astore 8
    //   256: aload_2
    //   257: ifnonnull +12 -> 269
    //   260: aload_0
    //   261: getstatic 556	org/eclipse/jdt/internal/compiler/lookup/Binding:NO_METHODS	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   264: dup
    //   265: astore_2
    //   266: putfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   269: aload 8
    //   271: areturn
    //   272: astore 7
    //   274: aload_2
    //   275: ifnonnull +12 -> 287
    //   278: aload_0
    //   279: getstatic 556	org/eclipse/jdt/internal/compiler/lookup/Binding:NO_METHODS	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   282: dup
    //   283: astore_2
    //   284: putfield 576	org/eclipse/jdt/internal/compiler/lookup/ParameterizedTypeBinding:methods	[Lorg/eclipse/jdt/internal/compiler/lookup/MethodBinding;
    //   287: aload 7
    //   289: athrow
    //
    // Exception table:
    //   from	to	target	type
    //   83	106	272	finally
    //   122	256	272	finally } 
  public boolean hasMemberTypes() { return this.type.hasMemberTypes(); }


  public boolean implementsMethod(MethodBinding method)
  {
    return this.type.implementsMethod(method);
  }

  void initialize(ReferenceBinding someType, TypeBinding[] someArguments) {
    this.type = someType;
    this.sourceName = someType.sourceName;
    this.compoundName = someType.compoundName;
    this.fPackage = someType.fPackage;
    this.fileName = someType.fileName;

    this.modifiers = (someType.modifiers & 0xBFFFFFFF);

    if (someArguments != null) {
      this.modifiers |= 1073741824;
    } else if (this.enclosingType != null) {
      this.modifiers |= this.enclosingType.modifiers & 0x40000000;
      this.tagBits |= this.enclosingType.tagBits & 0x20000000;
    }
    if (someArguments != null) {
      this.arguments = someArguments;
      int i = 0; for (int length = someArguments.length; i < length; i++) {
        TypeBinding someArgument = someArguments[i];
        boolean isWildcardArgument = someArgument.isWildcard();
        if (isWildcardArgument) {
          this.tagBits |= 1073741824L;
        }
        if ((!isWildcardArgument) || (((WildcardBinding)someArgument).boundKind != 0)) {
          this.tagBits |= 8388608L;
        }
        this.tagBits |= someArgument.tagBits & 0x20000000;
      }
    }
    this.tagBits |= someType.tagBits & 0x1C;
    this.tagBits &= -40961L;
  }

  protected void initializeArguments()
  {
  }

  public boolean isEquivalentTo(TypeBinding otherType) {
    if (this == otherType)
      return true;
    if (otherType == null)
      return false;
    switch (otherType.kind())
    {
    case 516:
      return ((WildcardBinding)otherType).boundCheck(this);
    case 260:
      ParameterizedTypeBinding otherParamType = (ParameterizedTypeBinding)otherType;
      if (this.type != otherParamType.type)
        return false;
      if (!isStatic()) {
        ReferenceBinding enclosing = enclosingType();
        if (enclosing != null) {
          ReferenceBinding otherEnclosing = otherParamType.enclosingType();
          if (otherEnclosing == null) return false;
          if ((otherEnclosing.tagBits & 0x40000000) == 0L) {
            if (enclosing != otherEnclosing) return false;
          }
          else if (!enclosing.isEquivalentTo(otherParamType.enclosingType())) return false;
        }
      }

      if (this.arguments == null) {
        return otherParamType.arguments == null;
      }
      int length = this.arguments.length;
      TypeBinding[] otherArguments = otherParamType.arguments;
      if ((otherArguments == null) || (otherArguments.length != length)) return false;
      for (int i = 0; i < length; i++) {
        if (!this.arguments[i].isTypeArgumentContainedBy(otherArguments[i]))
          return false;
      }
      return true;
    case 1028:
      return erasure() == otherType.erasure();
    }
    return false;
  }

  public boolean isIntersectingWith(TypeBinding otherType) {
    if (this == otherType)
      return true;
    if (otherType == null)
      return false;
    switch (otherType.kind())
    {
    case 260:
      ParameterizedTypeBinding otherParamType = (ParameterizedTypeBinding)otherType;
      if (this.type != otherParamType.type)
        return false;
      if (!isStatic()) {
        ReferenceBinding enclosing = enclosingType();
        if (enclosing != null) {
          ReferenceBinding otherEnclosing = otherParamType.enclosingType();
          if (otherEnclosing == null) return false;
          if ((otherEnclosing.tagBits & 0x40000000) == 0L) {
            if (enclosing != otherEnclosing) return false;
          }
          else if (!enclosing.isEquivalentTo(otherParamType.enclosingType())) return false;
        }
      }

      int length = this.arguments == null ? 0 : this.arguments.length;
      TypeBinding[] otherArguments = otherParamType.arguments;
      int otherLength = otherArguments == null ? 0 : otherArguments.length;
      if (otherLength != length)
        return false;
      for (int i = 0; i < length; i++) {
        if (!this.arguments[i].isTypeArgumentIntersecting(otherArguments[i]))
          return false;
      }
      return true;
    case 2052:
      SourceTypeBinding otherGenericType = (SourceTypeBinding)otherType;
      if (this.type != otherGenericType)
        return false;
      if (!isStatic()) {
        ReferenceBinding enclosing = enclosingType();
        if (enclosing != null) {
          ReferenceBinding otherEnclosing = otherGenericType.enclosingType();
          if (otherEnclosing == null) return false;
          if ((otherEnclosing.tagBits & 0x40000000) == 0L) {
            if (enclosing != otherEnclosing) return false;
          }
          else if (!enclosing.isEquivalentTo(otherGenericType.enclosingType())) return false;
        }
      }

      int length = this.arguments == null ? 0 : this.arguments.length;
      TypeBinding[] otherArguments = otherGenericType.typeVariables();
      int otherLength = otherArguments == null ? 0 : otherArguments.length;
      if (otherLength != length)
        return false;
      for (int i = 0; i < length; i++) {
        if (!this.arguments[i].isTypeArgumentIntersecting(otherArguments[i]))
          return false;
      }
      return true;
    case 1028:
      return erasure() == otherType.erasure();
    }
    return false;
  }

  public boolean isParameterizedType()
  {
    return true;
  }

  public boolean isRawSubstitution()
  {
    return isRawType();
  }

  public int kind() {
    return 260;
  }

  public ReferenceBinding[] memberTypes()
  {
    if (this.memberTypes == null) {
      try {
        ReferenceBinding[] originalMemberTypes = this.type.memberTypes();
        int length = originalMemberTypes.length;
        ReferenceBinding[] parameterizedMemberTypes = new ReferenceBinding[length];

        for (int i = 0; i < length; i++)
        {
          parameterizedMemberTypes[i] = 
            this.environment.createParameterizedType(originalMemberTypes[i], null, this);
        }this.memberTypes = parameterizedMemberTypes;
      }
      finally {
        if (this.memberTypes == null)
          this.memberTypes = Binding.NO_MEMBER_TYPES;
      }
    }
    return this.memberTypes;
  }

  public MethodBinding[] methods()
  {
    if ((this.tagBits & 0x8000) != 0L)
      return this.methods;
    try
    {
      MethodBinding[] originalMethods = this.type.methods();
      int length = originalMethods.length;
      MethodBinding[] parameterizedMethods = new MethodBinding[length];
      for (int i = 0; i < length; i++)
      {
        parameterizedMethods[i] = createParameterizedMethod(originalMethods[i]);
      }this.methods = parameterizedMethods;
    }
    finally {
      if (this.methods == null) {
        this.methods = Binding.NO_METHODS;
      }
      this.tagBits |= 32768L;
    }
    return this.methods;
  }

  public char[] qualifiedPackageName()
  {
    return this.type.qualifiedPackageName();
  }

  public char[] qualifiedSourceName()
  {
    return this.type.qualifiedSourceName();
  }

  public char[] readableName()
  {
    StringBuffer nameBuffer = new StringBuffer(10);
    if (isMemberType())
      nameBuffer.append(CharOperation.concat(enclosingType().readableName(), this.sourceName, '.'));
    else {
      nameBuffer.append(CharOperation.concatWith(this.type.compoundName, '.'));
    }
    if (this.arguments != null) {
      nameBuffer.append('<');
      int i = 0; for (int length = this.arguments.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(this.arguments[i].readableName());
      }
      nameBuffer.append('>');
    }
    int nameLength = nameBuffer.length();
    char[] readableName = new char[nameLength];
    nameBuffer.getChars(0, nameLength, readableName, 0);
    return readableName;
  }

  ReferenceBinding resolve() {
    if ((this.tagBits & 0x1000000) == 0L) {
      return this;
    }
    this.tagBits &= -16777217L;
    ReferenceBinding resolvedType = BinaryTypeBinding.resolveType(this.type, this.environment, false);
    if (this.arguments != null) {
      int argLength = this.arguments.length;
      for (int i = 0; i < argLength; i++) {
        BinaryTypeBinding.resolveType(this.arguments[i], this.environment, this, i);
      }
      TypeVariableBinding[] refTypeVariables = resolvedType.typeVariables();
      if (refTypeVariables == Binding.NO_TYPE_VARIABLES) {
        this.environment.problemReporter.nonGenericTypeCannotBeParameterized(null, resolvedType, this.arguments);
        return this;
      }if (argLength != refTypeVariables.length) {
        this.environment.problemReporter.incorrectArityForParameterizedType(null, resolvedType, this.arguments);
        return this;
      }

    }

    return this;
  }

  public char[] shortReadableName()
  {
    StringBuffer nameBuffer = new StringBuffer(10);
    if (isMemberType())
      nameBuffer.append(CharOperation.concat(enclosingType().shortReadableName(), this.sourceName, '.'));
    else {
      nameBuffer.append(this.type.sourceName);
    }
    if (this.arguments != null) {
      nameBuffer.append('<');
      int i = 0; for (int length = this.arguments.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(this.arguments[i].shortReadableName());
      }
      nameBuffer.append('>');
    }
    int nameLength = nameBuffer.length();
    char[] shortReadableName = new char[nameLength];
    nameBuffer.getChars(0, nameLength, shortReadableName, 0);
    return shortReadableName;
  }

  public char[] signature()
  {
    if (this.signature == null) {
      this.signature = this.type.signature();
    }
    return this.signature;
  }

  public char[] sourceName()
  {
    return this.type.sourceName();
  }

  public TypeBinding substitute(TypeVariableBinding originalVariable)
  {
    ParameterizedTypeBinding currentType = this;
    while (true) {
      TypeVariableBinding[] typeVariables = currentType.type.typeVariables();
      int length = typeVariables.length;

      if ((originalVariable.rank < length) && (typeVariables[originalVariable.rank] == originalVariable))
      {
        if (currentType.arguments == null)
          currentType.initializeArguments();
        if (currentType.arguments != null) {
          return currentType.arguments[originalVariable.rank];
        }
      }
      if (currentType.isStatic()) break;
      ReferenceBinding enclosing = currentType.enclosingType();
      if (!(enclosing instanceof ParameterizedTypeBinding))
        break;
      currentType = (ParameterizedTypeBinding)enclosing;
    }
    return originalVariable;
  }

  public ReferenceBinding superclass()
  {
    if (this.superclass == null)
    {
      ReferenceBinding genericSuperclass = this.type.superclass();
      if (genericSuperclass == null) return null;
      this.superclass = ((ReferenceBinding)Scope.substitute(this, genericSuperclass));
    }
    return this.superclass;
  }

  public ReferenceBinding[] superInterfaces()
  {
    if (this.superInterfaces == null) {
      if (this.type.isHierarchyBeingConnected())
        return Binding.NO_SUPERINTERFACES;
      this.superInterfaces = Scope.substitute(this, this.type.superInterfaces());
    }
    return this.superInterfaces;
  }

  public void swapUnresolved(UnresolvedReferenceBinding unresolvedType, ReferenceBinding resolvedType, LookupEnvironment env) {
    boolean update = false;
    if (this.type == unresolvedType) {
      this.type = resolvedType;
      update = true;
      ReferenceBinding enclosing = resolvedType.enclosingType();
      if (enclosing != null) {
        this.enclosingType = ((ReferenceBinding)env.convertUnresolvedBinaryToRawType(enclosing));
      }
    }
    if (this.arguments != null) {
      int i = 0; for (int l = this.arguments.length; i < l; i++) {
        if (this.arguments[i] == unresolvedType) {
          this.arguments[i] = env.convertUnresolvedBinaryToRawType(resolvedType);
          update = true;
        }
      }
    }
    if (update)
      initialize(this.type, this.arguments);
  }

  public ReferenceBinding[] syntheticEnclosingInstanceTypes()
  {
    return this.type.syntheticEnclosingInstanceTypes();
  }

  public SyntheticArgumentBinding[] syntheticOuterLocalVariables()
  {
    return this.type.syntheticOuterLocalVariables();
  }

  public String toString()
  {
    StringBuffer buffer = new StringBuffer(30);
    if (isDeprecated()) buffer.append("deprecated ");
    if (isPublic()) buffer.append("public ");
    if (isProtected()) buffer.append("protected ");
    if (isPrivate()) buffer.append("private ");
    if ((isAbstract()) && (isClass())) buffer.append("abstract ");
    if ((isStatic()) && (isNestedType())) buffer.append("static ");
    if (isFinal()) buffer.append("final ");

    if (isEnum()) buffer.append("enum ");
    else if (isAnnotationType()) buffer.append("@interface ");
    else if (isClass()) buffer.append("class "); else
      buffer.append("interface ");
    buffer.append(debugName());

    buffer.append("\n\textends ");
    buffer.append(this.superclass != null ? this.superclass.debugName() : "NULL TYPE");

    if (this.superInterfaces != null) {
      if (this.superInterfaces != Binding.NO_SUPERINTERFACES) {
        buffer.append("\n\timplements : ");
        int i = 0; for (int length = this.superInterfaces.length; i < length; i++) {
          if (i > 0)
            buffer.append(", ");
          buffer.append(this.superInterfaces[i] != null ? this.superInterfaces[i].debugName() : "NULL TYPE");
        }
      }
    }
    else buffer.append("NULL SUPERINTERFACES");

    if (enclosingType() != null) {
      buffer.append("\n\tenclosing type : ");
      buffer.append(enclosingType().debugName());
    }

    if (this.fields != null) {
      if (this.fields != Binding.NO_FIELDS) {
        buffer.append("\n/*   fields   */");
        int i = 0; for (int length = this.fields.length; i < length; i++)
          buffer.append('\n').append(this.fields[i] != null ? this.fields[i].toString() : "NULL FIELD");
      }
    }
    else buffer.append("NULL FIELDS");

    if (this.methods != null) {
      if (this.methods != Binding.NO_METHODS) {
        buffer.append("\n/*   methods   */");
        int i = 0; for (int length = this.methods.length; i < length; i++)
          buffer.append('\n').append(this.methods[i] != null ? this.methods[i].toString() : "NULL METHOD");
      }
    }
    else buffer.append("NULL METHODS");

    buffer.append("\n\n");
    return buffer.toString();
  }

  public TypeVariableBinding[] typeVariables()
  {
    if (this.arguments == null)
    {
      return this.type.typeVariables();
    }
    return Binding.NO_TYPE_VARIABLES;
  }
}