package org.eclipse.jdt.internal.compiler.lookup;

import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;
import org.eclipse.jdt.internal.compiler.util.Util;

public class SourceTypeBinding extends ReferenceBinding
{
  public ReferenceBinding superclass;
  public ReferenceBinding[] superInterfaces;
  private FieldBinding[] fields;
  private MethodBinding[] methods;
  public ReferenceBinding[] memberTypes;
  public TypeVariableBinding[] typeVariables;
  public ClassScope scope;
  public static final int METHOD_EMUL = 0;
  public static final int FIELD_EMUL = 1;
  public static final int CLASS_LITERAL_EMUL = 2;
  public static final int RECEIVER_TYPE_EMUL = 3;
  HashMap[] synthetics;
  char[] genericReferenceTypeSignature;
  private SimpleLookupTable storedAnnotations = null;

  public SourceTypeBinding(char[][] compoundName, PackageBinding fPackage, ClassScope scope) {
    this.compoundName = compoundName;
    this.fPackage = fPackage;
    this.fileName = scope.referenceCompilationUnit().getFileName();
    this.modifiers = scope.referenceContext.modifiers;
    this.sourceName = scope.referenceContext.name;
    this.scope = scope;

    this.fields = Binding.NO_FIELDS;
    this.methods = Binding.NO_METHODS;

    computeId();
  }

  private void addDefaultAbstractMethods() {
    if ((this.tagBits & 0x400) != 0L) return;

    this.tagBits |= 1024L;
    if ((isClass()) && (isAbstract())) {
      if (this.scope.compilerOptions().targetJDK >= 3014656L) {
        return;
      }
      ReferenceBinding[] itsInterfaces = superInterfaces();
      if (itsInterfaces != Binding.NO_SUPERINTERFACES) {
        MethodBinding[] defaultAbstracts = (MethodBinding[])null;
        int defaultAbstractsCount = 0;
        ReferenceBinding[] interfacesToVisit = itsInterfaces;
        int nextPosition = interfacesToVisit.length;
        for (int i = 0; i < nextPosition; i++) {
          ReferenceBinding superType = interfacesToVisit[i];
          if (superType.isValidBinding()) {
            MethodBinding[] superMethods = superType.methods();
            int m = superMethods.length;
            do { MethodBinding method = superMethods[m];

              if (!implementsMethod(method))
              {
                if (defaultAbstractsCount == 0) {
                  defaultAbstracts = new MethodBinding[5];
                }
                else {
                  for (int k = 0; k < defaultAbstractsCount; k++) {
                    MethodBinding alreadyAdded = defaultAbstracts[k];
                    if ((CharOperation.equals(alreadyAdded.selector, method.selector)) && (alreadyAdded.areParametersEqual(method)))
                      break;
                  }
                }
                MethodBinding defaultAbstract = new MethodBinding(
                  method.modifiers | 0x80000, 
                  method.selector, 
                  method.returnType, 
                  method.parameters, 
                  method.thrownExceptions, 
                  this);
                if (defaultAbstractsCount == defaultAbstracts.length)
                  System.arraycopy(defaultAbstracts, 0, defaultAbstracts = new MethodBinding[2 * defaultAbstractsCount], 0, defaultAbstractsCount);
                defaultAbstracts[(defaultAbstractsCount++)] = defaultAbstract;
              }
              m--; } while (m >= 0);

            if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
              int itsLength = itsInterfaces.length;
              if (nextPosition + itsLength >= interfacesToVisit.length)
                System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
              for (int a = 0; a < itsLength; a++) {
                ReferenceBinding next = itsInterfaces[a];
                int b = 0;
                while (next != interfacesToVisit[b])
                {
                  b++; if (b >= nextPosition)
                  {
                    interfacesToVisit[(nextPosition++)] = next;
                  }
                }
              }
            }
          }
        }
        if (defaultAbstractsCount > 0) {
          int length = this.methods.length;
          System.arraycopy(this.methods, 0, this.methods = new MethodBinding[length + defaultAbstractsCount], 0, length);
          System.arraycopy(defaultAbstracts, 0, this.methods, length, defaultAbstractsCount);

          length += defaultAbstractsCount;
          if (length > 1)
            ReferenceBinding.sortMethods(this.methods, 0, length);
        }
      }
    }
  }

  public FieldBinding addSyntheticFieldForInnerclass(LocalVariableBinding actualOuterLocalVariable)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[1] == null) {
      this.synthetics[1] = new HashMap(5);
    }
    FieldBinding synthField = (FieldBinding)this.synthetics[1].get(actualOuterLocalVariable);
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        CharOperation.concat(TypeConstants.SYNTHETIC_OUTER_LOCAL_PREFIX, actualOuterLocalVariable.name), 
        actualOuterLocalVariable.type, 
        4114, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[1].size());
      this.synthetics[1].put(actualOuterLocalVariable, synthField);
    }

    int index = 1;
    boolean needRecheck;
    do {
      needRecheck = false;
      FieldBinding existingField;
      if ((existingField = getField(synthField.name, true)) != null) {
        TypeDeclaration typeDecl = this.scope.referenceContext;
        int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
          FieldDeclaration fieldDecl = typeDecl.fields[i];
          if (fieldDecl.binding == existingField) {
            synthField.name = CharOperation.concat(
              TypeConstants.SYNTHETIC_OUTER_LOCAL_PREFIX, 
              actualOuterLocalVariable.name, 
              ("$" + String.valueOf(index++)).toCharArray());
            needRecheck = true;
            break;
          }
        }
      }
    }
    while (
      needRecheck);
    return synthField;
  }

  public FieldBinding addSyntheticFieldForInnerclass(ReferenceBinding enclosingType)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[1] == null) {
      this.synthetics[1] = new HashMap(5);
    }
    FieldBinding synthField = (FieldBinding)this.synthetics[1].get(enclosingType);
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        CharOperation.concat(
        TypeConstants.SYNTHETIC_ENCLOSING_INSTANCE_PREFIX, 
        String.valueOf(enclosingType.depth()).toCharArray()), 
        enclosingType, 
        4112, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[1].size());
      this.synthetics[1].put(enclosingType, synthField);
    }
    boolean needRecheck;
    do
    {
      needRecheck = false;
      FieldBinding existingField;
      if ((existingField = getField(synthField.name, true)) != null) {
        TypeDeclaration typeDecl = this.scope.referenceContext;
        int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
          FieldDeclaration fieldDecl = typeDecl.fields[i];
          if (fieldDecl.binding == existingField) {
            if (this.scope.compilerOptions().complianceLevel >= 3211264L) {
              synthField.name = CharOperation.concat(
                synthField.name, 
                "$".toCharArray());
              needRecheck = true; break;
            }
            this.scope.problemReporter().duplicateFieldInType(this, fieldDecl);

            break;
          }
        }
      }
    }
    while (
      needRecheck);
    return synthField;
  }

  public FieldBinding addSyntheticFieldForClassLiteral(TypeBinding targetType, BlockScope blockScope)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[2] == null) {
      this.synthetics[2] = new HashMap(5);
    }

    FieldBinding synthField = (FieldBinding)this.synthetics[2].get(targetType);
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        CharOperation.concat(
        TypeConstants.SYNTHETIC_CLASS, 
        String.valueOf(this.synthetics[2].size()).toCharArray()), 
        blockScope.getJavaLangClass(), 
        4104, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[2].size());
      this.synthetics[2].put(targetType, synthField);
    }
    FieldBinding existingField;
    if ((existingField = getField(synthField.name, true)) != null) {
      TypeDeclaration typeDecl = blockScope.referenceType();
      int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
        FieldDeclaration fieldDecl = typeDecl.fields[i];
        if (fieldDecl.binding == existingField) {
          blockScope.problemReporter().duplicateFieldInType(this, fieldDecl);
          break;
        }
      }
    }
    return synthField;
  }

  public FieldBinding addSyntheticFieldForAssert(BlockScope blockScope)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[1] == null) {
      this.synthetics[1] = new HashMap(5);
    }
    FieldBinding synthField = (FieldBinding)this.synthetics[1].get("assertionEmulation");
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        TypeConstants.SYNTHETIC_ASSERT_DISABLED, 
        TypeBinding.BOOLEAN, 
        4120, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[1].size());
      this.synthetics[1].put("assertionEmulation", synthField);
    }

    int index = 0;
    boolean needRecheck;
    do {
      needRecheck = false;
      FieldBinding existingField;
      if ((existingField = getField(synthField.name, true)) != null) {
        TypeDeclaration typeDecl = this.scope.referenceContext;
        int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
          FieldDeclaration fieldDecl = typeDecl.fields[i];
          if (fieldDecl.binding == existingField) {
            synthField.name = CharOperation.concat(
              TypeConstants.SYNTHETIC_ASSERT_DISABLED, 
              ("_" + String.valueOf(index++)).toCharArray());
            needRecheck = true;
            break;
          }
        }
      }
    }
    while (
      needRecheck);
    return synthField;
  }

  public FieldBinding addSyntheticFieldForEnumValues()
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[1] == null) {
      this.synthetics[1] = new HashMap(5);
    }
    FieldBinding synthField = (FieldBinding)this.synthetics[1].get("enumConstantValues");
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        TypeConstants.SYNTHETIC_ENUM_VALUES, 
        this.scope.createArrayType(this, 1), 
        4122, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[1].size());
      this.synthetics[1].put("enumConstantValues", synthField);
    }

    int index = 0;
    boolean needRecheck;
    do {
      needRecheck = false;
      FieldBinding existingField;
      if ((existingField = getField(synthField.name, true)) != null) {
        TypeDeclaration typeDecl = this.scope.referenceContext;
        int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
          FieldDeclaration fieldDecl = typeDecl.fields[i];
          if (fieldDecl.binding == existingField) {
            synthField.name = CharOperation.concat(
              TypeConstants.SYNTHETIC_ENUM_VALUES, 
              ("_" + String.valueOf(index++)).toCharArray());
            needRecheck = true;
            break;
          }
        }
      }
    }
    while (
      needRecheck);
    return synthField;
  }

  public SyntheticMethodBinding addSyntheticMethod(FieldBinding targetField, boolean isReadAccess)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[0] == null) {
      this.synthetics[0] = new HashMap(5);
    }
    SyntheticMethodBinding accessMethod = null;
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(targetField);
    if (accessors == null) {
      accessMethod = new SyntheticMethodBinding(targetField, isReadAccess, this);
      this.synthetics[0].put(targetField, accessors = new SyntheticMethodBinding[2]);
      accessors[(isReadAccess ? 0 : 1)] = accessMethod;
    }
    else if ((accessMethod = accessors[1]) == null) {
      accessMethod = new SyntheticMethodBinding(targetField, isReadAccess, this);
      accessors[(isReadAccess ? 0 : 1)] = accessMethod;
    }

    return accessMethod;
  }

  public SyntheticMethodBinding addSyntheticEnumMethod(char[] selector)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[0] == null) {
      this.synthetics[0] = new HashMap(5);
    }
    SyntheticMethodBinding accessMethod = null;
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(selector);
    if (accessors == null) {
      accessMethod = new SyntheticMethodBinding(this, selector);
      this.synthetics[0].put(selector, accessors = new SyntheticMethodBinding[2]);
      accessors[0] = accessMethod;
    }
    else if ((accessMethod = accessors[0]) == null) {
      accessMethod = new SyntheticMethodBinding(this, selector);
      accessors[0] = accessMethod;
    }

    return accessMethod;
  }

  public SyntheticFieldBinding addSyntheticFieldForSwitchEnum(char[] fieldName, String key)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[1] == null) {
      this.synthetics[1] = new HashMap(5);
    }
    SyntheticFieldBinding synthField = (SyntheticFieldBinding)this.synthetics[1].get(key);
    if (synthField == null) {
      synthField = new SyntheticFieldBinding(
        fieldName, 
        this.scope.createArrayType(TypeBinding.INT, 1), 
        4106, 
        this, 
        Constant.NotAConstant, 
        this.synthetics[1].size());
      this.synthetics[1].put(key, synthField);
    }
int index = 0;
    boolean needRecheck;
    do {
      needRecheck = false;
      FieldBinding existingField;
      if ((existingField = getField(synthField.name, true)) != null) {
        TypeDeclaration typeDecl = this.scope.referenceContext;
        int i = 0; for (int max = typeDecl.fields.length; i < max; i++) {
          FieldDeclaration fieldDecl = typeDecl.fields[i];
          if (fieldDecl.binding == existingField) {
            synthField.name = CharOperation.concat(
              fieldName, 
              ("_" + String.valueOf(index++)).toCharArray());
            needRecheck = true;
            break;
          }
        }
      }
    }
    while (
      needRecheck);
    return synthField;
  }

  public SyntheticMethodBinding addSyntheticMethodForSwitchEnum(TypeBinding enumBinding)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[0] == null) {
      this.synthetics[0] = new HashMap(5);
    }
    SyntheticMethodBinding accessMethod = null;
    char[] selector = CharOperation.concat(TypeConstants.SYNTHETIC_SWITCH_ENUM_TABLE, enumBinding.constantPoolName());
    CharOperation.replace(selector, '/', '$');
    String key = new String(selector);
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(key);

    if (accessors == null)
    {
      SyntheticFieldBinding fieldBinding = addSyntheticFieldForSwitchEnum(selector, key);
      accessMethod = new SyntheticMethodBinding(fieldBinding, this, enumBinding, selector);
      this.synthetics[0].put(key, accessors = new SyntheticMethodBinding[2]);
      accessors[0] = accessMethod;
    }
    else if ((accessMethod = accessors[0]) == null) {
      SyntheticFieldBinding fieldBinding = addSyntheticFieldForSwitchEnum(selector, key);
      accessMethod = new SyntheticMethodBinding(fieldBinding, this, enumBinding, selector);
      accessors[0] = accessMethod;
    }

    return accessMethod;
  }

  public SyntheticMethodBinding addSyntheticMethod(MethodBinding targetMethod, boolean isSuperAccess)
  {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[0] == null) {
      this.synthetics[0] = new HashMap(5);
    }
    SyntheticMethodBinding accessMethod = null;
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(targetMethod);
    if (accessors == null) {
      accessMethod = new SyntheticMethodBinding(targetMethod, isSuperAccess, this);
      this.synthetics[0].put(targetMethod, accessors = new SyntheticMethodBinding[2]);
      accessors[(isSuperAccess ? 0 : 1)] = accessMethod;
    }
    else if ((accessMethod = accessors[1]) == null) {
      accessMethod = new SyntheticMethodBinding(targetMethod, isSuperAccess, this);
      accessors[(isSuperAccess ? 0 : 1)] = accessMethod;
    }

    return accessMethod;
  }

  public SyntheticMethodBinding addSyntheticBridgeMethod(MethodBinding inheritedMethodToBridge, MethodBinding targetMethod)
  {
    if (isInterface()) return null;

    if ((inheritedMethodToBridge.returnType.erasure() == targetMethod.returnType.erasure()) && 
      (inheritedMethodToBridge.areParameterErasuresEqual(targetMethod))) {
      return null;
    }
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[0] == null) {
      this.synthetics[0] = new HashMap(5);
    }
    else {
      Iterator synthMethods = this.synthetics[0].keySet().iterator();
      while (synthMethods.hasNext()) {
        Object synthetic = synthMethods.next();
        if ((synthetic instanceof MethodBinding)) {
          MethodBinding method = (MethodBinding)synthetic;
          if ((CharOperation.equals(inheritedMethodToBridge.selector, method.selector)) && 
            (inheritedMethodToBridge.returnType.erasure() == method.returnType.erasure()) && 
            (inheritedMethodToBridge.areParameterErasuresEqual(method))) {
            return null;
          }
        }
      }
    }

    SyntheticMethodBinding accessMethod = null;
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(inheritedMethodToBridge);
    if (accessors == null) {
      accessMethod = new SyntheticMethodBinding(inheritedMethodToBridge, targetMethod, this);
      this.synthetics[0].put(inheritedMethodToBridge, accessors = new SyntheticMethodBinding[2]);
      accessors[1] = accessMethod;
    }
    else if ((accessMethod = accessors[1]) == null) {
      accessMethod = new SyntheticMethodBinding(inheritedMethodToBridge, targetMethod, this);
      accessors[1] = accessMethod;
    }

    return accessMethod;
  }
  public int kind() {
    if (this.typeVariables != Binding.NO_TYPE_VARIABLES) return 2052;
    return 4;
  }
  public char[] computeUniqueKey(boolean isLeaf) {
    char[] uniqueKey = super.computeUniqueKey(isLeaf);
    if (uniqueKey.length == 2) return uniqueKey;
    if (Util.isClassFileName(this.fileName)) return uniqueKey;

    int end = CharOperation.lastIndexOf('.', this.fileName);
    if (end != -1) {
      int start = CharOperation.lastIndexOf('/', this.fileName) + 1;
      char[] mainTypeName = CharOperation.subarray(this.fileName, start, end);
      start = CharOperation.lastIndexOf('/', uniqueKey) + 1;
      if (start == 0)
        start = 1;
      end = CharOperation.indexOf('$', uniqueKey, start);
      if (end == -1)
        end = CharOperation.indexOf('<', uniqueKey, start);
      if (end == -1)
        end = CharOperation.indexOf(';', uniqueKey, start);
      char[] topLevelType = CharOperation.subarray(uniqueKey, start, end);
      if (!CharOperation.equals(topLevelType, mainTypeName)) {
        StringBuffer buffer = new StringBuffer();
        buffer.append(uniqueKey, 0, start);
        buffer.append(mainTypeName);
        buffer.append('~');
        buffer.append(topLevelType);
        buffer.append(uniqueKey, end, uniqueKey.length - end);
        int length = buffer.length();
        uniqueKey = new char[length];
        buffer.getChars(0, length, uniqueKey, 0);
        return uniqueKey;
      }
    }
    return uniqueKey;
  }

  void faultInTypesForFieldsAndMethods() {
    getAnnotationTagBits();
    ReferenceBinding enclosingType = enclosingType();
    if ((enclosingType != null) && (enclosingType.isViewedAsDeprecated()) && (!isDeprecated()))
      this.modifiers |= 2097152;
    fields();
    methods();

    int i = 0; for (int length = this.memberTypes.length; i < length; i++)
      ((SourceTypeBinding)this.memberTypes[i]).faultInTypesForFieldsAndMethods();
  }

  public FieldBinding[] fields() {
    if ((this.tagBits & 0x2000) != 0L) {
      return this.fields; } int failed = 0;
    FieldBinding[] resolvedFields = this.fields;
    int newSize;
    FieldBinding[] newFields;
    int i;
    int j;
    int length;
    try { if ((this.tagBits & 0x1000) == 0L) {
        int length = this.fields.length;
        if (length > 1)
          ReferenceBinding.sortFields(this.fields, 0, length);
        this.tagBits |= 4096L;
      }
      int i = 0; for (int length = this.fields.length; i < length; i++)
        if (resolveTypeFor(this.fields[i]) == null)
        {
          if (resolvedFields == this.fields) {
            System.arraycopy(this.fields, 0, resolvedFields = new FieldBinding[length], 0, length);
          }
          resolvedFields[i] = null;
          failed++;
        }
    } finally
    {
      if (failed > 0)
      {
        int newSize = resolvedFields.length - failed;
        if (newSize == 0) {
          return this.fields = Binding.NO_FIELDS;
        }
        FieldBinding[] newFields = new FieldBinding[newSize];
        int i = 0; int j = 0; for (int length = resolvedFields.length; i < length; i++) {
          if (resolvedFields[i] != null)
            newFields[(j++)] = resolvedFields[i];
        }
        this.fields = newFields;
      }
    }
    this.tagBits |= 8192L;
    return this.fields;
  }

  public char[] genericTypeSignature()
  {
    if (this.genericReferenceTypeSignature == null)
      this.genericReferenceTypeSignature = computeGenericTypeSignature(this.typeVariables);
    return this.genericReferenceTypeSignature;
  }

  public char[] genericSignature()
  {
    StringBuffer sig = null;
    if (this.typeVariables != Binding.NO_TYPE_VARIABLES) {
      sig = new StringBuffer(10);
      sig.append('<');
      int i = 0; for (int length = this.typeVariables.length; i < length; i++)
        sig.append(this.typeVariables[i].genericSignature());
      sig.append('>');
    }
    else {
      if ((this.superclass == null) || (!this.superclass.isParameterizedType())) {
        int i = 0; int length = this.superInterfaces.length;
        while (!this.superInterfaces[i].isParameterizedType())
        {
          i++; if (i >= length)
          {
            return null;
          }
        }
      }
      sig = new StringBuffer(10);
    }
    if (this.superclass != null)
      sig.append(this.superclass.genericTypeSignature());
    else
      sig.append(this.scope.getJavaLangObject().genericTypeSignature());
    int i = 0; for (int length = this.superInterfaces.length; i < length; i++)
      sig.append(this.superInterfaces[i].genericTypeSignature());
    return sig.toString().toCharArray();
  }

  public long getAnnotationTagBits()
  {
    if (((this.tagBits & 0x0) == 0L) && (this.scope != null)) {
      TypeDeclaration typeDecl = this.scope.referenceContext;
      boolean old = typeDecl.staticInitializerScope.insideTypeAnnotation;
      try {
        typeDecl.staticInitializerScope.insideTypeAnnotation = true;
        ASTNode.resolveAnnotations(typeDecl.staticInitializerScope, typeDecl.annotations, this);
      } finally {
        typeDecl.staticInitializerScope.insideTypeAnnotation = old;
      }
      if ((this.tagBits & 0x0) != 0L)
        this.modifiers |= 1048576;
    }
    return this.tagBits;
  }
  public MethodBinding[] getDefaultAbstractMethods() {
    int count = 0;
    int i = this.methods.length;
    do { if (this.methods[i].isDefaultAbstract())
        count++;
      i--; } while (i >= 0);

    if (count == 0) return Binding.NO_METHODS;

    MethodBinding[] result = new MethodBinding[count];
    count = 0;
    int i = this.methods.length;
    do { if (this.methods[i].isDefaultAbstract())
        result[(count++)] = this.methods[i];
      i--; } while (i >= 0);

    return result;
  }

  public MethodBinding getExactConstructor(TypeBinding[] argumentTypes) {
    int argCount = argumentTypes.length;
    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(TypeConstants.INIT, this.methods)) >= 0L) {
        int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          if (method.parameters.length == argCount) {
            TypeBinding[] toMatch = method.parameters;
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                return method;
              }
            }
          }
        }
      }
    } else { if ((this.tagBits & 0x4000) == 0L) {
        int length = this.methods.length;
        if (length > 1)
          ReferenceBinding.sortMethods(this.methods, 0, length);
        this.tagBits |= 16384L;
      }
      long range;
      if ((range = ReferenceBinding.binarySearch(TypeConstants.INIT, this.methods)) >= 0L) {
        int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          if ((resolveTypesFor(method) == null) || (method.returnType == null)) {
            methods();
            return getExactConstructor(argumentTypes);
          }
          if (method.parameters.length == argCount) {
            TypeBinding[] toMatch = method.parameters;
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                return method;
              }
            }
          }
        }
      } } return null;
  }

  public MethodBinding getExactMethod(char[] selector, TypeBinding[] argumentTypes, CompilationUnitScope refScope)
  {
    int argCount = argumentTypes.length;
    boolean foundNothing = true;

    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
        int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          foundNothing = false;
          if (method.parameters.length == argCount) {
            TypeBinding[] toMatch = method.parameters;
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                return method;
              }
            }
          }
        }
      }
    } else { if ((this.tagBits & 0x4000) == 0L) {
        int length = this.methods.length;
        if (length > 1)
          ReferenceBinding.sortMethods(this.methods, 0, length);
        this.tagBits |= 16384L;
      }
      long range;
      if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L)
      {
        int start = (int)range; int end = (int)(range >> 32);
        for (int imethod = start; imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          if ((resolveTypesFor(method) == null) || (method.returnType == null)) {
            methods();
            return getExactMethod(selector, argumentTypes, refScope);
          }
        }

        boolean isSource15 = this.scope.compilerOptions().sourceLevel >= 3211264L;
        for (int i = start; i <= end; i++) {
          MethodBinding method1 = this.methods[i];
          for (int j = end; j > i; j--) {
            MethodBinding method2 = this.methods[j];
            boolean paramsMatch = isSource15 ? 
              method1.areParameterErasuresEqual(method2) : 
              method1.areParametersEqual(method2);
            if (paramsMatch) {
              methods();
              return getExactMethod(selector, argumentTypes, refScope);
            }
          }
        }
        for (int imethod = start; imethod <= end; imethod++) {
          MethodBinding method = this.methods[imethod];
          TypeBinding[] toMatch = method.parameters;
          if (toMatch.length == argCount) {
            int iarg = 0;
            while (toMatch[iarg] == argumentTypes[iarg])
            {
              iarg++; if (iarg >= argCount)
              {
                return method;
              }
            }
          }
        }
      } }
    if (foundNothing) {
      if (isInterface()) {
        if (this.superInterfaces.length == 1) {
          if (refScope != null)
            refScope.recordTypeReference(this.superInterfaces[0]);
          return this.superInterfaces[0].getExactMethod(selector, argumentTypes, refScope);
        }
      } else if (this.superclass != null) {
        if (refScope != null)
          refScope.recordTypeReference(this.superclass);
        return this.superclass.getExactMethod(selector, argumentTypes, refScope);
      }
    }
    return null;
  }

  public FieldBinding getField(char[] fieldName, boolean needResolve)
  {
    if ((this.tagBits & 0x2000) != 0L) {
      return ReferenceBinding.binarySearch(fieldName, this.fields);
    }

    if ((this.tagBits & 0x1000) == 0L) {
      int length = this.fields.length;
      if (length > 1)
        ReferenceBinding.sortFields(this.fields, 0, length);
      this.tagBits |= 4096L;
    }

    FieldBinding field = ReferenceBinding.binarySearch(fieldName, this.fields);
    if (field != null) {
      FieldBinding result = null;
      try {
        result = resolveTypeFor(field);
        int newSize;
        FieldBinding[] newFields;
        int index;
        int i;
        int length;
        FieldBinding f;
        return result;
      } finally {
        if (result == null)
        {
          int newSize = this.fields.length - 1;
          if (newSize == 0) {
            this.fields = Binding.NO_FIELDS;
          } else {
            FieldBinding[] newFields = new FieldBinding[newSize];
            int index = 0;
            int i = 0; for (int length = this.fields.length; i < length; i++) {
              FieldBinding f = this.fields[i];
              if (f != field)
                newFields[(index++)] = f;
            }
            this.fields = newFields;
          }
        }
      }
    }
    return null;
  }

  public MethodBinding[] getMethods(char[] selector)
  {
    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
        int start = (int)range; int end = (int)(range >> 32);
        int length = end - start + 1;
        MethodBinding[] result;
        System.arraycopy(this.methods, start, result = new MethodBinding[length], 0, length);
        return result;
      }
      return Binding.NO_METHODS;
    }

    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }
    long range;
    if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
      int start = (int)range; int end = (int)(range >> 32);
      for (int i = start; i <= end; i++) {
        MethodBinding method = this.methods[i];
        if ((resolveTypesFor(method) == null) || (method.returnType == null)) {
          methods();
          return getMethods(selector);
        }
      }
      int length = end - start + 1;
      MethodBinding[] result;
      System.arraycopy(this.methods, start, result = new MethodBinding[length], 0, length);
    } else {
      return Binding.NO_METHODS;
    }
    MethodBinding[] result;
    boolean isSource15 = this.scope.compilerOptions().sourceLevel >= 3211264L;
    int i = 0; for (int length = result.length - 1; i < length; i++) {
      MethodBinding method = result[i];
      for (int j = length; j > i; j--) {
        boolean paramsMatch = isSource15 ? 
          method.areParameterErasuresEqual(result[j]) : 
          method.areParametersEqual(result[j]);
        if (paramsMatch) {
          methods();
          return getMethods(selector);
        }
      }
    }
    return result;
  }

  public FieldBinding getSyntheticField(LocalVariableBinding actualOuterLocalVariable)
  {
    if ((this.synthetics == null) || (this.synthetics[1] == null)) return null;
    return (FieldBinding)this.synthetics[1].get(actualOuterLocalVariable);
  }

  public FieldBinding getSyntheticField(ReferenceBinding targetEnclosingType, boolean onlyExactMatch)
  {
    if ((this.synthetics == null) || (this.synthetics[1] == null)) return null;
    FieldBinding field = (FieldBinding)this.synthetics[1].get(targetEnclosingType);
    if (field != null) return field;

    if (!onlyExactMatch) {
      Iterator accessFields = this.synthetics[1].values().iterator();
      while (accessFields.hasNext()) {
        field = (FieldBinding)accessFields.next();
        if ((CharOperation.prefixEquals(TypeConstants.SYNTHETIC_ENCLOSING_INSTANCE_PREFIX, field.name)) && 
          (field.type.findSuperTypeWithSameErasure(targetEnclosingType) != null))
          return field;
      }
    }
    return null;
  }

  public SyntheticMethodBinding getSyntheticBridgeMethod(MethodBinding inheritedMethodToBridge)
  {
    if (this.synthetics == null) return null;
    if (this.synthetics[0] == null) return null;
    SyntheticMethodBinding[] accessors = (SyntheticMethodBinding[])this.synthetics[0].get(inheritedMethodToBridge);
    if (accessors == null) return null;
    return accessors[1];
  }

  public void initializeDeprecatedAnnotationTagBits()
  {
    if ((this.tagBits & 0x0) == 0L) {
      TypeDeclaration typeDecl = this.scope.referenceContext;
      boolean old = typeDecl.staticInitializerScope.insideTypeAnnotation;
      try {
        typeDecl.staticInitializerScope.insideTypeAnnotation = true;
        ASTNode.resolveDeprecatedAnnotations(typeDecl.staticInitializerScope, typeDecl.annotations, this);
        this.tagBits |= 17179869184L;
      } finally {
        typeDecl.staticInitializerScope.insideTypeAnnotation = old;
      }
      if ((this.tagBits & 0x0) != 0L)
        this.modifiers |= 1048576;
    }
  }

  public boolean isEquivalentTo(TypeBinding otherType)
  {
    if (this == otherType) return true;
    if (otherType == null) return false;
    switch (otherType.kind())
    {
    case 516:
      return ((WildcardBinding)otherType).boundCheck(this);
    case 260:
      if (((otherType.tagBits & 0x40000000) == 0L) && ((!isMemberType()) || (!otherType.isMemberType())))
        return false;
      ParameterizedTypeBinding otherParamType = (ParameterizedTypeBinding)otherType;
      if (this != otherParamType.genericType())
        return false;
      if (!isStatic()) {
        ReferenceBinding enclosing = enclosingType();
        if (enclosing != null) {
          ReferenceBinding otherEnclosing = otherParamType.enclosingType();
          if (otherEnclosing == null) return false;
          if ((otherEnclosing.tagBits & 0x40000000) == 0L) {
            if (enclosing != otherEnclosing) return false;
          }
          else if (!enclosing.isEquivalentTo(otherParamType.enclosingType())) return false;
        }
      }

      int length = this.typeVariables == null ? 0 : this.typeVariables.length;
      TypeBinding[] otherArguments = otherParamType.arguments;
      int otherLength = otherArguments == null ? 0 : otherArguments.length;
      if (otherLength != length)
        return false;
      for (int i = 0; i < length; i++)
        if (!this.typeVariables[i].isTypeArgumentContainedBy(otherArguments[i]))
          return false;
      return true;
    case 1028:
      return otherType.erasure() == this;
    }
    return false;
  }
  public boolean isGenericType() {
    return this.typeVariables != Binding.NO_TYPE_VARIABLES;
  }
  public ReferenceBinding[] memberTypes() {
    return this.memberTypes;
  }
  public FieldBinding getUpdatedFieldBinding(FieldBinding targetField, ReferenceBinding newDeclaringClass) {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[3] == null) {
      this.synthetics[3] = new HashMap(5);
    }
    Hashtable fieldMap = (Hashtable)this.synthetics[3].get(targetField);
    if (fieldMap == null) {
      fieldMap = new Hashtable(5);
      this.synthetics[3].put(targetField, fieldMap);
    }
    FieldBinding updatedField = (FieldBinding)fieldMap.get(newDeclaringClass);
    if (updatedField == null) {
      updatedField = new FieldBinding(targetField, newDeclaringClass);
      fieldMap.put(newDeclaringClass, updatedField);
    }
    return updatedField;
  }
  public MethodBinding getUpdatedMethodBinding(MethodBinding targetMethod, ReferenceBinding newDeclaringClass) {
    if (this.synthetics == null)
      this.synthetics = new HashMap[4];
    if (this.synthetics[3] == null) {
      this.synthetics[3] = new HashMap(5);
    }
    Hashtable methodMap = (Hashtable)this.synthetics[3].get(targetMethod);
    if (methodMap == null) {
      methodMap = new Hashtable(5);
      this.synthetics[3].put(targetMethod, methodMap);
    }
    MethodBinding updatedMethod = (MethodBinding)methodMap.get(newDeclaringClass);
    if (updatedMethod == null) {
      updatedMethod = new MethodBinding(targetMethod, newDeclaringClass);
      methodMap.put(newDeclaringClass, updatedMethod);
    }
    return updatedMethod;
  }
  public boolean hasMemberTypes() {
    return this.memberTypes.length > 0;
  }

  public MethodBinding[] methods() {
    if ((this.tagBits & 0x8000) != 0L) {
      return this.methods;
    }

    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L; } 
int failed = 0;
    MethodBinding[] resolvedMethods = this.methods;
    int newSize;
    MethodBinding[] newMethods;
    int i;
    int j;
    int length;
    try { int i = 0; for (int length = this.methods.length; i < length; i++) {
        if (resolveTypesFor(this.methods[i]) == null)
        {
          if (resolvedMethods == this.methods) {
            System.arraycopy(this.methods, 0, resolvedMethods = new MethodBinding[length], 0, length);
          }
          resolvedMethods[i] = null;
          failed++;
        }

      }

      boolean complyTo15 = this.scope.compilerOptions().sourceLevel >= 3211264L;
      int i = 0; for (int length = this.methods.length; i < length; i++) {
        MethodBinding method = resolvedMethods[i];
        if (method != null)
        {
          char[] selector = method.selector;
          AbstractMethodDeclaration methodDecl = null;
          for (int j = i + 1; j < length; j++) {
            MethodBinding method2 = resolvedMethods[j];
            if (method2 != null)
            {
              if (!CharOperation.equals(selector, method2.selector))
                break;
              TypeBinding[] params1;
              TypeBinding[] params2;
              int pLength;
              if ((complyTo15) && (method.returnType != null) && (method2.returnType != null))
              {
                params1 = method.parameters;
                params2 = method2.parameters;
                pLength = params1.length;
                if (pLength != params2.length) {
                  continue;
                }
                TypeVariableBinding[] vars = method.typeVariables;
                TypeVariableBinding[] vars2 = method2.typeVariables;
                boolean equalTypeVars = vars == vars2;
                MethodBinding subMethod = method2;
                if (!equalTypeVars) {
                  MethodBinding temp = method.computeSubstitutedMethod(method2, this.scope.environment());
                  if (temp != null) {
                    equalTypeVars = true;
                    subMethod = temp;
                  }
                }
                boolean equalParams = method.areParametersEqual(subMethod);
                if ((!equalParams) || (!equalTypeVars))
                {
                  if ((method.returnType.erasure() != subMethod.returnType.erasure()) || ((!equalParams) && (!method.areParameterErasuresEqual(method2))))
                  {
                    if ((!equalTypeVars) && (vars != Binding.NO_TYPE_VARIABLES) && (vars2 != Binding.NO_TYPE_VARIABLES)) {
                      continue;
                    }
                    if (pLength > 0)
                    {
                      int index = pLength;
                      do {
                        if (params1[index] != params2[index].erasure())
                          break;
                        if (params1[index] == params2[index]) {
                          TypeBinding type = params1[index].leafComponentType();
                          if (((type instanceof SourceTypeBinding)) && (type.typeVariables() != Binding.NO_TYPE_VARIABLES)) {
                            index = pLength;
                            break;
                          }
                        }
                        index--; } while (index >= 0);

                      if ((index >= 0) && (index < pLength)) {
                        index = pLength;
                        do { if (params1[index].erasure() != params2[index])
                            break;
                          index--; } while (index >= 0);
                      }

                      if (index >= 0)
                        continue;  } 
                  }
                } } else { if (!method.areParametersEqual(method2))
                  continue;
              }
              boolean isEnumSpecialMethod = (isEnum()) && ((CharOperation.equals(selector, TypeConstants.VALUEOF)) || (CharOperation.equals(selector, TypeConstants.VALUES)));

              if (methodDecl == null) {
                methodDecl = method.sourceMethod();
                if ((methodDecl != null) && (methodDecl.binding != null)) {
                  if (isEnumSpecialMethod)
                    this.scope.problemReporter().duplicateEnumSpecialMethod(this, methodDecl);
                  else {
                    this.scope.problemReporter().duplicateMethodInType(this, methodDecl);
                  }
                  methodDecl.binding = null;

                  if (resolvedMethods == this.methods) {
                    System.arraycopy(this.methods, 0, resolvedMethods = new MethodBinding[length], 0, length);
                  }
                  resolvedMethods[i] = null;
                  failed++;
                }
              }
              AbstractMethodDeclaration method2Decl = method2.sourceMethod();
              if ((method2Decl != null) && (method2Decl.binding != null)) {
                if (isEnumSpecialMethod)
                  this.scope.problemReporter().duplicateEnumSpecialMethod(this, method2Decl);
                else {
                  this.scope.problemReporter().duplicateMethodInType(this, method2Decl);
                }
                method2Decl.binding = null;

                if (resolvedMethods == this.methods) {
                  System.arraycopy(this.methods, 0, resolvedMethods = new MethodBinding[length], 0, length);
                }
                resolvedMethods[j] = null;
                failed++;
              }
            }
          }
          if ((method.returnType == null) && (methodDecl == null)) {
            methodDecl = method.sourceMethod();
            if (methodDecl != null) {
              methodDecl.binding = null;
            }

            if (resolvedMethods == this.methods) {
              System.arraycopy(this.methods, 0, resolvedMethods = new MethodBinding[length], 0, length);
            }
            resolvedMethods[i] = null;
            failed++;
          }
        }
      } } finally {
      if (failed > 0) {
        int newSize = resolvedMethods.length - failed;
        if (newSize == 0) {
          this.methods = Binding.NO_METHODS;
        } else {
          MethodBinding[] newMethods = new MethodBinding[newSize];
          int i = 0; int j = 0; for (int length = resolvedMethods.length; i < length; i++)
            if (resolvedMethods[i] != null)
              newMethods[(j++)] = resolvedMethods[i];
          this.methods = newMethods;
        }

      }

      addDefaultAbstractMethods();
      this.tagBits |= 32768L;
    }
    return this.methods;
  }
  private FieldBinding resolveTypeFor(FieldBinding field) {
    if ((field.modifiers & 0x2000000) == 0) {
      return field;
    }
    if ((this.scope.compilerOptions().sourceLevel >= 3211264L) && 
      ((field.getAnnotationTagBits() & 0x0) != 0L)) {
      field.modifiers |= 1048576;
    }
    if ((isViewedAsDeprecated()) && (!field.isDeprecated()))
      field.modifiers |= 2097152;
    if (hasRestrictedAccess())
      field.modifiers |= 262144;
    FieldDeclaration[] fieldDecls = this.scope.referenceContext.fields;
    int f = 0; for (int length = fieldDecls.length; f < length; f++)
      if (fieldDecls[f].binding == field)
      {
        MethodScope initializationScope = field.isStatic() ? 
          this.scope.referenceContext.staticInitializerScope : 
          this.scope.referenceContext.initializerScope;
        FieldBinding previousField = initializationScope.initializedField;
        try {
          initializationScope.initializedField = field;
          FieldDeclaration fieldDecl = fieldDecls[f];
          TypeBinding fieldType = 
            fieldDecl.getKind() == 3 ? 
            initializationScope.environment().convertToRawType(this) : 
            fieldDecl.type.resolveType(initializationScope, true);
          field.type = fieldType;
          field.modifiers &= -33554433;
          if (fieldType == null)
            fieldDecl.binding = null; while (true) {
            return null;

            if (fieldType == TypeBinding.VOID) {
              this.scope.problemReporter().variableTypeCannotBeVoid(fieldDecl);
              fieldDecl.binding = null;
            }
            else {
              if ((!fieldType.isArrayType()) || (((ArrayBinding)fieldType).leafComponentType != TypeBinding.VOID)) break;
              this.scope.problemReporter().variableTypeCannotBeVoidArray(fieldDecl);
              fieldDecl.binding = null;
            }
          }
          TypeBinding leafType = fieldType.leafComponentType();
          if (((leafType instanceof ReferenceBinding)) && ((((ReferenceBinding)leafType).modifiers & 0x40000000) != 0))
            field.modifiers |= 1073741824;
        }
        finally {
          initializationScope.initializedField = previousField; } initializationScope.initializedField = previousField;

        return field;
      }
    return null;
  }
  public MethodBinding resolveTypesFor(MethodBinding method) {
    if ((method.modifiers & 0x2000000) == 0) {
      return method;
    }
    if ((this.scope.compilerOptions().sourceLevel >= 3211264L) && 
      ((method.getAnnotationTagBits() & 0x0) != 0L)) {
      method.modifiers |= 1048576;
    }
    if ((isViewedAsDeprecated()) && (!method.isDeprecated()))
      method.modifiers |= 2097152;
    if (hasRestrictedAccess()) {
      method.modifiers |= 262144;
    }
    AbstractMethodDeclaration methodDecl = method.sourceMethod();
    if (methodDecl == null) return null;

    TypeParameter[] typeParameters = methodDecl.typeParameters();
    if (typeParameters != null) {
      methodDecl.scope.connectTypeVariables(typeParameters, true);

      int i = 0; for (int paramLength = typeParameters.length; i < paramLength; i++)
        typeParameters[i].checkBounds(methodDecl.scope);
    }
    TypeReference[] exceptionTypes = methodDecl.thrownExceptions;
    if (exceptionTypes != null) {
      int size = exceptionTypes.length;
      method.thrownExceptions = new ReferenceBinding[size];
      int count = 0;

      for (int i = 0; i < size; i++) {
        ReferenceBinding resolvedExceptionType = (ReferenceBinding)exceptionTypes[i].resolveType(methodDecl.scope, true);
        if (resolvedExceptionType != null)
        {
          if (resolvedExceptionType.isBoundParameterizedType()) {
            methodDecl.scope.problemReporter().invalidParameterizedExceptionType(resolvedExceptionType, exceptionTypes[i]);
          }
          else if (resolvedExceptionType.findSuperTypeErasingTo(21, true) == null) {
            methodDecl.scope.problemReporter().cannotThrowType(exceptionTypes[i], resolvedExceptionType);
          }
          else {
            if ((resolvedExceptionType.modifiers & 0x40000000) != 0)
              method.modifiers |= 1073741824;
            method.thrownExceptions[(count++)] = resolvedExceptionType;
          }
        }
      }
      if (count < size) {
        System.arraycopy(method.thrownExceptions, 0, method.thrownExceptions = new ReferenceBinding[count], 0, count);
      }
    }
    boolean foundArgProblem = false;
    Argument[] arguments = methodDecl.arguments;
    if (arguments != null) {
      int size = arguments.length;
      method.parameters = Binding.NO_PARAMETERS;
      TypeBinding[] newParameters = new TypeBinding[size];
      for (int i = 0; i < size; i++) {
        Argument arg = arguments[i];
        TypeBinding parameterType = arg.type.resolveType(methodDecl.scope, true);
        if (parameterType == null) {
          foundArgProblem = true;
        } else if (parameterType == TypeBinding.VOID) {
          methodDecl.scope.problemReporter().argumentTypeCannotBeVoid(this, methodDecl, arg);
          foundArgProblem = true;
        } else {
          TypeBinding leafType = parameterType.leafComponentType();
          if (((leafType instanceof ReferenceBinding)) && ((((ReferenceBinding)leafType).modifiers & 0x40000000) != 0))
            method.modifiers |= 1073741824;
          newParameters[i] = parameterType;
          arg.binding = new LocalVariableBinding(arg, parameterType, arg.modifiers, true);
        }
      }

      if (!foundArgProblem) {
        method.parameters = newParameters;
      }
    }

    boolean foundReturnTypeProblem = false;
    if (!method.isConstructor()) {
      TypeReference returnType = (methodDecl instanceof MethodDeclaration) ? 
        ((MethodDeclaration)methodDecl).returnType : 
        null;
      if (returnType == null) {
        methodDecl.scope.problemReporter().missingReturnType(methodDecl);
        method.returnType = null;
        foundReturnTypeProblem = true;
      } else {
        TypeBinding methodType = returnType.resolveType(methodDecl.scope, true);
        if (methodType == null) {
          foundReturnTypeProblem = true;
        } else if ((methodType.isArrayType()) && (((ArrayBinding)methodType).leafComponentType == TypeBinding.VOID)) {
          methodDecl.scope.problemReporter().returnTypeCannotBeVoidArray((MethodDeclaration)methodDecl);
          foundReturnTypeProblem = true;
        } else {
          method.returnType = methodType;
          TypeBinding leafType = methodType.leafComponentType();
          if (((leafType instanceof ReferenceBinding)) && ((((ReferenceBinding)leafType).modifiers & 0x40000000) != 0))
            method.modifiers |= 1073741824;
        }
      }
    }
    if (foundArgProblem) {
      methodDecl.binding = null;
      method.parameters = Binding.NO_PARAMETERS;

      if (typeParameters != null) {
        int i = 0; for (int length = typeParameters.length; i < length; i++)
          typeParameters[i].binding = null; 
      }
      return null;
    }
    if (foundReturnTypeProblem) {
      return method;
    }
    method.modifiers &= -33554433;
    return method;
  }
  public AnnotationHolder retrieveAnnotationHolder(Binding binding, boolean forceInitialization) {
    if (forceInitialization)
      binding.getAnnotationTagBits();
    return super.retrieveAnnotationHolder(binding, false);
  }
  public void setFields(FieldBinding[] fields) {
    this.fields = fields;
  }
  public void setMethods(MethodBinding[] methods) {
    this.methods = methods;
  }
  public final int sourceEnd() {
    return this.scope.referenceContext.sourceEnd;
  }
  public final int sourceStart() {
    return this.scope.referenceContext.sourceStart;
  }
  SimpleLookupTable storedAnnotations(boolean forceInitialize) {
    if ((forceInitialize) && (this.storedAnnotations == null) && (this.scope != null)) {
      this.scope.referenceCompilationUnit().compilationResult.hasAnnotations = true;
      if (!this.scope.environment().globalOptions.storeAnnotations)
        return null;
      this.storedAnnotations = new SimpleLookupTable(3);
    }
    return this.storedAnnotations;
  }
  public ReferenceBinding superclass() {
    return this.superclass;
  }
  public ReferenceBinding[] superInterfaces() {
    return this.superInterfaces;
  }

  public SyntheticMethodBinding[] syntheticMethods()
  {
    if ((this.synthetics == null) || (this.synthetics[0] == null) || (this.synthetics[0].size() == 0)) return null;

    int index = 0;
    SyntheticMethodBinding[] bindings = new SyntheticMethodBinding[1];
    Iterator fieldsOrMethods = this.synthetics[0].keySet().iterator();
    while (fieldsOrMethods.hasNext())
    {
      Object fieldOrMethod = fieldsOrMethods.next();

      if ((fieldOrMethod instanceof MethodBinding))
      {
        SyntheticMethodBinding[] methodAccessors = (SyntheticMethodBinding[])this.synthetics[0].get(fieldOrMethod);
        int numberOfAccessors = 0;
        if (methodAccessors[0] != null) numberOfAccessors++;
        if (methodAccessors[1] != null) numberOfAccessors++;
        if (index + numberOfAccessors > bindings.length)
          System.arraycopy(bindings, 0, bindings = new SyntheticMethodBinding[index + numberOfAccessors], 0, index);
        if (methodAccessors[0] != null)
          bindings[(index++)] = methodAccessors[0];
        if (methodAccessors[1] != null)
          bindings[(index++)] = methodAccessors[1];
      }
      else
      {
        SyntheticMethodBinding[] fieldAccessors = (SyntheticMethodBinding[])this.synthetics[0].get(fieldOrMethod);
        int numberOfAccessors = 0;
        if (fieldAccessors[0] != null) numberOfAccessors++;
        if (fieldAccessors[1] != null) numberOfAccessors++;
        if (index + numberOfAccessors > bindings.length)
          System.arraycopy(bindings, 0, bindings = new SyntheticMethodBinding[index + numberOfAccessors], 0, index);
        if (fieldAccessors[0] != null)
          bindings[(index++)] = fieldAccessors[0];
        if (fieldAccessors[1] != null)
          bindings[(index++)] = fieldAccessors[1];
      }
    }
    int length;
    SyntheticMethodBinding[] sortedBindings = new SyntheticMethodBinding[length = bindings.length];
    for (int i = 0; i < length; i++) {
      SyntheticMethodBinding binding = bindings[i];
      sortedBindings[binding.index] = binding;
    }
    return sortedBindings;
  }

  public FieldBinding[] syntheticFields()
  {
    if (this.synthetics == null) return null;

    int fieldSize = this.synthetics[1] == null ? 0 : this.synthetics[1].size();
    int literalSize = this.synthetics[2] == null ? 0 : this.synthetics[2].size();
    int totalSize = fieldSize + literalSize;
    if (totalSize == 0) return null;
    FieldBinding[] bindings = new FieldBinding[totalSize];

    if (this.synthetics[1] != null) {
      Iterator elements = this.synthetics[1].values().iterator();
      for (int i = 0; i < fieldSize; i++) {
        SyntheticFieldBinding synthBinding = (SyntheticFieldBinding)elements.next();
        bindings[synthBinding.index] = synthBinding;
      }
    }

    if (this.synthetics[2] != null) {
      Iterator elements = this.synthetics[2].values().iterator();
      for (int i = 0; i < literalSize; i++) {
        SyntheticFieldBinding synthBinding = (SyntheticFieldBinding)elements.next();
        bindings[(fieldSize + synthBinding.index)] = synthBinding;
      }
    }
    return bindings;
  }
  public String toString() {
    StringBuffer buffer = new StringBuffer(30);
    buffer.append("(id=");
    if (this.id == 2147483647)
      buffer.append("NoId");
    else
      buffer.append(this.id);
    buffer.append(")\n");
    if (isDeprecated()) buffer.append("deprecated ");
    if (isPublic()) buffer.append("public ");
    if (isProtected()) buffer.append("protected ");
    if (isPrivate()) buffer.append("private ");
    if ((isAbstract()) && (isClass())) buffer.append("abstract ");
    if ((isStatic()) && (isNestedType())) buffer.append("static ");
    if (isFinal()) buffer.append("final ");

    if (isEnum()) buffer.append("enum ");
    else if (isAnnotationType()) buffer.append("@interface ");
    else if (isClass()) buffer.append("class "); else
      buffer.append("interface ");
    buffer.append(this.compoundName != null ? CharOperation.toString(this.compoundName) : "UNNAMED TYPE");

    if (this.typeVariables == null) {
      buffer.append("<NULL TYPE VARIABLES>");
    } else if (this.typeVariables != Binding.NO_TYPE_VARIABLES) {
      buffer.append("\n\t<");
      int i = 0; for (int length = this.typeVariables.length; i < length; i++) {
        if (i > 0)
          buffer.append(", ");
        buffer.append(this.typeVariables[i] != null ? this.typeVariables[i].toString() : "NULL TYPE VARIABLE");
      }
      buffer.append(">");
    }
    buffer.append("\n\textends ");
    buffer.append(this.superclass != null ? this.superclass.debugName() : "NULL TYPE");

    if (this.superInterfaces != null) {
      if (this.superInterfaces != Binding.NO_SUPERINTERFACES) {
        buffer.append("\n\timplements : ");
        int i = 0; for (int length = this.superInterfaces.length; i < length; i++) {
          if (i > 0)
            buffer.append(", ");
          buffer.append(this.superInterfaces[i] != null ? this.superInterfaces[i].debugName() : "NULL TYPE");
        }
      }
    }
    else buffer.append("NULL SUPERINTERFACES");

    if (enclosingType() != null) {
      buffer.append("\n\tenclosing type : ");
      buffer.append(enclosingType().debugName());
    }

    if (this.fields != null) {
      if (this.fields != Binding.NO_FIELDS) {
        buffer.append("\n/*   fields   */");
        int i = 0; for (int length = this.fields.length; i < length; i++)
          buffer.append('\n').append(this.fields[i] != null ? this.fields[i].toString() : "NULL FIELD");
      }
    }
    else buffer.append("NULL FIELDS");

    if (this.methods != null) {
      if (this.methods != Binding.NO_METHODS) {
        buffer.append("\n/*   methods   */");
        int i = 0; for (int length = this.methods.length; i < length; i++)
          buffer.append('\n').append(this.methods[i] != null ? this.methods[i].toString() : "NULL METHOD");
      }
    }
    else buffer.append("NULL METHODS");

    if (this.memberTypes != null) {
      if (this.memberTypes != Binding.NO_MEMBER_TYPES) {
        buffer.append("\n/*   members   */");
        int i = 0; for (int length = this.memberTypes.length; i < length; i++)
          buffer.append('\n').append(this.memberTypes[i] != null ? this.memberTypes[i].toString() : "NULL TYPE");
      }
    }
    else buffer.append("NULL MEMBER TYPES");

    buffer.append("\n\n");
    return buffer.toString();
  }
  public TypeVariableBinding[] typeVariables() {
    return this.typeVariables;
  }
  void verifyMethods(MethodVerifier verifier) {
    verifier.verify(this);

    int i = this.memberTypes.length;
    do { ((SourceTypeBinding)this.memberTypes[i]).verifyMethods(verifier);

      i--; } while (i >= 0);
  }
}