package org.eclipse.jdt.internal.compiler.parser;

import java.util.List;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.core.compiler.InvalidInputException;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Javadoc;
import org.eclipse.jdt.internal.compiler.ast.JavadocAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.JavadocArgumentExpression;
import org.eclipse.jdt.internal.compiler.ast.JavadocArrayQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocArraySingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocFieldReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocImplicitTypeReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocMessageSend;
import org.eclipse.jdt.internal.compiler.ast.JavadocQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocReturnStatement;
import org.eclipse.jdt.internal.compiler.ast.JavadocSingleNameReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocSingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.Util;

public class JavadocParser extends AbstractCommentParser
{
  public Javadoc docComment;
  private int invalidParamReferencesPtr = -1;
  private ASTNode[] invalidParamReferencesStack;
  private long validValuePositions;
  private long invalidValuePositions;

  public JavadocParser(Parser sourceParser)
  {
    super(sourceParser);
    this.kind = 513;
  }

  public boolean checkDeprecation(int commentPtr)
  {
    this.javadocStart = this.sourceParser.scanner.commentStarts[commentPtr];
    this.javadocEnd = (this.sourceParser.scanner.commentStops[commentPtr] - 1);
    this.firstTagPosition = this.sourceParser.scanner.commentTagStarts[commentPtr];
    this.validValuePositions = -1L;
    this.invalidValuePositions = -1L;

    if (this.checkDocComment)
      this.docComment = new Javadoc(this.javadocStart, this.javadocEnd);
    else {
      this.docComment = null;
    }

    if (this.firstTagPosition == 0)
      switch (this.kind & 0xFF) {
      case 1:
      case 5:
        return false;
      case 2:
      case 3:
      case 4:
      }
    try {
      this.source = this.sourceParser.scanner.source;
      if (this.checkDocComment)
      {
        this.scanner.lineEnds = this.sourceParser.scanner.lineEnds;
        this.scanner.linePtr = this.sourceParser.scanner.linePtr;
        this.lineEnds = this.scanner.lineEnds;
        commentParse();
      }
      else
      {
        Scanner sourceScanner = this.sourceParser.scanner;
        int firstLineNumber = Util.getLineNumber(this.javadocStart, sourceScanner.lineEnds, 0, sourceScanner.linePtr);
        int lastLineNumber = Util.getLineNumber(this.javadocEnd, sourceScanner.lineEnds, 0, sourceScanner.linePtr);
        this.index = (this.javadocStart + 3);

        this.deprecated = false;
        label471: for (int line = firstLineNumber; line <= lastLineNumber; line++) {
          int lineStart = line == firstLineNumber ? 
            this.javadocStart + 3 : 
            this.sourceParser.scanner.getLineStart(line);
          this.index = lineStart;
          this.lineEnd = (line == lastLineNumber ? 
            this.javadocEnd - 2 : 
            this.sourceParser.scanner.getLineEnd(line));
          while (this.index < this.lineEnd) {
            char c = readChar();
            switch (c)
            {
            case '\t':
            case '\n':
            case '\f':
            case '\r':
            case ' ':
            case '*':
              break;
            case '@':
              parseSimpleTag();
              if ((this.tagValue != 1) || 
                (!this.abort)) break label471; break;
            default:
              break label471;
            }
          }
        }
        boolean bool = this.deprecated; return bool;
      }
    } finally {
      this.source = null;
    }
    return this.deprecated;
  }

  protected Object createArgumentReference(char[] name, int dim, boolean isVarargs, Object typeRef, long[] dimPositions, long argNamePos)
    throws InvalidInputException
  {
    try
    {
      TypeReference argTypeRef = (TypeReference)typeRef;
      if (dim > 0) {
        long pos = (argTypeRef.sourceStart << 32) + argTypeRef.sourceEnd;
        if ((typeRef instanceof JavadocSingleTypeReference)) {
          JavadocSingleTypeReference singleRef = (JavadocSingleTypeReference)typeRef;
          argTypeRef = new JavadocArraySingleTypeReference(singleRef.token, dim, pos);
        } else {
          JavadocQualifiedTypeReference qualifRef = (JavadocQualifiedTypeReference)typeRef;
          argTypeRef = new JavadocArrayQualifiedTypeReference(qualifRef, dim);
        }
      }
      int argEnd = argTypeRef.sourceEnd;
      if (dim > 0) {
        argEnd = (int)dimPositions[(dim - 1)];
        if (isVarargs) {
          argTypeRef.bits |= 16384;
        }
      }
      if (argNamePos >= 0L) argEnd = (int)argNamePos;
      return new JavadocArgumentExpression(name, argTypeRef.sourceStart, argEnd, argTypeRef);
    } catch (ClassCastException localClassCastException) {
    }
    throw new InvalidInputException();
  }

  protected Object createFieldReference(Object receiver)
    throws InvalidInputException
  {
    try
    {
      TypeReference typeRef = (TypeReference)receiver;
      if (typeRef == null) {
        char[] name = this.sourceParser.compilationUnit.getMainTypeName();
        typeRef = new JavadocImplicitTypeReference(name, this.memberStart);
      }

      JavadocFieldReference field = new JavadocFieldReference(this.identifierStack[0], this.identifierPositionStack[0]);
      field.receiver = typeRef;
      field.tagSourceStart = this.tagSourceStart;
      field.tagSourceEnd = this.tagSourceEnd;
      field.tagValue = this.tagValue;
      return field;
    } catch (ClassCastException localClassCastException) {
    }
    throw new InvalidInputException();
  }

  protected Object createMethodReference(Object receiver, List arguments)
    throws InvalidInputException
  {
    try
    {
      TypeReference typeRef = (TypeReference)receiver;

      boolean isConstructor = false;
      int length = this.identifierLengthStack[0];
      if (typeRef == null) {
        char[] name = this.sourceParser.compilationUnit.getMainTypeName();
        TypeDeclaration typeDecl = getParsedTypeDeclaration();
        if (typeDecl != null) {
          name = typeDecl.name;
        }
        isConstructor = CharOperation.equals(this.identifierStack[(length - 1)], name);
        typeRef = new JavadocImplicitTypeReference(name, this.memberStart);
      }
      else if ((typeRef instanceof JavadocSingleTypeReference)) {
        char[] name = ((JavadocSingleTypeReference)typeRef).token;
        isConstructor = CharOperation.equals(this.identifierStack[(length - 1)], name);
      } else if ((typeRef instanceof JavadocQualifiedTypeReference)) {
        char[][] tokens = ((JavadocQualifiedTypeReference)typeRef).tokens;
        int last = tokens.length - 1;
        isConstructor = CharOperation.equals(this.identifierStack[(length - 1)], tokens[last]);
        if (isConstructor) {
          boolean valid = true;
          if (valid) {
            for (int i = 0; (i < length - 1) && (valid); i++) {
              valid = CharOperation.equals(this.identifierStack[i], tokens[i]);
            }
          }
          if (!valid) {
            if (this.reportProblems) {
              this.sourceParser.problemReporter().javadocInvalidMemberTypeQualification((int)(this.identifierPositionStack[0] >>> 32), (int)this.identifierPositionStack[(length - 1)], -1);
            }
            return null;
          }
        }
      } else {
        throw new InvalidInputException();
      }

      if (arguments == null) {
        if (isConstructor) {
          JavadocAllocationExpression allocation = new JavadocAllocationExpression(this.identifierPositionStack[(length - 1)]);
          allocation.type = typeRef;
          allocation.tagValue = this.tagValue;
          allocation.sourceEnd = this.scanner.getCurrentTokenEndPosition();
          if (length == 1) {
            allocation.qualification = new char[][] { this.identifierStack[0] };
          } else {
            System.arraycopy(this.identifierStack, 0, allocation.qualification = new char[length][], 0, length);
            allocation.sourceStart = ((int)(this.identifierPositionStack[0] >>> 32));
          }
          allocation.memberStart = this.memberStart;
          return allocation;
        }
        JavadocMessageSend msg = new JavadocMessageSend(this.identifierStack[(length - 1)], this.identifierPositionStack[(length - 1)]);
        msg.receiver = typeRef;
        msg.tagValue = this.tagValue;
        msg.sourceEnd = this.scanner.getCurrentTokenEndPosition();
        return msg;
      }

      JavadocArgumentExpression[] expressions = new JavadocArgumentExpression[arguments.size()];
      arguments.toArray(expressions);
      if (isConstructor) {
        JavadocAllocationExpression allocation = new JavadocAllocationExpression(this.identifierPositionStack[(length - 1)]);
        allocation.arguments = expressions;
        allocation.type = typeRef;
        allocation.tagValue = this.tagValue;
        allocation.sourceEnd = this.scanner.getCurrentTokenEndPosition();
        if (length == 1) {
          allocation.qualification = new char[][] { this.identifierStack[0] };
        } else {
          System.arraycopy(this.identifierStack, 0, allocation.qualification = new char[length][], 0, length);
          allocation.sourceStart = ((int)(this.identifierPositionStack[0] >>> 32));
        }
        allocation.memberStart = this.memberStart;
        return allocation;
      }
      JavadocMessageSend msg = new JavadocMessageSend(this.identifierStack[(length - 1)], this.identifierPositionStack[(length - 1)], expressions);
      msg.receiver = typeRef;
      msg.tagValue = this.tagValue;
      msg.sourceEnd = this.scanner.getCurrentTokenEndPosition();
      return msg;
    }
    catch (ClassCastException localClassCastException)
    {
    }
    throw new InvalidInputException();
  }

  protected Object createReturnStatement()
  {
    return new JavadocReturnStatement(this.scanner.getCurrentTokenStartPosition(), 
      this.scanner.getCurrentTokenEndPosition());
  }

  protected void createTag()
  {
    this.tagValue = 100;
  }

  protected Object createTypeReference(int primitiveToken)
  {
    TypeReference typeRef = null;
    int size = this.identifierLengthStack[this.identifierLengthPtr];
    if (size == 1) {
      typeRef = new JavadocSingleTypeReference(
        this.identifierStack[this.identifierPtr], 
        this.identifierPositionStack[this.identifierPtr], 
        this.tagSourceStart, 
        this.tagSourceEnd);
    } else if (size > 1) {
      char[][] tokens = new char[size][];
      System.arraycopy(this.identifierStack, this.identifierPtr - size + 1, tokens, 0, size);
      long[] positions = new long[size];
      System.arraycopy(this.identifierPositionStack, this.identifierPtr - size + 1, positions, 0, size);
      typeRef = new JavadocQualifiedTypeReference(tokens, positions, this.tagSourceStart, this.tagSourceEnd);
    }
    return typeRef;
  }

  protected TypeDeclaration getParsedTypeDeclaration()
  {
    int ptr = this.sourceParser.astPtr;
    while (ptr >= 0) {
      Object node = this.sourceParser.astStack[ptr];
      if ((node instanceof TypeDeclaration)) {
        TypeDeclaration typeDecl = (TypeDeclaration)node;
        if (typeDecl.bodyEnd == 0) {
          return typeDecl;
        }
      }
      ptr--;
    }
    return null;
  }

  protected boolean parseReturn()
  {
    if (this.returnStatement == null) {
      this.returnStatement = createReturnStatement();
      return true;
    }
    if (this.reportProblems) {
      this.sourceParser.problemReporter().javadocDuplicatedReturnTag(
        this.scanner.getCurrentTokenStartPosition(), 
        this.scanner.getCurrentTokenEndPosition());
    }
    return false;
  }

  protected void parseSimpleTag()
  {
    char first = this.source[(this.index++)];
    if ((first == '\\') && (this.source[this.index] == 'u'))
    {
      int pos = this.index;
      this.index += 1;
      while (this.source[this.index] == 'u')
        this.index += 1;
      int c1;
      int c2;
      int c3;
      int c4;
      if (((c1 = ScannerHelper.getNumericValue(this.source[(this.index++)])) <= 15) && (c1 >= 0) && 
        ((c2 = ScannerHelper.getNumericValue(this.source[(this.index++)])) <= 15) && (c2 >= 0) && 
        ((c3 = ScannerHelper.getNumericValue(this.source[(this.index++)])) <= 15) && (c3 >= 0) && ((c4 = ScannerHelper.getNumericValue(this.source[(this.index++)])) <= 15) && (c4 >= 0))
        first = (char)(((c1 * 16 + c2) * 16 + c3) * 16 + c4);
      else {
        this.index = pos;
      }

    }

    switch (first) {
    case 'd':
      if ((readChar() == 'e') && 
        (readChar() == 'p') && (readChar() == 'r') && 
        (readChar() == 'e') && (readChar() == 'c') && 
        (readChar() == 'a') && (readChar() == 't') && 
        (readChar() == 'e') && (readChar() == 'd'))
      {
        char c = readChar();
        if ((ScannerHelper.isWhitespace(c)) || (c == '*')) {
          this.abort = true;
          this.deprecated = true;
          this.tagValue = 1;
        }
      }
      break;
    }
  }

  protected boolean parseTag(int previousPosition) throws InvalidInputException {
    boolean valid = false;

    int currentPosition = this.index;
    int token = readTokenAndConsume();
    if (currentPosition != this.scanner.startPosition) {
      this.tagSourceStart = previousPosition;
      this.tagSourceEnd = currentPosition;
      if (this.reportProblems) this.sourceParser.problemReporter().javadocInvalidTag(this.tagSourceStart, this.tagSourceEnd);
      return false;
    }
    if (this.index >= this.scanner.eofPosition) {
      this.tagSourceStart = previousPosition;
      this.tagSourceEnd = this.tokenPreviousPosition;
      if (this.reportProblems) this.sourceParser.problemReporter().javadocInvalidTag(this.tagSourceStart, this.tagSourceEnd);
      return false;
    }
    this.tagSourceStart = this.scanner.getCurrentTokenStartPosition();
    this.tagSourceEnd = this.scanner.getCurrentTokenEndPosition();
    char[] tagName = this.scanner.getCurrentIdentifierSource();

    if ((this.scanner.currentCharacter != ' ') && (!ScannerHelper.isWhitespace(this.scanner.currentCharacter))) {
      boolean validTag = true;
      while ((token != 68) && (this.index < this.scanner.eofPosition)) {
        int length = tagName.length;

        switch (this.scanner.currentCharacter) {
        case '*':
        case '}':
          break;
        case '!':
        case '"':
        case '#':
        case '%':
        case '&':
        case '\'':
        case ':':
        case '<':
        case '>':
        case '@':
          validTag = false;
          this.tagSourceEnd = this.scanner.getCurrentTokenEndPosition();
          this.index = this.scanner.currentPosition;
          break;
        case '-':
          System.arraycopy(tagName, 0, tagName = new char[length + 1], 0, length);
          tagName[length] = this.scanner.currentCharacter;
          this.tagSourceEnd = this.scanner.getCurrentTokenEndPosition();
          this.index = this.scanner.currentPosition;
          break;
        default:
          if ((this.scanner.currentCharacter == ' ') || (ScannerHelper.isWhitespace(this.scanner.currentCharacter))) {
            break label506;
          }
          token = readTokenAndConsume();
          char[] ident = this.scanner.getCurrentIdentifierSource();
          System.arraycopy(tagName, 0, tagName = new char[length + ident.length], 0, length);
          System.arraycopy(ident, 0, tagName, length, ident.length);
          this.tagSourceEnd = this.scanner.getCurrentTokenEndPosition();
        }

        this.scanner.getNextChar();
      }
      label506: if (!validTag) {
        if (this.reportProblems) this.sourceParser.problemReporter().javadocInvalidTag(this.tagSourceStart, this.tagSourceEnd);
        return false;
      }
    }
    int length = tagName.length;
    if (length == 0) return false;
    this.index = (this.tagSourceEnd + 1);
    this.scanner.currentPosition = (this.tagSourceEnd + 1);

    this.tagValue = 0;
    switch (token) {
    case 26:
      switch (tagName[0]) {
      case 'c':
        if ((length == TAG_CATEGORY_LENGTH) && (CharOperation.equals(TAG_CATEGORY, tagName))) {
          this.tagValue = 11;
          valid = parseIdentifierTag(false);
        }
        break;
      case 'd':
        if ((length == TAG_DEPRECATED_LENGTH) && (CharOperation.equals(TAG_DEPRECATED, tagName))) {
          this.deprecated = true;
          valid = true;
          this.tagValue = 1;
        }
        break;
      case 'e':
        if ((length == TAG_EXCEPTION_LENGTH) && (CharOperation.equals(TAG_EXCEPTION, tagName))) {
          this.tagValue = 5;
          valid = parseThrows();
        }
        break;
      case 'i':
        if ((length == TAG_INHERITDOC_LENGTH) && (CharOperation.equals(TAG_INHERITDOC, tagName)))
        {
          if (this.astPtr == -1) {
            this.inheritedPositions = ((this.tagSourceStart << 32) + this.tagSourceEnd);
          }
          valid = true;
          this.tagValue = 9;
        }
        break;
      case 'l':
        if ((length == TAG_LINK_LENGTH) && (CharOperation.equals(TAG_LINK, tagName))) {
          this.tagValue = 7;
          if ((this.inlineTagStarted) || ((this.kind & 0x4) != 0)) {
            valid = parseReference();
          }
          else
          {
            valid = false;
            if (this.reportProblems)
              this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd);
          }
        }
        else if ((length == TAG_LINKPLAIN_LENGTH) && (CharOperation.equals(TAG_LINKPLAIN, tagName))) {
          this.tagValue = 8;
          if (this.inlineTagStarted) {
            valid = parseReference();
          } else {
            valid = false;
            if (this.reportProblems) {
              this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd);
            }
          }
        }
        break;
      case 'p':
        if ((length == TAG_PARAM_LENGTH) && (CharOperation.equals(TAG_PARAM, tagName))) {
          this.tagValue = 2;
          valid = parseParam();
        }
        break;
      case 's':
        if ((length == TAG_SEE_LENGTH) && (CharOperation.equals(TAG_SEE, tagName))) {
          if (this.inlineTagStarted)
          {
            valid = false;
            if (this.reportProblems)
              this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd);
          }
          else {
            this.tagValue = 6;
            valid = parseReference();
          }
        }
        break;
      case 'v':
        if ((length == TAG_VALUE_LENGTH) && (CharOperation.equals(TAG_VALUE, tagName))) {
          this.tagValue = 10;
          if (this.sourceLevel >= 3211264L) {
            if (this.inlineTagStarted) {
              valid = parseReference();
            } else {
              valid = false;
              if (this.reportProblems) this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd);
            }
          }
          else if (this.validValuePositions == -1L) {
            if ((this.invalidValuePositions != -1L) && 
              (this.reportProblems)) this.sourceParser.problemReporter().javadocUnexpectedTag((int)(this.invalidValuePositions >>> 32), (int)this.invalidValuePositions);

            if (valid) {
              this.validValuePositions = ((this.tagSourceStart << 32) + this.tagSourceEnd);
              this.invalidValuePositions = -1L;
            } else {
              this.invalidValuePositions = ((this.tagSourceStart << 32) + this.tagSourceEnd);
            }
          }
          else if (this.reportProblems) { this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd); }
        }
        else
        {
          createTag();
        }
        break;
      case 'f':
      case 'g':
      case 'h':
      case 'j':
      case 'k':
      case 'm':
      case 'n':
      case 'o':
      case 'q':
      case 'r':
      case 't':
      case 'u':
      default:
        createTag();
      }

      break;
    case 80:
      this.tagValue = 3;
      valid = parseReturn();

      break;
    case 105:
      this.tagValue = 4;
      valid = parseThrows();
    }

    this.textStart = this.index;
    return valid;
  }

  protected boolean pushParamName(boolean isTypeParam)
  {
    ASTNode nameRef = null;
    if (isTypeParam) {
      JavadocSingleTypeReference ref = new JavadocSingleTypeReference(this.identifierStack[1], 
        this.identifierPositionStack[1], 
        this.tagSourceStart, 
        this.tagSourceEnd);
      nameRef = ref;
    } else {
      JavadocSingleNameReference ref = new JavadocSingleNameReference(this.identifierStack[0], 
        this.identifierPositionStack[0], 
        this.tagSourceStart, 
        this.tagSourceEnd);
      nameRef = ref;
    }

    if (this.astLengthPtr == -1) {
      pushOnAstStack(nameRef, true);
    }
    else {
      if (!isTypeParam) {
        for (int i = 1; i <= this.astLengthPtr; i += 3) {
          if (this.astLengthStack[i] != 0) {
            if (this.reportProblems) this.sourceParser.problemReporter().javadocUnexpectedTag(this.tagSourceStart, this.tagSourceEnd);

            if (this.invalidParamReferencesPtr == -1L) {
              this.invalidParamReferencesStack = new JavadocSingleNameReference[10];
            }
            int stackLength = this.invalidParamReferencesStack.length;
            if (++this.invalidParamReferencesPtr >= stackLength) {
              System.arraycopy(
                this.invalidParamReferencesStack, 0, 
                this.invalidParamReferencesStack = new JavadocSingleNameReference[stackLength + 10], 0, 
                stackLength);
            }
            this.invalidParamReferencesStack[this.invalidParamReferencesPtr] = nameRef;
            return false;
          }
        }
      }
      switch (this.astLengthPtr % 3)
      {
      case 0:
        pushOnAstStack(nameRef, false);
        break;
      case 2:
        pushOnAstStack(nameRef, true);
        break;
      case 1:
      default:
        return false;
      }
    }
    return true;
  }

  protected boolean pushSeeRef(Object statement)
  {
    if (this.astLengthPtr == -1) {
      pushOnAstStack(null, true);
      pushOnAstStack(null, true);
      pushOnAstStack(statement, true);
    } else {
      switch (this.astLengthPtr % 3)
      {
      case 0:
        pushOnAstStack(null, true);
        pushOnAstStack(statement, true);
        break;
      case 1:
        pushOnAstStack(statement, true);
        break;
      case 2:
        pushOnAstStack(statement, false);
        break;
      default:
        return false;
      }
    }
    return true;
  }

  protected boolean pushThrowName(Object typeRef)
  {
    if (this.astLengthPtr == -1) {
      pushOnAstStack(null, true);
      pushOnAstStack(typeRef, true);
    } else {
      switch (this.astLengthPtr % 3)
      {
      case 0:
        pushOnAstStack(typeRef, true);
        break;
      case 1:
        pushOnAstStack(typeRef, false);
        break;
      case 2:
        pushOnAstStack(null, true);
        pushOnAstStack(typeRef, true);
        break;
      default:
        return false;
      }
    }
    return true;
  }

  protected void refreshReturnStatement()
  {
    ((JavadocReturnStatement)this.returnStatement).bits &= -262145;
  }

  public String toString() {
    StringBuffer buffer = new StringBuffer();
    buffer.append("check javadoc: ").append(this.checkDocComment).append("\n");
    buffer.append("javadoc: ").append(this.docComment).append("\n");
    buffer.append(super.toString());
    return buffer.toString();
  }

  protected void updateDocComment()
  {
    this.docComment.inheritedPositions = this.inheritedPositions;
    this.docComment.valuePositions = (this.validValuePositions != -1L ? this.validValuePositions : this.invalidValuePositions);

    if (this.returnStatement != null) {
      this.docComment.returnStatement = ((JavadocReturnStatement)this.returnStatement);
    }

    if (this.invalidParamReferencesPtr >= 0) {
      this.docComment.invalidParameters = new JavadocSingleNameReference[this.invalidParamReferencesPtr + 1];
      System.arraycopy(this.invalidParamReferencesStack, 0, this.docComment.invalidParameters, 0, this.invalidParamReferencesPtr + 1);
    }

    if (this.astLengthPtr == -1) {
      return;
    }

    int[] sizes = new int[3];
    for (int i = 0; i <= this.astLengthPtr; i++) {
      sizes[(i % 3)] += this.astLengthStack[i];
    }
    this.docComment.seeReferences = new Expression[sizes[2]];
    this.docComment.exceptionReferences = new TypeReference[sizes[1]];
    int paramRefPtr = sizes[0];
    this.docComment.paramReferences = new JavadocSingleNameReference[paramRefPtr];
    int paramTypeParamPtr = sizes[0];
    this.docComment.paramTypeParameters = new JavadocSingleTypeReference[paramTypeParamPtr];

    while (this.astLengthPtr >= 0) {
      int ptr = this.astLengthPtr % 3;

      switch (ptr) {
      case 2:
        int size = this.astLengthStack[(this.astLengthPtr--)];
        for (int i = 0; i < size; i++)
        {
          int tmp278_276 = ptr;
          int[] tmp278_275 = sizes;
          int tmp282_281 = (tmp278_275[tmp278_276] - 1); tmp278_275[tmp278_276] = tmp282_281; this.docComment.seeReferences[tmp282_281] = ((Expression)this.astStack[(this.astPtr--)]);
        }
        break;
      case 1:
        int size = this.astLengthStack[(this.astLengthPtr--)];
        for (int i = 0; i < size; i++)
        {
          int tmp351_349 = ptr;
          int[] tmp351_348 = sizes;
          int tmp355_354 = (tmp351_348[tmp351_349] - 1); tmp351_348[tmp351_349] = tmp355_354; this.docComment.exceptionReferences[tmp355_354] = ((TypeReference)this.astStack[(this.astPtr--)]);
        }
        break;
      case 0:
        int size = this.astLengthStack[(this.astLengthPtr--)];
        for (int i = 0; i < size; i++) {
          Expression reference = (Expression)this.astStack[(this.astPtr--)];
          if ((reference instanceof JavadocSingleNameReference))
            this.docComment.paramReferences[(--paramRefPtr)] = ((JavadocSingleNameReference)reference);
          else if ((reference instanceof JavadocSingleTypeReference)) {
            this.docComment.paramTypeParameters[(--paramTypeParamPtr)] = ((JavadocSingleTypeReference)reference);
          }
        }
      }

    }

    if (paramRefPtr == 0) {
      this.docComment.paramTypeParameters = null;
    } else if (paramTypeParamPtr == 0) {
      this.docComment.paramReferences = null;
    } else {
      int size = sizes[0];
      System.arraycopy(this.docComment.paramReferences, paramRefPtr, this.docComment.paramReferences = new JavadocSingleNameReference[size - paramRefPtr], 0, size - paramRefPtr);
      System.arraycopy(this.docComment.paramTypeParameters, paramTypeParamPtr, this.docComment.paramTypeParameters = new JavadocSingleTypeReference[size - paramTypeParamPtr], 0, size - paramTypeParamPtr);
    }
  }
}