package org.eclipse.jdt.internal.compiler.util;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.eclipse.jdt.core.compiler.CharOperation;

public class Util
  implements SuffixConstants
{
  private static final int DEFAULT_READING_SIZE = 8192;
  public static final String UTF_8 = "UTF-8";
  public static final String LINE_SEPARATOR = System.getProperty("line.separator");

  public static final String EMPTY_STRING = new String(CharOperation.NO_CHAR);
  public static final int[] EMPTY_INT_ARRAY = new int[0];

  public static char[] bytesToChar(byte[] bytes, String encoding)
    throws IOException
  {
    return getInputStreamAsCharArray(new ByteArrayInputStream(bytes), bytes.length, encoding);
  }

  public static byte[] getFileByteContent(File file)
    throws IOException
  {
    InputStream stream = null;
    try {
      stream = new FileInputStream(file);
      return getInputStreamAsByteArray(stream, (int)file.length());
    } finally {
      if (stream != null)
        try {
          stream.close();
        }
        catch (IOException localIOException2)
        {
        }
    }
  }

  public static char[] getFileCharContent(File file, String encoding)
    throws IOException
  {
    InputStream stream = null;
    try {
      stream = new FileInputStream(file);
      return getInputStreamAsCharArray(stream, (int)file.length(), encoding);
    } finally {
      if (stream != null)
        try {
          stream.close();
        }
        catch (IOException localIOException2)
        {
        }
    }
  }

  public static byte[] getInputStreamAsByteArray(InputStream stream, int length)
    throws IOException
  {
    byte[] contents;
    if (length == -1) {
      byte[] contents = new byte[0];
      int contentsLength = 0;
      int amountRead = -1;
      do {
        int amountRequested = Math.max(stream.available(), 8192);

        if (contentsLength + amountRequested > contents.length) {
          System.arraycopy(
            contents, 
            0, 
            contents = new byte[contentsLength + amountRequested], 
            0, 
            contentsLength);
        }

        amountRead = stream.read(contents, contentsLength, amountRequested);

        if (amountRead > 0)
        {
          contentsLength += amountRead;
        }
      }
      while (amountRead != -1);

      if (contentsLength < contents.length)
        System.arraycopy(
          contents, 
          0, 
          contents = new byte[contentsLength], 
          0, 
          contentsLength);
    }
    else {
      contents = new byte[length];
      int len = 0;
      int readSize = 0;
      while ((readSize != -1) && (len != length))
      {
        len += readSize;
        readSize = stream.read(contents, len, length - len);
      }
    }

    return contents;
  }

  public static char[] getInputStreamAsCharArray(InputStream stream, int length, String encoding)
    throws IOException
  {
    InputStreamReader reader = null;
    try {
      reader = encoding == null ? 
        new InputStreamReader(stream) : 
        new InputStreamReader(stream, encoding);
    }
    catch (UnsupportedEncodingException localUnsupportedEncodingException) {
      reader = new InputStreamReader(stream);
    }

    int totalRead = 0;
    char[] contents;
    char[] contents;
    if (length == -1) {
      contents = CharOperation.NO_CHAR;
    }
    else
      contents = new char[length];
    while (true)
    {
      int amountRequested;
      int amountRequested;
      if (totalRead < length)
      {
        amountRequested = length - totalRead;
      }
      else {
        int current = reader.read();
        if (current < 0)
          break;
        amountRequested = Math.max(stream.available(), 8192);

        if (totalRead + 1 + amountRequested > contents.length) {
          System.arraycopy(contents, 0, contents = new char[totalRead + 1 + amountRequested], 0, totalRead);
        }

        contents[(totalRead++)] = ((char)current);
      }

      int amountRead = reader.read(contents, totalRead, amountRequested);
      if (amountRead < 0) break;
      totalRead += amountRead;
    }

    int start = 0;
    if ((totalRead > 0) && ("UTF-8".equals(encoding)) && 
      (contents[0] == 65279)) {
      totalRead--;
      start = 1;
    }

    if (totalRead < contents.length) {
      System.arraycopy(contents, start, contents = new char[totalRead], 0, totalRead);
    }
    return contents;
  }

  public static int getLineNumber(int position, int[] lineEnds, int g, int d) {
    if (lineEnds == null)
      return 1;
    if (d == -1)
      return 1;
    int m = g;
    while (g <= d) {
      m = g + (d - g) / 2;
      int start;
      if (position < (start = lineEnds[m]))
        d = m - 1;
      else if (position > start)
        g = m + 1;
      else {
        return m + 1;
      }
    }
    if (position < lineEnds[m]) {
      return m + 1;
    }
    return m + 2;
  }

  public static byte[] getZipEntryByteContent(ZipEntry ze, ZipFile zip)
    throws IOException
  {
    InputStream stream = null;
    try {
      stream = zip.getInputStream(ze);
      if (stream == null) throw new IOException("Invalid zip entry name : " + ze.getName());
      return getInputStreamAsByteArray(stream, (int)ze.getSize());
    } finally {
      if (stream != null)
        try {
          stream.close();
        }
        catch (IOException localIOException2)
        {
        }
    }
  }

  public static final boolean isArchiveFileName(String name)
  {
    int nameLength = name == null ? 0 : name.length();
    int suffixLength = SUFFIX_JAR.length;
    if (nameLength < suffixLength) return false;

    for (int i = 0; i < suffixLength; i++) {
      char c = name.charAt(nameLength - i - 1);
      int suffixIndex = suffixLength - i - 1;
      if ((c != SUFFIX_jar[suffixIndex]) && (c != SUFFIX_JAR[suffixIndex]))
      {
        suffixLength = SUFFIX_ZIP.length;
        if (nameLength < suffixLength) return false;
        for (int j = 0; j < suffixLength; j++) {
          c = name.charAt(nameLength - j - 1);
          suffixIndex = suffixLength - j - 1;
          if ((c != SUFFIX_zip[suffixIndex]) && (c != SUFFIX_ZIP[suffixIndex])) return false;
        }
        return true;
      }
    }
    return true;
  }

  public static final boolean isClassFileName(char[] name)
  {
    int nameLength = name == null ? 0 : name.length;
    int suffixLength = SUFFIX_CLASS.length;
    if (nameLength < suffixLength) return false;

    int i = 0; for (int offset = nameLength - suffixLength; i < suffixLength; i++) {
      char c = name[(offset + i)];
      if ((c != SUFFIX_class[i]) && (c != SUFFIX_CLASS[i])) return false;
    }
    return true;
  }

  public static final boolean isClassFileName(String name)
  {
    int nameLength = name == null ? 0 : name.length();
    int suffixLength = SUFFIX_CLASS.length;
    if (nameLength < suffixLength) return false;

    for (int i = 0; i < suffixLength; i++) {
      char c = name.charAt(nameLength - i - 1);
      int suffixIndex = suffixLength - i - 1;
      if ((c != SUFFIX_class[suffixIndex]) && (c != SUFFIX_CLASS[suffixIndex])) return false;
    }
    return true;
  }

  public static final boolean isExcluded(char[] path, char[][] inclusionPatterns, char[][] exclusionPatterns, boolean isFolderPath)
  {
    if ((inclusionPatterns == null) && (exclusionPatterns == null)) return false;

    if (inclusionPatterns != null) {
      int i = 0; for (int length = inclusionPatterns.length; i < length; i++) {
        char[] pattern = inclusionPatterns[i];
        char[] folderPattern = pattern;
        if (isFolderPath) {
          int lastSlash = CharOperation.lastIndexOf('/', pattern);
          if ((lastSlash != -1) && (lastSlash != pattern.length - 1)) {
            int star = CharOperation.indexOf('*', pattern, lastSlash);
            if ((star == -1) || 
              (star >= pattern.length - 1) || 
              (pattern[(star + 1)] != '*')) {
              folderPattern = CharOperation.subarray(pattern, 0, lastSlash);
            }
          }
        }
        if (CharOperation.pathMatch(folderPattern, path, true, '/')) {
          break;
        }
      }
      return true;
    }
    if (isFolderPath) {
      path = CharOperation.concat(path, new char[] { '*' }, '/');
    }
    if (exclusionPatterns != null) {
      int i = 0; for (int length = exclusionPatterns.length; i < length; i++) {
        if (CharOperation.pathMatch(exclusionPatterns[i], path, true, '/')) {
          return true;
        }
      }
    }
    return false;
  }

  public static final boolean isJavaFileName(char[] name)
  {
    int nameLength = name == null ? 0 : name.length;
    int suffixLength = SUFFIX_JAVA.length;
    if (nameLength < suffixLength) return false;

    int i = 0; for (int offset = nameLength - suffixLength; i < suffixLength; i++) {
      char c = name[(offset + i)];
      if ((c != SUFFIX_java[i]) && (c != SUFFIX_JAVA[i])) return false;
    }
    return true;
  }

  public static final boolean isJavaFileName(String name)
  {
    int nameLength = name == null ? 0 : name.length();
    int suffixLength = SUFFIX_JAVA.length;
    if (nameLength < suffixLength) return false;

    for (int i = 0; i < suffixLength; i++) {
      char c = name.charAt(nameLength - i - 1);
      int suffixIndex = suffixLength - i - 1;
      if ((c != SUFFIX_java[suffixIndex]) && (c != SUFFIX_JAVA[suffixIndex])) return false;
    }
    return true;
  }

  public static final int searchColumnNumber(int[] startLineIndexes, int lineNumber, int position)
  {
    switch (lineNumber) {
    case 1:
      return position + 1;
    case 2:
      return position - startLineIndexes[0];
    }
    int line = lineNumber - 2;
    int length = startLineIndexes.length;
    if (line >= length) {
      return position - startLineIndexes[(length - 1)];
    }
    return position - startLineIndexes[line];
  }

  public static Boolean toBoolean(boolean bool)
  {
    if (bool) {
      return Boolean.TRUE;
    }
    return Boolean.FALSE;
  }

  public static String toString(Object[] objects)
  {
    return toString(objects, 
      new Displayable() {
      public String displayString(Object o) {
        if (o == null) return "null";
        return o.toString();
      }
    });
  }

  public static String toString(Object[] objects, Displayable renderer)
  {
    if (objects == null) return "";
    StringBuffer buffer = new StringBuffer(10);
    for (int i = 0; i < objects.length; i++) {
      if (i > 0) buffer.append(", ");
      buffer.append(renderer.displayString(objects[i]));
    }
    return buffer.toString();
  }

  public static abstract interface Displayable
  {
    public abstract String displayString(Object paramObject);
  }
}