package org.quartz.ee.jmx.jboss;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Properties;
import javax.naming.InitialContext;
import javax.naming.Name;
import javax.naming.NameParser;
import javax.naming.NamingException;
import org.jboss.logging.Logger;
import org.jboss.naming.NonSerializableFactory;
import org.jboss.system.ServiceMBeanSupport;
import org.quartz.Scheduler;
import org.quartz.SchedulerConfigException;
import org.quartz.SchedulerException;
import org.quartz.impl.StdSchedulerFactory;

public class QuartzService extends ServiceMBeanSupport
  implements QuartzServiceMBean
{
  private Properties properties;
  private StdSchedulerFactory schedulerFactory;
  private String jndiName;
  private String propertiesFile;
  private boolean error;
  private boolean useProperties;
  private boolean usePropertiesFile;
  private boolean startScheduler = true;

  public QuartzService()
  {
    this.error = false;

    this.usePropertiesFile = false;
    this.propertiesFile = "";

    this.useProperties = false;
    this.properties = new Properties();

    this.jndiName = "Quartz";
  }

  public void setJndiName(String jndiName)
    throws Exception
  {
    String oldName = this.jndiName;
    this.jndiName = jndiName;

    if (super.getState() == 3) {
      unbind(oldName);
      try
      {
        rebind();
      } catch (NamingException ne) {
        this.log.error("Failed to rebind Scheduler", ne);

        throw new SchedulerConfigException("Failed to rebind Scheduler - ", ne);
      }
    }
  }

  public String getJndiName()
  {
    return this.jndiName;
  }

  public String getName()
  {
    return "QuartzService(" + this.jndiName + ")";
  }

  public void setProperties(String properties) {
    if (this.usePropertiesFile) {
      this.log.error("Must specify only one of 'Properties' or 'PropertiesFile'");

      this.error = true;

      return;
    }

    this.useProperties = true;
    try
    {
      properties = properties.replace(File.separator, "/");
      ByteArrayInputStream bais = new ByteArrayInputStream(properties.getBytes());

      this.properties = new Properties();
      this.properties.load(bais);
    }
    catch (IOException ioe) {
    }
  }

  public String getProperties() {
    try {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      this.properties.store(baos, "");

      return new String(baos.toByteArray());
    } catch (IOException ioe) {
    }
    return "";
  }

  public void setPropertiesFile(String propertiesFile)
  {
    if (this.useProperties) {
      this.log.error("Must specify only one of 'Properties' or 'PropertiesFile'");

      this.error = true;

      return;
    }

    this.usePropertiesFile = true;

    this.propertiesFile = propertiesFile;
  }

  public String getPropertiesFile() {
    return this.propertiesFile;
  }

  public void setStartScheduler(boolean startScheduler) {
    this.startScheduler = startScheduler;
  }

  public boolean getStartScheduler() {
    return this.startScheduler;
  }

  public void createService() throws Exception
  {
    this.log.info("Create QuartzService(" + this.jndiName + ")...");

    if (this.error) {
      this.log.error("Must specify only one of 'Properties' or 'PropertiesFile'");

      throw new Exception("Must specify only one of 'Properties' or 'PropertiesFile'");
    }

    this.schedulerFactory = new StdSchedulerFactory();
    try
    {
      if (this.useProperties) {
        this.schedulerFactory.initialize(this.properties);
      }

      if (this.usePropertiesFile)
        this.schedulerFactory.initialize(this.propertiesFile);
    }
    catch (Exception e) {
      this.log.error("Failed to initialize Scheduler", e);

      throw new SchedulerConfigException("Failed to initialize Scheduler - ", e);
    }

    this.log.info("QuartzService(" + this.jndiName + ") created.");
  }

  public void destroyService() throws Exception
  {
    this.log.info("Destroy QuartzService(" + this.jndiName + ")...");

    this.schedulerFactory = null;

    this.log.info("QuartzService(" + this.jndiName + ") destroyed.");
  }

  public void startService() throws Exception
  {
    this.log.info("Start QuartzService(" + this.jndiName + ")...");
    try
    {
      rebind();
    } catch (NamingException ne) {
      this.log.error("Failed to rebind Scheduler", ne);

      throw new SchedulerConfigException("Failed to rebind Scheduler - ", ne);
    }

    try
    {
      Scheduler scheduler = this.schedulerFactory.getScheduler();

      if (this.startScheduler)
        scheduler.start();
      else
        this.log.info("Skipping starting the scheduler (will not run jobs).");
    }
    catch (Exception e) {
      this.log.error("Failed to start Scheduler", e);

      throw new SchedulerConfigException("Failed to start Scheduler - ", e);
    }

    this.log.info("QuartzService(" + this.jndiName + ") started.");
  }

  public void stopService() throws Exception
  {
    this.log.info("Stop QuartzService(" + this.jndiName + ")...");
    try
    {
      Scheduler scheduler = this.schedulerFactory.getScheduler();

      scheduler.shutdown();
    } catch (Exception e) {
      this.log.error("Failed to shutdown Scheduler", e);

      throw new SchedulerConfigException("Failed to shutdown Scheduler - ", e);
    }

    unbind(this.jndiName);

    this.log.info("QuartzService(" + this.jndiName + ") stopped.");
  }

  private void rebind() throws NamingException, SchedulerException {
    InitialContext rootCtx = null;
    try {
      rootCtx = new InitialContext();
      Name fullName = rootCtx.getNameParser("").parse(this.jndiName);
      Scheduler scheduler = this.schedulerFactory.getScheduler();
      NonSerializableFactory.rebind(fullName, scheduler, true);
    } finally {
      if (rootCtx != null)
        try {
          rootCtx.close();
        } catch (NamingException ignore) {
        }
    }
  }

  private void unbind(String name) {
    InitialContext rootCtx = null;
    try {
      rootCtx = new InitialContext();
      rootCtx.unbind(name);
      NonSerializableFactory.unbind(name);
    } catch (NamingException e) {
      this.log.warn("Failed to unbind scheduler with jndiName: " + name, e);
    } finally {
      if (rootCtx != null)
        try {
          rootCtx.close();
        }
        catch (NamingException ignore)
        {
        }
    }
  }
}