package org.quartz.impl.jdbcjobstore;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Proxy;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.quartz.Calendar;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobKey;
import org.quartz.JobPersistenceException;
import org.quartz.ObjectAlreadyExistsException;
import org.quartz.SchedulerConfigException;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.Trigger.CompletedExecutionInstruction;
import org.quartz.Trigger.TriggerState;
import org.quartz.TriggerKey;
import org.quartz.impl.DefaultThreadExecutor;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.impl.matchers.StringMatcher.StringOperatorName;
import org.quartz.impl.triggers.SimpleTriggerImpl;
import org.quartz.spi.ClassLoadHelper;
import org.quartz.spi.JobStore;
import org.quartz.spi.OperableTrigger;
import org.quartz.spi.SchedulerSignaler;
import org.quartz.spi.ThreadExecutor;
import org.quartz.spi.TriggerFiredBundle;
import org.quartz.spi.TriggerFiredResult;
import org.quartz.utils.DBConnectionManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public abstract class JobStoreSupport
  implements JobStore, Constants
{
  protected static final String LOCK_TRIGGER_ACCESS = "TRIGGER_ACCESS";
  protected static final String LOCK_JOB_ACCESS = "JOB_ACCESS";
  protected static final String LOCK_CALENDAR_ACCESS = "CALENDAR_ACCESS";
  protected static final String LOCK_STATE_ACCESS = "STATE_ACCESS";
  protected static final String LOCK_MISFIRE_ACCESS = "MISFIRE_ACCESS";
  protected String dsName;
  protected String tablePrefix;
  protected boolean useProperties;
  protected String instanceId;
  protected String instanceName;
  protected String delegateClassName;
  protected String delegateInitString;
  protected Class<? extends DriverDelegate> delegateClass;
  protected HashMap<String, Calendar> calendarCache;
  private DriverDelegate delegate;
  private long misfireThreshold;
  private boolean dontSetAutoCommitFalse;
  private boolean isClustered;
  private boolean useDBLocks;
  private boolean lockOnInsert;
  private Semaphore lockHandler;
  private String selectWithLockSQL;
  private long clusterCheckinInterval;
  private ClusterManager clusterManagementThread;
  private MisfireHandler misfireHandler;
  private ClassLoadHelper classLoadHelper;
  private SchedulerSignaler schedSignaler;
  protected int maxToRecoverAtATime;
  private boolean setTxIsolationLevelSequential;
  private boolean acquireTriggersWithinLock;
  private long dbRetryInterval;
  private boolean makeThreadsDaemons;
  private boolean threadsInheritInitializersClassLoadContext;
  private ClassLoader initializersLoader;
  private boolean doubleCheckLockMisfireHandler;
  private final Logger log;
  private ThreadExecutor threadExecutor;
  private boolean schedulerRunning;
  private static long ftrCtr = System.currentTimeMillis();
  protected ThreadLocal<Long> sigChangeForTxCompletion;
  protected boolean firstCheckIn;
  protected long lastCheckin;

  public JobStoreSupport()
  {
    this.tablePrefix = "QRTZ_";

    this.useProperties = false;

    this.delegateClass = StdJDBCDelegate.class;

    this.calendarCache = new HashMap();

    this.misfireThreshold = 60000L;

    this.dontSetAutoCommitFalse = false;

    this.isClustered = false;

    this.useDBLocks = false;

    this.lockOnInsert = true;

    this.lockHandler = null;

    this.selectWithLockSQL = null;

    this.clusterCheckinInterval = 7500L;

    this.clusterManagementThread = null;

    this.misfireHandler = null;

    this.maxToRecoverAtATime = 20;

    this.setTxIsolationLevelSequential = false;

    this.acquireTriggersWithinLock = false;

    this.dbRetryInterval = 15000L;

    this.makeThreadsDaemons = false;

    this.threadsInheritInitializersClassLoadContext = false;
    this.initializersLoader = null;

    this.doubleCheckLockMisfireHandler = true;

    this.log = LoggerFactory.getLogger(getClass());

    this.threadExecutor = new DefaultThreadExecutor();

    this.schedulerRunning = false;

    this.sigChangeForTxCompletion = new ThreadLocal();

    this.firstCheckIn = true;

    this.lastCheckin = System.currentTimeMillis();
  }

  public void setDataSource(String dsName)
  {
    this.dsName = dsName;
  }

  public String getDataSource()
  {
    return this.dsName;
  }

  public void setTablePrefix(String prefix)
  {
    if (prefix == null) {
      prefix = "";
    }

    this.tablePrefix = prefix;
  }

  public String getTablePrefix()
  {
    return this.tablePrefix;
  }

  public void setUseProperties(String useProp)
  {
    if (useProp == null) {
      useProp = "false";
    }

    this.useProperties = Boolean.valueOf(useProp).booleanValue();
  }

  public boolean canUseProperties()
  {
    return this.useProperties;
  }

  public void setInstanceId(String instanceId)
  {
    this.instanceId = instanceId;
  }

  public String getInstanceId()
  {
    return this.instanceId;
  }

  public void setInstanceName(String instanceName)
  {
    this.instanceName = instanceName;
  }

  public void setThreadPoolSize(int poolSize)
  {
  }

  public void setThreadExecutor(ThreadExecutor threadExecutor) {
    this.threadExecutor = threadExecutor;
  }

  public ThreadExecutor getThreadExecutor() {
    return this.threadExecutor;
  }

  public String getInstanceName()
  {
    return this.instanceName;
  }

  public long getEstimatedTimeToReleaseAndAcquireTrigger() {
    return 70L;
  }

  public void setIsClustered(boolean isClustered)
  {
    this.isClustered = isClustered;
  }

  public boolean isClustered()
  {
    return this.isClustered;
  }

  public long getClusterCheckinInterval()
  {
    return this.clusterCheckinInterval;
  }

  public void setClusterCheckinInterval(long l)
  {
    this.clusterCheckinInterval = l;
  }

  public int getMaxMisfiresToHandleAtATime()
  {
    return this.maxToRecoverAtATime;
  }

  public void setMaxMisfiresToHandleAtATime(int maxToRecoverAtATime)
  {
    this.maxToRecoverAtATime = maxToRecoverAtATime;
  }

  public long getDbRetryInterval()
  {
    return this.dbRetryInterval;
  }

  public void setDbRetryInterval(long dbRetryInterval)
  {
    this.dbRetryInterval = dbRetryInterval;
  }

  public void setUseDBLocks(boolean useDBLocks)
  {
    this.useDBLocks = useDBLocks;
  }

  public boolean getUseDBLocks()
  {
    return this.useDBLocks;
  }

  public boolean isLockOnInsert() {
    return this.lockOnInsert;
  }

  public void setLockOnInsert(boolean lockOnInsert)
  {
    this.lockOnInsert = lockOnInsert;
  }

  public long getMisfireThreshold() {
    return this.misfireThreshold;
  }

  public void setMisfireThreshold(long misfireThreshold)
  {
    if (misfireThreshold < 1L) {
      throw new IllegalArgumentException("Misfirethreshold must be larger than 0");
    }

    this.misfireThreshold = misfireThreshold;
  }

  public boolean isDontSetAutoCommitFalse() {
    return this.dontSetAutoCommitFalse;
  }

  public void setDontSetAutoCommitFalse(boolean b)
  {
    this.dontSetAutoCommitFalse = b;
  }

  public boolean isTxIsolationLevelSerializable() {
    return this.setTxIsolationLevelSequential;
  }

  public void setTxIsolationLevelSerializable(boolean b)
  {
    this.setTxIsolationLevelSequential = b;
  }

  public boolean isAcquireTriggersWithinLock()
  {
    return this.acquireTriggersWithinLock;
  }

  public void setAcquireTriggersWithinLock(boolean acquireTriggersWithinLock)
  {
    this.acquireTriggersWithinLock = acquireTriggersWithinLock;
  }

  public void setDriverDelegateClass(String delegateClassName)
    throws InvalidConfigurationException
  {
    synchronized (this) {
      this.delegateClassName = delegateClassName;
    }
  }

  public String getDriverDelegateClass()
  {
    return this.delegateClassName;
  }

  public void setDriverDelegateInitString(String delegateInitString)
    throws InvalidConfigurationException
  {
    this.delegateInitString = delegateInitString;
  }

  public String getDriverDelegateInitString()
  {
    return this.delegateInitString;
  }

  public String getSelectWithLockSQL() {
    return this.selectWithLockSQL;
  }

  public void setSelectWithLockSQL(String string)
  {
    this.selectWithLockSQL = string;
  }

  protected ClassLoadHelper getClassLoadHelper() {
    return this.classLoadHelper;
  }

  public boolean getMakeThreadsDaemons()
  {
    return this.makeThreadsDaemons;
  }

  public void setMakeThreadsDaemons(boolean makeThreadsDaemons)
  {
    this.makeThreadsDaemons = makeThreadsDaemons;
  }

  public boolean isThreadsInheritInitializersClassLoadContext()
  {
    return this.threadsInheritInitializersClassLoadContext;
  }

  public void setThreadsInheritInitializersClassLoadContext(boolean threadsInheritInitializersClassLoadContext)
  {
    this.threadsInheritInitializersClassLoadContext = threadsInheritInitializersClassLoadContext;
  }

  public boolean getDoubleCheckLockMisfireHandler()
  {
    return this.doubleCheckLockMisfireHandler;
  }

  public void setDoubleCheckLockMisfireHandler(boolean doubleCheckLockMisfireHandler)
  {
    this.doubleCheckLockMisfireHandler = doubleCheckLockMisfireHandler;
  }

  protected Logger getLog()
  {
    return this.log;
  }

  public void initialize(ClassLoadHelper loadHelper, SchedulerSignaler signaler)
    throws SchedulerConfigException
  {
    if (this.dsName == null) {
      throw new SchedulerConfigException("DataSource name not set.");
    }

    this.classLoadHelper = loadHelper;
    if (isThreadsInheritInitializersClassLoadContext()) {
      this.log.info("JDBCJobStore threads will inherit ContextClassLoader of thread: " + Thread.currentThread().getName());
      this.initializersLoader = Thread.currentThread().getContextClassLoader();
    }

    this.schedSignaler = signaler;

    if (getLockHandler() == null)
    {
      if (isClustered()) {
        setUseDBLocks(true);
      }

      if (getUseDBLocks()) {
        if ((getDriverDelegateClass() != null) && (getDriverDelegateClass().equals(MSSQLDelegate.class.getName())) && 
          (getSelectWithLockSQL() == null)) {
          String msSqlDflt = "SELECT * FROM {0}LOCKS WITH (UPDLOCK,ROWLOCK) WHERE SCHED_NAME = {1} AND LOCK_NAME = ?";
          getLog().info("Detected usage of MSSQLDelegate class - defaulting 'selectWithLockSQL' to '" + msSqlDflt + "'.");
          setSelectWithLockSQL(msSqlDflt);
        }

        getLog().info("Using db table-based data access locking (synchronization).");
        setLockHandler(new StdRowLockSemaphore(getTablePrefix(), getInstanceName(), getSelectWithLockSQL()));
      } else {
        getLog().info("Using thread monitor-based data access locking (synchronization).");

        setLockHandler(new SimpleSemaphore());
      }
    }
  }

  public void schedulerStarted()
    throws SchedulerException
  {
    if (isClustered()) {
      this.clusterManagementThread = new ClusterManager();
      if (this.initializersLoader != null)
        this.clusterManagementThread.setContextClassLoader(this.initializersLoader);
      this.clusterManagementThread.initialize();
    } else {
      try {
        recoverJobs();
      } catch (SchedulerException se) {
        throw new SchedulerConfigException("Failure occured during job recovery.", se);
      }

    }

    this.misfireHandler = new MisfireHandler();
    if (this.initializersLoader != null)
      this.misfireHandler.setContextClassLoader(this.initializersLoader);
    this.misfireHandler.initialize();
    this.schedulerRunning = true;

    getLog().debug("JobStore background threads started (as scheduler was started).");
  }

  public void schedulerPaused() {
    this.schedulerRunning = false;
  }

  public void schedulerResumed() {
    this.schedulerRunning = true;
  }

  public void shutdown()
  {
    if (this.misfireHandler != null) {
      this.misfireHandler.shutdown();
      try {
        this.misfireHandler.join();
      }
      catch (InterruptedException ignoreInterruptionOfThisThread) {
      }
    }
    if (this.clusterManagementThread != null) {
      this.clusterManagementThread.shutdown();
      try {
        this.clusterManagementThread.join();
      }
      catch (InterruptedException ignoreInterruptionOfThisThread) {
      }
    }
    try {
      DBConnectionManager.getInstance().shutdown(getDataSource());
    } catch (SQLException sqle) {
      getLog().warn("Database connection shutdown unsuccessful.", sqle);
    }

    getLog().debug("JobStore background threads shutdown.");
  }

  public boolean supportsPersistence() {
    return true;
  }

  protected abstract Connection getNonManagedTXConnection()
    throws JobPersistenceException;

  protected Connection getAttributeRestoringConnection(Connection conn)
  {
    return (Connection)Proxy.newProxyInstance(Thread.currentThread().getContextClassLoader(), new Class[] { Connection.class }, new AttributeRestoringConnectionInvocationHandler(conn));
  }

  protected Connection getConnection()
    throws JobPersistenceException
  {
    Connection conn = null;
    try {
      conn = DBConnectionManager.getInstance().getConnection(getDataSource());
    }
    catch (SQLException sqle) {
      throw new JobPersistenceException("Failed to obtain DB connection from data source '" + getDataSource() + "': " + sqle.toString(), sqle);
    }
    catch (Throwable e)
    {
      throw new JobPersistenceException("Failed to obtain DB connection from data source '" + getDataSource() + "': " + e.toString(), e);
    }

    if (conn == null) {
      throw new JobPersistenceException("Could not get connection from DataSource '" + getDataSource() + "'");
    }

    conn = getAttributeRestoringConnection(conn);
    try
    {
      if (!isDontSetAutoCommitFalse()) {
        conn.setAutoCommit(false);
      }

      if (isTxIsolationLevelSerializable())
        conn.setTransactionIsolation(8);
    }
    catch (SQLException sqle) {
      getLog().warn("Failed to override connection auto commit/transaction isolation.", sqle); } catch (Throwable e) {
      try {
        conn.close(); } catch (Throwable tt) {
      }
      throw new JobPersistenceException("Failure setting up connection.", e);
    }

    return conn;
  }

  protected void releaseLock(Connection conn, String lockName, boolean doIt) {
    if ((doIt) && (conn != null))
      try {
        getLockHandler().releaseLock(conn, lockName);
      } catch (LockException le) {
        getLog().error("Error returning lock: " + le.getMessage(), le);
      }
  }

  protected void recoverJobs()
    throws JobPersistenceException
  {
    executeInNonManagedTXLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.recoverJobs(conn);
      }
    });
  }

  protected void recoverJobs(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      int rows = getDelegate().updateTriggerStatesFromOtherStates(conn, "WAITING", "ACQUIRED", "BLOCKED");

      rows += getDelegate().updateTriggerStatesFromOtherStates(conn, "PAUSED", "PAUSED_BLOCKED", "PAUSED_BLOCKED");

      getLog().info("Freed " + rows + " triggers from 'acquired' / 'blocked' state.");

      recoverMisfiredJobs(conn, true);

      List recoveringJobTriggers = getDelegate().selectTriggersForRecoveringJobs(conn);

      getLog().info("Recovering " + recoveringJobTriggers.size() + " jobs that were in-progress at the time of the last shut-down.");

      for (OperableTrigger recoveringJobTrigger : recoveringJobTriggers) {
        if (jobExists(conn, recoveringJobTrigger.getJobKey())) {
          recoveringJobTrigger.computeFirstFireTime(null);
          storeTrigger(conn, recoveringJobTrigger, null, false, "WAITING", false, true);
        }
      }

      getLog().info("Recovery complete.");

      List cts = getDelegate().selectTriggersInState(conn, "COMPLETE");
      for (TriggerKey ct : cts) {
        removeTrigger(conn, ct);
      }
      getLog().info("Removed " + cts.size() + " 'complete' triggers.");

      int n = getDelegate().deleteFiredTriggers(conn);
      getLog().info("Removed " + n + " stale fired job entries.");
    } catch (JobPersistenceException e) {
      throw e;
    } catch (Exception e) {
      throw new JobPersistenceException("Couldn't recover jobs: " + e.getMessage(), e);
    }
  }

  protected long getMisfireTime()
  {
    long misfireTime = System.currentTimeMillis();
    if (getMisfireThreshold() > 0L) {
      misfireTime -= getMisfireThreshold();
    }

    return misfireTime > 0L ? misfireTime : 0L;
  }

  protected RecoverMisfiredJobsResult recoverMisfiredJobs(Connection conn, boolean recovering)
    throws JobPersistenceException, SQLException
  {
    int maxMisfiresToHandleAtATime = recovering ? -1 : getMaxMisfiresToHandleAtATime();

    List misfiredTriggers = new LinkedList();
    long earliestNewTime = 9223372036854775807L;

    boolean hasMoreMisfiredTriggers = getDelegate().hasMisfiredTriggersInState(conn, "WAITING", getMisfireTime(), maxMisfiresToHandleAtATime, misfiredTriggers);

    if (hasMoreMisfiredTriggers) {
      getLog().info("Handling the first " + misfiredTriggers.size() + " triggers that missed their scheduled fire-time.  " + "More misfired triggers remain to be processed.");
    }
    else if (misfiredTriggers.size() > 0) {
      getLog().info("Handling " + misfiredTriggers.size() + " trigger(s) that missed their scheduled fire-time.");
    }
    else
    {
      getLog().debug("Found 0 triggers that missed their scheduled fire-time.");

      return RecoverMisfiredJobsResult.NO_OP;
    }

    for (TriggerKey triggerKey : misfiredTriggers)
    {
      OperableTrigger trig = retrieveTrigger(conn, triggerKey);

      if (trig != null)
      {
        doUpdateOfMisfiredTrigger(conn, trig, false, "WAITING", recovering);

        if ((trig.getNextFireTime() != null) && (trig.getNextFireTime().getTime() < earliestNewTime))
          earliestNewTime = trig.getNextFireTime().getTime();
      }
    }
    return new RecoverMisfiredJobsResult(hasMoreMisfiredTriggers, misfiredTriggers.size(), earliestNewTime);
  }

  protected boolean updateMisfiredTrigger(Connection conn, TriggerKey triggerKey, String newStateIfNotComplete, boolean forceState)
    throws JobPersistenceException
  {
    try
    {
      OperableTrigger trig = retrieveTrigger(conn, triggerKey);

      long misfireTime = System.currentTimeMillis();
      if (getMisfireThreshold() > 0L) {
        misfireTime -= getMisfireThreshold();
      }

      if (trig.getNextFireTime().getTime() > misfireTime) {
        return false;
      }

      doUpdateOfMisfiredTrigger(conn, trig, forceState, newStateIfNotComplete, false);

      return true;
    }
    catch (Exception e) {
      throw new JobPersistenceException("Couldn't update misfired trigger '" + triggerKey + "': " + e.getMessage(), e);
    }
  }

  private void doUpdateOfMisfiredTrigger(Connection conn, OperableTrigger trig, boolean forceState, String newStateIfNotComplete, boolean recovering) throws JobPersistenceException
  {
    Calendar cal = null;
    if (trig.getCalendarName() != null) {
      cal = retrieveCalendar(conn, trig.getCalendarName());
    }

    this.schedSignaler.notifyTriggerListenersMisfired(trig);

    trig.updateAfterMisfire(cal);

    if (trig.getNextFireTime() == null) {
      storeTrigger(conn, trig, null, true, "COMPLETE", forceState, recovering);

      this.schedSignaler.notifySchedulerListenersFinalized(trig);
    } else {
      storeTrigger(conn, trig, null, true, newStateIfNotComplete, forceState, false);
    }
  }

  public void storeJobAndTrigger(final JobDetail newJob, final OperableTrigger newTrigger)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    executeInLock(isLockOnInsert() ? "TRIGGER_ACCESS" : null, new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.storeJob(conn, newJob, false);
        JobStoreSupport.this.storeTrigger(conn, newTrigger, newJob, false, "WAITING", false, false);
      }
    });
  }

  public void storeJob(final JobDetail newJob, final boolean replaceExisting)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    executeInLock((isLockOnInsert()) || (replaceExisting) ? "TRIGGER_ACCESS" : null, new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.storeJob(conn, newJob, replaceExisting);
      }
    });
  }

  protected void storeJob(Connection conn, JobDetail newJob, boolean replaceExisting)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    boolean existingJob = jobExists(conn, newJob.getKey());
    try {
      if (existingJob) {
        if (!replaceExisting) {
          throw new ObjectAlreadyExistsException(newJob);
        }
        getDelegate().updateJobDetail(conn, newJob);
      } else {
        getDelegate().insertJobDetail(conn, newJob);
      }
    } catch (IOException e) {
      throw new JobPersistenceException("Couldn't store job: " + e.getMessage(), e);
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't store job: " + e.getMessage(), e);
    }
  }

  protected boolean jobExists(Connection conn, JobKey jobKey)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().jobExists(conn, jobKey);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine job existence (" + jobKey + "): " + e.getMessage(), e);
    }
  }

  public void storeTrigger(final OperableTrigger newTrigger, final boolean replaceExisting)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    executeInLock((isLockOnInsert()) || (replaceExisting) ? "TRIGGER_ACCESS" : null, new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.storeTrigger(conn, newTrigger, null, replaceExisting, "WAITING", false, false);
      }
    });
  }

  protected void storeTrigger(Connection conn, OperableTrigger newTrigger, JobDetail job, boolean replaceExisting, String state, boolean forceState, boolean recovering)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    boolean existingTrigger = triggerExists(conn, newTrigger.getKey());

    if ((existingTrigger) && (!replaceExisting)) {
      throw new ObjectAlreadyExistsException(newTrigger);
    }

    try
    {
      boolean shouldBepaused = false;

      if (!forceState) {
        shouldBepaused = getDelegate().isTriggerGroupPaused(conn, newTrigger.getKey().getGroup());

        if (!shouldBepaused) {
          shouldBepaused = getDelegate().isTriggerGroupPaused(conn, "_$_ALL_GROUPS_PAUSED_$_");

          if (shouldBepaused) {
            getDelegate().insertPausedTriggerGroup(conn, newTrigger.getKey().getGroup());
          }
        }

        if ((shouldBepaused) && ((state.equals("WAITING")) || (state.equals("ACQUIRED")))) {
          state = "PAUSED";
        }
      }

      if (job == null) {
        job = getDelegate().selectJobDetail(conn, newTrigger.getJobKey(), getClassLoadHelper());
      }
      if (job == null) {
        throw new JobPersistenceException("The job (" + newTrigger.getJobKey() + ") referenced by the trigger does not exist.");
      }

      if ((job.isConcurrentExectionDisallowed()) && (!recovering)) {
        state = checkBlockedState(conn, job.getKey(), state);
      }

      if (existingTrigger)
        getDelegate().updateTrigger(conn, newTrigger, state, job);
      else
        getDelegate().insertTrigger(conn, newTrigger, state, job);
    }
    catch (Exception e) {
      throw new JobPersistenceException("Couldn't store trigger '" + newTrigger.getKey() + "' for '" + newTrigger.getJobKey() + "' job:" + e.getMessage(), e);
    }
  }

  protected boolean triggerExists(Connection conn, TriggerKey key)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().triggerExists(conn, key);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine trigger existence (" + key + "): " + e.getMessage(), e);
    }
  }

  public boolean removeJob(final JobKey jobKey)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.removeJob(conn, jobKey, true) ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  protected boolean removeJob(Connection conn, JobKey jobKey, boolean activeDeleteSafe)
    throws JobPersistenceException
  {
    try
    {
      List jobTriggers = getDelegate().selectTriggerKeysForJob(conn, jobKey);
      for (TriggerKey jobTrigger : jobTriggers) {
        deleteTriggerAndChildren(conn, jobTrigger);
      }

      return deleteJobAndChildren(conn, jobKey);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't remove job: " + e.getMessage(), e);
    }
  }

  public boolean removeJobs(final List<JobKey> jobKeys)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        boolean allFound = true;

        for (JobKey jobKey : jobKeys) {
          allFound = (JobStoreSupport.this.removeJob(conn, jobKey, true)) && (allFound);
        }
        return allFound ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  public boolean removeTriggers(final List<TriggerKey> triggerKeys)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        boolean allFound = true;

        for (TriggerKey triggerKey : triggerKeys) {
          allFound = (JobStoreSupport.this.removeTrigger(conn, triggerKey)) && (allFound);
        }
        return allFound ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  public void storeJobsAndTriggers(final Map<JobDetail, List<Trigger>> triggersAndJobs, final boolean replace)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    executeInLock((isLockOnInsert()) || (replace) ? "TRIGGER_ACCESS" : null, new VoidTransactionCallback()
    {
      public void execute(Connection conn)
        throws JobPersistenceException
      {
        for (Iterator i$ = triggersAndJobs.keySet().iterator(); i$.hasNext(); ) { job = (JobDetail)i$.next();
          JobStoreSupport.this.storeJob(conn, job, replace);
          for (Trigger trigger : (List)triggersAndJobs.get(job))
            JobStoreSupport.this.storeTrigger(conn, (OperableTrigger)trigger, job, replace, "WAITING", false, false);
        }
        JobDetail job;
      }
    });
  }

  private boolean deleteJobAndChildren(Connection conn, JobKey key)
    throws NoSuchDelegateException, SQLException
  {
    return getDelegate().deleteJobDetail(conn, key) > 0;
  }

  private boolean deleteTriggerAndChildren(Connection conn, TriggerKey key)
    throws SQLException, NoSuchDelegateException
  {
    return getDelegate().deleteTrigger(conn, key) > 0;
  }

  public JobDetail retrieveJob(final JobKey jobKey)
    throws JobPersistenceException
  {
    return (JobDetail)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.retrieveJob(conn, jobKey);
      }
    });
  }

  protected JobDetail retrieveJob(Connection conn, JobKey key) throws JobPersistenceException {
    try {
      return getDelegate().selectJobDetail(conn, key, getClassLoadHelper());
    }
    catch (ClassNotFoundException e)
    {
      throw new JobPersistenceException("Couldn't retrieve job because a required class was not found: " + e.getMessage(), e);
    }
    catch (IOException e)
    {
      throw new JobPersistenceException("Couldn't retrieve job because the BLOB couldn't be deserialized: " + e.getMessage(), e);
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't retrieve job: " + e.getMessage(), e);
    }
  }

  public boolean removeTrigger(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.removeTrigger(conn, triggerKey) ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  protected boolean removeTrigger(Connection conn, TriggerKey key)
    throws JobPersistenceException
  {
    boolean removedTrigger = false;
    try
    {
      JobDetail job = getDelegate().selectJobForTrigger(conn, getClassLoadHelper(), key);

      removedTrigger = deleteTriggerAndChildren(conn, key);

      if ((null != job) && (!job.isDurable())) {
        int numTriggers = getDelegate().selectNumTriggersForJob(conn, job.getKey());

        if (numTriggers == 0)
        {
          deleteJobAndChildren(conn, job.getKey());
        }
      }
    } catch (ClassNotFoundException e) {
      throw new JobPersistenceException("Couldn't remove trigger: " + e.getMessage(), e);
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't remove trigger: " + e.getMessage(), e);
    }

    return removedTrigger;
  }

  public boolean replaceTrigger(final TriggerKey triggerKey, final OperableTrigger newTrigger)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.replaceTrigger(conn, triggerKey, newTrigger) ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  protected boolean replaceTrigger(Connection conn, TriggerKey key, OperableTrigger newTrigger)
    throws JobPersistenceException
  {
    try
    {
      JobDetail job = getDelegate().selectJobForTrigger(conn, getClassLoadHelper(), key);

      if (job == null) {
        return false;
      }

      if (!newTrigger.getJobKey().equals(job.getKey())) {
        throw new JobPersistenceException("New trigger is not related to the same job as the old trigger.");
      }

      boolean removedTrigger = deleteTriggerAndChildren(conn, key);

      storeTrigger(conn, newTrigger, job, false, "WAITING", false, false);

      return removedTrigger;
    } catch (ClassNotFoundException e) {
      throw new JobPersistenceException("Couldn't remove trigger: " + e.getMessage(), e);
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't remove trigger: " + e.getMessage(), e);
    }
  }

  public OperableTrigger retrieveTrigger(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    return (OperableTrigger)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.retrieveTrigger(conn, triggerKey);
      }
    });
  }

  protected OperableTrigger retrieveTrigger(Connection conn, TriggerKey key) throws JobPersistenceException
  {
    try {
      return getDelegate().selectTrigger(conn, key);
    }
    catch (Exception e)
    {
      throw new JobPersistenceException("Couldn't retrieve trigger: " + e.getMessage(), e);
    }
  }

  public Trigger.TriggerState getTriggerState(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    return (Trigger.TriggerState)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getTriggerState(conn, triggerKey);
      }
    });
  }

  public Trigger.TriggerState getTriggerState(Connection conn, TriggerKey key) throws JobPersistenceException
  {
    try {
      String ts = getDelegate().selectTriggerState(conn, key);

      if (ts == null) {
        return Trigger.TriggerState.NONE;
      }

      if (ts.equals("DELETED")) {
        return Trigger.TriggerState.NONE;
      }

      if (ts.equals("COMPLETE")) {
        return Trigger.TriggerState.COMPLETE;
      }

      if (ts.equals("PAUSED")) {
        return Trigger.TriggerState.PAUSED;
      }

      if (ts.equals("PAUSED_BLOCKED")) {
        return Trigger.TriggerState.PAUSED;
      }

      if (ts.equals("ERROR")) {
        return Trigger.TriggerState.ERROR;
      }

      if (ts.equals("BLOCKED")) {
        return Trigger.TriggerState.BLOCKED;
      }

      return Trigger.TriggerState.NORMAL;
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine state of trigger (" + key + "): " + e.getMessage(), e);
    }
  }

  public void storeCalendar(final String calName, final Calendar calendar, final boolean replaceExisting, final boolean updateTriggers)
    throws ObjectAlreadyExistsException, JobPersistenceException
  {
    executeInLock((isLockOnInsert()) || (updateTriggers) ? "TRIGGER_ACCESS" : null, new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.storeCalendar(conn, calName, calendar, replaceExisting, updateTriggers);
      }
    });
  }

  protected void storeCalendar(Connection conn, String calName, Calendar calendar, boolean replaceExisting, boolean updateTriggers) throws ObjectAlreadyExistsException, JobPersistenceException
  {
    try
    {
      boolean existingCal = calendarExists(conn, calName);
      if ((existingCal) && (!replaceExisting)) {
        throw new ObjectAlreadyExistsException("Calendar with name '" + calName + "' already exists.");
      }

      if (existingCal) {
        if (getDelegate().updateCalendar(conn, calName, calendar) < 1) {
          throw new JobPersistenceException("Couldn't store calendar.  Update failed.");
        }

        if (updateTriggers) {
          List trigs = getDelegate().selectTriggersForCalendar(conn, calName);

          for (OperableTrigger trigger : trigs) {
            trigger.updateWithNewCalendar(calendar, getMisfireThreshold());
            storeTrigger(conn, trigger, null, true, "WAITING", false, false);
          }
        }
      }
      else if (getDelegate().insertCalendar(conn, calName, calendar) < 1) {
        throw new JobPersistenceException("Couldn't store calendar.  Insert failed.");
      }

      if (!this.isClustered)
        this.calendarCache.put(calName, calendar);
    }
    catch (IOException e)
    {
      throw new JobPersistenceException("Couldn't store calendar because the BLOB couldn't be serialized: " + e.getMessage(), e);
    }
    catch (ClassNotFoundException e)
    {
      throw new JobPersistenceException("Couldn't store calendar: " + e.getMessage(), e);
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't store calendar: " + e.getMessage(), e);
    }
  }

  protected boolean calendarExists(Connection conn, String calName) throws JobPersistenceException
  {
    try
    {
      return getDelegate().calendarExists(conn, calName);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine calendar existence (" + calName + "): " + e.getMessage(), e);
    }
  }

  public boolean removeCalendar(final String calName)
    throws JobPersistenceException
  {
    return ((Boolean)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.removeCalendar(conn, calName) ? Boolean.TRUE : Boolean.FALSE;
      }
    })).booleanValue();
  }

  protected boolean removeCalendar(Connection conn, String calName)
    throws JobPersistenceException
  {
    try
    {
      if (getDelegate().calendarIsReferenced(conn, calName)) {
        throw new JobPersistenceException("Calender cannot be removed if it referenced by a trigger!");
      }

      if (!this.isClustered) {
        this.calendarCache.remove(calName);
      }

      return getDelegate().deleteCalendar(conn, calName) > 0;
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't remove calendar: " + e.getMessage(), e);
    }
  }

  public Calendar retrieveCalendar(final String calName)
    throws JobPersistenceException
  {
    return (Calendar)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.retrieveCalendar(conn, calName);
      }
    });
  }

  protected Calendar retrieveCalendar(Connection conn, String calName)
    throws JobPersistenceException
  {
    Calendar cal = this.isClustered ? null : (Calendar)this.calendarCache.get(calName);
    if (cal != null) {
      return cal;
    }
    try
    {
      cal = getDelegate().selectCalendar(conn, calName);
      if (!this.isClustered) {
        this.calendarCache.put(calName, cal);
      }
      return cal;
    } catch (ClassNotFoundException e) {
      throw new JobPersistenceException("Couldn't retrieve calendar because a required class was not found: " + e.getMessage(), e);
    }
    catch (IOException e)
    {
      throw new JobPersistenceException("Couldn't retrieve calendar because the BLOB couldn't be deserialized: " + e.getMessage(), e);
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't retrieve calendar: " + e.getMessage(), e);
    }
  }

  public int getNumberOfJobs()
    throws JobPersistenceException
  {
    return ((Integer)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return Integer.valueOf(JobStoreSupport.this.getNumberOfJobs(conn));
      }
    })).intValue();
  }

  protected int getNumberOfJobs(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().selectNumJobs(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain number of jobs: " + e.getMessage(), e);
    }
  }

  public int getNumberOfTriggers()
    throws JobPersistenceException
  {
    return ((Integer)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return Integer.valueOf(JobStoreSupport.this.getNumberOfTriggers(conn));
      }
    })).intValue();
  }

  protected int getNumberOfTriggers(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().selectNumTriggers(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain number of triggers: " + e.getMessage(), e);
    }
  }

  public int getNumberOfCalendars()
    throws JobPersistenceException
  {
    return ((Integer)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return Integer.valueOf(JobStoreSupport.this.getNumberOfCalendars(conn));
      }
    })).intValue();
  }

  protected int getNumberOfCalendars(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().selectNumCalendars(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain number of calendars: " + e.getMessage(), e);
    }
  }

  public Set<JobKey> getJobKeys(final GroupMatcher<JobKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getJobNames(conn, matcher);
      }
    });
  }

  protected Set<JobKey> getJobNames(Connection conn, GroupMatcher<JobKey> matcher) throws JobPersistenceException
  {
    Set jobNames;
    try
    {
      jobNames = getDelegate().selectJobsInGroup(conn, matcher);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain job names: " + e.getMessage(), e);
    }

    return jobNames;
  }

  public boolean checkExists(final JobKey jobKey)
    throws JobPersistenceException
  {
    return ((Boolean)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return Boolean.valueOf(JobStoreSupport.this.checkExists(conn, jobKey));
      }
    })).booleanValue();
  }

  protected boolean checkExists(Connection conn, JobKey jobKey)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().jobExists(conn, jobKey);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't check for existence of job: " + e.getMessage(), e);
    }
  }

  public boolean checkExists(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    return ((Boolean)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return Boolean.valueOf(JobStoreSupport.this.checkExists(conn, triggerKey));
      }
    })).booleanValue();
  }

  protected boolean checkExists(Connection conn, TriggerKey triggerKey)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().triggerExists(conn, triggerKey);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't check for existence of job: " + e.getMessage(), e);
    }
  }

  public void clearAllSchedulingData()
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.clearAllSchedulingData(conn);
      }
    });
  }

  protected void clearAllSchedulingData(Connection conn) throws JobPersistenceException {
    try {
      getDelegate().clearData(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Error clearing scheduling data: " + e.getMessage(), e);
    }
  }

  public Set<TriggerKey> getTriggerKeys(final GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getTriggerNames(conn, matcher);
      }
    });
  }

  protected Set<TriggerKey> getTriggerNames(Connection conn, GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    Set trigNames;
    try
    {
      trigNames = getDelegate().selectTriggersInGroup(conn, matcher);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain trigger names: " + e.getMessage(), e);
    }

    return trigNames;
  }

  public List<String> getJobGroupNames()
    throws JobPersistenceException
  {
    return (List)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getJobGroupNames(conn);
      }
    });
  }

  protected List<String> getJobGroupNames(Connection conn)
    throws JobPersistenceException
  {
    List groupNames = null;
    try
    {
      groupNames = getDelegate().selectJobGroups(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain job groups: " + e.getMessage(), e);
    }

    return groupNames;
  }

  public List<String> getTriggerGroupNames()
    throws JobPersistenceException
  {
    return (List)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getTriggerGroupNames(conn);
      }
    });
  }

  protected List<String> getTriggerGroupNames(Connection conn) throws JobPersistenceException
  {
    List groupNames = null;
    try
    {
      groupNames = getDelegate().selectTriggerGroups(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain trigger groups: " + e.getMessage(), e);
    }

    return groupNames;
  }

  public List<String> getCalendarNames()
    throws JobPersistenceException
  {
    return (List)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getCalendarNames(conn);
      }
    });
  }

  protected List<String> getCalendarNames(Connection conn) throws JobPersistenceException
  {
    try {
      return getDelegate().selectCalendars(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't obtain trigger groups: " + e.getMessage(), e);
    }
  }

  public List<OperableTrigger> getTriggersForJob(final JobKey jobKey)
    throws JobPersistenceException
  {
    return (List)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getTriggersForJob(conn, jobKey);
      }
    });
  }

  protected List<OperableTrigger> getTriggersForJob(Connection conn, JobKey key)
    throws JobPersistenceException
  {
    List list = null;
    try
    {
      list = getDelegate().selectTriggersForJob(conn, key);
    }
    catch (Exception e) {
      throw new JobPersistenceException("Couldn't obtain triggers for job: " + e.getMessage(), e);
    }

    return list;
  }

  public void pauseTrigger(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.pauseTrigger(conn, triggerKey);
      }
    });
  }

  public void pauseTrigger(Connection conn, TriggerKey triggerKey)
    throws JobPersistenceException
  {
    try
    {
      String oldState = getDelegate().selectTriggerState(conn, triggerKey);

      if ((oldState.equals("WAITING")) || (oldState.equals("ACQUIRED")))
      {
        getDelegate().updateTriggerState(conn, triggerKey, "PAUSED");
      }
      else if (oldState.equals("BLOCKED"))
        getDelegate().updateTriggerState(conn, triggerKey, "PAUSED_BLOCKED");
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't pause trigger '" + triggerKey + "': " + e.getMessage(), e);
    }
  }

  public void pauseJob(final JobKey jobKey)
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        List triggers = JobStoreSupport.this.getTriggersForJob(conn, jobKey);
        for (OperableTrigger trigger : triggers)
          JobStoreSupport.this.pauseTrigger(conn, trigger.getKey());
      }
    });
  }

  public Set<String> pauseJobs(final GroupMatcher<JobKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Set<String> execute(Connection conn) throws JobPersistenceException
      {
        Set groupNames = new HashSet();
        Set jobNames = JobStoreSupport.this.getJobNames(conn, matcher);

        for (JobKey jobKey : jobNames) {
          List triggers = JobStoreSupport.this.getTriggersForJob(conn, jobKey);
          for (OperableTrigger trigger : triggers) {
            JobStoreSupport.this.pauseTrigger(conn, trigger.getKey());
          }
          groupNames.add(jobKey.getGroup());
        }

        return groupNames;
      }
    });
  }

  protected String checkBlockedState(Connection conn, JobKey jobKey, String currentState)
    throws JobPersistenceException
  {
    if ((!currentState.equals("WAITING")) && (!currentState.equals("PAUSED")))
    {
      return currentState;
    }
    try
    {
      List lst = getDelegate().selectFiredTriggerRecordsByJob(conn, jobKey.getName(), jobKey.getGroup());

      if (lst.size() > 0) {
        FiredTriggerRecord rec = (FiredTriggerRecord)lst.get(0);
        if (rec.isJobDisallowsConcurrentExecution()) {
          return "PAUSED".equals(currentState) ? "PAUSED_BLOCKED" : "BLOCKED";
        }
      }

      return currentState;
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine if trigger should be in a blocked state '" + jobKey + "': " + e.getMessage(), e);
    }
  }

  public void resumeTrigger(final TriggerKey triggerKey)
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.resumeTrigger(conn, triggerKey);
      }
    });
  }

  public void resumeTrigger(Connection conn, TriggerKey key)
    throws JobPersistenceException
  {
    try
    {
      TriggerStatus status = getDelegate().selectTriggerStatus(conn, key);

      if ((status == null) || (status.getNextFireTime() == null)) {
        return;
      }

      boolean blocked = false;
      if ("PAUSED_BLOCKED".equals(status.getStatus())) {
        blocked = true;
      }

      String newState = checkBlockedState(conn, status.getJobKey(), "WAITING");

      boolean misfired = false;

      if ((this.schedulerRunning) && (status.getNextFireTime().before(new Date()))) {
        misfired = updateMisfiredTrigger(conn, key, newState, true);
      }

      if (!misfired) {
        if (blocked) {
          getDelegate().updateTriggerStateFromOtherState(conn, key, newState, "PAUSED_BLOCKED");
        }
        else {
          getDelegate().updateTriggerStateFromOtherState(conn, key, newState, "PAUSED");
        }
      }
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't resume trigger '" + key + "': " + e.getMessage(), e);
    }
  }

  public void resumeJob(final JobKey jobKey)
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        List triggers = JobStoreSupport.this.getTriggersForJob(conn, jobKey);
        for (OperableTrigger trigger : triggers)
          JobStoreSupport.this.resumeTrigger(conn, trigger.getKey());
      }
    });
  }

  public Set<String> resumeJobs(final GroupMatcher<JobKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Set<String> execute(Connection conn) throws JobPersistenceException
      {
        Set jobKeys = JobStoreSupport.this.getJobNames(conn, matcher);
        Set groupNames = new HashSet();

        for (JobKey jobKey : jobKeys) {
          List triggers = JobStoreSupport.this.getTriggersForJob(conn, jobKey);
          for (OperableTrigger trigger : triggers) {
            JobStoreSupport.this.resumeTrigger(conn, trigger.getKey());
          }
          groupNames.add(jobKey.getGroup());
        }
        return groupNames;
      }
    });
  }

  public Set<String> pauseTriggers(final GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Set<String> execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.pauseTriggerGroup(conn, matcher);
      }
    });
  }

  public Set<String> pauseTriggerGroup(Connection conn, GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    try
    {
      getDelegate().updateTriggerGroupStateFromOtherStates(conn, matcher, "PAUSED", "ACQUIRED", "WAITING", "WAITING");

      getDelegate().updateTriggerGroupStateFromOtherState(conn, matcher, "PAUSED_BLOCKED", "BLOCKED");

      List groups = getDelegate().selectTriggerGroups(conn, matcher);

      StringMatcher.StringOperatorName operator = matcher.getCompareWithOperator();
      if ((operator.equals(StringMatcher.StringOperatorName.EQUALS)) && (!groups.contains(matcher.getCompareToValue()))) {
        groups.add(matcher.getCompareToValue());
      }

      for (String group : groups) {
        if (!getDelegate().isTriggerGroupPaused(conn, group)) {
          getDelegate().insertPausedTriggerGroup(conn, group);
        }
      }

      return new HashSet(groups);
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't pause trigger group '" + matcher + "': " + e.getMessage(), e);
    }
  }

  public Set<String> getPausedTriggerGroups()
    throws JobPersistenceException
  {
    return (Set)executeWithoutLock(new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException {
        return JobStoreSupport.this.getPausedTriggerGroups(conn);
      }
    });
  }

  public Set<String> getPausedTriggerGroups(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      return getDelegate().selectPausedTriggerGroups(conn);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't determine paused trigger groups: " + e.getMessage(), e);
    }
  }

  public Set<String> resumeTriggers(final GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    return (Set)executeInLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Set<String> execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.resumeTriggerGroup(conn, matcher);
      }
    });
  }

  public Set<String> resumeTriggerGroup(Connection conn, GroupMatcher<TriggerKey> matcher)
    throws JobPersistenceException
  {
    try
    {
      getDelegate().deletePausedTriggerGroup(conn, matcher);
      HashSet groups = new HashSet();

      Set keys = getDelegate().selectTriggersInGroup(conn, matcher);

      for (TriggerKey key : keys) {
        resumeTrigger(conn, key);
        groups.add(key.getGroup());
      }

      return groups;
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't pause trigger group '" + matcher + "': " + e.getMessage(), e);
    }
  }

  public void pauseAll()
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.pauseAll(conn);
      }
    });
  }

  public void pauseAll(Connection conn)
    throws JobPersistenceException
  {
    List names = getTriggerGroupNames(conn);

    for (String name : names) {
      pauseTriggerGroup(conn, GroupMatcher.triggerGroupEquals(name));
    }
    try
    {
      if (!getDelegate().isTriggerGroupPaused(conn, "_$_ALL_GROUPS_PAUSED_$_"))
        getDelegate().insertPausedTriggerGroup(conn, "_$_ALL_GROUPS_PAUSED_$_");
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't pause all trigger groups: " + e.getMessage(), e);
    }
  }

  public void resumeAll()
    throws JobPersistenceException
  {
    executeInLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.resumeAll(conn);
      }
    });
  }

  public void resumeAll(Connection conn)
    throws JobPersistenceException
  {
    List names = getTriggerGroupNames(conn);

    for (String name : names) {
      resumeTriggerGroup(conn, GroupMatcher.triggerGroupEquals(name));
    }
    try
    {
      getDelegate().deletePausedTriggerGroup(conn, "_$_ALL_GROUPS_PAUSED_$_");
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't resume all trigger groups: " + e.getMessage(), e);
    }
  }

  protected synchronized String getFiredTriggerRecordId()
  {
    return getInstanceId() + ftrCtr++;
  }

  public List<OperableTrigger> acquireNextTriggers(final long noLaterThan, int maxCount, final long timeWindow)
    throws JobPersistenceException
  {
    if ((isAcquireTriggersWithinLock()) || (maxCount > 1)) {
      return (List)executeInNonManagedTXLock("TRIGGER_ACCESS", new TransactionCallback()
      {
        public Object execute(Connection conn) throws JobPersistenceException
        {
          return JobStoreSupport.this.acquireNextTrigger(conn, noLaterThan, timeWindow, this.val$timeWindow);
        }
      });
    }

    return (List)executeInNonManagedTXLock(null, new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        return JobStoreSupport.this.acquireNextTrigger(conn, noLaterThan, timeWindow, this.val$timeWindow);
      }
    });
  }

  protected List<OperableTrigger> acquireNextTrigger(Connection conn, long noLaterThan, int maxCount, long timeWindow)
    throws JobPersistenceException
  {
    List acquiredTriggers = new ArrayList();
    Set acquiredJobKeysForNoConcurrentExec = new HashSet();
    int MAX_DO_LOOP_RETRY = 3;
    int currentLoopCount = 0;
    long firstAcquiredTriggerFireTime = 0L;
    while (true)
    {
      currentLoopCount++;
      try {
        List keys = null;

        if (timeWindow > 0L)
          keys = getDelegate().selectTriggerToAcquire(conn, noLaterThan + timeWindow, getMisfireTime(), maxCount);
        else {
          keys = getDelegate().selectTriggerToAcquire(conn, noLaterThan, getMisfireTime(), maxCount);
        }

        if ((keys == null) || (keys.size() == 0)) {
          return acquiredTriggers;
        }
        for (TriggerKey triggerKey : keys)
        {
          OperableTrigger nextTrigger = retrieveTrigger(conn, triggerKey);
          if (nextTrigger != null)
          {
            if ((firstAcquiredTriggerFireTime > 0L) && (nextTrigger.getNextFireTime().getTime() > firstAcquiredTriggerFireTime + timeWindow))
            {
              break;
            }

            JobKey jobKey = nextTrigger.getJobKey();
            JobDetail job = getDelegate().selectJobDetail(conn, jobKey, getClassLoadHelper());
            if (job.isConcurrentExectionDisallowed()) {
              if (!acquiredJobKeysForNoConcurrentExec.contains(jobKey))
              {
                acquiredJobKeysForNoConcurrentExec.add(jobKey);
              }

            }
            else
            {
              int rowsUpdated = getDelegate().updateTriggerStateFromOtherState(conn, triggerKey, "ACQUIRED", "WAITING");
              if (rowsUpdated > 0)
              {
                nextTrigger.setFireInstanceId(getFiredTriggerRecordId());
                getDelegate().insertFiredTrigger(conn, nextTrigger, "ACQUIRED", null);

                acquiredTriggers.add(nextTrigger);
                if (firstAcquiredTriggerFireTime == 0L) {
                  firstAcquiredTriggerFireTime = nextTrigger.getNextFireTime().getTime();
                }
              }
            }
          }

        }

        if ((acquiredTriggers.size() == 0) && (currentLoopCount < 3));
      }
      catch (Exception e)
      {
        throw new JobPersistenceException("Couldn't acquire next trigger: " + e.getMessage(), e);
      }

    }

    return acquiredTriggers;
  }

  public void releaseAcquiredTrigger(final OperableTrigger trigger)
    throws JobPersistenceException
  {
    executeInNonManagedTXLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.releaseAcquiredTrigger(conn, trigger);
      }
    });
  }

  protected void releaseAcquiredTrigger(Connection conn, OperableTrigger trigger) throws JobPersistenceException
  {
    try
    {
      getDelegate().updateTriggerStateFromOtherState(conn, trigger.getKey(), "WAITING", "ACQUIRED");

      getDelegate().deleteFiredTrigger(conn, trigger.getFireInstanceId());
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't release acquired trigger: " + e.getMessage(), e);
    }
  }

  public List<TriggerFiredResult> triggersFired(final List<OperableTrigger> triggers)
    throws JobPersistenceException
  {
    return (List)executeInNonManagedTXLock("TRIGGER_ACCESS", new TransactionCallback()
    {
      public Object execute(Connection conn)
        throws JobPersistenceException
      {
        List results = new ArrayList();

        for (OperableTrigger trigger : triggers) {
          TriggerFiredResult result;
          try { TriggerFiredBundle bundle = JobStoreSupport.this.triggerFired(conn, trigger);
            result = new TriggerFiredResult(bundle);
          } catch (JobPersistenceException jpe) {
            result = new TriggerFiredResult(jpe);
          } catch (RuntimeException re) {
            result = new TriggerFiredResult(re);
          }
          results.add(result);
        }

        return results;
      }
    });
  }

  protected TriggerFiredBundle triggerFired(Connection conn, OperableTrigger trigger)
    throws JobPersistenceException
  {
    JobDetail job = null;
    Calendar cal = null;
    try
    {
      String state = getDelegate().selectTriggerState(conn, trigger.getKey());

      if (!state.equals("ACQUIRED"))
        return null;
    }
    catch (SQLException e) {
      throw new JobPersistenceException("Couldn't select trigger state: " + e.getMessage(), e);
    }

    try
    {
      job = retrieveJob(conn, trigger.getJobKey());
      if (job == null) return null; 
    }
    catch (JobPersistenceException jpe) {
      try { getLog().error("Error retrieving job, setting trigger state to ERROR.", jpe);
        getDelegate().updateTriggerState(conn, trigger.getKey(), "ERROR");
      } catch (SQLException sqle)
      {
        getLog().error("Unable to set trigger state to ERROR.", sqle);
      }
      throw jpe;
    }

    if (trigger.getCalendarName() != null) {
      cal = retrieveCalendar(conn, trigger.getCalendarName());
      if (cal == null) return null;
    }
    try
    {
      getDelegate().updateFiredTrigger(conn, trigger, "EXECUTING", job);
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't insert fired trigger: " + e.getMessage(), e);
    }

    Date prevFireTime = trigger.getPreviousFireTime();

    trigger.triggered(cal);

    String state = "WAITING";
    boolean force = true;

    if (job.isConcurrentExectionDisallowed()) {
      state = "BLOCKED";
      force = false;
      try {
        getDelegate().updateTriggerStatesForJobFromOtherState(conn, job.getKey(), "BLOCKED", "WAITING");

        getDelegate().updateTriggerStatesForJobFromOtherState(conn, job.getKey(), "BLOCKED", "ACQUIRED");

        getDelegate().updateTriggerStatesForJobFromOtherState(conn, job.getKey(), "PAUSED_BLOCKED", "PAUSED");
      }
      catch (SQLException e) {
        throw new JobPersistenceException("Couldn't update states of blocked triggers: " + e.getMessage(), e);
      }

    }

    if (trigger.getNextFireTime() == null) {
      state = "COMPLETE";
      force = true;
    }

    storeTrigger(conn, trigger, job, true, state, force, false);

    job.getJobDataMap().clearDirtyFlag();

    return new TriggerFiredBundle(job, trigger, cal, trigger.getKey().getGroup().equals("RECOVERING_JOBS"), new Date(), trigger.getPreviousFireTime(), prevFireTime, trigger.getNextFireTime());
  }

  public void triggeredJobComplete(final OperableTrigger trigger, final JobDetail jobDetail, final Trigger.CompletedExecutionInstruction triggerInstCode)
    throws JobPersistenceException
  {
    executeInNonManagedTXLock("TRIGGER_ACCESS", new VoidTransactionCallback()
    {
      public void execute(Connection conn) throws JobPersistenceException
      {
        JobStoreSupport.this.triggeredJobComplete(conn, trigger, jobDetail, triggerInstCode);
      }
    });
  }

  protected void triggeredJobComplete(Connection conn, OperableTrigger trigger, JobDetail jobDetail, Trigger.CompletedExecutionInstruction triggerInstCode) throws JobPersistenceException
  {
    try
    {
      if (triggerInstCode == Trigger.CompletedExecutionInstruction.DELETE_TRIGGER) {
        if (trigger.getNextFireTime() == null)
        {
          TriggerStatus stat = getDelegate().selectTriggerStatus(conn, trigger.getKey());

          if ((stat != null) && (stat.getNextFireTime() == null))
            removeTrigger(conn, trigger.getKey());
        }
        else {
          removeTrigger(conn, trigger.getKey());
          signalSchedulingChangeOnTxCompletion(0L);
        }
      } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_TRIGGER_COMPLETE) {
        getDelegate().updateTriggerState(conn, trigger.getKey(), "COMPLETE");

        signalSchedulingChangeOnTxCompletion(0L);
      } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_TRIGGER_ERROR) {
        getLog().info("Trigger " + trigger.getKey() + " set to ERROR state.");
        getDelegate().updateTriggerState(conn, trigger.getKey(), "ERROR");

        signalSchedulingChangeOnTxCompletion(0L);
      } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_COMPLETE) {
        getDelegate().updateTriggerStatesForJob(conn, trigger.getJobKey(), "COMPLETE");

        signalSchedulingChangeOnTxCompletion(0L);
      } else if (triggerInstCode == Trigger.CompletedExecutionInstruction.SET_ALL_JOB_TRIGGERS_ERROR) {
        getLog().info("All triggers of Job " + trigger.getKey() + " set to ERROR state.");

        getDelegate().updateTriggerStatesForJob(conn, trigger.getJobKey(), "ERROR");

        signalSchedulingChangeOnTxCompletion(0L);
      }

      if (jobDetail.isConcurrentExectionDisallowed()) {
        getDelegate().updateTriggerStatesForJobFromOtherState(conn, jobDetail.getKey(), "WAITING", "BLOCKED");

        getDelegate().updateTriggerStatesForJobFromOtherState(conn, jobDetail.getKey(), "PAUSED", "PAUSED_BLOCKED");

        signalSchedulingChangeOnTxCompletion(0L);
      }
      if (jobDetail.isPersistJobDataAfterExecution())
        try {
          if (jobDetail.getJobDataMap().isDirty())
            getDelegate().updateJobData(conn, jobDetail);
        }
        catch (IOException e) {
          throw new JobPersistenceException("Couldn't serialize job data: " + e.getMessage(), e);
        }
        catch (SQLException e) {
          throw new JobPersistenceException("Couldn't update job data: " + e.getMessage(), e);
        }
    }
    catch (SQLException e)
    {
      throw new JobPersistenceException("Couldn't update trigger state(s): " + e.getMessage(), e);
    }

    try
    {
      getDelegate().deleteFiredTrigger(conn, trigger.getFireInstanceId());
    } catch (SQLException e) {
      throw new JobPersistenceException("Couldn't delete fired trigger: " + e.getMessage(), e);
    }
  }

  protected DriverDelegate getDelegate()
    throws NoSuchDelegateException
  {
    synchronized (this) {
      if (null == this.delegate) {
        try {
          if (this.delegateClassName != null) {
            this.delegateClass = getClassLoadHelper().loadClass(this.delegateClassName, DriverDelegate.class);
          }

          Constructor ctor = null;
          Object[] ctorParams = null;
          if (canUseProperties()) {
            Class[] ctorParamTypes = { Logger.class, String.class, String.class, String.class, ClassLoadHelper.class, Boolean.class };

            ctor = this.delegateClass.getConstructor(ctorParamTypes);
            ctorParams = new Object[] { getLog(), this.tablePrefix, this.instanceName, this.instanceId, getClassLoadHelper(), Boolean.valueOf(canUseProperties()) };
          }
          else {
            Class[] ctorParamTypes = { Logger.class, String.class, String.class, String.class, ClassLoadHelper.class };

            ctor = this.delegateClass.getConstructor(ctorParamTypes);
            ctorParams = new Object[] { getLog(), this.tablePrefix, this.instanceName, this.instanceId, getClassLoadHelper() };
          }

          this.delegate = ((DriverDelegate)ctor.newInstance(ctorParams));

          this.delegate.initialize(getDriverDelegateInitString());
        }
        catch (NoSuchMethodException e) {
          throw new NoSuchDelegateException("Couldn't find delegate constructor: " + e.getMessage(), e);
        }
        catch (InstantiationException e) {
          throw new NoSuchDelegateException("Couldn't create delegate: " + e.getMessage(), e);
        }
        catch (IllegalAccessException e) {
          throw new NoSuchDelegateException("Couldn't create delegate: " + e.getMessage(), e);
        }
        catch (InvocationTargetException e) {
          throw new NoSuchDelegateException("Couldn't create delegate: " + e.getMessage(), e);
        }
        catch (ClassNotFoundException e) {
          throw new NoSuchDelegateException("Couldn't load delegate class: " + e.getMessage(), e);
        }
      }

      return this.delegate;
    }
  }

  protected Semaphore getLockHandler() {
    return this.lockHandler;
  }

  public void setLockHandler(Semaphore lockHandler) {
    this.lockHandler = lockHandler;
  }

  protected RecoverMisfiredJobsResult doRecoverMisfires()
    throws JobPersistenceException
  {
    boolean transOwner = false;
    Connection conn = getNonManagedTXConnection();
    try {
      RecoverMisfiredJobsResult result = RecoverMisfiredJobsResult.NO_OP;

      int misfireCount = getDoubleCheckLockMisfireHandler() ? getDelegate().countMisfiredTriggersInState(conn, "WAITING", getMisfireTime()) : 2147483647;

      if (misfireCount == 0) {
        getLog().debug("Found 0 triggers that missed their scheduled fire-time.");
      }
      else {
        transOwner = getLockHandler().obtainLock(conn, "TRIGGER_ACCESS");

        result = recoverMisfiredJobs(conn, false);
      }

      commitConnection(conn);
      return result;
    } catch (JobPersistenceException e) {
      rollbackConnection(conn);
      throw e;
    } catch (SQLException e) {
      rollbackConnection(conn);
      throw new JobPersistenceException("Database error recovering from misfires.", e);
    } catch (RuntimeException e) {
      rollbackConnection(conn);
      throw new JobPersistenceException("Unexpected runtime exception: " + e.getMessage(), e);
    }
    finally {
      try {
        releaseLock(conn, "TRIGGER_ACCESS", transOwner);
      } finally {
        cleanupConnection(conn);
      }
    }
  }

  protected void signalSchedulingChangeOnTxCompletion(long candidateNewNextFireTime)
  {
    Long sigTime = (Long)this.sigChangeForTxCompletion.get();
    if ((sigTime == null) && (candidateNewNextFireTime >= 0L)) {
      this.sigChangeForTxCompletion.set(Long.valueOf(candidateNewNextFireTime));
    }
    else if ((sigTime == null) || (candidateNewNextFireTime < sigTime.longValue()))
      this.sigChangeForTxCompletion.set(Long.valueOf(candidateNewNextFireTime));
  }

  protected Long clearAndGetSignalSchedulingChangeOnTxCompletion()
  {
    Long t = (Long)this.sigChangeForTxCompletion.get();
    this.sigChangeForTxCompletion.set(null);
    return t;
  }

  protected void signalSchedulingChangeImmediately(long candidateNewNextFireTime) {
    this.schedSignaler.signalSchedulingChange(candidateNewNextFireTime);
  }

  protected boolean doCheckin()
    throws JobPersistenceException
  {
    boolean transOwner = false;
    boolean transStateOwner = false;
    boolean recovered = false;

    Connection conn = getNonManagedTXConnection();
    try
    {
      List failedRecords = null;
      if (!this.firstCheckIn) {
        failedRecords = clusterCheckIn(conn);
        commitConnection(conn);
      }

      if ((this.firstCheckIn) || (failedRecords.size() > 0)) {
        getLockHandler().obtainLock(conn, "STATE_ACCESS");
        transStateOwner = true;

        failedRecords = this.firstCheckIn ? clusterCheckIn(conn) : findFailedInstances(conn);

        if (failedRecords.size() > 0) {
          getLockHandler().obtainLock(conn, "TRIGGER_ACCESS");

          transOwner = true;

          clusterRecover(conn, failedRecords);
          recovered = true;
        }
      }

      commitConnection(conn);
    } catch (JobPersistenceException e) {
      rollbackConnection(conn);
      throw e;
    } finally {
      try {
        releaseLock(conn, "TRIGGER_ACCESS", transOwner);
      } finally {
        try {
          releaseLock(conn, "STATE_ACCESS", transStateOwner);
        } finally {
          cleanupConnection(conn);
        }
      }
    }

    this.firstCheckIn = false;

    return recovered;
  }

  protected List<SchedulerStateRecord> findFailedInstances(Connection conn)
    throws JobPersistenceException
  {
    try
    {
      List failedInstances = new LinkedList();
      boolean foundThisScheduler = false;
      long timeNow = System.currentTimeMillis();

      List states = getDelegate().selectSchedulerStateRecords(conn, null);

      for (SchedulerStateRecord rec : states)
      {
        if (rec.getSchedulerInstanceId().equals(getInstanceId())) {
          foundThisScheduler = true;
          if (this.firstCheckIn) {
            failedInstances.add(rec);
          }

        }
        else if (calcFailedIfAfter(rec) < timeNow) {
          failedInstances.add(rec);
        }

      }

      if (this.firstCheckIn) {
        failedInstances.addAll(findOrphanedFailedInstances(conn, states));
      }

      if ((!foundThisScheduler) && (!this.firstCheckIn))
      {
        getLog().warn("This scheduler instance (" + getInstanceId() + ") is still " + "active but was recovered by another instance in the cluster.  " + "This may cause inconsistent behavior.");
      }

      return failedInstances;
    } catch (Exception e) {
      this.lastCheckin = System.currentTimeMillis();
      throw new JobPersistenceException("Failure identifying failed instances when checking-in: " + e.getMessage(), e);
    }
  }

  private List<SchedulerStateRecord> findOrphanedFailedInstances(Connection conn, List<SchedulerStateRecord> schedulerStateRecords)
    throws SQLException, NoSuchDelegateException
  {
    List orphanedInstances = new ArrayList();

    Set allFiredTriggerInstanceNames = getDelegate().selectFiredTriggerInstanceNames(conn);
    if (!allFiredTriggerInstanceNames.isEmpty()) {
      for (SchedulerStateRecord rec : schedulerStateRecords)
      {
        allFiredTriggerInstanceNames.remove(rec.getSchedulerInstanceId());
      }

      for (String inst : allFiredTriggerInstanceNames)
      {
        SchedulerStateRecord orphanedInstance = new SchedulerStateRecord();
        orphanedInstance.setSchedulerInstanceId(inst);

        orphanedInstances.add(orphanedInstance);

        getLog().warn("Found orphaned fired triggers for instance: " + orphanedInstance.getSchedulerInstanceId());
      }

    }

    return orphanedInstances;
  }

  protected long calcFailedIfAfter(SchedulerStateRecord rec) {
    return rec.getCheckinTimestamp() + Math.max(rec.getCheckinInterval(), System.currentTimeMillis() - this.lastCheckin) + 7500L;
  }

  protected List<SchedulerStateRecord> clusterCheckIn(Connection conn)
    throws JobPersistenceException
  {
    List failedInstances = findFailedInstances(conn);
    try
    {
      this.lastCheckin = System.currentTimeMillis();
      if (getDelegate().updateSchedulerState(conn, getInstanceId(), this.lastCheckin) == 0) {
        getDelegate().insertSchedulerState(conn, getInstanceId(), this.lastCheckin, getClusterCheckinInterval());
      }
    }
    catch (Exception e)
    {
      throw new JobPersistenceException("Failure updating scheduler state when checking-in: " + e.getMessage(), e);
    }

    return failedInstances;
  }

  protected void clusterRecover(Connection conn, List<SchedulerStateRecord> failedInstances)
    throws JobPersistenceException
  {
    if (failedInstances.size() > 0)
    {
      long recoverIds = System.currentTimeMillis();

      logWarnIfNonZero(failedInstances.size(), "ClusterManager: detected " + failedInstances.size() + " failed or restarted instances.");
      try
      {
        Iterator itr = failedInstances.iterator();
        while (itr.hasNext()) {
          SchedulerStateRecord rec = (SchedulerStateRecord)itr.next();

          getLog().info("ClusterManager: Scanning for instance \"" + rec.getSchedulerInstanceId() + "\"'s failed in-progress jobs.");

          List firedTriggerRecs = getDelegate().selectInstancesFiredTriggerRecords(conn, rec.getSchedulerInstanceId());

          int acquiredCount = 0;
          int recoveredCount = 0;
          int otherCount = 0;

          Set triggerKeys = new HashSet();

          for (FiredTriggerRecord ftRec : firedTriggerRecs)
          {
            TriggerKey tKey = ftRec.getTriggerKey();
            JobKey jKey = ftRec.getJobKey();

            triggerKeys.add(tKey);

            if (ftRec.getFireInstanceState().equals("BLOCKED")) {
              getDelegate().updateTriggerStatesForJobFromOtherState(conn, jKey, "WAITING", "BLOCKED");
            }
            else if (ftRec.getFireInstanceState().equals("PAUSED_BLOCKED")) {
              getDelegate().updateTriggerStatesForJobFromOtherState(conn, jKey, "PAUSED", "PAUSED_BLOCKED");
            }

            if (ftRec.getFireInstanceState().equals("ACQUIRED")) {
              getDelegate().updateTriggerStateFromOtherState(conn, tKey, "WAITING", "ACQUIRED");

              acquiredCount++;
            } else if (ftRec.isJobRequestsRecovery())
            {
              if (jobExists(conn, jKey)) {
                SimpleTriggerImpl rcvryTrig = new SimpleTriggerImpl("recover_" + rec.getSchedulerInstanceId() + "_" + String.valueOf(recoverIds++), "RECOVERING_JOBS", new Date(ftRec.getFireTimestamp()));

                rcvryTrig.setJobName(jKey.getName());
                rcvryTrig.setJobGroup(jKey.getGroup());
                rcvryTrig.setMisfireInstruction(1);
                rcvryTrig.setPriority(ftRec.getPriority());
                JobDataMap jd = getDelegate().selectTriggerJobDataMap(conn, tKey.getName(), tKey.getGroup());
                jd.put("QRTZ_FAILED_JOB_ORIG_TRIGGER_NAME", tKey.getName());
                jd.put("QRTZ_FAILED_JOB_ORIG_TRIGGER_GROUP", tKey.getGroup());
                jd.put("QRTZ_FAILED_JOB_ORIG_TRIGGER_FIRETIME_IN_MILLISECONDS_AS_STRING", String.valueOf(ftRec.getFireTimestamp()));
                rcvryTrig.setJobDataMap(jd);

                rcvryTrig.computeFirstFireTime(null);
                storeTrigger(conn, rcvryTrig, null, false, "WAITING", false, true);

                recoveredCount++;
              } else {
                getLog().warn("ClusterManager: failed job '" + jKey + "' no longer exists, cannot schedule recovery.");

                otherCount++;
              }
            } else {
              otherCount++;
            }

            if (ftRec.isJobDisallowsConcurrentExecution()) {
              getDelegate().updateTriggerStatesForJobFromOtherState(conn, jKey, "WAITING", "BLOCKED");

              getDelegate().updateTriggerStatesForJobFromOtherState(conn, jKey, "PAUSED", "PAUSED_BLOCKED");
            }

          }

          getDelegate().deleteFiredTriggers(conn, rec.getSchedulerInstanceId());

          int completeCount = 0;
          for (TriggerKey triggerKey : triggerKeys)
          {
            if (getDelegate().selectTriggerState(conn, triggerKey).equals("COMPLETE"))
            {
              List firedTriggers = getDelegate().selectFiredTriggerRecords(conn, triggerKey.getName(), triggerKey.getGroup());

              if (firedTriggers.isEmpty())
              {
                if (removeTrigger(conn, triggerKey)) {
                  completeCount++;
                }
              }
            }
          }

          logWarnIfNonZero(acquiredCount, "ClusterManager: ......Freed " + acquiredCount + " acquired trigger(s).");

          logWarnIfNonZero(completeCount, "ClusterManager: ......Deleted " + completeCount + " complete triggers(s).");

          logWarnIfNonZero(recoveredCount, "ClusterManager: ......Scheduled " + recoveredCount + " recoverable job(s) for recovery.");

          logWarnIfNonZero(otherCount, "ClusterManager: ......Cleaned-up " + otherCount + " other failed job(s).");

          if (!rec.getSchedulerInstanceId().equals(getInstanceId()))
            getDelegate().deleteSchedulerState(conn, rec.getSchedulerInstanceId());
        }
      }
      catch (Throwable e)
      {
        throw new JobPersistenceException("Failure recovering jobs: " + e.getMessage(), e);
      }
    }
  }

  protected void logWarnIfNonZero(int val, String warning)
  {
    if (val > 0)
      getLog().info(warning);
    else
      getLog().debug(warning);
  }

  protected void cleanupConnection(Connection conn)
  {
    if (conn != null) {
      if ((conn instanceof Proxy)) {
        Proxy connProxy = (Proxy)conn;

        InvocationHandler invocationHandler = Proxy.getInvocationHandler(connProxy);

        if ((invocationHandler instanceof AttributeRestoringConnectionInvocationHandler)) {
          AttributeRestoringConnectionInvocationHandler connHandler = (AttributeRestoringConnectionInvocationHandler)invocationHandler;

          connHandler.restoreOriginalAtributes();
          closeConnection(connHandler.getWrappedConnection());
          return;
        }

      }

      closeConnection(conn);
    }
  }

  protected void closeConnection(Connection conn)
  {
    if (conn != null)
      try {
        conn.close();
      } catch (SQLException e) {
        getLog().error("Failed to close Connection", e);
      } catch (Throwable e) {
        getLog().error("Unexpected exception closing Connection.  This is often due to a Connection being returned after or during shutdown.", e);
      }
  }

  protected void rollbackConnection(Connection conn)
  {
    if (conn != null)
      try {
        conn.rollback();
      } catch (SQLException e) {
        getLog().error("Couldn't rollback jdbc connection. " + e.getMessage(), e);
      }
  }

  protected void commitConnection(Connection conn)
    throws JobPersistenceException
  {
    if (conn != null)
      try {
        conn.commit();
      } catch (SQLException e) {
        throw new JobPersistenceException("Couldn't commit jdbc connection. " + e.getMessage(), e);
      }
  }

  public Object executeWithoutLock(TransactionCallback txCallback)
    throws JobPersistenceException
  {
    return executeInLock(null, txCallback);
  }

  protected void executeInLock(String lockName, final VoidTransactionCallback txCallback)
    throws JobPersistenceException
  {
    executeInLock(lockName, new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        txCallback.execute(conn);
        return null;
      }
    });
  }

  protected abstract Object executeInLock(String paramString, TransactionCallback paramTransactionCallback)
    throws JobPersistenceException;

  protected void executeInNonManagedTXLock(String lockName, final VoidTransactionCallback txCallback)
    throws JobPersistenceException
  {
    executeInNonManagedTXLock(lockName, new TransactionCallback()
    {
      public Object execute(Connection conn) throws JobPersistenceException
      {
        txCallback.execute(conn);
        return null;
      }
    });
  }

  protected Object executeInNonManagedTXLock(String lockName, TransactionCallback txCallback)
    throws JobPersistenceException
  {
    boolean transOwner = false;
    Connection conn = null;
    try {
      if (lockName != null)
      {
        if (getLockHandler().requiresConnection()) {
          conn = getNonManagedTXConnection();
        }

        transOwner = getLockHandler().obtainLock(conn, lockName);
      }

      if (conn == null) {
        conn = getNonManagedTXConnection();
      }

      Object result = txCallback.execute(conn);
      commitConnection(conn);

      Long sigTime = clearAndGetSignalSchedulingChangeOnTxCompletion();
      if ((sigTime != null) && (sigTime.longValue() >= 0L)) {
        signalSchedulingChangeImmediately(sigTime.longValue());
      }

      return result;
    } catch (JobPersistenceException e) {
      rollbackConnection(conn);
      throw e;
    } catch (RuntimeException e) {
      rollbackConnection(conn);
      throw new JobPersistenceException("Unexpected runtime exception: " + e.getMessage(), e);
    }
    finally {
      try {
        releaseLock(conn, lockName, transOwner);
      } finally {
        cleanupConnection(conn);
      }
    }
  }

  class MisfireHandler extends Thread
  {
    private volatile boolean shutdown = false;

    private int numFails = 0;

    MisfireHandler()
    {
      setName("QuartzScheduler_" + JobStoreSupport.this.instanceName + "-" + JobStoreSupport.this.instanceId + "_MisfireHandler");
      setDaemon(JobStoreSupport.this.getMakeThreadsDaemons());
    }

    public void initialize() {
      ThreadExecutor executor = JobStoreSupport.this.getThreadExecutor();
      executor.execute(this);
    }

    public void shutdown() {
      this.shutdown = true;
      interrupt();
    }

    private JobStoreSupport.RecoverMisfiredJobsResult manage() {
      try {
        JobStoreSupport.this.getLog().debug("MisfireHandler: scanning for misfires...");

        JobStoreSupport.RecoverMisfiredJobsResult res = JobStoreSupport.this.doRecoverMisfires();
        this.numFails = 0;
        return res;
      } catch (Exception e) {
        if (this.numFails % 4 == 0) {
          JobStoreSupport.this.getLog().error("MisfireHandler: Error handling misfires: " + e.getMessage(), e);
        }

        this.numFails += 1;
      }
      return JobStoreSupport.RecoverMisfiredJobsResult.NO_OP;
    }

    public void run()
    {
      while (!this.shutdown)
      {
        long sTime = System.currentTimeMillis();

        JobStoreSupport.RecoverMisfiredJobsResult recoverMisfiredJobsResult = manage();

        if (recoverMisfiredJobsResult.getProcessedMisfiredTriggerCount() > 0) {
          JobStoreSupport.this.signalSchedulingChangeImmediately(recoverMisfiredJobsResult.getEarliestNewTime());
        }

        if (!this.shutdown) {
          long timeToSleep = 50L;
          if (!recoverMisfiredJobsResult.hasMoreMisfiredTriggers()) {
            timeToSleep = JobStoreSupport.this.getMisfireThreshold() - (System.currentTimeMillis() - sTime);
            if (timeToSleep <= 0L) {
              timeToSleep = 50L;
            }

            if (this.numFails > 0) {
              timeToSleep = Math.max(JobStoreSupport.this.getDbRetryInterval(), timeToSleep);
            }
          }
          try
          {
            Thread.sleep(timeToSleep);
          }
          catch (Exception ignore)
          {
          }
        }
      }
    }
  }

  class ClusterManager extends Thread
  {
    private volatile boolean shutdown = false;

    private int numFails = 0;

    ClusterManager() {
      setPriority(7);
      setName("QuartzScheduler_" + JobStoreSupport.this.instanceName + "-" + JobStoreSupport.this.instanceId + "_ClusterManager");
      setDaemon(JobStoreSupport.this.getMakeThreadsDaemons());
    }

    public void initialize() {
      manage();

      ThreadExecutor executor = JobStoreSupport.this.getThreadExecutor();
      executor.execute(this);
    }

    public void shutdown() {
      this.shutdown = true;
      interrupt();
    }

    private boolean manage() {
      boolean res = false;
      try
      {
        res = JobStoreSupport.this.doCheckin();

        this.numFails = 0;
        JobStoreSupport.this.getLog().debug("ClusterManager: Check-in complete.");
      } catch (Exception e) {
        if (this.numFails % 4 == 0) {
          JobStoreSupport.this.getLog().error("ClusterManager: Error managing cluster: " + e.getMessage(), e);
        }

        this.numFails += 1;
      }
      return res;
    }

    public void run()
    {
      while (!this.shutdown)
      {
        if (!this.shutdown) {
          long timeToSleep = JobStoreSupport.this.getClusterCheckinInterval();
          long transpiredTime = System.currentTimeMillis() - JobStoreSupport.this.lastCheckin;
          timeToSleep -= transpiredTime;
          if (timeToSleep <= 0L) {
            timeToSleep = 100L;
          }

          if (this.numFails > 0) {
            timeToSleep = Math.max(JobStoreSupport.this.getDbRetryInterval(), timeToSleep);
          }
          try
          {
            Thread.sleep(timeToSleep);
          }
          catch (Exception ignore) {
          }
        }
        if ((!this.shutdown) && (manage()))
          JobStoreSupport.this.signalSchedulingChangeImmediately(0L);
      }
    }
  }

  protected static abstract interface VoidTransactionCallback
  {
    public abstract void execute(Connection paramConnection)
      throws JobPersistenceException;
  }

  protected static abstract interface TransactionCallback
  {
    public abstract Object execute(Connection paramConnection)
      throws JobPersistenceException;
  }

  protected static class RecoverMisfiredJobsResult
  {
    public static final RecoverMisfiredJobsResult NO_OP = new RecoverMisfiredJobsResult(false, 0, 9223372036854775807L);
    private boolean _hasMoreMisfiredTriggers;
    private int _processedMisfiredTriggerCount;
    private long _earliestNewTime;

    public RecoverMisfiredJobsResult(boolean hasMoreMisfiredTriggers, int processedMisfiredTriggerCount, long earliestNewTime)
    {
      this._hasMoreMisfiredTriggers = hasMoreMisfiredTriggers;
      this._processedMisfiredTriggerCount = processedMisfiredTriggerCount;
      this._earliestNewTime = earliestNewTime;
    }

    public boolean hasMoreMisfiredTriggers() {
      return this._hasMoreMisfiredTriggers;
    }
    public int getProcessedMisfiredTriggerCount() {
      return this._processedMisfiredTriggerCount;
    }
    public long getEarliestNewTime() {
      return this._earliestNewTime;
    }
  }
}