package org.quartz.impl.triggers;

import java.util.Date;
import java.util.TimeZone;
import org.quartz.CalendarIntervalScheduleBuilder;
import org.quartz.CalendarIntervalTrigger;
import org.quartz.DateBuilder.IntervalUnit;
import org.quartz.ScheduleBuilder;
import org.quartz.SchedulerException;

public class CalendarIntervalTriggerImpl extends AbstractTrigger<CalendarIntervalTrigger>
  implements CalendarIntervalTrigger, CoreTrigger
{
  private static final long serialVersionUID = -2635982274232850343L;
  private static final int YEAR_TO_GIVEUP_SCHEDULING_AT = java.util.Calendar.getInstance().get(1) + 100;

  private Date startTime = null;

  private Date endTime = null;

  private Date nextFireTime = null;

  private Date previousFireTime = null;

  private int repeatInterval = 0;

  private DateBuilder.IntervalUnit repeatIntervalUnit = DateBuilder.IntervalUnit.DAY;
  private TimeZone timeZone;
  private boolean preserveHourOfDayAcrossDaylightSavings = false;

  private boolean skipDayIfHourDoesNotExist = false;

  private int timesTriggered = 0;

  private boolean complete = false;

  public CalendarIntervalTriggerImpl()
  {
  }

  public CalendarIntervalTriggerImpl(String name, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, null, intervalUnit, repeatInterval);
  }

  public CalendarIntervalTriggerImpl(String name, String group, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, group, new Date(), null, intervalUnit, repeatInterval);
  }

  public CalendarIntervalTriggerImpl(String name, Date startTime, Date endTime, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, null, startTime, endTime, intervalUnit, repeatInterval);
  }

  public CalendarIntervalTriggerImpl(String name, String group, Date startTime, Date endTime, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    super(name, group);

    setStartTime(startTime);
    setEndTime(endTime);
    setRepeatIntervalUnit(intervalUnit);
    setRepeatInterval(repeatInterval);
  }

  public CalendarIntervalTriggerImpl(String name, String group, String jobName, String jobGroup, Date startTime, Date endTime, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    super(name, group, jobName, jobGroup);

    setStartTime(startTime);
    setEndTime(endTime);
    setRepeatIntervalUnit(intervalUnit);
    setRepeatInterval(repeatInterval);
  }

  public Date getStartTime()
  {
    if (this.startTime == null)
      this.startTime = new Date();
    return this.startTime;
  }

  public void setStartTime(Date startTime)
  {
    if (startTime == null) {
      throw new IllegalArgumentException("Start time cannot be null");
    }

    Date eTime = getEndTime();
    if ((eTime != null) && (startTime != null) && (eTime.before(startTime))) {
      throw new IllegalArgumentException("End time cannot be before start time");
    }

    this.startTime = startTime;
  }

  public Date getEndTime()
  {
    return this.endTime;
  }

  public void setEndTime(Date endTime)
  {
    Date sTime = getStartTime();
    if ((sTime != null) && (endTime != null) && (sTime.after(endTime))) {
      throw new IllegalArgumentException("End time cannot be before start time");
    }

    this.endTime = endTime;
  }

  public DateBuilder.IntervalUnit getRepeatIntervalUnit()
  {
    return this.repeatIntervalUnit;
  }

  public void setRepeatIntervalUnit(DateBuilder.IntervalUnit intervalUnit)
  {
    this.repeatIntervalUnit = intervalUnit;
  }

  public int getRepeatInterval()
  {
    return this.repeatInterval;
  }

  public void setRepeatInterval(int repeatInterval)
  {
    if (repeatInterval < 0) {
      throw new IllegalArgumentException("Repeat interval must be >= 1");
    }

    this.repeatInterval = repeatInterval;
  }

  public TimeZone getTimeZone()
  {
    if (this.timeZone == null) {
      this.timeZone = TimeZone.getDefault();
    }
    return this.timeZone;
  }

  public void setTimeZone(TimeZone timeZone)
  {
    this.timeZone = timeZone;
  }

  public boolean isPreserveHourOfDayAcrossDaylightSavings()
  {
    return this.preserveHourOfDayAcrossDaylightSavings;
  }

  public void setPreserveHourOfDayAcrossDaylightSavings(boolean preserveHourOfDayAcrossDaylightSavings) {
    this.preserveHourOfDayAcrossDaylightSavings = preserveHourOfDayAcrossDaylightSavings;
  }

  public boolean isSkipDayIfHourDoesNotExist()
  {
    return this.skipDayIfHourDoesNotExist;
  }

  public void setSkipDayIfHourDoesNotExist(boolean skipDayIfHourDoesNotExist) {
    this.skipDayIfHourDoesNotExist = skipDayIfHourDoesNotExist;
  }

  public int getTimesTriggered()
  {
    return this.timesTriggered;
  }

  public void setTimesTriggered(int timesTriggered)
  {
    this.timesTriggered = timesTriggered;
  }

  protected boolean validateMisfireInstruction(int misfireInstruction)
  {
    if (misfireInstruction < -1) {
      return false;
    }

    if (misfireInstruction > 2) {
      return false;
    }

    return true;
  }

  public void updateAfterMisfire(org.quartz.Calendar cal)
  {
    int instr = getMisfireInstruction();

    if (instr == -1) {
      return;
    }
    if (instr == 0) {
      instr = 1;
    }

    if (instr == 2) {
      Date newFireTime = getFireTimeAfter(new Date());

      while ((newFireTime != null) && (cal != null) && (!cal.isTimeIncluded(newFireTime.getTime()))) {
        newFireTime = getFireTimeAfter(newFireTime);
      }
      setNextFireTime(newFireTime);
    } else if (instr == 1)
    {
      setNextFireTime(new Date());
    }
  }

  public void triggered(org.quartz.Calendar calendar)
  {
    this.timesTriggered += 1;
    this.previousFireTime = this.nextFireTime;
    this.nextFireTime = getFireTimeAfter(this.nextFireTime);

    while ((this.nextFireTime != null) && (calendar != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT)
        this.nextFireTime = null;
    }
  }

  public void updateWithNewCalendar(org.quartz.Calendar calendar, long misfireThreshold)
  {
    this.nextFireTime = getFireTimeAfter(this.previousFireTime);

    if ((this.nextFireTime == null) || (calendar == null)) {
      return;
    }

    Date now = new Date();
    while ((this.nextFireTime != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT) {
        this.nextFireTime = null;
      }

      if ((this.nextFireTime != null) && (this.nextFireTime.before(now))) {
        long diff = now.getTime() - this.nextFireTime.getTime();
        if (diff >= misfireThreshold)
          this.nextFireTime = getFireTimeAfter(this.nextFireTime);
      }
    }
  }

  public Date computeFirstFireTime(org.quartz.Calendar calendar)
  {
    this.nextFireTime = getStartTime();

    while ((this.nextFireTime != null) && (calendar != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT) {
        return null;
      }
    }

    return this.nextFireTime;
  }

  public Date getNextFireTime()
  {
    return this.nextFireTime;
  }

  public Date getPreviousFireTime()
  {
    return this.previousFireTime;
  }

  public void setNextFireTime(Date nextFireTime)
  {
    this.nextFireTime = nextFireTime;
  }

  public void setPreviousFireTime(Date previousFireTime)
  {
    this.previousFireTime = previousFireTime;
  }

  public Date getFireTimeAfter(Date afterTime)
  {
    return getFireTimeAfter(afterTime, false);
  }

  protected Date getFireTimeAfter(Date afterTime, boolean ignoreEndTime) {
    if (this.complete) {
      return null;
    }

    if (afterTime == null) {
      afterTime = new Date(System.currentTimeMillis() + 1000L);
    }
    else {
      afterTime = new Date(afterTime.getTime() + 1000L);
    }

    long startMillis = getStartTime().getTime();
    long afterMillis = afterTime.getTime();
    long endMillis = getEndTime() == null ? 9223372036854775807L : getEndTime().getTime();

    if ((!ignoreEndTime) && (endMillis <= afterMillis)) {
      return null;
    }

    if (afterMillis < startMillis) {
      return new Date(startMillis);
    }

    long secondsAfterStart = (afterMillis - startMillis) / 1000L;

    Date time = null;
    long repeatLong = getRepeatInterval();

    java.util.Calendar aTime = java.util.Calendar.getInstance();
    aTime.setTime(afterTime);

    java.util.Calendar sTime = java.util.Calendar.getInstance();
    if (this.timeZone != null)
      sTime.setTimeZone(this.timeZone);
    sTime.setTime(getStartTime());
    sTime.setLenient(true);

    if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.SECOND)) {
      long jumpCount = secondsAfterStart / repeatLong;
      if (secondsAfterStart % repeatLong != 0L)
        jumpCount += 1L;
      sTime.add(13, getRepeatInterval() * (int)jumpCount);
      time = sTime.getTime();
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.MINUTE)) {
      long jumpCount = secondsAfterStart / (repeatLong * 60L);
      if (secondsAfterStart % (repeatLong * 60L) != 0L)
        jumpCount += 1L;
      sTime.add(12, getRepeatInterval() * (int)jumpCount);
      time = sTime.getTime();
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.HOUR)) {
      long jumpCount = secondsAfterStart / (repeatLong * 60L * 60L);
      if (secondsAfterStart % (repeatLong * 60L * 60L) != 0L)
        jumpCount += 1L;
      sTime.add(11, getRepeatInterval() * (int)jumpCount);
      time = sTime.getTime();
    }
    else
    {
      int initialHourOfDay = sTime.get(11);

      if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.DAY)) {
        sTime.setLenient(true);

        long jumpCount = secondsAfterStart / (repeatLong * 24L * 60L * 60L);

        if (jumpCount > 20L) {
          if (jumpCount < 50L)
            jumpCount = ()(jumpCount * 0.8D);
          else if (jumpCount < 500L)
            jumpCount = ()(jumpCount * 0.9D);
          else
            jumpCount = ()(jumpCount * 0.95D);
          sTime.add(6, (int)(getRepeatInterval() * jumpCount));
        }

        while ((sTime.getTime().before(afterTime)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(6, getRepeatInterval());
        }
        while ((daylightSavingHourShiftOccuredAndAdvanceNeeded(sTime, initialHourOfDay)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(6, getRepeatInterval());
        }
        time = sTime.getTime();
      }
      else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.WEEK)) {
        sTime.setLenient(true);

        long jumpCount = secondsAfterStart / (repeatLong * 7L * 24L * 60L * 60L);

        if (jumpCount > 20L) {
          if (jumpCount < 50L)
            jumpCount = ()(jumpCount * 0.8D);
          else if (jumpCount < 500L)
            jumpCount = ()(jumpCount * 0.9D);
          else
            jumpCount = ()(jumpCount * 0.95D);
          sTime.add(3, (int)(getRepeatInterval() * jumpCount));
        }

        while ((sTime.getTime().before(afterTime)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(3, getRepeatInterval());
        }
        while ((daylightSavingHourShiftOccuredAndAdvanceNeeded(sTime, initialHourOfDay)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(3, getRepeatInterval());
        }
        time = sTime.getTime();
      }
      else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.MONTH)) {
        sTime.setLenient(true);

        while ((sTime.getTime().before(afterTime)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(2, getRepeatInterval());
        }
        while ((daylightSavingHourShiftOccuredAndAdvanceNeeded(sTime, initialHourOfDay)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(2, getRepeatInterval());
        }
        time = sTime.getTime();
      }
      else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.YEAR))
      {
        while ((sTime.getTime().before(afterTime)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(1, getRepeatInterval());
        }
        while ((daylightSavingHourShiftOccuredAndAdvanceNeeded(sTime, initialHourOfDay)) && (sTime.get(1) < YEAR_TO_GIVEUP_SCHEDULING_AT))
        {
          sTime.add(1, getRepeatInterval());
        }
        time = sTime.getTime();
      }
    }

    if ((!ignoreEndTime) && (endMillis <= time.getTime())) {
      return null;
    }

    return time;
  }

  private boolean daylightSavingHourShiftOccuredAndAdvanceNeeded(java.util.Calendar newTime, int initialHourOfDay) {
    if ((isPreserveHourOfDayAcrossDaylightSavings()) && (newTime.get(11) != initialHourOfDay)) {
      newTime.set(11, initialHourOfDay);
      if (newTime.get(11) != initialHourOfDay)
        return isSkipDayIfHourDoesNotExist();
    }
    return false;
  }

  public Date getFinalFireTime()
  {
    if ((this.complete) || (getEndTime() == null)) {
      return null;
    }

    Date fTime = new Date(getEndTime().getTime() - 1000L);

    fTime = getFireTimeAfter(fTime, true);

    if (fTime.equals(getEndTime())) {
      return fTime;
    }

    java.util.Calendar lTime = java.util.Calendar.getInstance();
    if (this.timeZone != null)
      lTime.setTimeZone(this.timeZone);
    lTime.setTime(fTime);
    lTime.setLenient(true);

    if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.SECOND)) {
      lTime.add(13, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.MINUTE)) {
      lTime.add(12, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.HOUR)) {
      lTime.add(11, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.DAY)) {
      lTime.add(6, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.WEEK)) {
      lTime.add(3, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.MONTH)) {
      lTime.add(2, -1 * getRepeatInterval());
    }
    else if (getRepeatIntervalUnit().equals(DateBuilder.IntervalUnit.YEAR)) {
      lTime.add(1, -1 * getRepeatInterval());
    }

    return lTime.getTime();
  }

  public boolean mayFireAgain()
  {
    return getNextFireTime() != null;
  }

  public void validate()
    throws SchedulerException
  {
    super.validate();

    if (this.repeatInterval < 1)
      throw new SchedulerException("Repeat Interval cannot be zero.");
  }

  public ScheduleBuilder<CalendarIntervalTrigger> getScheduleBuilder()
  {
    CalendarIntervalScheduleBuilder cb = CalendarIntervalScheduleBuilder.calendarIntervalSchedule().withInterval(getRepeatInterval(), getRepeatIntervalUnit());

    switch (getMisfireInstruction()) { case 2:
      cb.withMisfireHandlingInstructionDoNothing();
      break;
    case 1:
      cb.withMisfireHandlingInstructionFireAndProceed();
    }

    return cb;
  }

  public boolean hasAdditionalProperties() {
    return false;
  }
}