package org.quartz.impl.triggers;

import java.util.Date;
import java.util.Set;
import org.quartz.DailyTimeIntervalScheduleBuilder;
import org.quartz.DailyTimeIntervalTrigger;
import org.quartz.DateBuilder.IntervalUnit;
import org.quartz.ScheduleBuilder;
import org.quartz.SchedulerException;
import org.quartz.TimeOfDay;

public class DailyTimeIntervalTriggerImpl extends AbstractTrigger<DailyTimeIntervalTrigger>
  implements DailyTimeIntervalTrigger, CoreTrigger
{
  private static final int YEAR_TO_GIVEUP_SCHEDULING_AT = java.util.Calendar.getInstance().get(1) + 100;

  private Date startTime = null;

  private Date endTime = null;

  private Date nextFireTime = null;

  private Date previousFireTime = null;

  private int repeatCount = -1;

  private int repeatInterval = 1;

  private DateBuilder.IntervalUnit repeatIntervalUnit = DateBuilder.IntervalUnit.MINUTE;
  private Set<Integer> daysOfWeek;
  private TimeOfDay startTimeOfDay;
  private TimeOfDay endTimeOfDay;
  private int timesTriggered = 0;

  private boolean complete = false;

  public DailyTimeIntervalTriggerImpl()
  {
  }

  public DailyTimeIntervalTriggerImpl(String name, TimeOfDay startTimeOfDay, TimeOfDay endTimeOfDay, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, null, startTimeOfDay, endTimeOfDay, intervalUnit, repeatInterval);
  }

  public DailyTimeIntervalTriggerImpl(String name, String group, TimeOfDay startTimeOfDay, TimeOfDay endTimeOfDay, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, group, new Date(), null, startTimeOfDay, endTimeOfDay, intervalUnit, repeatInterval);
  }

  public DailyTimeIntervalTriggerImpl(String name, Date startTime, Date endTime, TimeOfDay startTimeOfDay, TimeOfDay endTimeOfDay, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    this(name, null, startTime, endTime, startTimeOfDay, endTimeOfDay, intervalUnit, repeatInterval);
  }

  public DailyTimeIntervalTriggerImpl(String name, String group, Date startTime, Date endTime, TimeOfDay startTimeOfDay, TimeOfDay endTimeOfDay, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    super(name, group);

    setStartTime(startTime);
    setEndTime(endTime);
    setRepeatIntervalUnit(intervalUnit);
    setRepeatInterval(repeatInterval);
  }

  public DailyTimeIntervalTriggerImpl(String name, String group, String jobName, String jobGroup, Date startTime, Date endTime, TimeOfDay startTimeOfDay, TimeOfDay endTimeOfDay, DateBuilder.IntervalUnit intervalUnit, int repeatInterval)
  {
    super(name, group, jobName, jobGroup);

    setStartTime(startTime);
    setEndTime(endTime);
    setRepeatIntervalUnit(intervalUnit);
    setRepeatInterval(repeatInterval);
  }

  public Date getStartTime()
  {
    if (this.startTime == null) {
      this.startTime = new Date();
    }
    return this.startTime;
  }

  public void setStartTime(Date startTime)
  {
    if (startTime == null) {
      throw new IllegalArgumentException("Start time cannot be null");
    }

    Date eTime = getEndTime();
    if ((eTime != null) && (startTime != null) && (eTime.before(startTime))) {
      throw new IllegalArgumentException("End time cannot be before start time");
    }

    this.startTime = startTime;
  }

  public Date getEndTime()
  {
    return this.endTime;
  }

  public void setEndTime(Date endTime)
  {
    Date sTime = getStartTime();
    if ((sTime != null) && (endTime != null) && (sTime.after(endTime))) {
      throw new IllegalArgumentException("End time cannot be before start time");
    }

    this.endTime = endTime;
  }

  public DateBuilder.IntervalUnit getRepeatIntervalUnit()
  {
    return this.repeatIntervalUnit;
  }

  public void setRepeatIntervalUnit(DateBuilder.IntervalUnit intervalUnit)
  {
    if ((this.repeatIntervalUnit == null) || ((!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.SECOND)) && (!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.MINUTE)) && (!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.HOUR))))
    {
      throw new IllegalArgumentException("Invalid repeat IntervalUnit (must be SECOND, MINUTE or HOUR).");
    }this.repeatIntervalUnit = intervalUnit;
  }

  public int getRepeatInterval()
  {
    return this.repeatInterval;
  }

  public void setRepeatInterval(int repeatInterval)
  {
    if (repeatInterval < 0) {
      throw new IllegalArgumentException("Repeat interval must be >= 1");
    }

    this.repeatInterval = repeatInterval;
  }

  public int getTimesTriggered()
  {
    return this.timesTriggered;
  }

  public void setTimesTriggered(int timesTriggered)
  {
    this.timesTriggered = timesTriggered;
  }

  protected boolean validateMisfireInstruction(int misfireInstruction)
  {
    if (misfireInstruction < -1) {
      return false;
    }

    if (misfireInstruction > 2) {
      return false;
    }

    return true;
  }

  public void updateAfterMisfire(org.quartz.Calendar cal)
  {
    int instr = getMisfireInstruction();

    if (instr == -1) {
      return;
    }
    if (instr == 0) {
      instr = 1;
    }

    if (instr == 2) {
      Date newFireTime = getFireTimeAfter(new Date());

      while ((newFireTime != null) && (cal != null) && (!cal.isTimeIncluded(newFireTime.getTime()))) {
        newFireTime = getFireTimeAfter(newFireTime);
      }
      setNextFireTime(newFireTime);
    } else if (instr == 1)
    {
      setNextFireTime(new Date());
    }
  }

  public void triggered(org.quartz.Calendar calendar)
  {
    this.timesTriggered += 1;
    this.previousFireTime = this.nextFireTime;
    this.nextFireTime = getFireTimeAfter(this.nextFireTime);

    while ((this.nextFireTime != null) && (calendar != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT) {
        this.nextFireTime = null;
      }
    }

    if (this.nextFireTime == null)
      this.complete = true;
  }

  public void updateWithNewCalendar(org.quartz.Calendar calendar, long misfireThreshold)
  {
    this.nextFireTime = getFireTimeAfter(this.previousFireTime);

    if ((this.nextFireTime == null) || (calendar == null)) {
      return;
    }

    Date now = new Date();
    while ((this.nextFireTime != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT) {
        this.nextFireTime = null;
      }

      if ((this.nextFireTime != null) && (this.nextFireTime.before(now))) {
        long diff = now.getTime() - this.nextFireTime.getTime();
        if (diff >= misfireThreshold)
          this.nextFireTime = getFireTimeAfter(this.nextFireTime);
      }
    }
  }

  public Date computeFirstFireTime(org.quartz.Calendar calendar)
  {
    Date sTime = getStartTime();
    Date startTimeOfDayDate = getStartTimeOfDay().getTimeOfDayForDate(sTime);

    if (sTime.getTime() > startTimeOfDayDate.getTime())
      this.nextFireTime = getFireTimeAfter(sTime);
    else {
      this.nextFireTime = advanceToNextDayOfWeek(startTimeOfDayDate, false);
    }

    while ((this.nextFireTime != null) && (calendar != null) && (!calendar.isTimeIncluded(this.nextFireTime.getTime())))
    {
      this.nextFireTime = getFireTimeAfter(this.nextFireTime);

      if (this.nextFireTime == null)
      {
        break;
      }
      java.util.Calendar c = java.util.Calendar.getInstance();
      c.setTime(this.nextFireTime);
      if (c.get(1) > YEAR_TO_GIVEUP_SCHEDULING_AT) {
        return null;
      }
    }

    return this.nextFireTime;
  }

  private java.util.Calendar createCalendarTime(Date dateTime) {
    java.util.Calendar cal = java.util.Calendar.getInstance();
    cal.setTime(dateTime);
    return cal;
  }

  private java.util.Calendar createCalendarTimeWithoutMillis(Date dateTime) {
    java.util.Calendar cal = java.util.Calendar.getInstance();
    cal.setTime(dateTime);
    cal.clear(14);
    return cal;
  }

  public Date getNextFireTime()
  {
    return this.nextFireTime;
  }

  public Date getPreviousFireTime()
  {
    return this.previousFireTime;
  }

  public void setNextFireTime(Date nextFireTime)
  {
    this.nextFireTime = nextFireTime;
  }

  public void setPreviousFireTime(Date previousFireTime)
  {
    this.previousFireTime = previousFireTime;
  }

  public Date getFireTimeAfter(Date afterTime)
  {
    if (this.complete) {
      return null;
    }

    if ((this.repeatCount != -1) && (this.timesTriggered > this.repeatCount)) {
      return null;
    }

    if (afterTime == null)
      afterTime = new Date(System.currentTimeMillis() + 1000L);
    else {
      afterTime = new Date(afterTime.getTime() + 1000L);
    }

    boolean afterTimePassEndTimeOfDay = false;
    if (this.endTimeOfDay != null) {
      afterTimePassEndTimeOfDay = afterTime.getTime() > this.endTimeOfDay.getTimeOfDayForDate(afterTime).getTime();
    }
    Date fireTime = advanceToNextDayOfWeek(afterTime, afterTimePassEndTimeOfDay);
    if (fireTime == null) {
      return null;
    }

    Date fireTimeEndDate = null;
    if (this.endTimeOfDay == null)
      fireTimeEndDate = new TimeOfDay(23, 59, 59).getTimeOfDayForDate(fireTime);
    else {
      fireTimeEndDate = this.endTimeOfDay.getTimeOfDayForDate(fireTime);
    }

    Date fireTimeStartDate = this.startTimeOfDay.getTimeOfDayForDate(fireTime);
    if ((fireTime.getTime() < this.startTime.getTime()) && (this.startTime.getTime() < fireTimeStartDate.getTime()))
      return fireTimeStartDate;
    if ((fireTime.getTime() < this.startTime.getTime()) && (this.startTime.getTime() > fireTimeStartDate.getTime()))
      return this.startTime;
    if ((fireTime.getTime() > this.startTime.getTime()) && (fireTime.getTime() < fireTimeStartDate.getTime())) {
      return fireTimeStartDate;
    }

    this.startTime = fireTimeStartDate;

    long fireMillis = fireTime.getTime();
    long startMillis = this.startTime.getTime();
    long secondsAfterStart = (fireMillis - startMillis) / 1000L;
    long repeatLong = getRepeatInterval();
    java.util.Calendar sTime = createCalendarTime(this.startTime);
    DateBuilder.IntervalUnit repeatUnit = getRepeatIntervalUnit();
    if (repeatUnit.equals(DateBuilder.IntervalUnit.SECOND)) {
      long jumpCount = secondsAfterStart / repeatLong;
      if (secondsAfterStart % repeatLong != 0L)
        jumpCount += 1L;
      sTime.add(13, getRepeatInterval() * (int)jumpCount);
      fireTime = sTime.getTime();
    } else if (repeatUnit.equals(DateBuilder.IntervalUnit.MINUTE)) {
      long jumpCount = secondsAfterStart / (repeatLong * 60L);
      if (secondsAfterStart % (repeatLong * 60L) != 0L)
        jumpCount += 1L;
      sTime.add(12, getRepeatInterval() * (int)jumpCount);
      fireTime = sTime.getTime();
    } else if (repeatUnit.equals(DateBuilder.IntervalUnit.HOUR)) {
      long jumpCount = secondsAfterStart / (repeatLong * 60L * 60L);
      if (secondsAfterStart % (repeatLong * 60L * 60L) != 0L)
        jumpCount += 1L;
      sTime.add(11, getRepeatInterval() * (int)jumpCount);
      fireTime = sTime.getTime();
    }

    if (fireTime.getTime() > fireTimeEndDate.getTime())
    {
      Date fireTimeEndOfDay = new TimeOfDay(23, 59, 59).getTimeOfDayForDate(fireTimeEndDate);
      if (fireTime.getTime() > fireTimeEndOfDay.getTime())
        fireTime = advanceToNextDayOfWeek(fireTime, false);
      else {
        fireTime = advanceToNextDayOfWeek(fireTime, true);
      }
      if (fireTime == null) {
        return null;
      }

      Date nextDayfireTimeStartDate = getStartTimeOfDay().getTimeOfDayForDate(fireTime);
      if (fireTime.getTime() < nextDayfireTimeStartDate.getTime()) {
        fireTime = nextDayfireTimeStartDate;
      }

    }

    return fireTime;
  }

  private Date advanceToNextDayOfWeek(Date fireTime, boolean forceToAdvanceNextDay)
  {
    TimeOfDay sTimeOfDay = getStartTimeOfDay();
    Date fireTimeStartDate = sTimeOfDay.getTimeOfDayForDate(fireTime);
    java.util.Calendar fireTimeStartDateCal = createCalendarTime(fireTimeStartDate);
    int dayOfWeekOfFireTime = fireTimeStartDateCal.get(7);

    Set daysOfWeekToFire = getDaysOfWeek();
    if ((forceToAdvanceNextDay) || (!daysOfWeekToFire.contains(Integer.valueOf(dayOfWeekOfFireTime))))
    {
      for (int i = 1; i <= 7; i++) {
        fireTimeStartDateCal.add(5, 1);
        dayOfWeekOfFireTime = fireTimeStartDateCal.get(7);
        if (daysOfWeekToFire.contains(Integer.valueOf(dayOfWeekOfFireTime))) {
          fireTime = fireTimeStartDateCal.getTime();
          break;
        }
      }

    }

    Date eTime = getEndTime();
    if ((eTime != null) && (fireTime.getTime() > eTime.getTime())) {
      return null;
    }

    return fireTime;
  }

  public Date getFinalFireTime()
  {
    if ((this.complete) || (getEndTime() == null)) {
      return null;
    }

    Date eTime = getEndTime();
    if (this.endTimeOfDay != null) {
      Date endTimeOfDayDate = this.endTimeOfDay.getTimeOfDayForDate(eTime);
      if (eTime.getTime() < endTimeOfDayDate.getTime()) {
        eTime = endTimeOfDayDate;
      }
    }
    return eTime;
  }

  public boolean mayFireAgain()
  {
    return getNextFireTime() != null;
  }

  public void validate()
    throws SchedulerException
  {
    super.validate();

    if ((this.repeatIntervalUnit == null) || ((!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.SECOND)) && (!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.MINUTE)) && (!this.repeatIntervalUnit.equals(DateBuilder.IntervalUnit.HOUR))))
    {
      throw new SchedulerException("Invalid repeat IntervalUnit (must be SECOND, MINUTE or HOUR).");
    }if (this.repeatInterval < 1) {
      throw new SchedulerException("Repeat Interval cannot be zero.");
    }

    long secondsInHour = 86400L;
    if ((this.repeatIntervalUnit == DateBuilder.IntervalUnit.SECOND) && (this.repeatInterval > secondsInHour)) {
      throw new SchedulerException("repeatInterval can not exceed 24 hours (" + secondsInHour + " seconds). Given " + this.repeatInterval);
    }
    if ((this.repeatIntervalUnit == DateBuilder.IntervalUnit.MINUTE) && (this.repeatInterval > secondsInHour / 60L)) {
      throw new SchedulerException("repeatInterval can not exceed 24 hours (" + secondsInHour / 60L + " minutes). Given " + this.repeatInterval);
    }
    if ((this.repeatIntervalUnit == DateBuilder.IntervalUnit.HOUR) && (this.repeatInterval > 24)) {
      throw new SchedulerException("repeatInterval can not exceed 24 hours. Given " + this.repeatInterval + " hours.");
    }

    if ((getEndTimeOfDay() != null) && (!getStartTimeOfDay().before(getEndTimeOfDay())))
      throw new SchedulerException("StartTimeOfDay " + this.startTimeOfDay + " should not come after endTimeOfDay " + this.endTimeOfDay);
  }

  public Set<Integer> getDaysOfWeek()
  {
    if (this.daysOfWeek == null) {
      this.daysOfWeek = DailyTimeIntervalScheduleBuilder.ALL_DAYS_OF_THE_WEEK;
    }
    return this.daysOfWeek;
  }

  public void setDaysOfWeek(Set<Integer> daysOfWeek) {
    if ((daysOfWeek == null) || (daysOfWeek.size() == 0))
      throw new IllegalArgumentException("DaysOfWeek set must be a set that contains at least one day.");
    if (daysOfWeek.size() == 0) {
      throw new IllegalArgumentException("DaysOfWeek set must contain at least one day.");
    }
    this.daysOfWeek = daysOfWeek;
  }

  public TimeOfDay getStartTimeOfDay()
  {
    if (this.startTimeOfDay == null) {
      this.startTimeOfDay = new TimeOfDay(0, 0, 0);
    }
    return this.startTimeOfDay;
  }

  public void setStartTimeOfDay(TimeOfDay startTimeOfDay) {
    if (startTimeOfDay == null) {
      throw new IllegalArgumentException("Start time of day cannot be null");
    }
    TimeOfDay eTime = getEndTimeOfDay();
    if ((eTime != null) && (startTimeOfDay != null) && (eTime.before(startTimeOfDay))) {
      throw new IllegalArgumentException("End time of day cannot be before start time of day");
    }

    this.startTimeOfDay = startTimeOfDay;
  }

  public TimeOfDay getEndTimeOfDay()
  {
    return this.endTimeOfDay;
  }

  public void setEndTimeOfDay(TimeOfDay endTimeOfDay) {
    if (endTimeOfDay == null) {
      throw new IllegalArgumentException("End time of day cannot be null");
    }
    TimeOfDay sTime = getStartTimeOfDay();
    if ((sTime != null) && (endTimeOfDay != null) && (endTimeOfDay.before(endTimeOfDay))) {
      throw new IllegalArgumentException("End time of day cannot be before start time of day");
    }

    this.endTimeOfDay = endTimeOfDay;
  }

  public ScheduleBuilder<DailyTimeIntervalTrigger> getScheduleBuilder()
  {
    DailyTimeIntervalScheduleBuilder cb = DailyTimeIntervalScheduleBuilder.dailyTimeIntervalSchedule().withInterval(getRepeatInterval(), getRepeatIntervalUnit()).onDaysOfTheWeek(getDaysOfWeek()).startingDailyAt(getStartTimeOfDay()).endingDailyAt(getEndTimeOfDay());

    switch (getMisfireInstruction()) { case 2:
      cb.withMisfireHandlingInstructionDoNothing();
      break;
    case 1:
      cb.withMisfireHandlingInstructionFireAndProceed();
    }

    return cb;
  }

  public boolean hasAdditionalProperties()
  {
    return false;
  }

  public int getRepeatCount() {
    return this.repeatCount;
  }

  public void setRepeatCount(int repeatCount) {
    if ((repeatCount < 0) && (repeatCount != -1)) {
      throw new IllegalArgumentException("Repeat count must be >= 0, use the constant REPEAT_INDEFINITELY for infinite.");
    }

    this.repeatCount = repeatCount;
  }
}