package org.quartz.jobs;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import org.quartz.Job;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class NativeJob
  implements Job
{
  private final Logger log;
  public static final String PROP_COMMAND = "command";
  public static final String PROP_PARAMETERS = "parameters";
  public static final String PROP_WAIT_FOR_PROCESS = "waitForProcess";
  public static final String PROP_CONSUME_STREAMS = "consumeStreams";

  public NativeJob()
  {
    this.log = LoggerFactory.getLogger(getClass());
  }

  public void execute(JobExecutionContext context)
    throws JobExecutionException
  {
    JobDataMap data = context.getMergedJobDataMap();

    String command = data.getString("command");

    String parameters = data.getString("parameters");

    if (parameters == null) {
      parameters = "";
    }

    boolean wait = true;
    if (data.containsKey("waitForProcess")) {
      wait = data.getBooleanValue("waitForProcess");
    }
    boolean consumeStreams = false;
    if (data.containsKey("consumeStreams")) {
      consumeStreams = data.getBooleanValue("consumeStreams");
    }

    Integer exitCode = runNativeCommand(command, parameters, wait, consumeStreams);
    context.setResult(exitCode);
  }

  protected Logger getLog()
  {
    return this.log;
  }

  private Integer runNativeCommand(String command, String parameters, boolean wait, boolean consumeStreams) throws JobExecutionException
  {
    String[] cmd = null;
    String[] args = new String[2];
    Integer result = null;
    args[0] = command;
    args[1] = parameters;
    try
    {
      String osName = System.getProperty("os.name");

      if (osName.startsWith("Windows")) {
        cmd = new String[args.length + 2];
        if (osName.equals("Windows 95"))
          cmd[0] = "command.com";
        else {
          cmd[0] = "cmd.exe";
        }
        cmd[1] = "/C";
        for (int i = 0; i < args.length; i++)
          cmd[(i + 2)] = args[i];
      }
      else if (osName.equals("Linux")) {
        if (cmd == null) {
          cmd = new String[3];
        }
        cmd[0] = "/bin/sh";
        cmd[1] = "-c";
        cmd[2] = (args[0] + " " + args[1]);
      } else {
        cmd = args;
      }

      Runtime rt = Runtime.getRuntime();

      getLog().info("About to run " + cmd[0] + " " + cmd[1] + " " + (cmd.length > 2 ? cmd[2] : "") + " ...");
      Process proc = rt.exec(cmd);

      StreamConsumer stdoutConsumer = new StreamConsumer(proc.getInputStream(), "stdout");

      if (consumeStreams) {
        StreamConsumer stderrConsumer = new StreamConsumer(proc.getErrorStream(), "stderr");
        stdoutConsumer.start();
        stderrConsumer.start();
      }

      if (wait) {
        result = Integer.valueOf(proc.waitFor());
      }
    }
    catch (Throwable x)
    {
      throw new JobExecutionException("Error launching native command: ", x, false);
    }

    return result;
  }

  class StreamConsumer extends Thread
  {
    InputStream is;
    String type;

    public StreamConsumer(InputStream inputStream, String type)
    {
      this.is = inputStream;
      this.type = type;
    }

    public void run()
    {
      BufferedReader br = null;
      try {
        br = new BufferedReader(new InputStreamReader(this.is));
        String line = null;

        while ((line = br.readLine()) != null)
          if (this.type.equalsIgnoreCase("stderr"))
            NativeJob.this.getLog().warn(this.type + ">" + line);
          else
            NativeJob.this.getLog().info(this.type + ">" + line);
      }
      catch (IOException ioe)
      {
        NativeJob.this.getLog().error("Error consuming " + this.type + " stream of spawned process.", ioe);
      } finally {
        if (br != null) try {
            br.close();
          }
          catch (Exception ignore)
          {
          }
      }
    }
  }
}