package org.quartz.plugins.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import javax.transaction.UserTransaction;
import org.quartz.JobDataMap;
import org.quartz.Scheduler;
import org.quartz.SchedulerContext;
import org.quartz.SchedulerException;
import org.quartz.TriggerKey;
import org.quartz.impl.JobDetailImpl;
import org.quartz.impl.triggers.SimpleTriggerImpl;
import org.quartz.jobs.FileScanJob;
import org.quartz.jobs.FileScanListener;
import org.quartz.plugins.SchedulerPluginWithUserTransactionSupport;
import org.quartz.simpl.CascadingClassLoadHelper;
import org.quartz.spi.ClassLoadHelper;
import org.quartz.xml.XMLSchedulingDataProcessor;
import org.slf4j.Logger;

public class XMLSchedulingDataProcessorPlugin extends SchedulerPluginWithUserTransactionSupport
  implements FileScanListener
{
  private static final int MAX_JOB_TRIGGER_NAME_LEN = 80;
  private static final String JOB_INITIALIZATION_PLUGIN_NAME = "JobSchedulingDataLoaderPlugin";
  private static final String FILE_NAME_DELIMITERS = ",";
  private boolean failOnFileNotFound = true;

  private String fileNames = "quartz_data.xml";

  private Map<String, JobFile> jobFiles = new LinkedHashMap();

  private long scanInterval = 0L;

  boolean started = false;

  protected ClassLoadHelper classLoadHelper = null;

  private Set<String> jobTriggerNameSet = new HashSet();

  public String getFileNames()
  {
    return this.fileNames;
  }

  public void setFileNames(String fileNames)
  {
    this.fileNames = fileNames;
  }

  public long getScanInterval()
  {
    return this.scanInterval / 1000L;
  }

  public void setScanInterval(long scanInterval)
  {
    this.scanInterval = (scanInterval * 1000L);
  }

  public boolean isFailOnFileNotFound()
  {
    return this.failOnFileNotFound;
  }

  public void setFailOnFileNotFound(boolean failOnFileNotFound)
  {
    this.failOnFileNotFound = failOnFileNotFound;
  }

  public void initialize(String name, Scheduler scheduler)
    throws SchedulerException
  {
    super.initialize(name, scheduler);

    this.classLoadHelper = new CascadingClassLoadHelper();
    this.classLoadHelper.initialize();

    getLog().info("Registering Quartz Job Initialization Plug-in.");

    StringTokenizer stok = new StringTokenizer(this.fileNames, ",");
    while (stok.hasMoreTokens()) {
      String fileName = stok.nextToken();
      JobFile jobFile = new JobFile(fileName);
      this.jobFiles.put(fileName, jobFile);
    }
  }

  public void start(UserTransaction userTransaction)
  {
    try
    {
      if (!this.jobFiles.isEmpty())
      {
        if (this.scanInterval > 0L) {
          getScheduler().getContext().put("JobSchedulingDataLoaderPlugin_" + getName(), this);
        }

        Iterator iterator = this.jobFiles.values().iterator();
        while (iterator.hasNext()) {
          JobFile jobFile = (JobFile)iterator.next();

          if (this.scanInterval > 0L) {
            String jobTriggerName = buildJobTriggerName(jobFile.getFileBasename());
            TriggerKey tKey = new TriggerKey(jobTriggerName, "JobSchedulingDataLoaderPlugin");

            getScheduler().unscheduleJob(tKey);

            SimpleTriggerImpl trig = (SimpleTriggerImpl)getScheduler().getTrigger(tKey);
            trig = new SimpleTriggerImpl();
            trig.setName(jobTriggerName);
            trig.setGroup("JobSchedulingDataLoaderPlugin");
            trig.setStartTime(new Date());
            trig.setEndTime(null);
            trig.setRepeatCount(-1);
            trig.setRepeatInterval(this.scanInterval);

            JobDetailImpl job = new JobDetailImpl(jobTriggerName, "JobSchedulingDataLoaderPlugin", FileScanJob.class);

            job.getJobDataMap().put("FILE_NAME", jobFile.getFileName());
            job.getJobDataMap().put("FILE_SCAN_LISTENER_NAME", "JobSchedulingDataLoaderPlugin_" + getName());

            getScheduler().scheduleJob(job, trig);
            getLog().debug("Scheduled file scan job for data file: {}, at interval: {}", jobFile.getFileName(), Long.valueOf(this.scanInterval));
          }

          processFile(jobFile);
        }
      }
    } catch (SchedulerException se) {
      getLog().error("Error starting background-task for watching jobs file.", se);
    } finally {
      this.started = true;
    }
  }

  private String buildJobTriggerName(String fileBasename)
  {
    String jobTriggerName = "JobSchedulingDataLoaderPlugin_" + getName() + '_' + fileBasename.replace('.', '_');

    if (jobTriggerName.length() > 80) {
      jobTriggerName = jobTriggerName.substring(0, 80);
    }

    int currentIndex = 1;
    while (!this.jobTriggerNameSet.add(jobTriggerName))
    {
      if (currentIndex > 1) {
        jobTriggerName = jobTriggerName.substring(0, jobTriggerName.lastIndexOf('_'));
      }

      String numericSuffix = "_" + currentIndex++;

      if (jobTriggerName.length() > 80 - numericSuffix.length()) {
        jobTriggerName = jobTriggerName.substring(0, 80 - numericSuffix.length());
      }

      jobTriggerName = jobTriggerName + numericSuffix;
    }

    return jobTriggerName;
  }

  public void shutdown()
  {
  }

  private void processFile(JobFile jobFile)
  {
    if ((jobFile == null) || (!jobFile.getFileFound())) {
      return;
    }

    try
    {
      XMLSchedulingDataProcessor processor = new XMLSchedulingDataProcessor(this.classLoadHelper);

      processor.addJobGroupToNeverDelete("JobSchedulingDataLoaderPlugin");
      processor.addTriggerGroupToNeverDelete("JobSchedulingDataLoaderPlugin");

      processor.processFileAndScheduleJobs(jobFile.getFileName(), jobFile.getFileName(), getScheduler());
    }
    catch (Exception e)
    {
      getLog().error("Error scheduling jobs: " + e.getMessage(), e);
    }
  }

  public void processFile(String filePath) {
    processFile((JobFile)this.jobFiles.get(filePath));
  }

  public void fileUpdated(String fileName)
  {
    if (this.started)
      processFile(fileName);
  }

  class JobFile {
    private String fileName;
    private String filePath;
    private String fileBasename;
    private boolean fileFound;

    protected JobFile(String fileName) throws SchedulerException {
      this.fileName = fileName;
      initialize();
    }

    protected String getFileName() {
      return this.fileName;
    }

    protected boolean getFileFound() {
      return this.fileFound;
    }

    protected String getFilePath() {
      return this.filePath;
    }

    protected String getFileBasename() {
      return this.fileBasename;
    }

    private void initialize() throws SchedulerException {
      InputStream f = null;
      try {
        String furl = null;

        File file = new File(getFileName());
        if (!file.exists()) {
          URL url = XMLSchedulingDataProcessorPlugin.this.classLoadHelper.getResource(getFileName());
          if (url != null) {
            try {
              furl = URLDecoder.decode(url.getPath(), "UTF-8");
            } catch (UnsupportedEncodingException e) {
              furl = url.getPath();
            }
            file = new File(furl);
            try {
              f = url.openStream();
            } catch (IOException ignor) {
            }
          }
        }
        else {
          try {
            f = new FileInputStream(file);
          }
          catch (FileNotFoundException e)
          {
          }
        }
        if (f == null) {
          if (XMLSchedulingDataProcessorPlugin.this.isFailOnFileNotFound()) {
            throw new SchedulerException("File named '" + getFileName() + "' does not exist.");
          }

          XMLSchedulingDataProcessorPlugin.this.getLog().warn("File named '" + getFileName() + "' does not exist.");
        }
        else {
          this.fileFound = true;
        }
        this.filePath = (furl != null ? furl : file.getAbsolutePath());
        this.fileBasename = file.getName();
      } finally {
        try {
          if (f != null)
            f.close();
        }
        catch (IOException ioe) {
          XMLSchedulingDataProcessorPlugin.this.getLog().warn("Error closing jobs file " + getFileName(), ioe);
        }
      }
    }
  }
}