package org.quartz.simpl;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import org.quartz.SchedulerConfigException;
import org.quartz.spi.ThreadPool;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SimpleThreadPool
  implements ThreadPool
{
  private int count = -1;

  private int prio = 5;

  private boolean isShutdown = false;
  private boolean handoffPending = false;

  private boolean inheritLoader = false;

  private boolean inheritGroup = true;

  private boolean makeThreadsDaemons = false;
  private ThreadGroup threadGroup;
  private final Object nextRunnableLock = new Object();
  private List<WorkerThread> workers;
  private LinkedList<WorkerThread> availWorkers = new LinkedList();
  private LinkedList<WorkerThread> busyWorkers = new LinkedList();
  private String threadNamePrefix;
  private final Logger log = LoggerFactory.getLogger(getClass());
  private String schedulerInstanceName;

  public SimpleThreadPool()
  {
  }

  public SimpleThreadPool(int threadCount, int threadPriority)
  {
    setThreadCount(threadCount);
    setThreadPriority(threadPriority);
  }

  public Logger getLog()
  {
    return this.log;
  }

  public int getPoolSize() {
    return getThreadCount();
  }

  public void setThreadCount(int count)
  {
    this.count = count;
  }

  public int getThreadCount()
  {
    return this.count;
  }

  public void setThreadPriority(int prio)
  {
    this.prio = prio;
  }

  public int getThreadPriority()
  {
    return this.prio;
  }

  public void setThreadNamePrefix(String prfx) {
    this.threadNamePrefix = prfx;
  }

  public String getThreadNamePrefix() {
    if (this.threadNamePrefix == null) {
      this.threadNamePrefix = (this.schedulerInstanceName + "-SimpleThreadPoolWorker");
    }
    return this.threadNamePrefix;
  }

  public boolean isThreadsInheritContextClassLoaderOfInitializingThread()
  {
    return this.inheritLoader;
  }

  public void setThreadsInheritContextClassLoaderOfInitializingThread(boolean inheritLoader)
  {
    this.inheritLoader = inheritLoader;
  }

  public boolean isThreadsInheritGroupOfInitializingThread() {
    return this.inheritGroup;
  }

  public void setThreadsInheritGroupOfInitializingThread(boolean inheritGroup)
  {
    this.inheritGroup = inheritGroup;
  }

  public boolean isMakeThreadsDaemons()
  {
    return this.makeThreadsDaemons;
  }

  public void setMakeThreadsDaemons(boolean makeThreadsDaemons)
  {
    this.makeThreadsDaemons = makeThreadsDaemons;
  }

  public void setInstanceId(String schedInstId) {
  }

  public void setInstanceName(String schedName) {
    this.schedulerInstanceName = schedName;
  }

  public void initialize() throws SchedulerConfigException
  {
    if ((this.workers != null) && (this.workers.size() > 0)) {
      return;
    }
    if (this.count <= 0) {
      throw new SchedulerConfigException("Thread count must be > 0");
    }

    if ((this.prio <= 0) || (this.prio > 9)) {
      throw new SchedulerConfigException("Thread priority must be > 0 and <= 9");
    }

    if (isThreadsInheritGroupOfInitializingThread()) {
      this.threadGroup = Thread.currentThread().getThreadGroup();
    }
    else {
      this.threadGroup = Thread.currentThread().getThreadGroup();
      ThreadGroup parent = this.threadGroup;
      while (!parent.getName().equals("main")) {
        this.threadGroup = parent;
        parent = this.threadGroup.getParent();
      }
      this.threadGroup = new ThreadGroup(parent, this.schedulerInstanceName + "-SimpleThreadPool");
      if (isMakeThreadsDaemons()) {
        this.threadGroup.setDaemon(true);
      }

    }

    if (isThreadsInheritContextClassLoaderOfInitializingThread()) {
      getLog().info("Job execution threads will use class loader of thread: " + Thread.currentThread().getName());
    }

    Iterator workerThreads = createWorkerThreads(this.count).iterator();
    while (workerThreads.hasNext()) {
      WorkerThread wt = (WorkerThread)workerThreads.next();
      wt.start();
      this.availWorkers.add(wt);
    }
  }

  protected List<WorkerThread> createWorkerThreads(int createCount) {
    this.workers = new LinkedList();
    for (int i = 1; i <= createCount; i++) {
      WorkerThread wt = new WorkerThread(this, this.threadGroup, getThreadNamePrefix() + "-" + i, getThreadPriority(), isMakeThreadsDaemons());

      if (isThreadsInheritContextClassLoaderOfInitializingThread()) {
        wt.setContextClassLoader(Thread.currentThread().getContextClassLoader());
      }

      this.workers.add(wt);
    }

    return this.workers;
  }

  public void shutdown()
  {
    shutdown(true);
  }

  public void shutdown(boolean waitForJobsToComplete)
  {
    synchronized (this.nextRunnableLock) {
      getLog().debug("Shutting down threadpool...");

      this.isShutdown = true;

      if (this.workers == null) {
        return;
      }

      Iterator workerThreads = this.workers.iterator();
      while (workerThreads.hasNext()) {
        WorkerThread wt = (WorkerThread)workerThreads.next();
        wt.shutdown();
        this.availWorkers.remove(wt);
      }

      this.nextRunnableLock.notifyAll();

      if (waitForJobsToComplete == true)
      {
        while (this.handoffPending) try {
            this.nextRunnableLock.wait(100L);
          }
          catch (Throwable t)
          {
          } while (this.busyWorkers.size() > 0) {
          WorkerThread wt = (WorkerThread)this.busyWorkers.getFirst();
          try {
            getLog().debug("Waiting for thread " + wt.getName() + " to shut down");

            this.nextRunnableLock.wait(2000L);
          }
          catch (InterruptedException ex) {
          }
        }
        workerThreads = this.workers.iterator();
        while (workerThreads.hasNext()) {
          WorkerThread wt = (WorkerThread)workerThreads.next();
          try {
            wt.join();
            workerThreads.remove();
          }
          catch (InterruptedException ignore) {
          }
        }
        getLog().debug("No executing jobs remaining, all threads stopped.");
      }
      getLog().debug("Shutdown of threadpool complete.");
    }
  }

  public boolean runInThread(Runnable runnable)
  {
    if (runnable == null) {
      return false;
    }

    synchronized (this.nextRunnableLock)
    {
      this.handoffPending = true;

      while ((this.availWorkers.size() < 1) && (!this.isShutdown))
        try {
          this.nextRunnableLock.wait(500L);
        }
        catch (InterruptedException ignore)
        {
        }
      if (!this.isShutdown) {
        WorkerThread wt = (WorkerThread)this.availWorkers.removeFirst();
        this.busyWorkers.add(wt);
        wt.run(runnable);
      }
      else
      {
        WorkerThread wt = new WorkerThread(this, this.threadGroup, "WorkerThread-LastJob", this.prio, isMakeThreadsDaemons(), runnable);

        this.busyWorkers.add(wt);
        this.workers.add(wt);
        wt.start();
      }
      this.nextRunnableLock.notifyAll();
      this.handoffPending = false;
    }

    return true;
  }

  public int blockForAvailableThreads() {
    synchronized (this.nextRunnableLock)
    {
      while (((this.availWorkers.size() < 1) || (this.handoffPending)) && (!this.isShutdown))
        try {
          this.nextRunnableLock.wait(500L);
        }
        catch (InterruptedException ignore)
        {
        }
      return this.availWorkers.size();
    }
  }

  protected void makeAvailable(WorkerThread wt) {
    synchronized (this.nextRunnableLock) {
      if (!this.isShutdown) {
        this.availWorkers.add(wt);
      }
      this.busyWorkers.remove(wt);
      this.nextRunnableLock.notifyAll();
    }
  }

  protected void clearFromBusyWorkersList(WorkerThread wt) {
    synchronized (this.nextRunnableLock) {
      this.busyWorkers.remove(wt);
      this.nextRunnableLock.notifyAll();
    }
  }

  class WorkerThread extends Thread
  {
    private AtomicBoolean run = new AtomicBoolean(true);
    private SimpleThreadPool tp;
    private Runnable runnable = null;

    private boolean runOnce = false;

    WorkerThread(SimpleThreadPool tp, ThreadGroup threadGroup, String name, int prio, boolean isDaemon)
    {
      this(tp, threadGroup, name, prio, isDaemon, null);
    }

    WorkerThread(SimpleThreadPool tp, ThreadGroup threadGroup, String name, int prio, boolean isDaemon, Runnable runnable)
    {
      super(name);
      this.tp = tp;
      this.runnable = runnable;
      if (runnable != null)
        this.runOnce = true;
      setPriority(prio);
      setDaemon(isDaemon);
    }

    void shutdown()
    {
      this.run.set(false);
    }

    public void run(Runnable newRunnable) {
      synchronized (this) {
        if (this.runnable != null) {
          throw new IllegalStateException("Already running a Runnable!");
        }

        this.runnable = newRunnable;
        notifyAll();
      }
    }

    public void run()
    {
      boolean ran = false;

      while (this.run.get()) {
        try {
          synchronized (this) {
            while ((this.runnable == null) && (this.run.get())) {
              wait(500L);
            }

            if (this.runnable != null) {
              ran = true;
              this.runnable.run();
            }
          }
        }
        catch (InterruptedException unblock) {
          try {
            SimpleThreadPool.this.getLog().error("Worker thread was interrupt()'ed.", unblock);
          }
          catch (Exception e)
          {
            jsr 40;
          } } catch (Throwable exceptionInRunnable) {
          try { SimpleThreadPool.this.getLog().error("Error while executing the Runnable: ", exceptionInRunnable); }
          catch (Exception e)
          {
            jsr 14;
          } } finally { synchronized (this) {
            this.runnable = null;
          }

          if (getPriority() != this.tp.getThreadPriority()) {
            setPriority(this.tp.getThreadPriority());
          }

          if (this.runOnce) {
            this.run.set(false);
            SimpleThreadPool.this.clearFromBusyWorkersList(this);
          } else if (ran) {
            ran = false;
            SimpleThreadPool.this.makeAvailable(this);
          }
        }

      }

      try
      {
        SimpleThreadPool.this.getLog().debug("WorkerThread is shut down.");
      }
      catch (Exception e)
      {
      }
    }
  }
}