package org.quartz.utils;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

public class DirtyFlagMap<K, V>
  implements Map<K, V>, Cloneable, Serializable
{
  private static final long serialVersionUID = 1433884852607126222L;
  private boolean dirty = false;
  private Map<K, V> map;

  public DirtyFlagMap()
  {
    this.map = new HashMap();
  }

  public DirtyFlagMap(int initialCapacity)
  {
    this.map = new HashMap(initialCapacity);
  }

  public DirtyFlagMap(int initialCapacity, float loadFactor)
  {
    this.map = new HashMap(initialCapacity, loadFactor);
  }

  public void clearDirtyFlag()
  {
    this.dirty = false;
  }

  public boolean isDirty()
  {
    return this.dirty;
  }

  public Map<K, V> getWrappedMap()
  {
    return this.map;
  }

  public void clear() {
    if (!this.map.isEmpty()) {
      this.dirty = true;
    }
    this.map.clear();
  }

  public boolean containsKey(Object key) {
    return this.map.containsKey(key);
  }

  public boolean containsValue(Object val) {
    return this.map.containsValue(val);
  }

  public Set<Map.Entry<K, V>> entrySet() {
    return new DirtyFlagMapEntrySet(this.map.entrySet());
  }

  public boolean equals(Object obj)
  {
    if ((obj == null) || (!(obj instanceof DirtyFlagMap))) {
      return false;
    }

    return this.map.equals(((DirtyFlagMap)obj).getWrappedMap());
  }

  public int hashCode()
  {
    return this.map.hashCode();
  }

  public V get(Object key) {
    return this.map.get(key);
  }

  public boolean isEmpty() {
    return this.map.isEmpty();
  }

  public Set<K> keySet() {
    return new DirtyFlagSet(this.map.keySet());
  }

  public V put(K key, V val) {
    this.dirty = true;

    return this.map.put(key, val);
  }

  public void putAll(Map<? extends K, ? extends V> t) {
    if (!t.isEmpty()) {
      this.dirty = true;
    }

    this.map.putAll(t);
  }

  public V remove(Object key) {
    Object obj = this.map.remove(key);

    if (obj != null) {
      this.dirty = true;
    }

    return obj;
  }

  public int size() {
    return this.map.size();
  }

  public Collection<V> values() {
    return new DirtyFlagCollection(this.map.values());
  }

  public Object clone()
  {
    DirtyFlagMap copy;
    try
    {
      copy = (DirtyFlagMap)super.clone();
      if ((this.map instanceof HashMap))
        copy.map = ((Map)((HashMap)this.map).clone());
    }
    catch (CloneNotSupportedException ex) {
      throw new IncompatibleClassChangeError("Not Cloneable.");
    }

    return copy;
  }

  private class DirtyFlagMapEntry
    implements Map.Entry<K, V>
  {
    private Map.Entry<K, V> entry;

    public DirtyFlagMapEntry()
    {
      this.entry = entry;
    }

    public V setValue(V o) {
      DirtyFlagMap.this.dirty = true;
      return this.entry.setValue(o);
    }

    public K getKey() {
      return this.entry.getKey(); } 
    public V getValue() { return this.entry.getValue(); }

  }

  private class DirtyFlagMapEntryIterator extends DirtyFlagMap<K, V>.DirtyFlagIterator<Map.Entry<K, V>>
  {
    public DirtyFlagMapEntryIterator()
    {
      super(iterator);
    }

    public DirtyFlagMap<K, V>.DirtyFlagMapEntry next()
    {
      return new DirtyFlagMap.DirtyFlagMapEntry(DirtyFlagMap.this, (Map.Entry)super.next());
    }
  }

  private class DirtyFlagMapEntrySet extends DirtyFlagMap<K, V>.DirtyFlagSet<Map.Entry<K, V>>
  {
    public DirtyFlagMapEntrySet()
    {
      super(set);
    }

    public Iterator<Map.Entry<K, V>> iterator()
    {
      return new DirtyFlagMap.DirtyFlagMapEntryIterator(DirtyFlagMap.this, getWrappedSet().iterator());
    }

    public Object[] toArray()
    {
      return toArray(new Object[super.size()]);
    }

    public <U> U[] toArray(U[] array)
    {
      if (!array.getClass().getComponentType().isAssignableFrom(Map.Entry.class)) {
        throw new IllegalArgumentException("Array must be of type assignable from Map.Entry");
      }

      int size = super.size();

      Object[] result = array.length < size ? (Object[])Array.newInstance(array.getClass().getComponentType(), size) : array;

      Iterator entryIter = iterator();
      for (int i = 0; i < size; i++) {
        result[i] = entryIter.next();
      }

      if (result.length > size) {
        result[size] = null;
      }

      return result;
    }
  }

  private class DirtyFlagIterator<T>
    implements Iterator<T>
  {
    private Iterator<T> iterator;

    public DirtyFlagIterator()
    {
      this.iterator = iterator;
    }

    public void remove() {
      DirtyFlagMap.this.dirty = true;
      this.iterator.remove();
    }

    public boolean hasNext() {
      return this.iterator.hasNext(); } 
    public T next() { return this.iterator.next(); }

  }

  private class DirtyFlagSet<T> extends DirtyFlagMap<K, V>.DirtyFlagCollection<T>
    implements Set<T>
  {
    public DirtyFlagSet()
    {
      super(set);
    }

    protected Set<T> getWrappedSet() {
      return (Set)getWrappedCollection();
    }
  }

  private class DirtyFlagCollection<T>
    implements Collection<T>
  {
    private Collection<T> collection;

    public DirtyFlagCollection()
    {
      this.collection = c;
    }

    protected Collection<T> getWrappedCollection() {
      return this.collection;
    }

    public Iterator<T> iterator() {
      return new DirtyFlagMap.DirtyFlagIterator(DirtyFlagMap.this, this.collection.iterator());
    }

    public boolean remove(Object o) {
      boolean removed = this.collection.remove(o);
      if (removed) {
        DirtyFlagMap.this.dirty = true;
      }
      return removed;
    }

    public boolean removeAll(Collection<?> c) {
      boolean changed = this.collection.removeAll(c);
      if (changed) {
        DirtyFlagMap.this.dirty = true;
      }
      return changed;
    }

    public boolean retainAll(Collection<?> c) {
      boolean changed = this.collection.retainAll(c);
      if (changed) {
        DirtyFlagMap.this.dirty = true;
      }
      return changed;
    }

    public void clear() {
      if (!this.collection.isEmpty()) {
        DirtyFlagMap.this.dirty = true;
      }
      this.collection.clear();
    }

    public int size() {
      return this.collection.size(); } 
    public boolean isEmpty() { return this.collection.isEmpty(); } 
    public boolean contains(Object o) { return this.collection.contains(o); } 
    public boolean add(T o) { return this.collection.add(o); } 
    public boolean addAll(Collection<? extends T> c) { return this.collection.addAll(c); } 
    public boolean containsAll(Collection<?> c) { return this.collection.containsAll(c); } 
    public Object[] toArray() { return this.collection.toArray(); } 
    public <U> U[] toArray(U[] array) { return this.collection.toArray(array); }

  }
}